@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -68,6 +68,7 @@
 #include <termios.h>
 #include <unistd.h>
 #include <sys/select.h>
+#include <sys/time.h>
 #include <sys/types.h>
 #include <time.h>
 
@@ -95,8 +96,9 @@ static int console_stdout_pending;
 
 #define	CONSOLE_FIFO_LEN	4096
 
-static int console_mouse_x;		/*  absolute x, 0-based  */
-static int console_mouse_y;		/*  absolute y, 0-based  */
+static struct timeval console_mouse_lastupdate;
+static int console_mouse_dx;		/*  relative since last call  */
+static int console_mouse_dy;		/*  relative since last call  */
 static int console_mouse_fb_nr;		/*  framebuffer number of
 					    host movement, 0-based  */
 static int console_mouse_buttons;	/*  left=4, middle=2, right=1  */
@@ -203,23 +205,23 @@ static void start_xterm(int handle)
 	/*  printf("filedesB = %i,%i\n", filedesB[0], filedesB[1]);  */
 
 	/*  NOTE/warning: Hardcoded max nr of args!  */
-	CHECK_ALLOCATION(a = malloc(sizeof(char *) * 20));
+	CHECK_ALLOCATION(a = (char **) malloc(sizeof(char *) * 20));
 
 	a[0] = getenv("XTERM");
 	if (a[0] == NULL)
-		a[0] = "xterm";
-	a[1] = "-geometry";
-	a[2] = "80x25";
-	a[3] = "-title";
+		a[0] = strdup("xterm");
+	a[1] = strdup("-geometry");
+	a[2] = strdup("80x25");
+	a[3] = strdup("-title");
 	mlen = strlen(console_handles[handle].name) +
 	    strlen(console_handles[handle].machine_name) + 30;
-	CHECK_ALLOCATION(a[4] = malloc(mlen));
+	CHECK_ALLOCATION(a[4] = (char *) malloc(mlen));
 	snprintf(a[4], mlen, "GXemul: %s %s",
 	    console_handles[handle].machine_name,
 	    console_handles[handle].name);
-	a[5] = "-e";
+	a[5] = strdup("-e");
 	a[6] = progname;
-	CHECK_ALLOCATION(a[7] = malloc(80));
+	CHECK_ALLOCATION(a[7] = (char *) malloc(80));
 	snprintf(a[7], 80, "-WW@S%i,%i", filedes[0], filedesB[1]);
 	a[8] = NULL;
 
@@ -283,7 +285,13 @@ static int d_avail(int d)
 	FD_SET(d, &rfds);
 	tv.tv_sec = 0;
 	tv.tv_usec = 0;
-	return select(d+1, &rfds, NULL, NULL, &tv);
+
+	for (;;) {
+		// Select may return -1 if interrupted.
+		int r = select(d+1, &rfds, NULL, NULL, &tv);
+		if (r >= 0)
+			return r;
+	}
 }
 
 
@@ -328,10 +336,21 @@ static int console_stdin_avail(int handle)
 }
 
 
+static int console_room_left_in_fifo(int handle)
+{
+	int roomLeftInFIFO = console_handles[handle].fifo_tail
+	    - console_handles[handle].fifo_head;
+	if (roomLeftInFIFO <= 0)
+		roomLeftInFIFO += CONSOLE_FIFO_LEN;
+
+	return roomLeftInFIFO;
+}
+
+
 /*
  *  console_charavail():
  *
- *  Returns 1 if a char is available in the fifo, 0 otherwise.
+ *  Returns the number of chararacters available in the fifo.
  */
 int console_charavail(int handle)
 {
@@ -340,6 +359,12 @@ int console_charavail(int handle)
 		ssize_t len;
 		int i, d;
 
+		// If adding more would lead to a full FIFO, then let's
+		// wait.
+		int roomLeftInFIFO = console_room_left_in_fifo(handle);
+		if (roomLeftInFIFO < (int)sizeof(ch) + 1)
+			break;
+
 		if (!allow_slaves)
 			d = STDIN_FILENO;
 		else
@@ -351,21 +376,58 @@ int console_charavail(int handle)
 			/*  printf("[ %i: %i ]\n", i, ch[i]);  */
 
 			if (!allow_slaves) {
-				/*  Ugly hack: convert ctrl-b into ctrl-c.
-				    (TODO: fix)  */
+				/*  Convert CTRL-B intro CTRL-C.  */
 				if (ch[i] == 2)
 					ch[i] = 3;
 			}
 
+			// CTRL-T (SIGINFO-like) info. This does not work
+			// well when running with -x, since noone checks
+			// for charavail at stdin then...
+			//
+			// if (ch[i] == 20 && d == STDIN_FILENO) {
+			//	printf("\n");
+			//	cpu_show_cycles(machine..., true);
+			// }
+
 			console_makeavail(handle, ch[i]);
 		}
 	}
 
-	if (console_handles[handle].fifo_head ==
-	    console_handles[handle].fifo_tail)
-		return 0;
+	return CONSOLE_FIFO_LEN - console_room_left_in_fifo(handle);
+}
 
-	return 1;
+
+/*
+ *  console_any_input_available():
+ *
+ *  Return true if keyboard or mouse input is available.
+ *
+ *  TODO: Mouse buttons are not checked yet.
+ */
+bool console_any_input_available(struct emul *emul)
+{
+	bool somethingAvailable = false;
+
+	// Any non-zero mouse delta in the last second counts as input available
+	if (console_mouse_dx != 0 || console_mouse_dy != 0) {
+		struct timeval tv;
+		gettimeofday(&tv, NULL);
+
+		uint64_t t1 = console_mouse_lastupdate.tv_sec * 1000000 +
+			console_mouse_lastupdate.tv_usec;
+		uint64_t t2 = tv.tv_sec * 1000000 + tv.tv_usec;
+
+		if (t2 - t1 < 1000000)
+			somethingAvailable = true;
+	}
+
+	if (emul->n_machines == 1) {
+		if (console_charavail(emul->machines[0]->main_console_handle))
+			somethingAvailable = true;
+	}
+
+	return somethingAvailable;
 }
 
 
@@ -446,18 +508,17 @@ void console_flush(void)
 
 
 /*
- *  console_mouse_coordinates():
+ *  console_mouse_coordinate_update():
  *
- *  Sets mouse coordinates. Called by for example an X11 event handler.
- *  x and y are absolute coordinates, fb_nr is where the mouse movement
+ *  Updates mouse coordinates. Called by for example an X11 event handler.
+ *  dx and dy are relative coordinates, fb_nr is where the mouse movement
  *  took place.
  */
-void console_mouse_coordinates(int x, int y, int fb_nr)
+void console_mouse_coordinate_update(int dx, int dy, int fb_nr)
 {
-	/*  TODO: fb_nr isn't used yet.  */
-
-	console_mouse_x = x;
-	console_mouse_y = y;
+	gettimeofday(&console_mouse_lastupdate, NULL);
+	console_mouse_dx += dx;
+	console_mouse_dy += dy;
 	console_mouse_fb_nr = fb_nr;
 }
 
@@ -483,15 +544,18 @@ void console_mouse_button(int button, int pressed)
 /*
  *  console_getmouse():
  *
- *  Puts current mouse data into the variables pointed to by
- *  the arguments.
+ *  Gets current mouse data into the variables pointed to by the arguments.
+ *  The mouse delta x and y are then cleared.
  */
-void console_getmouse(int *x, int *y, int *buttons, int *fb_nr)
+void console_getmouse(int *dx, int *dy, int *buttons, int *fb_nr)
 {
-	*x = console_mouse_x;
-	*y = console_mouse_y;
+	*dx = console_mouse_dx;
+	*dy = console_mouse_dy;
 	*buttons = console_mouse_buttons;
 	*fb_nr = console_mouse_fb_nr;
+
+	console_mouse_dx = 0;
+	console_mouse_dy = 0;
 }
 
 
@@ -533,11 +597,11 @@ static void console_slave_sigcont(int x)
  *  This function is used when running with X11, and gxemul opens up
  *  separate xterms for each emulated terminal or serial port.
  */
-void console_slave(char *arg)
+void console_slave(const char *arg)
 {
 	int inputd;
 	int len;
-	char *p;
+	const char *p;
 	char buf[16384];
 
 	/*  arg = '3,6' or similar, input and output descriptors  */
@@ -597,7 +661,7 @@ void console_slave(char *arg)
  *
  *  For internal use.
  */
-static struct console_handle *console_new_handle(char *name, int *handlep)
+static struct console_handle *console_new_handle(const char *name, int *handlep)
 {
 	struct console_handle *chp;
 	int i, n, found_free = -1;
@@ -613,7 +677,7 @@ static struct console_handle *console_new_handle(char *name, int *handlep)
 	if (found_free == -1) {
 		/*  Let's realloc console_handles[], to make room
 		    for the new one:  */
-		CHECK_ALLOCATION(console_handles =
+		CHECK_ALLOCATION(console_handles = (struct console_handle *)
 		    realloc(console_handles, sizeof(
 		    struct console_handle) * (n_console_handles + 1)));
 		found_free = n_console_handles;
@@ -624,7 +688,7 @@ static struct console_handle *console_new_handle(char *name, int *handlep)
 	memset(chp, 0, sizeof(struct console_handle));
 
 	chp->in_use = 1;
-	chp->machine_name = "";
+	chp->machine_name = strdup("");
 	CHECK_ALLOCATION(chp->name = strdup(name));
 
 	*handlep = found_free;
@@ -657,7 +721,7 @@ static struct console_handle *console_new_handle(char *name, int *handlep)
  *
  *  On failure, -1 is returned.
  */
-int console_start_slave(struct machine *machine, char *consolename,
+int console_start_slave(struct machine *machine, const char *consolename,
 	int use_for_input)
 {
 	struct console_handle *chp;
@@ -703,7 +767,7 @@ int console_start_slave(struct machine *machine, char *consolename,
  *
  *  On failure, -1 is returned.
  */
-int console_start_slave_inputonly(struct machine *machine, char *consolename,
+int console_start_slave_inputonly(struct machine *machine, const char *consolename,
 	int use_for_input)
 {
 	struct console_handle *chp;
@@ -813,8 +877,8 @@ void console_init_main(struct emul *emul)
 	console_handles[MAIN_CONSOLE].fifo_head = 0;
 	console_handles[MAIN_CONSOLE].fifo_tail = 0;
 
-	console_mouse_x = 0;
-	console_mouse_y = 0;
+	console_mouse_dx = 0;
+	console_mouse_dy = 0;
 	console_mouse_buttons = 0;
 
 	console_initialized = 1;
@@ -828,7 +892,7 @@ void console_init_main(struct emul *emul)
  */
 void console_debug_dump(struct machine *machine)
 {
-	int i, iadd = DEBUG_INDENTATION, listed_main = 0;
+	int i, iadd = 1, listed_main = 0;
 
 	if (verbose < 2)
 		return;
@@ -872,6 +936,12 @@ void console_debug_dump(struct machine *machine)
 void console_allow_slaves(int allow)
 {
 	allow_slaves = allow;
+
+	if (allow_slaves) {
+		for (int i = 0; i < n_console_handles; i++)
+			if (i != MAIN_CONSOLE)
+				console_handles[i].using_xterm = USING_XTERM_BUT_NOT_YET_OPEN;
+	}
 }
 
 
