@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -31,6 +31,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "console.h"
 #include "emul.h"
@@ -60,6 +61,158 @@ void x11_check_event(struct emul *emul) { }
 #include <X11/Xutil.h>
 #include <X11/cursorfont.h>
 
+static bool left_ctrl = false;
+static bool left_alt = false;
+static struct fb_window *grabbed = NULL;
+static bool mouseExplicityMoved = false;
+static int mouseXbeforeGrab = 0;
+static int mouseYbeforeGrab = 0;
+static int mouseXofLastEvent = 0;
+static int mouseYofLastEvent = 0;
+
+static bool mouseCursorHidden = false;
+
+
+static void x11_unhide_cursor()
+{
+	if (!mouseCursorHidden)
+		return;
+
+	struct fb_window *fbwin = grabbed;
+
+	/*  Remove the old X11 host cursor:  */
+	XUndefineCursor(fbwin->x11_display, fbwin->x11_fb_window);
+	XFreeCursor(fbwin->x11_display, fbwin->host_cursor);
+	fbwin->host_cursor = 0;
+
+	mouseCursorHidden = false;
+}
+
+
+static void x11_hide_cursor()
+{
+	if (mouseCursorHidden)
+		return;
+
+	struct fb_window *fbwin = grabbed;
+
+	/*  Create a new "empy" X11 host cursor:  */
+	if (fbwin->host_cursor_pixmap != 0) {
+		XFreePixmap(fbwin->x11_display, fbwin->host_cursor_pixmap);
+		fbwin->host_cursor_pixmap = 0;
+	}
+
+	fbwin->host_cursor_pixmap = XCreatePixmap(fbwin->x11_display, fbwin->x11_fb_window, 1, 1, 1);
+	XSetForeground(fbwin->x11_display, fbwin->x11_fb_gc, fbwin->x11_graycolor[0].pixel);
+
+	GC tmpgc = XCreateGC(fbwin->x11_display, fbwin->host_cursor_pixmap, 0,0);
+
+	XDrawPoint(fbwin->x11_display, fbwin->host_cursor_pixmap, tmpgc, 0, 0);
+
+	XFreeGC(fbwin->x11_display, tmpgc);
+
+	fbwin->host_cursor =
+	    XCreatePixmapCursor(fbwin->x11_display,
+	    fbwin->host_cursor_pixmap,
+	    fbwin->host_cursor_pixmap,
+	    &fbwin->x11_graycolor[N_GRAYCOLORS-1],
+	    &fbwin->x11_graycolor[N_GRAYCOLORS-1],
+	    0, 0);
+
+	// For testing:
+	// fbwin->host_cursor = XCreateFontCursor(fbwin->x11_display, XC_coffee_mug);
+
+	if (fbwin->host_cursor != 0)
+		XDefineCursor(fbwin->x11_display, fbwin->x11_fb_window, fbwin->host_cursor);
+
+	mouseCursorHidden = true;
+}
+
+
+static void setMousePointerCoordinates(struct fb_window *fbwin, int x, int y)
+{
+	XWarpPointer(fbwin->x11_display, None,
+	    DefaultRootWindow(fbwin->x11_display), 0, 0, 0, 0, x, y);
+	XFlush(fbwin->x11_display);
+
+	mouseExplicityMoved = true;
+}
+
+
+static void mouseMouseToCenterOfScreen(struct fb_window *fbwin)
+{
+	Screen *screen = XDefaultScreenOfDisplay(fbwin->x11_display);
+
+	int screenWidth = XWidthOfScreen(screen);
+	int screenHeight = XHeightOfScreen(screen);
+	setMousePointerCoordinates(fbwin, screenWidth / 2, screenHeight / 2);
+}
+
+static void grab(struct fb_window *fbwin)
+{
+	if (grabbed != NULL)
+		return;
+
+	Window xqpWindow;
+	int rootx, rooty, x, y;
+	unsigned int mask;
+	int res = XQueryPointer(fbwin->x11_display,
+	    RootWindow(fbwin->x11_display, DefaultScreen(fbwin->x11_display)),
+	    &xqpWindow,
+            &xqpWindow, &rootx, &rooty, &x, &y,
+            &mask);
+        if (res != True)
+        	return;
+
+	debugmsg(SUBSYS_X11, "grab", VERBOSITY_DEBUG, "Mouse coordinates "
+	    "before grab: %i, %i", rootx, rooty);
+	mouseXbeforeGrab = rootx;
+	mouseYbeforeGrab = rooty;
+
+	res = XGrabPointer(fbwin->x11_display, fbwin->x11_fb_window, False,
+	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask | FocusChangeMask |
+	    EnterWindowMask | LeaveWindowMask,
+	    GrabModeAsync, GrabModeAsync,
+	    RootWindow(fbwin->x11_display, DefaultScreen(fbwin->x11_display)),
+	    None, CurrentTime);
+
+	if (res == GrabSuccess)
+		grabbed = fbwin;
+
+	debugmsg(SUBSYS_X11, "grab", VERBOSITY_DEBUG, "Grab mouse pointer: %s",
+	    res == GrabSuccess ? "success" : "FAILURE");
+
+	if (res != GrabSuccess)
+		return;
+
+	x11_hide_cursor();
+
+	mouseMouseToCenterOfScreen(fbwin);
+
+	x11_set_standard_properties(fbwin);
+}
+
+
+static void ungrab()
+{
+	if (grabbed == NULL)
+		return;
+
+	struct fb_window *fbwin = grabbed;
+
+	x11_unhide_cursor();
+
+	grabbed = NULL;
+
+	x11_set_standard_properties(fbwin);
+
+	debugmsg(SUBSYS_X11, "grab", VERBOSITY_DEBUG, "Releasing grab.");
+
+	XUngrabPointer(fbwin->x11_display, CurrentTime);
+
+	setMousePointerCoordinates(fbwin, mouseXbeforeGrab, mouseYbeforeGrab);
+}
+
 
 /*
  *  x11_redraw_cursor():
@@ -70,12 +223,13 @@ void x11_check_event(struct emul *emul) { }
  */
 void x11_redraw_cursor(struct machine *m, int i)
 {
-	int last_color_used = 0;
-	int n_colors_used = 0;
 	struct fb_window *fbwin = m->x11_md.fb_windows[i];
 
+	if (fbwin->x11_display == NULL)
+		return;
+
 	/*  Remove old cursor, if any:  */
-	if (fbwin->x11_display != NULL && fbwin->OLD_cursor_on) {
+	if (fbwin->OLD_cursor_on) {
 		XPutImage(fbwin->x11_display, fbwin->x11_fb_window,
 		    fbwin->x11_fb_gc, fbwin->fb_ximage,
 		    fbwin->OLD_cursor_x/fbwin->scaledown,
@@ -86,141 +240,73 @@ void x11_redraw_cursor(struct machine *m, int i)
 		    fbwin->OLD_cursor_ysize/fbwin->scaledown + 1);
 	}
 
-	if (fbwin->x11_display != NULL && fbwin->cursor_on) {
-		int x, y, subx, suby;
-		XImage *xtmp;
+	if (!fbwin->cursor_on)
+		return;
 
-		CHECK_ALLOCATION(xtmp = XSubImage(fbwin->fb_ximage,
-		    fbwin->cursor_x/fbwin->scaledown,
-		    fbwin->cursor_y/fbwin->scaledown,
-		    fbwin->cursor_xsize/fbwin->scaledown + 1,
-		    fbwin->cursor_ysize/fbwin->scaledown + 1));
+	XImage *xtmp;
+	CHECK_ALLOCATION(xtmp = XSubImage(fbwin->fb_ximage,
+	    fbwin->cursor_x/fbwin->scaledown,
+	    fbwin->cursor_y/fbwin->scaledown,
+	    fbwin->cursor_xsize/fbwin->scaledown + 1,
+	    fbwin->cursor_ysize/fbwin->scaledown + 1));
 
-		for (y=0; y<fbwin->cursor_ysize;
-		    y+=fbwin->scaledown)
-			for (x=0; x<fbwin->cursor_xsize;
-			    x+=fbwin->scaledown) {
-				int px = x/fbwin->scaledown;
-				int py = y/fbwin->scaledown;
-				int p = 0, n = 0, c = 0;
-				unsigned long oldcol;
+	for (int y = 0; y < fbwin->cursor_ysize; y += fbwin->scaledown) {
+		for (int x = 0; x < fbwin->cursor_xsize; x += fbwin->scaledown) {
+			int px = x/fbwin->scaledown;
+			int py = y/fbwin->scaledown;
+			int p = 0, n = 0, c = 0;
+			unsigned long oldcol;
 
-				for (suby=0; suby<fbwin->scaledown;
-				    suby++)
-					for (subx=0; subx<fbwin->
-					    scaledown; subx++) {
-						c = fbwin->
-						    cursor_pixels[y+suby]
-						    [x+subx];
-						if (c >= 0) {
-							p += c;
-							n++;
-						}
+			for (int suby = 0; suby < fbwin->scaledown; suby++)
+				for (int subx = 0; subx < fbwin->scaledown; subx++) {
+					c = fbwin->cursor_pixels[y+suby][x+subx];
+					if (c >= 0) {
+						p += c;
+						n++;
 					}
-				if (n > 0)
-					p /= n;
-				else
-					p = c;
-
-				if (n_colors_used == 0) {
-					last_color_used = p;
-					n_colors_used = 1;
-				} else
-					if (p != last_color_used)
-						n_colors_used = 2;
-
-				switch (p) {
-				case CURSOR_COLOR_TRANSPARENT:
-					break;
-				case CURSOR_COLOR_INVERT:
-					oldcol = XGetPixel(xtmp, px, py);
-					if (oldcol != fbwin->
-					    x11_graycolor[N_GRAYCOLORS-1].pixel)
-						oldcol = fbwin->
-						    x11_graycolor[N_GRAYCOLORS
-						    -1].pixel;
-					else
-						oldcol = fbwin->
-						    x11_graycolor[0].pixel;
-					XPutPixel(xtmp, px, py, oldcol);
-					break;
-				default:	/*  Normal grayscale:  */
-					XPutPixel(xtmp, px, py, fbwin->
-					    x11_graycolor[p].pixel);
 				}
+			if (n > 0)
+				p /= n;
+			else
+				p = c;
+
+			switch (p) {
+			case CURSOR_COLOR_TRANSPARENT:
+				break;
+
+			case CURSOR_COLOR_INVERT:
+				oldcol = XGetPixel(xtmp, px, py);
+				if (oldcol != fbwin->
+				    x11_graycolor[N_GRAYCOLORS-1].pixel)
+					oldcol = fbwin->x11_graycolor[N_GRAYCOLORS-1].pixel;
+				else
+					oldcol = fbwin->x11_graycolor[0].pixel;
+
+				XPutPixel(xtmp, px, py, oldcol);
+				break;
+
+			default:	/*  Normal grayscale:  */
+				XPutPixel(xtmp, px, py, fbwin->x11_graycolor[p].pixel);
 			}
-
-		XPutImage(fbwin->x11_display,
-		    fbwin->x11_fb_window,
-		    fbwin->x11_fb_gc,
-		    xtmp, 0, 0,
-		    fbwin->cursor_x/fbwin->scaledown,
-		    fbwin->cursor_y/fbwin->scaledown,
-		    fbwin->cursor_xsize/fbwin->scaledown,
-		    fbwin->cursor_ysize/fbwin->scaledown);
-
-		XDestroyImage(xtmp);
-
-		fbwin->OLD_cursor_on = fbwin->cursor_on;
-		fbwin->OLD_cursor_x = fbwin->cursor_x;
-		fbwin->OLD_cursor_y = fbwin->cursor_y;
-		fbwin->OLD_cursor_xsize =
-		    fbwin->cursor_xsize;
-		fbwin->OLD_cursor_ysize =
-		    fbwin->cursor_ysize;
-	}
-
-	/*  printf("n_colors_used = %i\n", n_colors_used);  */
-
-	if (fbwin->host_cursor != 0 && n_colors_used < 2) {
-		/*  Remove the old X11 host cursor:  */
-		XUndefineCursor(fbwin->x11_display,
-		    fbwin->x11_fb_window);
-		XFreeCursor(fbwin->x11_display,
-		    fbwin->host_cursor);
-		fbwin->host_cursor = 0;
-	}
-
-	if (n_colors_used >= 2 && fbwin->host_cursor == 0) {
-		GC tmpgc;
-
-		/*  Create a new X11 host cursor:  */
-		/*  cursor = XCreateFontCursor(fbwin->x11_display,
-		    XC_coffee_mug);  :-)  */
-		if (fbwin->host_cursor_pixmap != 0) {
-			XFreePixmap(fbwin->x11_display,
-			    fbwin->host_cursor_pixmap);
-			fbwin->host_cursor_pixmap = 0;
-		}
-		fbwin->host_cursor_pixmap =
-		    XCreatePixmap(fbwin->x11_display,
-		    fbwin->x11_fb_window, 1, 1, 1);
-		XSetForeground(fbwin->x11_display,
-		    fbwin->x11_fb_gc,
-		    fbwin->x11_graycolor[0].pixel);
-
-		tmpgc = XCreateGC(fbwin->x11_display,
-		    fbwin->host_cursor_pixmap, 0,0);
-
-		XDrawPoint(fbwin->x11_display,
-		    fbwin->host_cursor_pixmap,
-		    tmpgc, 0, 0);
-
-		XFreeGC(fbwin->x11_display, tmpgc);
-
-		fbwin->host_cursor =
-		    XCreatePixmapCursor(fbwin->x11_display,
-		    fbwin->host_cursor_pixmap,
-		    fbwin->host_cursor_pixmap,
-		    &fbwin->x11_graycolor[N_GRAYCOLORS-1],
-		    &fbwin->x11_graycolor[N_GRAYCOLORS-1],
-		    0, 0);
-		if (fbwin->host_cursor != 0) {
-			XDefineCursor(fbwin->x11_display,
-			    fbwin->x11_fb_window,
-			    fbwin->host_cursor);
 		}
 	}
+
+	XPutImage(fbwin->x11_display,
+	    fbwin->x11_fb_window,
+	    fbwin->x11_fb_gc,
+	    xtmp, 0, 0,
+	    fbwin->cursor_x/fbwin->scaledown,
+	    fbwin->cursor_y/fbwin->scaledown,
+	    fbwin->cursor_xsize/fbwin->scaledown,
+	    fbwin->cursor_ysize/fbwin->scaledown);
+
+	XDestroyImage(xtmp);
+
+	fbwin->OLD_cursor_on = fbwin->cursor_on;
+	fbwin->OLD_cursor_x = fbwin->cursor_x;
+	fbwin->OLD_cursor_y = fbwin->cursor_y;
+	fbwin->OLD_cursor_xsize = fbwin->cursor_xsize;
+	fbwin->OLD_cursor_ysize = fbwin->cursor_ysize;
 }
 
 
@@ -312,8 +398,8 @@ void x11_init(struct machine *m)
 	if (m->x11_md.n_display_names > 0) {
 		int i;
 		for (i=0; i<m->x11_md.n_display_names; i++)
-			fatal("Using X11 display: %s\n",
-			    m->x11_md.display_names[i]);
+			debugmsg(SUBSYS_X11, "init", VERBOSITY_INFO,
+			    "using X11 display: %s", m->x11_md.display_names[i]);
 	}
 
 	m->x11_md.current_display_name_nr = 0;
@@ -332,7 +418,7 @@ void x11_fb_resize(struct fb_window *win, int new_xsize, int new_ysize)
 	int alloc_depth;
 
 	if (win == NULL) {
-		fatal("x11_fb_resize(): win == NULL\n");
+		debugmsg(SUBSYS_X11, "resize", VERBOSITY_ERROR, "win == NULL");
 		return;
 	}
 
@@ -348,7 +434,7 @@ void x11_fb_resize(struct fb_window *win, int new_xsize, int new_ysize)
 	/*  Note: ximage_data seems to be freed by XDestroyImage below.  */
 	/*  if (win->ximage_data != NULL)
 		free(win->ximage_data);  */
-	CHECK_ALLOCATION(win->ximage_data = malloc(
+	CHECK_ALLOCATION(win->ximage_data = (unsigned char *) malloc(
 	    new_xsize * new_ysize * alloc_depth / 8));
 
 	/*  TODO: clear for non-truecolor modes  */
@@ -371,11 +457,20 @@ void x11_fb_resize(struct fb_window *win, int new_xsize, int new_ysize)
  *
  *  Right now, this only sets the title of a window.
  */
-void x11_set_standard_properties(struct fb_window *fb_window, char *name)
+void x11_set_standard_properties(struct fb_window *fb_window)
 {
+	size_t title_maxlen = strlen(fb_window->name) + 100;
+	char *title;
+	CHECK_ALLOCATION(title = malloc(title_maxlen));
+
+	snprintf(title, title_maxlen, "%s%s", fb_window->name,
+	    grabbed != NULL ? " (Left CTRL+ALT to ungrab)" : "");
+
 	XSetStandardProperties(fb_window->x11_display,
-	    fb_window->x11_fb_window, name, "GXemul "VERSION,
+	    fb_window->x11_fb_window, title, "GXemul " VERSION,
 	    None, NULL, 0, NULL);
+
+	free(title);
 }
 
 
@@ -398,10 +493,11 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 
 	fb_number = m->x11_md.n_fb_windows;
 
-	CHECK_ALLOCATION(m->x11_md.fb_windows = realloc(m->x11_md.fb_windows,
+	CHECK_ALLOCATION(m->x11_md.fb_windows = 
+	    (struct fb_window **) realloc(m->x11_md.fb_windows,
 	    sizeof(struct fb_window *) * (m->x11_md.n_fb_windows + 1)));
 	CHECK_ALLOCATION(fbwin = m->x11_md.fb_windows[fb_number] =
-	    malloc(sizeof(struct fb_window)));
+	    (struct fb_window *) malloc(sizeof(struct fb_window)));
 
 	m->x11_md.n_fb_windows ++;
 
@@ -420,15 +516,18 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 	}
 
 	if (display_name != NULL)
-		debug("[ x11_fb_init(): framebuffer window %i, %ix%i, DISPLAY"
-		    "=%s ]\n", fb_number, xsize, ysize, display_name);
+		debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_INFO,
+		    "framebuffer window %i, %ix%i, DISPLAY"
+		    "=%s", fb_number, xsize, ysize, display_name);
 
 	x11_display = XOpenDisplay(display_name);
 
 	if (x11_display == NULL) {
-		fatal("x11_fb_init(\"%s\"): couldn't open display\n", name);
+		debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_ERROR,
+		    "couldn't open display '%s'", name);
 		if (display_name != NULL)
-			fatal("display_name = '%s'\n", display_name);
+			debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_ERROR, "display_name = '%s'", display_name);
+
 		exit(1);
 	}
 
@@ -442,17 +541,19 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 	    fbwin->x11_screen_depth != 15 &&
 	    fbwin->x11_screen_depth != 16 &&
 	    fbwin->x11_screen_depth != 24) {
-		fatal("\n***\n***  WARNING! Your X server is running %i-bit "
+		debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_WARNING,
+		    "***\n***  WARNING! Your X server is running %i-bit "
 		    "color mode. This is not really\n",
-		    fbwin->x11_screen_depth);
-		fatal("***  supported yet.  8, 15, 16, and 24 bits should "
+		    "***  supported yet.  8, 15, 16, and 24 bits should "
 		    "work.\n***  24-bit server gives color.  Any other bit "
-		    "depth gives undefined result!\n***\n\n");
+		    "depth gives undefined result!\n***",
+		    fbwin->x11_screen_depth);
 	}
 
 	if (fbwin->x11_screen_depth <= 8)
-		debug("WARNING! X11 screen depth is not enough for color; "
-		    "using only 16 grayscales instead\n");
+		debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_WARNING,
+		    "screen depth is not enough for color; "
+		    "using only 16 grayscales instead");
 
 	strlcpy(bg, "Black", sizeof(bg));
 	strlcpy(fg, "White", sizeof(fg));
@@ -500,12 +601,14 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 
 	fbwin->x11_display = x11_display;
 
-	x11_set_standard_properties(fbwin, name);
+	fbwin->name = strdup(name);
+
+	x11_set_standard_properties(fbwin);
 
 	XSelectInput(x11_display,
 	    fbwin->x11_fb_window,
-	    StructureNotifyMask | ExposureMask | ButtonPressMask |
-	    ButtonReleaseMask | PointerMotionMask | KeyPressMask);
+	    StructureNotifyMask | ExposureMask | ButtonPressMask | FocusChangeMask |
+	    ButtonReleaseMask | PointerMotionMask | KeyPressMask | KeyReleaseMask);
 	fbwin->x11_fb_gc = XCreateGC(x11_display,
 	    fbwin->x11_fb_window, 0,0);
 
@@ -522,7 +625,7 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 	fbwin->fb_number = fb_number;
 
 	alloclen = xsize * ysize * alloc_depth / 8;
-	CHECK_ALLOCATION(fbwin->ximage_data = malloc(alloclen));
+	CHECK_ALLOCATION(fbwin->ximage_data = (unsigned char *) malloc(alloclen));
 
 	fbwin->fb_ximage = XCreateImage(fbwin->x11_display, CopyFromParent,
 	    fbwin->x11_screen_depth, ZPixmap, 0, (char *)fbwin->ximage_data,
@@ -533,7 +636,8 @@ struct fb_window *x11_fb_init(int xsize, int ysize, char *name,
 	if (fbwin->x11_screen_depth > 8)
 		memset(fbwin->ximage_data, 0, alloclen);
 	else {
-		debug("x11_fb_init(): clearing the XImage\n");
+		debugmsg(SUBSYS_X11, "fb_init", VERBOSITY_DEBUG,
+		    "clearing the XImage\n");
 		for (y=0; y<ysize; y++)
 			for (x=0; x<xsize; x++)
 				XPutPixel(fbwin->fb_ximage, x, y,
@@ -570,16 +674,19 @@ static void x11_check_events_machine(struct emul *emul, struct machine *m)
 	for (fb_nr = 0; fb_nr < m->x11_md.n_fb_windows; fb_nr ++) {
 		struct fb_window *fbwin = m->x11_md.fb_windows[fb_nr];
 		XEvent event;
-		int need_redraw = 0, found, i, j;
+		bool need_redraw = false;
 
 		while (XPending(fbwin->x11_display)) {
 			XNextEvent(fbwin->x11_display, &event);
 
-			if (event.type==ConfigureNotify) {
-				need_redraw = 1;
+			if (event.type == ConfigureNotify) {
+				need_redraw = true;
 			}
 
-			if (event.type==Expose && event.xexpose.count==0) {
+			if (event.type == FocusOut)
+				ungrab();
+
+			if (event.type == Expose && event.xexpose.count == 0) {
 				/*
 				 *  TODO:  the xexpose struct has x,y,width,
 				 *  height. Those could be used to only redraw
@@ -589,57 +696,98 @@ static void x11_check_events_machine(struct emul *emul, struct machine *m)
 				 */
 				/*  x11_winxsize = event.xexpose.width;
 				    x11_winysize = event.xexpose.height;  */
-				need_redraw = 1;
+				need_redraw = true;
 			}
 
-			if (event.type == MotionNotify) {
-				/*  debug("[ X11 MotionNotify: %i,%i ]\n",
-				    event.xmotion.x, event.xmotion.y);  */
+			if (event.type == MotionNotify && mouseExplicityMoved) {
+				// debugmsg(SUBSYS_X11, "event", VERBOSITY_WARNING,
+				//    "mouse explicitly moved to screen center; ignored.");
 
-				/*  Which window in which machine in
-				    which emulation?  */
-				found = -1;
-				for (j=0; j<emul->n_machines; j++) {
-					struct machine *m2 = emul->machines[j];
-					for (i=0; i<m2->x11_md.
-					    n_fb_windows; i++)
-						if (m->x11_md.
-						    fb_windows[fb_nr]->
-						    x11_display == m2->
-						    x11_md.
-						    fb_windows[i]->
-						    x11_display &&
-						    event.xmotion.
-						    window == m2->
-						    x11_md.
-						    fb_windows[i]->
-						    x11_fb_window)
-							found = i;
+				mouseExplicityMoved = false;
+
+				mouseXofLastEvent = event.xmotion.x;
+				mouseYofLastEvent = event.xmotion.y;
+
+			} else if (event.type == MotionNotify) {
+				// debugmsg(SUBSYS_X11, "event", VERBOSITY_WARNING,
+				//    "mouse moved to %i, %i",
+				//    event.xmotion.x, event.xmotion.y);
+
+				int dx = event.xmotion.x - mouseXofLastEvent;
+				int dy = event.xmotion.y - mouseYofLastEvent;
+
+				mouseXofLastEvent = event.xmotion.x;
+				mouseYofLastEvent = event.xmotion.y;
+
+				if (grabbed == fbwin && (dx != 0 || dy != 0)) {
+					// debugmsg(SUBSYS_X11, "event", VERBOSITY_WARNING,
+					//    "mouse moved dx %i, %i", dx, dy);
+
+					dx *= fbwin->scaledown;
+					dy *= fbwin->scaledown;
+					console_mouse_coordinate_update(dx, dy, fb_nr);
+
+					// Hack for keeping the mouse pointer away
+					// from the edges of the screen.
+					Window xqpWindow;
+					int rootx, rooty, x, y;
+					unsigned int mask;
+					int res = XQueryPointer(fbwin->x11_display,
+					    RootWindow(fbwin->x11_display, DefaultScreen(fbwin->x11_display)),
+					    &xqpWindow,
+					    &xqpWindow, &rootx, &rooty, &x, &y,
+					    &mask);
+
+					Screen *screen = XDefaultScreenOfDisplay(fbwin->x11_display);
+
+					int w = XWidthOfScreen(screen);
+					int h = XHeightOfScreen(screen);
+					int x1 = w * 1 / 5;
+					int y1 = h * 1 / 5;
+					int x2 = w * 4 / 5;
+					int y2 = h * 4 / 5;
+
+					if (res == True && (rootx < x1 || rooty < y1
+					    || rootx >= x2 || rooty >= y2))
+						mouseMouseToCenterOfScreen(fbwin);
 				}
-				if (found < 0) {
-					printf("Internal error in x11.c.\n");
-					exit(1);
-				}
-				console_mouse_coordinates(event.xmotion.x *
-				    m->x11_md.fb_windows[found]->scaledown,
-				    event.xmotion.y * m->x11_md.fb_windows[
-				    found]->scaledown, found);
 			}
 
 			if (event.type == ButtonPress) {
-				debug("[ X11 ButtonPress: %i ]\n",
-				    event.xbutton.button);
+				/*  debug("[ X11 ButtonPress: %i ]\n",
+				    event.xbutton.button);  */
 				/*  button = 1,2,3 = left,middle,right  */
 
-				console_mouse_button(event.xbutton.button, 1);
+				// TODO: console_mouse_button with multiple machines?!
+				if (grabbed == fbwin)
+					console_mouse_button(event.xbutton.button, 1);
 			}
 
 			if (event.type == ButtonRelease) {
-				debug("[ X11 ButtonRelease: %i ]\n",
-				    event.xbutton.button);
+				/*  debug("[ X11 ButtonRelease: %i ]\n",
+				    event.xbutton.button);  */
 				/*  button = 1,2,3 = left,middle,right  */
 
-				console_mouse_button(event.xbutton.button, 0);
+				// TODO: console_mouse_button with multiple machines?!
+				if (grabbed == fbwin)
+					console_mouse_button(event.xbutton.button, 0);
+
+				grab(fbwin);
+			}
+
+			if (event.type==KeyRelease) {
+				XKeyPressedEvent *ke = &event.xkey;
+				int x = ke->keycode;
+				switch (x) {
+					case 37:left_ctrl = false;
+						break;
+					case 64:left_alt = false;
+						break;
+					//default:
+						// 37 = left CTRL, 64 = left ALT
+						//debugmsg(SUBSYS_X11, "event", VERBOSITY_DEBUG,
+						//    "RELEASE: unimplemented keycode %i", x);
+				}
 			}
 
 			if (event.type==KeyPress) {
@@ -816,9 +964,19 @@ static void x11_check_events_machine(struct emul *emul, struct machine *m)
 						console_makeavail(m->
 						    main_console_handle, '~');
 						break;
-					default:
-						debug("[ unimplemented X11 "
-						    "keycode %i ]\n", x);
+
+					case 37:left_ctrl = true;
+						if (left_ctrl && left_alt)
+							ungrab();
+						break;
+					case 64:left_alt = true;
+						if (left_ctrl && left_alt)
+							ungrab();
+						break;
+
+					//default:
+						//debugmsg(SUBSYS_X11, "event", VERBOSITY_DEBUG,
+						//    "unimplemented keycode %i", x);
 					}
 				}
 			}
