@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -58,21 +58,19 @@ static struct cpu_family *first_cpu_family = NULL;
 struct cpu *cpu_new(struct memory *mem, struct machine *machine,
         int cpu_id, char *name)
 {
-	struct cpu *cpu;
-	struct cpu_family *fp;
 	char *cpu_type_name;
 	char tmpstr[30];
 
 	if (name == NULL) {
-		fprintf(stderr, "cpu_new(): cpu name = NULL?\n");
-		exit(1);
+		debugmsg(SUBSYS_CPU, "", VERBOSITY_ERROR, "cpu name = NULL?");
+		return NULL;
 	}
 
 	CHECK_ALLOCATION(cpu_type_name = strdup(name));
 
-	cpu = zeroed_alloc(sizeof(struct cpu));
+	struct cpu *cpu = (struct cpu *) zeroed_alloc(sizeof(struct cpu));
 
-	CHECK_ALLOCATION(cpu->path = malloc(strlen(machine->path) + 15));
+	CHECK_ALLOCATION(cpu->path = (char *) malloc(strlen(machine->path) + 15));
 	snprintf(cpu->path, strlen(machine->path) + 15,
 	    "%s.cpu[%i]", machine->path, cpu_id);
 
@@ -97,35 +95,51 @@ struct cpu *cpu_new(struct memory *mem, struct machine *machine,
 
 	cpu_create_or_reset_tc(cpu);
 
-	fp = first_cpu_family;
+	struct cpu_family *fp = first_cpu_family;
 
 	while (fp != NULL) {
 		if (fp->cpu_new != NULL) {
-			if (fp->cpu_new(cpu, mem, machine, cpu_id,
-			    cpu_type_name)) {
-				/*  Sanity check:  */
-				if (cpu->memory_rw == NULL) {
-					fatal("\ncpu_new(): memory_rw == "
-					    "NULL\n");
-					exit(1);
-				}
+			if (fp->cpu_new(cpu, mem, machine, cpu_id, cpu_type_name))
 				break;
-			}
 		}
 
 		fp = fp->next;
 	}
 
 	if (fp == NULL) {
-		fatal("\ncpu_new(): unknown cpu type '%s'\n", cpu_type_name);
+		debugmsg(SUBSYS_CPU, "", VERBOSITY_ERROR,
+		    "unknown cpu type '%s'", cpu_type_name);
+		free(cpu);
+		return NULL;
+	}
+
+	cpu->cpu_family = fp;
+
+	if (cpu->memory_rw == NULL) {
+		debugmsg_cpu(cpu, SUBSYS_CPU, "", VERBOSITY_ERROR,
+		    "memory_rw == NULL");
+		free(cpu);
 		return NULL;
 	}
 
 	fp->init_tables(cpu);
 
 	if (cpu->byte_order == EMUL_UNDEFINED_ENDIAN) {
-		fatal("\ncpu_new(): Internal bug: Endianness not set.\n");
-		exit(1);
+		debugmsg_cpu(cpu, SUBSYS_CPU, "endianness", VERBOSITY_ERROR,
+		    "Internal bug: Endianness not set!");
+		free(cpu);
+		return NULL;
+	}
+	
+	if (cpu->vaddr_mask == 0) {
+		if (cpu->is_32bit)
+			cpu->vaddr_mask = 0x00000000ffffffffULL;
+		else
+			cpu->vaddr_mask = (int64_t)-1;
+
+		debugmsg_cpu(cpu, SUBSYS_CPU, "vaddr_mask", VERBOSITY_DEBUG,
+		    "Warning: vaddr_mask should be set in the CPU family's "
+		    "cpu_new()! Assuming 0x%16llx", (long long)cpu->vaddr_mask);
 	}
 
 	return cpu;
@@ -165,12 +179,12 @@ void cpu_destroy(struct cpu *cpu)
  *  If rawflag is nonzero, then the TLB contents isn't formated nicely,
  *  just dumped.
  */
-void cpu_tlbdump(struct machine *m, int x, int rawflag)
+void cpu_tlbdump(struct cpu *cpu, int rawflag)
 {
-	if (m->cpu_family == NULL || m->cpu_family->tlbdump == NULL)
+	if (cpu->cpu_family == NULL || cpu->cpu_family->tlbdump == NULL)
 		fatal("cpu_tlbdump(): NULL\n");
 	else
-		m->cpu_family->tlbdump(m, x, rawflag);
+		cpu->cpu_family->tlbdump(cpu, rawflag);
 }
 
 
@@ -183,11 +197,11 @@ void cpu_tlbdump(struct machine *m, int x, int rawflag)
 int cpu_disassemble_instr(struct machine *m, struct cpu *cpu,
 	unsigned char *instr, int running, uint64_t addr)
 {
-	if (m->cpu_family == NULL || m->cpu_family->disassemble_instr == NULL) {
+	if (cpu->cpu_family == NULL || cpu->cpu_family->disassemble_instr == NULL) {
 		fatal("cpu_disassemble_instr(): NULL\n");
 		return 0;
 	} else
-		return m->cpu_family->disassemble_instr(cpu, instr,
+		return cpu->cpu_family->disassemble_instr(cpu, instr,
 		    running, addr);
 }
 
@@ -203,10 +217,10 @@ int cpu_disassemble_instr(struct machine *m, struct cpu *cpu,
 void cpu_register_dump(struct machine *m, struct cpu *cpu,
 	int gprs, int coprocs)
 {
-	if (m->cpu_family == NULL || m->cpu_family->register_dump == NULL)
+	if (cpu->cpu_family == NULL || cpu->cpu_family->register_dump == NULL)
 		fatal("cpu_register_dump(): NULL\n");
 	else
-		m->cpu_family->register_dump(cpu, gprs, coprocs);
+		cpu->cpu_family->register_dump(cpu, gprs, coprocs);
 }
 
 
@@ -224,7 +238,7 @@ void cpu_functioncall_trace(struct cpu *cpu, uint64_t f)
 	uint64_t offset;
 
 	/*  Special hack for M88K userspace:  */
-	if (cpu->machine->arch == ARCH_M88K &&
+	if (cpu->cpu_family->arch == ARCH_M88K &&
 	    !(cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE))
 		show_symbolic_function_name = 0;
 
@@ -241,25 +255,20 @@ void cpu_functioncall_trace(struct cpu *cpu, uint64_t f)
 	fatal("<");
 	symbol = get_symbol_name_and_n_args(&cpu->machine->symbol_context,
 	    f, &offset, &n_args);
-	if (symbol != NULL && show_symbolic_function_name)
+	if (symbol != NULL && show_symbolic_function_name && offset == 0)
 		fatal("%s", symbol);
 	else {
 		if (cpu->is_32bit)
-			fatal("0x%"PRIx32, (uint32_t) f);
+			fatal("0x%" PRIx32, (uint32_t) f);
 		else
-			fatal("0x%"PRIx64, (uint64_t) f);
+			fatal("0x%" PRIx64, (uint64_t) f);
 	}
 	fatal("(");
 
-	if (cpu->machine->cpu_family->functioncall_trace != NULL)
-		cpu->machine->cpu_family->functioncall_trace(cpu, n_args);
+	if (cpu->cpu_family->functioncall_trace != NULL)
+		cpu->cpu_family->functioncall_trace(cpu, n_args);
 
 	fatal(")>\n");
-
-#ifdef PRINT_MEMORY_CHECKSUM
-	/*  Temporary hack for finding bugs:  */
-	fatal("call chksum=%016"PRIx64"\n", memory_checksum(cpu->mem));
-#endif
 }
 
 
@@ -291,7 +300,7 @@ void cpu_create_or_reset_tc(struct cpu *cpu)
 	size_t s = dyntrans_cache_size + DYNTRANS_CACHE_MARGIN;
 
 	if (cpu->translation_cache == NULL)
-		cpu->translation_cache = zeroed_alloc(s);
+		cpu->translation_cache = (unsigned char *) zeroed_alloc(s);
 
 	/*  Create an empty table at the beginning of the translation cache:  */
 	memset(cpu->translation_cache, 0, sizeof(uint32_t)
@@ -315,15 +324,27 @@ void cpu_create_or_reset_tc(struct cpu *cpu)
  *  Dumps info about a CPU using debug(). "cpu0: CPUNAME, running" (or similar)
  *  is outputed, and it is up to CPU dependent code to complete the line.
  */
-void cpu_dumpinfo(struct machine *m, struct cpu *cpu)
+void cpu_dumpinfo(struct machine *m, struct cpu *cpu, bool verbose)
 {
-	debug("cpu%i: %s, %s", cpu->cpu_id, cpu->name,
-	    cpu->running? "running" : "stopped");
+	char cpuname[100];
+	snprintf(cpuname, sizeof(cpuname), "cpu%i", cpu->cpu_id);
 
-	if (m->cpu_family == NULL || m->cpu_family->dumpinfo == NULL)
-		fatal("cpu_dumpinfo(): NULL\n");
-	else
-		m->cpu_family->dumpinfo(cpu);
+	if (verbose)
+		debugmsg(SUBSYS_MACHINE, cpuname, VERBOSITY_INFO,
+		    "%s", cpu->running? "running" : "stopped");
+
+	if (cpu->cpu_family == NULL || cpu->cpu_family->dumpinfo == NULL) {
+		debugmsg(SUBSYS_MACHINE, cpuname, VERBOSITY_ERROR,
+		    "cpu_dumpinfo(): NULL");
+	} else {
+		if (verbose)
+			debug_indentation(1);
+
+		cpu->cpu_family->dumpinfo(cpu, verbose);
+
+		if (verbose)
+			debug_indentation(-1);
+	}
 }
 
 
@@ -334,10 +355,7 @@ void cpu_dumpinfo(struct machine *m, struct cpu *cpu)
  */
 void cpu_list_available_types(void)
 {
-	struct cpu_family *fp;
-	int iadd = DEBUG_INDENTATION;
-
-	fp = first_cpu_family;
+	struct cpu_family *fp = first_cpu_family;
 
 	if (fp == NULL) {
 		debug("No CPUs defined!\n");
@@ -346,13 +364,13 @@ void cpu_list_available_types(void)
 
 	while (fp != NULL) {
 		debug("%s:\n", fp->name);
-		debug_indentation(iadd);
+		debug_indentation(1);
 		if (fp->list_available_types != NULL)
 			fp->list_available_types();
 		else
 			debug("(internal error: list_available_types"
 			    " = NULL)\n");
-		debug_indentation(-iadd);
+		debug_indentation(-1);
 
 		fp = fp->next;
 	}
@@ -384,19 +402,37 @@ void cpu_run_deinit(struct machine *machine)
 	}
 
 	if (machine->show_nr_of_instructions)
-		cpu_show_cycles(machine, 1);
+		cpu_show_cycles(machine, true);
 
 	fflush(stdout);
 }
 
 
+/*
+ *  cpu_print_pc_indicator_in_disassembly():
+ *
+ *  Helper which shows an arrow indicating the current instruction, during
+ *  disassembly.
+ */
+void cpu_print_pc_indicator_in_disassembly(struct cpu *cpu, int running, uint64_t dumpaddr)
+{
+	if (!running && cpu->pc == dumpaddr) {
+		color_pc_indicator();
+		debug(" <- ");
+		color_normal();
+	} else {
+		debug("    ");
+	}
+}
+
+
 /*
  *  cpu_show_cycles():
  *
  *  If show_nr_of_instructions is on, then print a line to stdout about how
  *  many instructions/cycles have been executed so far.
  */
-void cpu_show_cycles(struct machine *machine, int forced)
+void cpu_show_cycles(struct machine *machine, bool forced)
 {
 	uint64_t offset, pc;
 	char *symbol;
@@ -425,7 +461,7 @@ void cpu_show_cycles(struct machine *machine, int forced)
 	if (!machine->show_nr_of_instructions && !forced)
 		goto do_return;
 
-	printf("[ %"PRIi64" instrs", (int64_t) cpu->ninstrs);
+	printf("[ %" PRIi64" instrs", (int64_t) cpu->ninstrs);
 
 	/*  Instructions per second, and average so far:  */
 	is = 1000 * (ninstrs-ninstrs_last) / (mseconds-mseconds_last);
@@ -437,21 +473,21 @@ void cpu_show_cycles(struct machine *machine, int forced)
 
 	if (cpu->has_been_idling) {
 		printf("; idling");
-		cpu->has_been_idling = 0;
+		cpu->has_been_idling = false;
 	} else
-		printf("; i/s=%"PRIi64" avg=%"PRIi64, is, avg);
+		printf("; i/s=%" PRIi64" avg=%" PRIi64, is, avg);
 
 	symbol = get_symbol_name(&machine->symbol_context, pc, &offset);
 
 	if (machine->ncpus == 1) {
 		if (cpu->is_32bit)
-			printf("; pc=0x%08"PRIx32, (uint32_t) pc);
+			printf("; pc=0x%08" PRIx32, (uint32_t) pc);
 		else
-			printf("; pc=0x%016"PRIx64, (uint64_t) pc);
+			printf("; pc=0x%016" PRIx64, (uint64_t) pc);
 	}
 
 	/*  Special hack for M88K userland:  (Don't show symbols.)  */
-	if (cpu->machine->arch == ARCH_M88K &&
+	if (cpu->cpu_family->arch == ARCH_M88K &&
 	    !(cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE))
 		symbol = NULL;
 
@@ -500,29 +536,20 @@ void cpu_run_init(struct machine *machine)
  *  Allocates a cpu_family struct and calls an init function for the
  *  family to fill in reasonable data and pointers.
  */
-static void add_cpu_family(int (*family_init)(struct cpu_family *), int arch)
+static void add_cpu_family(void (*family_init)(struct cpu_family *), int arch)
 {
-	struct cpu_family *fp, *tmp;
-	int res;
+	struct cpu_family *fp;
 
-	CHECK_ALLOCATION(fp = malloc(sizeof(struct cpu_family)));
+	CHECK_ALLOCATION(fp = (struct cpu_family *) malloc(sizeof(struct cpu_family)));
 	memset(fp, 0, sizeof(struct cpu_family));
 
-	/*
-	 *  family_init() returns 1 if the struct has been filled with
-	 *  valid data, 0 if suppor for the cpu family isn't compiled
-	 *  into the emulator.
-	 */
-	res = family_init(fp);
-	if (!res) {
-		free(fp);
-		return;
-	}
+	family_init(fp);
+
 	fp->arch = arch;
 	fp->next = NULL;
 
 	/*  Add last in family chain:  */
-	tmp = first_cpu_family;
+	struct cpu_family* tmp = first_cpu_family;
 	if (tmp == NULL) {
 		first_cpu_family = fp;
 	} else {
@@ -533,28 +560,6 @@ static void add_cpu_family(int (*family_init)(struct cpu_family *), int arch)
 }
 
 
-/*
- *  cpu_family_ptr_by_number():
- *
- *  Returns a pointer to a CPU family based on the ARCH_* integers.
- */
-struct cpu_family *cpu_family_ptr_by_number(int arch)
-{
-	struct cpu_family *fp;
-	fp = first_cpu_family;
-
-	/*  YUCK! This is too hardcoded! TODO  */
-
-	while (fp != NULL) {
-		if (arch == fp->arch)
-			return fp;
-		fp = fp->next;
-	}
-
-	return NULL;
-}
-
-
 /*
  *  cpu_init():
  *
