@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -33,10 +33,11 @@
  *  on instruction formats etc.
  */
 
+#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <ctype.h>
+#include <unistd.h>
 
 #include "cpu.h"
 #include "interrupt.h"
@@ -52,7 +53,7 @@
 
 
 /*  Alpha symbolic register names:  */
-static char *alpha_regname[N_ALPHA_REGS] = ALPHA_REG_NAMES; 
+static const char *alpha_regname[N_ALPHA_REGS] = ALPHA_REG_NAMES; 
 
 void alpha_irq_interrupt_assert(struct interrupt *interrupt);
 void alpha_irq_interrupt_deassert(struct interrupt *interrupt);
@@ -93,11 +94,6 @@ int alpha_cpu_new(struct cpu *cpu, struct memory *mem,
 
 	cpu->cd.alpha.cpu_type = cpu_type_defs[i];
 
-	/*  Only show name and caches etc for CPU nr 0:  */
-	if (cpu_id == 0) {
-		debug("%s", cpu->name);
-	}
-
 	cpu->cd.alpha.r[ALPHA_SP] = 0xfffffc000000ff00ULL;
 
 	/*  Set up dummy kentry pointers to something which crashes
@@ -117,15 +113,15 @@ int alpha_cpu_new(struct cpu *cpu, struct memory *mem,
 
 	/*  Register the CPU interrupt pin:  */
 	{
-		struct interrupt template;
+		struct interrupt templ;
 
-		memset(&template, 0, sizeof(template));
-		template.line = 0;
-		template.name = cpu->path;
-		template.extra = cpu;
-		template.interrupt_assert = alpha_irq_interrupt_assert; 
-		template.interrupt_deassert = alpha_irq_interrupt_deassert; 
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 0;
+		templ.name = cpu->path;
+		templ.extra = cpu;
+		templ.interrupt_assert = alpha_irq_interrupt_assert; 
+		templ.interrupt_deassert = alpha_irq_interrupt_deassert; 
+		interrupt_handler_register(&templ);
 	}
 
 	return 1;
@@ -135,10 +131,9 @@ int alpha_cpu_new(struct cpu *cpu, struct memory *mem,
 /*
  *  alpha_cpu_dumpinfo():
  */
-void alpha_cpu_dumpinfo(struct cpu *cpu)
+void alpha_cpu_dumpinfo(struct cpu *cpu, bool verbose)
 {
-	/*  TODO  */
-	debug("\n");
+	debugmsg(SUBSYS_MACHINE, "cpu", VERBOSITY_INFO, "%s", cpu->name);
 }
 
 
@@ -181,16 +176,19 @@ void alpha_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 	if (gprs) {
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->pc, &offset);
-		debug("cpu%i:\t pc = 0x%016"PRIx64, x, (uint64_t) cpu->pc);
+		debug("cpu%i:\t pc = 0x%016" PRIx64, x, (uint64_t) cpu->pc);
 		debug("  <%s>\n", symbol != NULL? symbol : " no symbol ");
 		for (i=0; i<N_ALPHA_REGS; i++) {
 			int r = (i >> 1) + ((i & 1) << 4);
 			if ((i % 2) == 0)
 				debug("cpu%i:\t", x);
 			if (r != ALPHA_ZERO)
-				debug("%3s = 0x%016"PRIx64, alpha_regname[r],
+				debug("%3s = 0x%016" PRIx64, alpha_regname[r],
 				    (uint64_t) cpu->cd.alpha.r[r]);
-			debug((i % 2) == 1? "\n" : "   ");
+			if ((i % 2) == 1)
+				debug("\n");
+			else
+				debug("   ");
 		}
 	}
 }
@@ -205,7 +203,7 @@ void alpha_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
  *  If rawflag is nonzero, then the TLB contents isn't formated nicely,
  *  just dumped.
  */
-void alpha_cpu_tlbdump(struct machine *m, int x, int rawflag)
+void alpha_cpu_tlbdump(struct cpu* cpu, int rawflag)
 {
 }
 
@@ -266,7 +264,7 @@ int alpha_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	uint32_t iw;
 	uint64_t offset, tmp;
 	int opcode, ra, rb, func, rc, imm, floating, rbrc = 0, indir = 0;
-	char *symbol, *mnem = NULL;
+	const char *symbol, *mnem = NULL;
 	char palcode_name[30];
 
 	if (running)
@@ -280,10 +278,12 @@ int alpha_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	if (cpu->machine->ncpus > 1 && running)
 		debug("cpu%i:\t", cpu->cpu_id);
 
-	debug("%016"PRIx64":  ", (uint64_t) dumpaddr);
+	debug("%016" PRIx64":  ", (uint64_t) dumpaddr);
 
 	iw = ib[0] + (ib[1]<<8) + (ib[2]<<16) + (ib[3]<<24);
-	debug("%08x\t", (int)iw);
+	debug("%08x", (int)iw);
+
+	cpu_print_pc_indicator_in_disassembly(cpu, running, dumpaddr);
 
 	opcode = iw >> 26;
 	ra = (iw >> 21) & 31;
@@ -591,7 +591,7 @@ int alpha_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				debug("jsr");
 			debug("\t%s,", alpha_regname[ra]);
 			debug("(%s),", alpha_regname[rb]);
-			debug("0x%"PRIx64, (uint64_t) tmp);
+			debug("0x%" PRIx64, (uint64_t) tmp);
 			symbol = get_symbol_name(&cpu->machine->symbol_context,
 			    tmp, &offset);
 			if (symbol != NULL)
@@ -613,7 +613,7 @@ int alpha_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 		debug("%s\t", opcode==0x30? "br" : "bsr");
 		if (ra != ALPHA_ZERO)
 			debug("%s,", alpha_regname[ra]);
-		debug("0x%"PRIx64, (uint64_t) tmp);
+		debug("0x%" PRIx64, (uint64_t) tmp);
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    tmp, &offset);
 		if (symbol != NULL)
@@ -653,7 +653,7 @@ int alpha_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 			debug("f%i,", ra);
 		else
 			debug("%s,", alpha_regname[ra]);
-		debug("0x%"PRIx64, (uint64_t) tmp);
+		debug("0x%" PRIx64, (uint64_t) tmp);
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    tmp, &offset);
 		if (symbol != NULL)
