@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -476,8 +476,7 @@ X(cvttq_c)
  */
 X(cvtqt)
 {
-	reg(ic->arg[0]) = ieee_store_float_value(reg(ic->arg[2]),
-	    IEEE_FMT_D, 0);
+	reg(ic->arg[0]) = ieee_store_float_value(reg(ic->arg[2]), IEEE_FMT_D);
 }
 
 
@@ -507,54 +506,30 @@ X(fneg)
 X(addt)
 {
 	struct ieee_float_value fa, fb;
-	double res;
 	ieee_interpret_float_value(reg(ic->arg[1]), &fa, IEEE_FMT_D);
 	ieee_interpret_float_value(reg(ic->arg[2]), &fb, IEEE_FMT_D);
-	if (fa.nan | fb.nan)
-		res = 0.0;
-	else
-		res = fa.f + fb.f;
-	reg(ic->arg[0]) = ieee_store_float_value(res,
-	    IEEE_FMT_D, fa.nan | fb.nan);
+	reg(ic->arg[0]) = ieee_store_float_value(fa.f + fb.f, IEEE_FMT_D);
 }
 X(subt)
 {
 	struct ieee_float_value fa, fb;
-	double res;
 	ieee_interpret_float_value(reg(ic->arg[1]), &fa, IEEE_FMT_D);
 	ieee_interpret_float_value(reg(ic->arg[2]), &fb, IEEE_FMT_D);
-	if (fa.nan | fb.nan)
-		res = 0.0;
-	else
-		res = fa.f - fb.f;
-	reg(ic->arg[0]) = ieee_store_float_value(res,
-	    IEEE_FMT_D, fa.nan | fb.nan);
+	reg(ic->arg[0]) = ieee_store_float_value(fa.f - fb.f, IEEE_FMT_D);
 }
 X(mult)
 {
 	struct ieee_float_value fa, fb;
-	double res;
 	ieee_interpret_float_value(reg(ic->arg[1]), &fa, IEEE_FMT_D);
 	ieee_interpret_float_value(reg(ic->arg[2]), &fb, IEEE_FMT_D);
-	if (fa.nan | fb.nan)
-		res = 0.0;
-	else
-		res = fa.f * fb.f;
-	reg(ic->arg[0]) = ieee_store_float_value(res,
-	    IEEE_FMT_D, fa.nan | fb.nan);
+	reg(ic->arg[0]) = ieee_store_float_value(fa.f * fb.f, IEEE_FMT_D);
 }
 X(divt)
 {
 	struct ieee_float_value fa, fb;
-	double res;
 	ieee_interpret_float_value(reg(ic->arg[1]), &fa, IEEE_FMT_D);
 	ieee_interpret_float_value(reg(ic->arg[2]), &fb, IEEE_FMT_D);
-	if (fa.nan | fb.nan || fb.f == 0)
-		res = 0.0;
-	else
-		res = fa.f / fb.f;
-	reg(ic->arg[0]) = ieee_store_float_value(res,
-	    IEEE_FMT_D, fa.nan | fb.nan || fb.f == 0);
+	reg(ic->arg[0]) = ieee_store_float_value(fa.f / fb.f, IEEE_FMT_D);
 }
 X(cmpteq)
 {
@@ -606,11 +581,11 @@ X(implver)
 
 
 /*
- *  mull:  Signed Multiply 32x32 => 32.
+ *  mull, mull_imm:  Signed Multiply 32x32 => 32.
  *
  *  arg[0] = pointer to destination uint64_t
  *  arg[1] = pointer to source uint64_t
- *  arg[2] = pointer to source uint64_t
+ *  arg[2] = pointer to source uint64_t or immediate
  */
 X(mull)
 {
@@ -618,19 +593,29 @@ X(mull)
 	int32_t b = reg(ic->arg[2]);
 	reg(ic->arg[0]) = (int64_t)(int32_t)(a * b);
 }
+X(mull_imm)
+{
+	int32_t a = reg(ic->arg[1]);
+	int32_t b = ic->arg[2];
+	reg(ic->arg[0]) = (int64_t)(int32_t)(a * b);
+}
 
 
 /*
- *  mulq:  Unsigned Multiply 64x64 => 64.
+ *  mulq, mulq_imm:  Unsigned Multiply 64x64 => 64.
  *
  *  arg[0] = pointer to destination uint64_t
  *  arg[1] = pointer to source uint64_t
- *  arg[2] = pointer to source uint64_t
+ *  arg[2] = pointer to source uint64_t or immediate
  */
 X(mulq)
 {
 	reg(ic->arg[0]) = reg(ic->arg[1]) * reg(ic->arg[2]);
 }
+X(mulq_imm)
+{
+	reg(ic->arg[0]) = reg(ic->arg[1]) * ic->arg[2];
+}
 
 
 /*
@@ -932,6 +917,7 @@ X(to_be_translated)
 		case 0x3b: ic->f = instr(s8subq); break;
 		case 0x3d: ic->f = instr(cmpule); break;
 		case 0x4d: ic->f = instr(cmplt); break;
+		// case 0x69: ic->f = instr(subq_v); break;
 		case 0x6d: ic->f = instr(cmple); break;
 
 		case 0x80: ic->f = instr(addl_imm); break;
@@ -1093,10 +1079,14 @@ X(to_be_translated)
 			ic->arg[2] = (size_t)((rb << 3) + (func >> 8));
 		else
 			ic->arg[2] = (size_t) &cpu->cd.alpha.r[rb];
+		// TODO: mulq/v etc? overflow detection
+		// bit 0..6 are function, but 7 is "imm" bit.
 		switch (func & 0xff) {
 		case 0x00: ic->f = instr(mull); break;
 		case 0x20: ic->f = instr(mulq); break;
 		case 0x30: ic->f = instr(umulh); break;
+		case 0x80: ic->f = instr(mull_imm); break;
+		case 0xa0: ic->f = instr(mulq_imm); break;
 		default:if (!cpu->translation_readahead)
 				fatal("[ Alpha: unimplemented function 0x%03x "
 				    "for opcode 0x%02x ]\n", func, opcode);
@@ -1229,6 +1219,7 @@ X(to_be_translated)
 			break;
 		case 0x31:
 			fp = 1;
+			// fall through
 		case 0x39:
 			ic->f = instr(beq);
 			samepage_function = instr(beq_samepage);
@@ -1247,6 +1238,7 @@ X(to_be_translated)
 			break;
 		case 0x35:
 			fp = 1;
+			// fall through
 		case 0x3d:
 			ic->f = instr(bne);
 			samepage_function = instr(bne_samepage);
