@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -194,7 +194,7 @@ uint8_t condition_gt[16] = { 1,0,1,0, 0,0,0,0, 0,1,0,1, 0,0,0,0 };
 		arm_instr_ ## n ## __hi, arm_instr_ ## n ## __ls,	\
 		arm_instr_ ## n ## __ge, arm_instr_ ## n ## __lt,	\
 		arm_instr_ ## n ## __gt, arm_instr_ ## n ## __le,	\
-		arm_instr_ ## n , arm_instr_nop };
+		arm_instr_ ## n , arm_instr_never };
 
 #define cond_instr(n)	( arm_cond_instr_ ## n  [condition_code] )
 
@@ -214,14 +214,32 @@ X(invalid) {
 	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
 
 	fatal("FATAL ERROR: An internal error occured in the ARM"
-	    " dyntrans code. Please contact the author with detailed"
-	    " repro steps on how to trigger this bug. pc = 0x%08"PRIx32"\n",
+	    " dyntrans code. This could be due to an unimplemented instruction"
+	    " encoding. pc = 0x%08" PRIx32"\n",
 	    (uint32_t)cpu->pc);
 
 	cpu->cd.arm.next_ic = &nothing_call;
 }
 
 
+/*
+ *  never:  So far unimplemented "never" instructions end up here.
+ *  (Those are the ones using the "0xf" condition prefix.)
+ */
+X(never) {
+	uint32_t low_pc;
+	low_pc = ((size_t)ic - (size_t)
+	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1)
+	    << ARM_INSTR_ALIGNMENT_SHIFT);
+	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+
+	fatal("[ ARM: unimplemented 0xf instruction at pc = 0x%08" PRIx32" ]\n", (uint32_t)cpu->pc);
+
+	cpu->cd.arm.next_ic = &nothing_call;
+}
+
+
 /*
  *  nop:  Do nothing.
  */
@@ -240,7 +258,7 @@ X(b)
 	cpu->pc = (uint32_t)((cpu->pc & 0xfffff000) + (int32_t)ic->arg[0]);
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(b)
 
@@ -334,15 +352,26 @@ void (*arm_cond_instr_b_samepage[16])(struct cpu *,
  */
 X(bx)
 {
+	uint32_t old_cpsr = cpu->cd.arm.cpsr;
 	cpu->pc = reg(ic->arg[0]);
-	if (cpu->pc & 1) {
-		fatal("thumb: TODO\n");
-		exit(1);
+	if (cpu->pc & 1)
+		cpu->cd.arm.cpsr |= ARM_FLAG_T;
+	else
+		cpu->cd.arm.cpsr &= ~ARM_FLAG_T;
+
+	if (cpu->cd.arm.cpsr != old_cpsr)
+		cpu->cd.arm.next_ic = &nothing_call;
+
+	if (cpu->pc & 2 && ((cpu->pc & 1) == 0)) {
+		fatal("[ ARM pc misaligned? 0x%08x ]\n", (int)cpu->pc);
+		cpu->running = 0;
+		cpu->n_translated_instrs --;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
 	}
-	cpu->pc &= ~3;
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(bx)
 
@@ -354,17 +383,28 @@ Y(bx)
  */
 X(bx_trace)
 {
+	uint32_t old_cpsr = cpu->cd.arm.cpsr;
 	cpu->pc = cpu->cd.arm.r[ARM_LR];
-	if (cpu->pc & 1) {
-		fatal("thumb: TODO\n");
-		exit(1);
+	if (cpu->pc & 1)
+		cpu->cd.arm.cpsr |= ARM_FLAG_T;
+	else
+		cpu->cd.arm.cpsr &= ~ARM_FLAG_T;
+
+	if (cpu->cd.arm.cpsr != old_cpsr)
+		cpu->cd.arm.next_ic = &nothing_call;
+
+	if (cpu->pc & 2 && ((cpu->pc & 1) == 0)) {
+		fatal("[ ARM pc misaligned? 0x%08x ]\n", (int)cpu->pc);
+		cpu->running = 0;
+		cpu->n_translated_instrs --;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
 	}
-	cpu->pc &= ~3;
 
 	cpu_functioncall_trace_return(cpu);
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(bx_trace)
 
@@ -373,6 +413,7 @@ Y(bx_trace)
  *  bl:  Branch and Link (to a different translated page)
  *
  *  arg[0] = relative address
+ *  arg[1] = offset within current page to the instruction
  */
 X(bl)
 {
@@ -383,31 +424,88 @@ X(bl)
 	cpu->pc = pc + (int32_t)ic->arg[0];
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(bl)
 
 
 /*
- *  blx:  Branch and Link, potentially exchanging Thumb/ARM encoding
+ *  blx_imm:  Branch and Link, always switching to THUMB encoding
+ *
+ *  arg[0] = relative address
+ *  arg[1] = offset within current page to the current instruction
+ */
+X(blx_imm)
+{
+	uint32_t pc = ((uint32_t)cpu->pc & 0xfffff000) + (int32_t)ic->arg[1];
+	uint32_t old_cpsr = cpu->cd.arm.cpsr;
+	cpu->cd.arm.r[ARM_LR] = pc + 4;
+
+	/*  Calculate new PC from this instruction + arg[0]  */
+	cpu->pc = pc + (int32_t)ic->arg[0];
+
+	if (cpu->pc & 1)
+		cpu->cd.arm.cpsr |= ARM_FLAG_T;
+	else {
+		fatal("[ blx_imm internal error. Should have switched to THUMB! 0x%08x ]\n", (int)cpu->pc);
+		cpu->running = 0;
+		cpu->n_translated_instrs --;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
+	}
+
+	if (cpu->cd.arm.cpsr != old_cpsr)
+		cpu->cd.arm.next_ic = &nothing_call;
+
+	if (cpu->pc & 2 && ((cpu->pc & 1) == 0)) {
+		fatal("[ ARM pc misaligned? 0x%08x ]\n", (int)cpu->pc);
+		cpu->running = 0;
+		cpu->n_translated_instrs --;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
+	}
+
+	if (cpu->machine->show_trace_tree)
+		cpu_functioncall_trace(cpu, cpu->pc);
+
+	/*  Find the new physical page and update the translation pointers:  */
+	quick_pc_to_pointers_arm(cpu);
+}
+
+
+/*
+ *  blx_reg:  Branch and Link, potentially exchanging Thumb/ARM encoding
  *
  *  arg[0] = ptr to rm
+ *  arg[2] = offset within current page to the instruction to return to
  */
-X(blx)
+X(blx_reg)
 {
 	uint32_t lr = ((uint32_t)cpu->pc & 0xfffff000) + (int32_t)ic->arg[2];
 	cpu->cd.arm.r[ARM_LR] = lr;
 	cpu->pc = reg(ic->arg[0]);
-	if (cpu->pc & 1) {
-		fatal("thumb: TODO\n");
-		exit(1);
+
+	uint32_t old_cpsr = cpu->cd.arm.cpsr;
+	if (cpu->pc & 1)
+		cpu->cd.arm.cpsr |= ARM_FLAG_T;
+	else
+		cpu->cd.arm.cpsr &= ~ARM_FLAG_T;
+
+	if (cpu->cd.arm.cpsr != old_cpsr)
+		cpu->cd.arm.next_ic = &nothing_call;
+
+	if (cpu->pc & 2 && ((cpu->pc & 1) == 0)) {
+		fatal("[ ARM pc misaligned? 0x%08x ]\n", (int)cpu->pc);
+		cpu->running = 0;
+		cpu->n_translated_instrs --;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
 	}
-	cpu->pc &= ~3;
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
-Y(blx)
+Y(blx_reg)
 
 
 /*
@@ -426,7 +524,7 @@ X(bl_trace)
 	cpu_functioncall_trace(cpu, cpu->pc);
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(bl_trace)
 
@@ -561,6 +659,26 @@ X(mlas)
 Y(mlas)
 
 
+/*
+ *  mla: Multiplication with subtraction
+ *
+ *  arg[0] = copy of instruction word
+ */
+X(mls)
+{
+	/*  xxxx0000 0110dddd aaaammmm 1001nnnn  mls Rd,Rn,Rm,Ra  */
+	uint32_t iw = ic->arg[0];
+	int rd = (iw >> 16) & 15;
+	int rn = (iw >>  0) & 15;
+	int rm = (iw >>  8) & 15;
+	int ra = (iw >> 12) & 15;
+
+	cpu->cd.arm.r[rd] = cpu->cd.arm.r[ra] -
+	    cpu->cd.arm.r[rn] * cpu->cd.arm.r[rm];
+}
+Y(mls)
+
+
 /*
  *  mull: Long multiplication
  *
@@ -679,14 +797,14 @@ X(ret_trace)
 		    ((cpu->pc & mask_within_page) >> ARM_INSTR_ALIGNMENT_SHIFT);
 	} else {
 		/*  Find the new physical page and update pointers:  */
-		quick_pc_to_pointers(cpu);
+		quick_pc_to_pointers_arm(cpu);
 	}
 }
 Y(ret_trace)
 X(ret)
 {
 	cpu->pc = cpu->cd.arm.r[ARM_LR];
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 Y(ret)
 
@@ -704,6 +822,11 @@ Y(ret)
 X(msr_imm)
 {
 	uint32_t mask = ic->arg[1];
+
+	if ((cpu->cd.arm.cpsr & ARM_FLAG_MODE) == ARM_MODE_USR32) {
+		mask &= 0xff000000;
+	}
+
 	int switch_register_banks = (mask & ARM_FLAG_MODE) &&
 	    ((cpu->cd.arm.cpsr & ARM_FLAG_MODE) !=
 	    (ic->arg[0] & ARM_FLAG_MODE));
@@ -757,16 +880,17 @@ X(msr_imm_spsr)
 		break;
 	default:fatal("msr_spsr: unimplemented mode %i\n",
 		    cpu->cd.arm.cpsr & ARM_FLAG_MODE);
-{
-	/*  Synchronize the program counter:  */
-	uint32_t old_pc, low_pc = ((size_t)ic - (size_t)
-	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
-	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
-	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
-	old_pc = cpu->pc;
-	printf("msr_spsr: old pc = 0x%08"PRIx32"\n", old_pc);
-}
-		exit(1);
+		{
+			/*  Synchronize the program counter:  */
+			uint32_t old_pc, low_pc = ((size_t)ic - (size_t)
+			    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+			cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
+			cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+			old_pc = cpu->pc;
+			printf("msr_spsr: old pc = 0x%08" PRIx32"\n", old_pc);
+		}
+		cpu->running = 0;
+		cpu->cd.arm.next_ic = &nothing_call;
 	}
 }
 Y(msr_imm_spsr)
@@ -807,9 +931,9 @@ X(mrs_spsr)
 	case ARM_MODE_SVC32: reg(ic->arg[0]) = cpu->cd.arm.spsr_svc; break;
 	case ARM_MODE_USR32:
 	case ARM_MODE_SYS32: reg(ic->arg[0]) = 0; break;
-	default:fatal("mrs_spsr: unimplemented mode %i\n",
-		    cpu->cd.arm.cpsr & ARM_FLAG_MODE);
-		exit(1);
+	default:fatal("mrs_spsr: unimplemented mode %i\n", cpu->cd.arm.cpsr & ARM_FLAG_MODE);
+		cpu->running = 0;
+		cpu->cd.arm.next_ic = &nothing_call;
 	}
 }
 Y(mrs_spsr)
@@ -849,7 +973,7 @@ X(openfirmware)
 	cpu->pc = cpu->cd.arm.r[ARM_LR];
 	if (cpu->machine->show_trace_tree)
 		cpu_functioncall_trace_return(cpu);
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 }
 
 
@@ -864,35 +988,6 @@ X(reboot)
 }
 
 
-/*
- *  swi_useremul: Syscall.
- *
- *  arg[0] = swi number
- */
-X(swi_useremul)
-{
-	/*  Synchronize the program counter:  */
-	uint32_t old_pc, low_pc = ((size_t)ic - (size_t)
-	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
-	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1)
-	    << ARM_INSTR_ALIGNMENT_SHIFT);
-	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
-	old_pc = cpu->pc;
-
-	useremul_syscall(cpu, ic->arg[0]);
-
-	if (!cpu->running) {
-		cpu->n_translated_instrs --;
-		cpu->cd.arm.next_ic = &nothing_call;
-	} else if (cpu->pc != old_pc) {
-		/*  PC was changed by the SWI call. Find the new physical
-		    page and update the translation pointers:  */
-		quick_pc_to_pointers(cpu);
-	}
-}
-Y(swi_useremul)
-
-
 /*
  *  swi:  Software interrupt.
  */
@@ -932,6 +1027,222 @@ X(und)
 Y(und)
 
 
+/*
+ *  movt:  Move Top.
+ *
+ *  arg[1] = 32-bit immediate value. Top 16 bits are those of interest.
+ *  arg[2] = ptr to rd
+ */
+X(movt)
+{
+	reg(ic->arg[2]) &= 0xffff;
+	reg(ic->arg[2]) |= ic->arg[1];
+}
+Y(movt)
+
+
+/*
+ *  movw:  Move (Word).
+ *
+ *  arg[1] = 32-bit immediate value.
+ *  arg[2] = ptr to rd
+ */
+X(movw)
+{
+	reg(ic->arg[2]) = ic->arg[1];
+}
+Y(movw)
+
+
+/*
+ *  rev:  Reverse endian.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rm
+ */
+X(rev)
+{
+	uint32_t v = reg(ic->arg[1]);
+
+	reg(ic->arg[0]) = (v >> 24) | ((v & 0x00ff0000) >> 8)
+		| ((v & 0x0000ff00) << 8) | ((v & 0xff) << 24);
+}
+Y(rev)
+
+
+/*
+ *  uxtb:  Unsigned Extend Byte.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rm
+ *  arg[2] = rotation amount (shift works too)
+ */
+X(uxtb)
+{
+	uint32_t x = reg(ic->arg[1]);
+	reg(ic->arg[0]) = (uint8_t)(x >> ic->arg[2]);
+}
+Y(uxtb)
+
+
+/*
+ *  sxth:  Signed Extend Halfword.
+ *
+ *  TODO: Could be optimized so that the rotation amount is "inlined".
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rm
+ *  arg[2] = rotation amount
+ */
+X(sxth)
+{
+	uint32_t x = reg(ic->arg[1]);
+
+	switch (ic->arg[2]) {
+	case 0:	break;
+	case 8: x >>= 8; break;
+	case 16: x >>= 16; break;
+	case 24: x = (x >> 24) | ((x & 255) << 8); break;
+	}
+
+	int16_t rotated = x;
+	reg(ic->arg[0]) = (int32_t)rotated;
+}
+Y(sxth)
+
+
+/*
+ *  uxth:  Unsigned Extend Halfword.
+ *
+ *  TODO: Could be optimized so that the rotation amount is "inlined".
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rm
+ *  arg[2] = rotation amount
+ */
+X(uxth)
+{
+	uint32_t x = reg(ic->arg[1]);
+
+	switch (ic->arg[2]) {
+	case 0:	break;
+	case 8: x >>= 8; break;
+	case 16: x >>= 16; break;
+	case 24: x = (x >> 24) | ((x & 255) << 8); break;
+	}
+
+	uint16_t rotated = x;
+	reg(ic->arg[0]) = (uint32_t)rotated;
+}
+Y(uxth)
+
+
+/*
+ *  uxtah:  Unsigned Extend and Add Byte.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = ptr to rm
+ */
+X(uxtab)
+{
+	reg(ic->arg[0]) = reg(ic->arg[1]) + (uint8_t)reg(ic->arg[2]);
+}
+Y(uxtab)
+
+
+/*
+ *  uxtah:  Unsigned Extend and Add Halfword.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = ptr to rm
+ */
+X(uxtah)
+{
+	reg(ic->arg[0]) = reg(ic->arg[1]) + (uint16_t)reg(ic->arg[2]);
+}
+Y(uxtah)
+
+
+/*
+ *  ubfx:  Unsigned Bit-Field Extract.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = (width << 16) + lsb
+ */
+X(ubfx)
+{
+	uint32_t x = reg(ic->arg[1]);
+
+	int lsb = (uint8_t)ic->arg[2];
+	int width = ic->arg[2] >> 16;
+
+	uint32_t mask = (1 << width) - 1;
+
+	x >>= lsb;
+	x &= mask;
+
+	reg(ic->arg[0]) = x;
+}
+Y(ubfx)
+
+
+/*
+ *  sbfx:  Signed Bit-Field Extract.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = (width << 16) + lsb
+ */
+X(sbfx)
+{
+	uint32_t x = reg(ic->arg[1]);
+
+	int lsb = (uint8_t)ic->arg[2];
+	int width = ic->arg[2] >> 16;
+
+	uint32_t mask = (1 << width) - 1;
+	x >>= lsb;
+	x &= mask;
+
+	uint32_t topBitMask = 1 << (width-1);
+	if (x & topBitMask && width < 32)
+		x |= ~mask;
+
+	reg(ic->arg[0]) = x;
+}
+Y(sbfx)
+
+
+/*
+ *  bfi:  Bit-Field Insert.
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = (msb << 16) + lsb
+ */
+X(bfi)
+{
+	uint32_t x = reg(ic->arg[1]);
+
+	int lsb = (uint8_t)ic->arg[2];
+	int msb = ic->arg[2] >> 16;
+	int width = msb - lsb + 1;
+
+	x <<= lsb;
+
+	uint32_t mask = (1 << width) - 1;
+
+	mask <<= lsb;
+
+	reg(ic->arg[0]) &= ~mask;
+	reg(ic->arg[0]) |= (x & mask);
+}
+Y(bfi)
+
+
 /*
  *  swp, swpb:  Swap (word or byte).
  *
@@ -1017,7 +1328,7 @@ extern void (*arm_load_store_instr_3_pc[2048])(struct cpu *,
 	struct arm_instr_call *);
 
 extern uint32_t (*arm_r[8192])(struct cpu *, struct arm_instr_call *);
-extern void arm_r_r3_t0_c0(void);
+extern uint32_t arm_r_r3_t0_c0(struct cpu *cpu, struct arm_instr_call *ic);
 
 extern void (*arm_dpi_instr[2 * 2 * 2 * 16 * 16])(struct cpu *,
 	struct arm_instr_call *);
@@ -1037,36 +1348,16 @@ X(cmps_regshort);
 
 
 /*
- *  bdt_load:  Block Data Transfer, Load
- *
- *  arg[0] = pointer to uint32_t in host memory, pointing to the base register
- *  arg[1] = 32-bit instruction word. Most bits are read from this.
+ *  Shared between regular ARM and the THUMB encoded 'pop'.
  */
-X(bdt_load)
+void arm_pop(struct cpu* cpu, uint32_t* np, int p_bit, int u_bit, int s_bit, int w_bit, uint32_t iw)
 {
+	uint32_t addr = *np;
 	unsigned char data[4];
-	uint32_t *np = (uint32_t *)ic->arg[0];
-	uint32_t addr = *np, low_pc;
 	unsigned char *page;
-	uint32_t iw = ic->arg[1];  /*  xxxx100P USWLnnnn llllllll llllllll  */
-	int p_bit = iw & 0x01000000;
-	int u_bit = iw & 0x00800000;
-	int s_bit = iw & 0x00400000;
-	int w_bit = iw & 0x00200000;
 	int i, return_flag = 0;
 	uint32_t new_values[16];
 
-#ifdef GATHER_BDT_STATISTICS
-	if (!s_bit)
-		update_bdt_statistics(iw);
-#endif
-
-	/*  Synchronize the program counter:  */
-	low_pc = ((size_t)ic - (size_t)
-	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
-	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
-	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
-
 	if (s_bit) {
 		/*  Load to USR registers:  */
 		if ((cpu->cd.arm.cpsr & ARM_FLAG_MODE) == ARM_MODE_USR32) {
@@ -1191,7 +1482,9 @@ X(bdt_load)
 			new_cpsr = cpu->cd.arm.spsr_svc; break;
 		default:fatal("bdt_load: unimplemented mode %i\n",
 			    cpu->cd.arm.cpsr & ARM_FLAG_MODE);
-			exit(1);
+			cpu->running = 0;
+			cpu->cd.arm.next_ic = &nothing_call;
+			return;
 		}
 
 		switch_register_banks = (cpu->cd.arm.cpsr & ARM_FLAG_MODE) !=
@@ -1217,44 +1510,23 @@ X(bdt_load)
 		    same page!  */
 		/*  Find the new physical page and update the
 		    translation pointers:  */
-		quick_pc_to_pointers(cpu);
+		quick_pc_to_pointers_arm(cpu);
 	}
 }
-Y(bdt_load)
 
 
 /*
- *  bdt_store:  Block Data Transfer, Store
- *
- *  arg[0] = pointer to uint32_t in host memory, pointing to the base register
- *  arg[1] = 32-bit instruction word. Most bits are read from this.
+ *  Shared between regular ARM and the THUMB encoded 'push'.
  */
-X(bdt_store)
+void arm_push(struct cpu* cpu, uint32_t* np, int p_bit, int u_bit, int s_bit, int w_bit, uint16_t regs)
 {
-	unsigned char data[4];
-	uint32_t *np = (uint32_t *)ic->arg[0];
-	uint32_t low_pc, value, addr = *np;
-	uint32_t iw = ic->arg[1];  /*  xxxx100P USWLnnnn llllllll llllllll  */
-	unsigned char *page;
-	int p_bit = iw & 0x01000000;
-	int u_bit = iw & 0x00800000;
-	int s_bit = iw & 0x00400000;
-	int w_bit = iw & 0x00200000;
 	int i;
-
-#ifdef GATHER_BDT_STATISTICS
-	if (!s_bit)
-		update_bdt_statistics(iw);
-#endif
-
-	/*  Synchronize the program counter:  */
-	low_pc = ((size_t)ic - (size_t)
-	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
-	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
-	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+	uint32_t value, addr = *np;
+	unsigned char data[4];
+	unsigned char *page;
 
 	for (i=(u_bit? 0 : 15); i>=0 && i<=15; i+=(u_bit? 1 : -1)) {
-		if (!((iw >> i) & 1)) {
+		if (!((regs >> i) & 1)) {
 			/*  Skip register i:  */
 			continue;
 		}
@@ -1336,9 +1608,193 @@ X(bdt_store)
 	if (w_bit)
 		*np = addr;
 }
+
+
+/*
+ *  bdt_load:  Block Data Transfer, Load
+ *
+ *  arg[0] = pointer to uint32_t in host memory, pointing to the base register
+ *  arg[1] = 32-bit instruction word. Most bits are read from this.
+ */
+X(bdt_load)
+{
+	uint32_t *np = (uint32_t *)ic->arg[0];
+	uint32_t low_pc;
+	uint32_t iw = ic->arg[1];  /*  xxxx100P USWLnnnn llllllll llllllll  */
+	int p_bit = iw & 0x01000000;
+	int u_bit = iw & 0x00800000;
+	int s_bit = iw & 0x00400000;
+	int w_bit = iw & 0x00200000;
+
+#ifdef GATHER_BDT_STATISTICS
+	if (!s_bit)
+		update_bdt_statistics(iw);
+#endif
+
+	/*  Synchronize the program counter:  */
+	low_pc = ((size_t)ic - (size_t)
+	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
+	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+
+	arm_pop(cpu, np, p_bit, u_bit, s_bit, w_bit, (uint16_t)iw);
+}
+Y(bdt_load)
+
+
+/*
+ *  bdt_store:  Block Data Transfer, Store
+ *
+ *  arg[0] = pointer to uint32_t in host memory, pointing to the base register
+ *  arg[1] = 32-bit instruction word. Most bits are read from this.
+ */
+X(bdt_store)
+{
+	uint32_t *np = (uint32_t *)ic->arg[0];
+	uint32_t low_pc;
+	uint32_t iw = ic->arg[1];  /*  xxxx100P USWLnnnn llllllll llllllll  */
+	int p_bit = iw & 0x01000000;
+	int u_bit = iw & 0x00800000;
+	int s_bit = iw & 0x00400000;
+	int w_bit = iw & 0x00200000;
+
+#ifdef GATHER_BDT_STATISTICS
+	if (!s_bit)
+		update_bdt_statistics(iw);
+#endif
+
+	/*  Synchronize the program counter:  */
+	low_pc = ((size_t)ic - (size_t)
+	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
+	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+
+	arm_push(cpu, np, p_bit, u_bit, s_bit, w_bit, (uint16_t)iw);
+}
 Y(bdt_store)
 
 
+/*
+ *  Load Register Exclusive (ARM "load linked"):
+ *
+ *  A Load Register Exclusive instruction initiates a RMW (read-modify-write)
+ *  sequence.
+ *
+ *  A Store Register Exclusive instruction ends the sequence.
+ *
+ *  arg[0] = ptr to rt
+ *  arg[1] = ptr to rn
+ *  arg[2] = int32_t imm offset
+ */
+X(ldrex)
+{
+	uint32_t addr = reg(ic->arg[1]) + (int32_t)ic->arg[2];
+	int low_pc;
+	uint8_t word[sizeof(uint32_t)];
+
+	/*  Synchronize the program counter:  */
+	low_pc = ((size_t)ic - (size_t)
+	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
+	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+
+	if (addr & (sizeof(word)-1)) {
+		fatal("TODO: ldrex unaligned access: exception\n");
+		cpu->running = 0;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
+	}
+
+	if (!cpu->memory_rw(cpu, cpu->mem, addr, word,
+	    sizeof(word), MEM_READ, CACHE_DATA)) {
+		/*  An exception occurred.  */
+		return;
+	}
+
+	cpu->cd.arm.rmw = 1;
+	cpu->cd.arm.rmw_addr = addr;
+	cpu->cd.arm.rmw_len = sizeof(word);
+
+	if (cpu->byte_order == EMUL_LITTLE_ENDIAN)
+		reg(ic->arg[0]) = word[0] + (word[1] << 8)
+		    + (word[2] << 16) + (word[3] << 24);
+	else
+		reg(ic->arg[0]) = word[3] + (word[2] << 8)
+		    + (word[1] << 16) + (word[0] << 24);
+}
+Y(ldrex)
+/*
+ *  Store Register Exclusive
+ *
+ *  arg[0] = ptr to rd
+ *  arg[1] = ptr to rn
+ *  arg[2] = ptr to rt
+ */
+X(strex)
+{
+	uint32_t addr = reg(ic->arg[1]);
+	uint64_t r = reg(ic->arg[2]);
+	int low_pc, i;
+	uint8_t word[sizeof(uint32_t)];
+	
+	/*  Synchronize the program counter:  */
+	low_pc = ((size_t)ic - (size_t)
+	    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
+	cpu->pc &= ~((ARM_IC_ENTRIES_PER_PAGE-1) << ARM_INSTR_ALIGNMENT_SHIFT);
+	cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
+
+	if (addr & (sizeof(word)-1)) {
+		fatal("TODO: strex unaligned access: exception\n");
+		cpu->running = 0;
+		cpu->cd.arm.next_ic = &nothing_call;
+		return;
+	}
+
+	if (cpu->byte_order == EMUL_LITTLE_ENDIAN) {
+		word[0]=r; word[1]=r>>8; word[2]=r>>16; word[3]=r>>24;
+	} else {
+		word[3]=r; word[2]=r>>8; word[1]=r>>16; word[0]=r>>24;
+	}
+
+	/*  If rmw is 0, then the store failed.  (This cache-line was written
+	    to by someone else.)  */
+	if (cpu->cd.arm.rmw == 0 || cpu->cd.arm.rmw_addr != addr
+	    || cpu->cd.arm.rmw_len != sizeof(word)) {
+		reg(ic->arg[0]) = 1;	// 1 = fail.
+		cpu->cd.arm.rmw = 0;
+		return;
+	}
+
+	if (!cpu->memory_rw(cpu, cpu->mem, addr, word,
+	    sizeof(word), MEM_WRITE, CACHE_DATA)) {
+		/*  An exception occurred.  */
+		return;
+	}
+
+	/*  We succeeded. Let's invalidate everybody else's store to this
+	    cache line:  */
+	for (i=0; i<cpu->machine->ncpus; i++) {
+		if (cpu->machine->cpus[i]->cd.arm.rmw) {
+			uint64_t yaddr = addr, xaddr = cpu->machine->cpus[i]->
+			    cd.arm.rmw_addr;
+
+			/*  8-2048 bytes, implementation dependent :-(  */			
+			/*  https://stackoverflow.com/questions/11383125/do-the-arm-instructions-ldrex-strex-have-to-operate-on-cache-aligned-data  */
+			uint64_t mask = 2047;
+
+			xaddr &= mask;
+			yaddr &= mask;
+			if (xaddr == yaddr)
+				cpu->machine->cpus[i]->cd.arm.rmw = 0;
+		}
+	}
+
+	reg(ic->arg[0]) = 0;	// 0 = success
+	cpu->cd.arm.rmw = 0;
+}
+Y(strex)
+
+
 /*  Various load/store multiple instructions:  */
 extern uint32_t *multi_opcode[256];
 extern void (**multi_opcode_f[256])(struct cpu *, struct arm_instr_call *);
@@ -1580,8 +2036,6 @@ X(netbsd_idle)
 	}
 
 	if (rZ == 0) {
-		static int x = 0;
-
 		/*  Synch the program counter.  */
 		uint32_t low_pc = ((size_t)ic - (size_t)
 		    cpu->cd.arm.cur_ic_page) / sizeof(struct arm_instr_call);
@@ -1589,14 +2043,8 @@ X(netbsd_idle)
 		    << ARM_INSTR_ALIGNMENT_SHIFT);
 		cpu->pc += (low_pc << ARM_INSTR_ALIGNMENT_SHIFT);
 
-		/*  Quasi-idle for a while:  */
-		cpu->has_been_idling = 1;
-	        if (cpu->machine->ncpus == 1 && (++x) == 100) {
-			usleep(50);
-			x = 0;
-		}
-
-		cpu->n_translated_instrs += N_SAFE_DYNTRANS_LIMIT / 6;
+		cpu->wants_to_idle = true;
+		cpu->n_translated_instrs += N_DYNTRANS_IDLE_BREAK;
 		cpu->cd.arm.next_ic = &nothing_call;
 		return;
 	}
@@ -1782,7 +2230,7 @@ X(cmps_0_beq)
 		cpu->cd.arm.flags = ARM_F_Z | ARM_F_C;
 		cpu->pc = (uint32_t)(((uint32_t)cpu->pc & 0xfffff000)
 		    + (int32_t)ic[1].arg[0]);
-		quick_pc_to_pointers(cpu);
+		quick_pc_to_pointers_arm(cpu);
 	} else {
 		/*  Semi-ugly hack which sets the negative-bit if a < 0:  */
 		cpu->cd.arm.flags = ARM_F_C | ((a >> 28) & 8);
@@ -1800,7 +2248,7 @@ X(cmps_pos_beq)
 		cpu->cd.arm.flags |= ARM_F_Z;
 		cpu->pc = (uint32_t)(((uint32_t)cpu->pc & 0xfffff000)
 		    + (int32_t)ic[1].arg[0]);
-		quick_pc_to_pointers(cpu);
+		quick_pc_to_pointers_arm(cpu);
 	} else {
 		cpu->cd.arm.next_ic = &ic[2];
 		if (c & 0x80000000)
@@ -1818,7 +2266,7 @@ X(cmps_neg_beq)
 		cpu->cd.arm.flags |= ARM_F_Z;
 		cpu->pc = (uint32_t)(((uint32_t)cpu->pc & 0xfffff000)
 		    + (int32_t)ic[1].arg[0]);
-		quick_pc_to_pointers(cpu);
+		quick_pc_to_pointers_arm(cpu);
 	} else {
 		cpu->cd.arm.next_ic = &ic[2];
 		if (c & 0x80000000)
@@ -2091,7 +2539,7 @@ X(end_of_page)
 	cpu->pc += (ARM_IC_ENTRIES_PER_PAGE << ARM_INSTR_ALIGNMENT_SHIFT);
 
 	/*  Find the new physical page and update the translation pointers:  */
-	quick_pc_to_pointers(cpu);
+	quick_pc_to_pointers_arm(cpu);
 
 	/*  end_of_page doesn't count as an executed instruction:  */
 	cpu->n_translated_instrs --;
@@ -2530,7 +2978,7 @@ X(to_be_translated)
 	unsigned char *page;
 	unsigned char ib[4];
 	int condition_code, main_opcode, secondary_opcode, s_bit, rn, rd, r8;
-	int p_bit, u_bit, w_bit, l_bit, regform, rm, c, t, any_pc_reg;
+	int p_bit, u_bit, w_bit, l_bit, regform, rm, any_pc_reg; // , c, t
 	void (*samepage_function)(struct cpu *, struct arm_instr_call *);
 
 	/*  Figure out the address of the instruction:  */
@@ -2580,28 +3028,75 @@ X(to_be_translated)
 	rn    = (iword >> 16) & 15;
 	rd    = (iword >> 12) & 15;
 	r8    = (iword >> 8) & 15;
-	c     = (iword >> 7) & 31;
-	t     = (iword >> 4) & 7;
+	// c     = (iword >> 7) & 31;
+	// t     = (iword >> 4) & 7;
 	rm    = iword & 15;
 
-	if (condition_code == 0xf) {
+	/*
+	 *  Translate the instruction:
+	 */
+
+	if ((iword >> 28) == 0xf) {
+		/*  The "never" condition is nowadays used for special encodings.  */
 		if ((iword & 0xfc70f000) == 0xf450f000) {
 			/*  Preload:  TODO.  Treat as NOP for now.  */
 			ic->f = instr(nop);
 			goto okay;
 		}
 
-		if (!cpu->translation_readahead)
-			fatal("TODO: ARM condition code 0x%x\n",
-			    condition_code);
-		goto bad;
+		if (iword == 0xf10c0040) {
+			/*  cpsid f. Treat as NOP for now.  */
+			ic->f = instr(nop);
+			goto okay;
+		}
+
+		if (iword == 0xf10c0080) {
+			/*  cpsid i. Treat as NOP for now.  */
+			ic->f = instr(nop);
+			goto okay;
+		}
+
+		if (iword == 0xf57ff04f) {
+			/*  dsb sy. Treat as NOP for now.  */
+			ic->f = instr(nop);
+			goto okay;
+		}
+
+		if (iword == 0xf57ff05f) {
+			/*  dmb sy. Treat as NOP for now.  */
+			ic->f = instr(nop);
+			goto okay;
+		}
+
+		if (iword == 0xf57ff06f) {
+			/*  isb sy. Treat as NOP for now.  */
+			ic->f = instr(nop);
+			goto okay;
+		}
+
+		switch (main_opcode) {
+		case 0xa:
+		case 0xb:
+			ic->f = instr(blx_imm);
+
+			/*  arg 1 = offset of current instruction  */
+			ic->arg[1] = addr & 0xffc;
+
+			/*  arg 0 = relative jump distance + 1 (to enable THUMB)  */
+			ic->arg[0] = (iword & 0x00ffffff) << 2;
+			/*  Sign-extend:  */
+			if (ic->arg[0] & 0x02000000)
+				ic->arg[0] |= 0xfc000000;
+			if (main_opcode == 0xb)
+				ic->arg[0] += 2;
+			ic->arg[0] = (int32_t)(ic->arg[0] + 8 + 1);
+			break;
+		default:
+			goto bad;
+		}
+		goto okay;
 	}
 
-
-	/*
-	 *  Translate the instruction:
-	 */
-
 	switch (main_opcode) {
 
 	case 0x0:
@@ -2609,6 +3104,43 @@ X(to_be_translated)
 	case 0x2:
 	case 0x3:
 		/*  Check special cases first:  */
+		if ((iword & 0x0ff00fff) == 0x01900f9f) {
+			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+			ic->arg[1] = (size_t)(&cpu->cd.arm.r[rn]);
+			ic->arg[2] = 0;
+
+			if (rd == ARM_PC || rn == ARM_PC) {
+				if (!cpu->translation_readahead)
+					fatal("ldrex with pc register: TODO\n");
+				goto bad;
+			}
+			
+			ic->f = cond_instr(ldrex);
+			break;
+		}
+
+		if ((iword & 0x0ff00ff0) == 0x01800f90) {
+			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+			ic->arg[1] = (size_t)(&cpu->cd.arm.r[rn]);
+			ic->arg[2] = (size_t)(&cpu->cd.arm.r[rm]);
+
+			if (rd == ARM_PC || rm == ARM_PC || rn == ARM_PC ||
+			    rd == rm || rd == rn) {
+				if (!cpu->translation_readahead)
+					fatal("strex with bad register: TODO\n");
+				goto bad;
+			}
+			
+			ic->f = cond_instr(strex);
+			break;
+		}
+
+		if ((iword & 0x0ff000f0) == 0x00600090) {
+			ic->arg[0] = iword;
+			ic->f = cond_instr(mls);
+			break;
+		}
+		
 		if ((iword & 0x0fc000f0) == 0x00000090) {
 			/*
 			 *  Multiplication:
@@ -2651,7 +3183,7 @@ X(to_be_translated)
 		if ((iword & 0x0ff000d0) == 0x01200010) {
 			/*  bx or blx  */
 			if (iword & 0x20)
-				ic->f = cond_instr(blx);
+				ic->f = cond_instr(blx_reg);
 			else {
 				if (cpu->machine->show_trace_tree &&
 				    rm == ARM_LR)
@@ -2660,6 +3192,7 @@ X(to_be_translated)
 					ic->f = cond_instr(bx);
 			}
 			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rm]);
+			ic->arg[2] = (addr & 0xffc) + 4;
                         break;
                 }
 		if ((iword & 0x0fb00ff0) == 0x1000090) {
@@ -2732,13 +3265,15 @@ X(to_be_translated)
 				imm = (imm >> 2) | ((imm & 3) << 30);
 			ic->arg[0] = imm;
 			ic->arg[2] = (size_t)(&cpu->cd.arm.r[rm]);
-			switch ((iword >> 16) & 15) {
-			case 1:	ic->arg[1] = 0x000000ff; break;
-			case 8:	ic->arg[1] = 0xff000000; break;
-			case 9:	ic->arg[1] = 0xff0000ff; break;
-			default:if (!cpu->translation_readahead)
-					fatal("unimpl a: msr regform\n");
-				goto bad;
+			{
+				uint32_t arg1 = 0;
+				if (iword & (1<<16)) arg1 |= 0x000000ff;
+				if (iword & (1<<17)) arg1 |= 0x0000ff00;
+				if (iword & (1<<18)) arg1 |= 0x00ff0000;
+				if (iword & (1<<19)) arg1 |= 0xff000000;
+				ic->arg[1] = arg1;
+				if (arg1 == 0)
+					ic->f = instr(nop);
 			}
 			break;
 		}
@@ -2757,8 +3292,8 @@ X(to_be_translated)
 			break;
 		}
 		if ((iword & 0x0e000090) == 0x00000090) {
-			int imm = ((iword >> 4) & 0xf0) | (iword & 0xf);
-			int regform = !(iword & 0x00400000);
+			regform = !(iword & 0x00400000);
+			imm = ((iword >> 4) & 0xf0) | (iword & 0xf);
 			p_bit = main_opcode & 1;
 			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rn]);
 			ic->arg[2] = (size_t)(&cpu->cd.arm.r[rd]);
@@ -2845,6 +3380,30 @@ X(to_be_translated)
 			break;
 		}
 
+		if ((iword & 0x0ff00000) == 0x03000000) {
+			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rn]);
+			ic->arg[1] = (((iword & 0xf0000) >> 4) | (iword & 0xfff));
+			ic->arg[2] = (size_t)(&cpu->cd.arm.r[rd]);
+			ic->f = cond_instr(movw);
+			if (rd == ARM_PC) {
+				if (!cpu->translation_readahead)
+					fatal("movw with rd = pc?\n");
+				goto bad;
+			}
+			break;
+		} else if ((iword & 0x0ff00000) == 0x03400000) {
+			ic->arg[0] = (size_t)(&cpu->cd.arm.r[rn]);
+			ic->arg[1] = (((iword & 0xf0000) >> 4) | (iword & 0xfff)) << 16;
+			ic->arg[2] = (size_t)(&cpu->cd.arm.r[rd]);
+			ic->f = cond_instr(movt);
+			if (rd == ARM_PC) {
+				if (!cpu->translation_readahead)
+					fatal("movt with rd = pc?\n");
+				goto bad;
+			}
+			break;
+		}
+
 		/*
 		 *  Generic Data Processing Instructions:
 		 */
@@ -2866,9 +3425,19 @@ X(to_be_translated)
 				q = 0;
 			ic->arg[1] = (size_t)(void *)arm_r[(iword & 0xfff) + q];
 		} else {
+			int steps = r8;
+
 			imm = iword & 0xff;
+			
 			while (r8-- > 0)
 				imm = (imm >> 2) | ((imm & 3) << 30);
+
+			if (steps != 0 && (imm != 0 && imm < 256)) {
+				if (!cpu->translation_readahead)
+					fatal("TODO: see cpu_arm_instr_dpi; non-zero steps but still under 256 is not implemented yet\n");
+				goto bad;
+			}
+
 			ic->arg[1] = imm;
 		}
 
@@ -2903,6 +3472,76 @@ X(to_be_translated)
 	case 0x5:	/*  xxxx010P UBWLnnnn ddddoooo oooooooo  Immediate  */
 	case 0x6:	/*  xxxx011P UBWLnnnn ddddcccc ctt0mmmm  Register  */
 	case 0x7:
+		// Special non-loadstore encodings:
+		if (main_opcode >= 6 && iword & 0x10) {
+			if ((iword & 0x0fff0ff0) == 0x06bf0f30) {
+				ic->f = cond_instr(rev);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+			} else if ((iword & 0x0fff03f0) == 0x06bf0070) {
+				ic->f = cond_instr(sxth);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				ic->arg[2] = ((iword & 0xc00) >> 10) << 3;
+			} else if ((iword & 0x0fff03f0) == 0x06ef0070) {
+				ic->f = cond_instr(uxtb);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				ic->arg[2] = ((iword & 0xc00) >> 10) << 3;
+			} else if ((iword & 0x0ff003f0) == 0x06e00070) {
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rn]);
+				ic->arg[2] = (size_t)(&cpu->cd.arm.r[rm]);
+				ic->f = cond_instr(uxtab);
+				if (iword & 0xc00) {
+					if (!cpu->translation_readahead)
+						fatal("unimplemented uxtab with rotate != 0\n");
+					goto bad;
+				}
+			} else if ((iword & 0x0fff03f0) == 0x06ff0070) {
+				ic->f = cond_instr(uxth);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				ic->arg[2] = ((iword & 0xc00) >> 10) << 3;
+			} else if ((iword & 0x0ff003f0) == 0x06f00070) {
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rn]);
+				ic->arg[2] = (size_t)(&cpu->cd.arm.r[rm]);
+				ic->f = cond_instr(uxtah);
+				if (iword & 0xc00) {
+					if (!cpu->translation_readahead)
+						fatal("unimplemented uxtah with rotate != 0\n");
+					goto bad;
+				}
+			} else if ((iword & 0x0fe00070) == 0x07c00010) {
+				ic->f = cond_instr(bfi);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				int lsb = (iword >> 7) & 31;
+				int msb = (iword >> 16) & 31;
+				ic->arg[2] = (msb << 16) + lsb;
+			} else if ((iword & 0x0fe00070) == 0x07e00050) {
+				ic->f = cond_instr(ubfx);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				int lsb = (iword >> 7) & 31;
+				int width = 1 + ((iword >> 16) & 31);
+				ic->arg[2] = (width << 16) + lsb;
+			} else if ((iword & 0x0fe00070) == 0x07a00050) {
+				ic->f = cond_instr(sbfx);
+				ic->arg[0] = (size_t)(&cpu->cd.arm.r[rd]);
+				ic->arg[1] = (size_t)(&cpu->cd.arm.r[rm]);
+				int lsb = (iword >> 7) & 31;
+				int width = 1 + ((iword >> 16) & 31);
+				ic->arg[2] = (width << 16) + lsb;
+			} else {
+				if (!cpu->translation_readahead)
+					fatal("unimplemented special non-loadstore encoding!\n");
+				goto bad;
+			}
+			break;
+		}
+	
 		ic->arg[0] = (size_t)(&cpu->cd.arm.r[rn]);
 		ic->arg[2] = (size_t)(&cpu->cd.arm.r[rd]);
 		if (rd == ARM_PC || rn == ARM_PC) {
@@ -2940,25 +3579,26 @@ X(to_be_translated)
 			/*  NOTE/TODO: This assumes 4KB pages,
 			    it will not work with 1KB pages.  */
 			if (ofs >= 0 && ofs <= max && p != NULL) {
-				unsigned char c[4];
+				unsigned char cbuf[4];
 				int len = b_bit? 1 : 4;
 				uint32_t x, a = (addr & 0xfffff000) | ofs;
 				/*  ic->f = cond_instr(mov);  */
 				ic->f = arm_dpi_instr[condition_code + 16*0xd];
 				ic->arg[2] = (size_t)(&cpu->cd.arm.r[rd]);
 
-				memcpy(c, p + (a & 0xfff), len);
+				memcpy(cbuf, p + (a & 0xfff), len);
 
 				if (b_bit) {
-					x = c[0];
+					x = cbuf[0];
 				} else {
 					if (cpu->byte_order == EMUL_LITTLE_ENDIAN)
-						x = c[0] + (c[1]<<8) +
-						    (c[2]<<16) + (c[3]<<24);
+						x = cbuf[0] + (cbuf[1]<<8) +
+						    (cbuf[2]<<16) + (cbuf[3]<<24);
 					else
-						x = c[3] + (c[2]<<8) +
-						    (c[1]<<16) + (c[0]<<24);
+						x = cbuf[3] + (cbuf[2]<<8) +
+						    (cbuf[1]<<16) + (cbuf[0]<<24);
 				}
+				
 				ic->arg[1] = x;
 			}
 		}
@@ -3167,15 +3807,6 @@ X(to_be_translated)
 		} else if (iword == 0xef8c64be) {
 			/*  Hack for openfirmware prom emulation:  */
 			ic->f = instr(openfirmware);
-		} else if (cpu->machine->userland_emul != NULL) {
-			if ((iword & 0x00f00000) == 0x00a00000) {
-				ic->arg[0] = iword & 0x00ffffff;
-				ic->f = cond_instr(swi_useremul);
-			} else {
-				if (!cpu->translation_readahead)
-					fatal("Bad userland SWI?\n");
-				goto bad;
-			}
 		}
 		break;
 
