@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -71,7 +71,7 @@ static void gather_statistics(struct cpu *cpu)
 			break;
 		case 'p':
 			/*  Physical program counter address:  */
-			cpu->cd.DYNTRANS_ARCH.cur_physpage = (void *)
+			cpu->cd.DYNTRANS_ARCH.cur_physpage = (struct DYNTRANS_TC_PHYSPAGE *)
 			    cpu->cd.DYNTRANS_ARCH.cur_ic_page;
 			a = cpu->cd.DYNTRANS_ARCH.cur_physpage->physaddr;
 			a &= ~((DYNTRANS_IC_ENTRIES_PER_PAGE-1) <<
@@ -79,10 +79,10 @@ static void gather_statistics(struct cpu *cpu)
 			a += low_pc << DYNTRANS_INSTR_ALIGNMENT_SHIFT;
 			if (cpu->is_32bit)
 				snprintf(buf + strlen(buf), sizeof(buf),
-				    "0x%08"PRIx32, (uint32_t)a);
+				    "0x%08" PRIx32, (uint32_t)a);
 			else
 				snprintf(buf + strlen(buf), sizeof(buf),
-				    "0x%016"PRIx64, (uint64_t)a);
+				    "0x%016" PRIx64, (uint64_t)a);
 			break;
 		case 'v':
 			/*  Virtual program counter address:  */
@@ -92,10 +92,10 @@ static void gather_statistics(struct cpu *cpu)
 			a += low_pc << DYNTRANS_INSTR_ALIGNMENT_SHIFT;
 			if (cpu->is_32bit)
 				snprintf(buf + strlen(buf), sizeof(buf),
-				    "0x%08"PRIx32, (uint32_t)a);
+				    "0x%08" PRIx32, (uint32_t)a);
 			else
 				snprintf(buf + strlen(buf), sizeof(buf),
-				    "0x%016"PRIx64, (uint64_t)a);
+				    "0x%016" PRIx64, (uint64_t)a);
 			break;
 		}
 		i++;
@@ -141,7 +141,7 @@ static void gather_statistics(struct cpu *cpu)
 			cpu->pc &= ~((DYNTRANS_IC_ENTRIES_PER_PAGE-1) << \
 			    DYNTRANS_INSTR_ALIGNMENT_SHIFT);		\
 			cpu->pc += (low_pc << DYNTRANS_INSTR_ALIGNMENT_SHIFT);\
-			printf("Crash at %016"PRIx64"\n", cpu->pc);	\
+			printf("Crash at %016" PRIx64"\n", cpu->pc);	\
 			printf("nr of I calls: %lli\n", nr_of_I_calls);	\
 			printf("Next ic = %p\n", cpu->cd.		\
 				DYNTRANS_ARCH.next_ic);			\
@@ -190,6 +190,8 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 	DYNTRANS_PC_TO_POINTERS(cpu);
 #endif
 
+	cpu->wants_to_idle = false;
+
 	/*
 	 *  Interrupt assertion?  (This is _below_ the initial PC to pointer
 	 *  conversion; if the conversion caused an exception of some kind
@@ -249,11 +251,22 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 #endif
 	}
 
+#ifdef DYNTRANS_ARM
+	if (cpu->cd.arm.cpsr & ARM_FLAG_T) {
+		// fatal("THUMB execution not implemented.\n");
+		// cpu->running = false;
+		// return 0;
+		
+		arm_cpu_interpret_thumb_SLOW(cpu);
+		return 1;
+	}
+#endif
+
 	cached_pc = cpu->pc;
 
 	cpu->n_translated_instrs = 0;
 
-	cpu->cd.DYNTRANS_ARCH.cur_physpage = (void *)
+	cpu->cd.DYNTRANS_ARCH.cur_physpage = (struct DYNTRANS_TC_PHYSPAGE *)
 	    cpu->cd.DYNTRANS_ARCH.cur_ic_page;
 
 	if (single_step || cpu->machine->instruction_trace
@@ -266,7 +279,7 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 			debug("\n");
 			cpu_register_dump(cpu->machine, cpu, 1, 0x1);
 		}
-		if (cpu->machine->instruction_trace) {
+		if (single_step || cpu->machine->instruction_trace) {
 			/*  TODO/Note: This must be large enough to hold
 			    any instruction for any ISA:  */
 			unsigned char instr[1 <<
@@ -280,7 +293,7 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 				int len =
 #endif
 				    cpu_disassemble_instr(
-				    cpu->machine, cpu, instr, 1, 0);
+				    cpu->machine, cpu, instr, 1, cpu->pc);
 #ifdef DYNTRANS_DELAYSLOT
 				/*  Show the instruction in the delay slot,
 				    if any:  */
@@ -297,7 +310,7 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 					cpu->delay_slot = DELAYED;
 					cpu->pc += len;
 					cpu_disassemble_instr(cpu->machine,
-					    cpu, instr, 1, 0);
+					    cpu, instr, 1, cpu->pc);
 					cpu->delay_slot = saved_delayslot;
 					cpu->pc -= len;
 				}
@@ -323,10 +336,8 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 			S; I; S; I; S; I; S; I; S; I; S; I;
 			S; I; S; I; S; I; S; I; S; I; S; I;
 
-			n_instrs += 24;
-
-			if (n_instrs + cpu->n_translated_instrs >=
-			    N_SAFE_DYNTRANS_LIMIT)
+			cpu->n_translated_instrs += 24;
+			if (cpu->n_translated_instrs >= N_SAFE_DYNTRANS_LIMIT)
 				break;
 		}
 	} else {
@@ -362,6 +373,22 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 		}
 	}
 
+	if (cpu->wants_to_idle) {
+		cpu->n_translated_instrs -= N_DYNTRANS_IDLE_BREAK;
+
+		// TODO: More arch specific interrupt checks
+		if (false
+#ifdef DYNTRANS_M88K
+		    || (cpu->cd.m88k.irq_asserted && !(cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_IND))
+#endif
+		    ) {
+			debugmsg_cpu(cpu, SUBSYS_CPU, "idle", VERBOSITY_DEBUG, "not idling; exception");
+			cpu->wants_to_idle = false;
+		}
+
+		cpu->has_been_idling = true;
+	}
+
 	n_instrs += cpu->n_translated_instrs;
 
 	/*  Synchronize the program counter:  */
@@ -389,9 +416,12 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 #ifdef DYNTRANS_MIPS
 	/*  Update the count register (on everything except EXC3K):  */
 	if (cpu->cd.mips.cpu_type.exc_model != EXC3K) {
-		uint32_t old = cpu->cd.mips.coproc[0]->reg[COP0_COUNT];
-		int32_t diff1 = cpu->cd.mips.coproc[0]->reg[COP0_COMPARE] - old;
-		int32_t diff2;
+		uint32_t old;
+		int32_t diff1, diff2;
+		cpu->cd.mips.coproc[0]->reg[COP0_COUNT] -= cpu->cd.mips.count_register_read_count;
+		cpu->cd.mips.count_register_read_count = 0;
+		old = cpu->cd.mips.coproc[0]->reg[COP0_COUNT];
+		diff1 = cpu->cd.mips.coproc[0]->reg[COP0_COMPARE] - old;
 		cpu->cd.mips.coproc[0]->reg[COP0_COUNT] =
 		    (int32_t) (old + n_instrs);
 		diff2 = cpu->cd.mips.coproc[0]->reg[COP0_COMPARE] -
@@ -449,15 +479,15 @@ int DYNTRANS_RUN_INSTR_DEF(struct cpu *cpu)
 void DYNTRANS_FUNCTION_TRACE_DEF(struct cpu *cpu, int n_args)
 {
 	int show_symbolic_function_name = 1;
-        char strbuf[50];
+        char strbuf[55];
 	char *symbol;
 	uint64_t ot;
 	int x, print_dots = 1, n_args_to_print =
-#if defined(DYNTRANS_ALPHA) || defined(DYNTRANS_SPARC)
+#if defined(DYNTRANS_ALPHA)
 	    6
 #else
-#if defined(DYNTRANS_SH) || defined(DYNTRANS_M88K)
-	    8	/*  Both for 32-bit and 64-bit SuperH, and M88K  */
+#if defined(DYNTRANS_SH) || defined(DYNTRANS_M88K) || defined(DYNTRANS_RISCV) || defined(DYNTRANS_I960)
+	    8	/*  For 32-bit and 64-bit SuperH, M88K, i960, and RISC-V  */
 #else
 	    4	/*  Default value for most archs  */
 #endif
@@ -495,29 +525,24 @@ void DYNTRANS_FUNCTION_TRACE_DEF(struct cpu *cpu, int n_args)
 #ifdef DYNTRANS_ARM
 		    r[0
 #endif
-#ifdef DYNTRANS_AVR
-		    /*  TODO: 24,25 = first register, but then
-			they go downwards, ie. 22,23 and so on  */
-		    r[24
+#ifdef DYNTRANS_I960
+		    r[I960_G0
 #endif
 #ifdef DYNTRANS_MIPS
 		    gpr[MIPS_GPR_A0
 #endif
-#ifdef DYNTRANS_M32R
-		    r[0		/*  r0..r3?  */
-#endif
 #ifdef DYNTRANS_M88K
 		    r[2		/*  r2..r9  */
 #endif
 #ifdef DYNTRANS_PPC
 		    gpr[3
 #endif
+#ifdef DYNTRANS_RISCV
+		    x[10
+#endif
 #ifdef DYNTRANS_SH
 		    r[4		/*  NetBSD seems to use 4? But 2 seems
 					to be used by other code? TODO  */
-#endif
-#ifdef DYNTRANS_SPARC
-		    r[8		/*  o0..o5  */
 #endif
 		    + x];
 
@@ -525,17 +550,25 @@ void DYNTRANS_FUNCTION_TRACE_DEF(struct cpu *cpu, int n_args)
 
 		if (d > -256 && d < 256)
 			fatal("%i", (int)d);
-		else if (memory_points_to_string(cpu, cpu->mem, d, 1))
-			fatal("\"%s\"", memory_conv_to_string(cpu,
-			    cpu->mem, d, strbuf, sizeof(strbuf)));
-		else if (symbol != NULL && ot == 0 &&
+		else if (memory_points_to_string(cpu, cpu->mem, d, 1)) {
+			memset(strbuf, 0, sizeof(strbuf));
+			memory_conv_to_string(cpu,
+			    cpu->mem, d, strbuf, sizeof(strbuf));
+			fatal("\"%s\"", strbuf);
+			// TODO: This shows "..." when the string is longer
+			// _or exactly the max length_. An improvement would
+			// be to detect the case where the string is exactly
+			// the max length and not show "..." then.
+			if (strlen(strbuf) >= sizeof(strbuf)-1)
+				fatal("...");
+		} else if (symbol != NULL && ot == 0 &&
 		    show_symbolic_function_name)
 			fatal("&%s", symbol);
 		else {
 			if (cpu->is_32bit)
-				fatal("0x%"PRIx32, (uint32_t)d);
+				fatal("0x%" PRIx32, (uint32_t)d);
 			else
-				fatal("0x%"PRIx64, (uint64_t)d);
+				fatal("0x%" PRIx64, (uint64_t)d);
 		}
 
 		if (x < n_args_to_print - 1)
@@ -612,7 +645,7 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 	x1 = (cached_pc >> (64-DYNTRANS_L1N)) & mask1;
 	x2 = (cached_pc >> (64-DYNTRANS_L1N-DYNTRANS_L2N)) & mask2;
 	x3 = (cached_pc >> (64-DYNTRANS_L1N-DYNTRANS_L2N-DYNTRANS_L3N)) & mask3;
-	/*  fatal("X3: cached_pc=%016"PRIx64" x1=%x x2=%x x3=%x\n",
+	/*  fatal("X3: cached_pc=%016" PRIx64" x1=%x x2=%x x3=%x\n",
 	    (uint64_t)cached_pc, (int)x1, (int)x2, (int)x3);  */
 	l2 = cpu->cd.DYNTRANS_ARCH.l1_64[x1];
 	/*  fatal("  l2 = %p\n", l2);  */
@@ -657,8 +690,8 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 			 *  exception handler.
 			 */
 			/*  fatal("TODO: instruction vaddr=>paddr translation "
-			    "failed. vaddr=0x%"PRIx64"\n", (uint64_t)cached_pc);
-			fatal("!! cpu->pc=0x%"PRIx64"\n", (uint64_t)cpu->pc); */
+			    "failed. vaddr=0x%" PRIx64"\n", (uint64_t)cached_pc);
+			fatal("!! cpu->pc=0x%" PRIx64"\n", (uint64_t)cpu->pc); */
 
 			/*  If there was an exception, the PC has changed.
 			    Update cached_pc:  */
@@ -691,7 +724,7 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 
 			/*  printf("EXCEPTION HANDLER: vaddr = 0x%x ==> "
 			    "paddr = 0x%x\n", (int)cpu->pc, (int)paddr);
-			fatal("!? cpu->pc=0x%"PRIx64"\n", (uint64_t)cpu->pc); */
+			fatal("!? cpu->pc=0x%" PRIx64"\n", (uint64_t)cpu->pc); */
 
 			if (!ok) {
 				fatal("FATAL: could not find physical"
@@ -720,9 +753,9 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 	}
 
 	if (cpu->translation_cache_cur_ofs >= dyntrans_cache_size) {
-#ifdef UNSTABLE_DEVEL
-		fatal("[ dyntrans: resetting the translation cache ]\n");
-#endif
+		debugmsg(SUBSYS_CPU, "dyntrans", VERBOSITY_INFO,
+		    "resetting the translation cache");
+
 		cpu_create_or_reset_tc(cpu);
 	}
 
@@ -754,7 +787,7 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 	if (physpage_ofs == 0) {
 		uint32_t previous_first_page_in_chain;
 
-		/*  fatal("CREATING page %lli (physaddr 0x%"PRIx64"), table "
+		/*  fatal("CREATING page %lli (physaddr 0x%" PRIx64"), table "
 		    "index %i\n", (long long)pagenr, (uint64_t)physaddr,
 		    (int)table_index);  */
 
@@ -799,8 +832,8 @@ void DYNTRANS_PC_TO_POINTERS_GENERIC(struct cpu *cpu)
 	cpu->cd.DYNTRANS_ARCH.next_ic = cpu->cd.DYNTRANS_ARCH.cur_ic_page +
 	    DYNTRANS_PC_TO_IC_ENTRY(cached_pc);
 
-	/*  printf("cached_pc=0x%016"PRIx64"  pagenr=%lli  table_index=%lli, "
-	    "physpage_ofs=0x%016"PRIx64"\n", (uint64_t)cached_pc, (long long)
+	/*  printf("cached_pc=0x%016" PRIx64"  pagenr=%lli  table_index=%lli, "
+	    "physpage_ofs=0x%016" PRIx64"\n", (uint64_t)cached_pc, (long long)
 	    pagenr, (long long)table_index, (uint64_t)physpage_ofs);  */
 }
 
@@ -861,8 +894,8 @@ have_it:
 	cpu->cd.DYNTRANS_ARCH.next_ic = cpu->cd.DYNTRANS_ARCH.cur_ic_page +
 	    DYNTRANS_PC_TO_IC_ENTRY(cached_pc);
 
-	/*  printf("cached_pc=0x%016"PRIx64"  pagenr=%lli  table_index=%lli, "
-	    "physpage_ofs=0x%016"PRIx64"\n", (uint64_t)cached_pc, (long long)
+	/*  printf("cached_pc=0x%016" PRIx64"  pagenr=%lli  table_index=%lli, "
+	    "physpage_ofs=0x%016" PRIx64"\n", (uint64_t)cached_pc, (long long)
 	    pagenr, (long long)table_index, (uint64_t)physpage_ofs);  */
 }
 #endif	/*  DYNTRANS_PC_TO_POINTERS_FUNC  */
@@ -909,7 +942,8 @@ void DYNTRANS_INIT_TABLES(struct cpu *cpu)
 	int i;
 	struct DYNTRANS_TC_PHYSPAGE *ppp;
 
-	CHECK_ALLOCATION(ppp = malloc(sizeof(struct DYNTRANS_TC_PHYSPAGE)));
+	CHECK_ALLOCATION(ppp =
+	    (struct DYNTRANS_TC_PHYSPAGE *) malloc(sizeof(struct DYNTRANS_TC_PHYSPAGE)));
 
 	ppp->next_ofs = 0;
 	ppp->translations_bitmap = 0;
@@ -943,8 +977,8 @@ void DYNTRANS_INIT_TABLES(struct cpu *cpu)
 	if (cpu->is_32bit)
 		return;
 
-	dummy_l2 = zeroed_alloc(sizeof(struct DYNTRANS_L2_64_TABLE));
-	dummy_l3 = zeroed_alloc(sizeof(struct DYNTRANS_L3_64_TABLE));
+	dummy_l2 = (struct DYNTRANS_L2_64_TABLE *) zeroed_alloc(sizeof(struct DYNTRANS_L2_64_TABLE));
+	dummy_l3 = (struct DYNTRANS_L3_64_TABLE *) zeroed_alloc(sizeof(struct DYNTRANS_L3_64_TABLE));
 
 	cpu->cd.DYNTRANS_ARCH.l2_64_dummy = dummy_l2;
 	cpu->cd.DYNTRANS_ARCH.l3_64_dummy = dummy_l3;
@@ -1000,6 +1034,7 @@ static void DYNTRANS_INVALIDATE_TLB_ENTRY(struct cpu *cpu,
 		cpu->cd.DYNTRANS_ARCH.vaddr_to_tlbindex[index] = 0;
 	}
 #else
+	// 64-bit:
 	const uint32_t mask1 = (1 << DYNTRANS_L1N) - 1;
 	const uint32_t mask2 = (1 << DYNTRANS_L2N) - 1;
 	const uint32_t mask3 = (1 << DYNTRANS_L3N) - 1;
@@ -1068,6 +1103,34 @@ static void DYNTRANS_INVALIDATE_TLB_ENTRY(struct cpu *cpu,
 }
 #endif
 
+/*
+	int found = -1;
+	for (int q = 0; q < 192; ++q)
+	{
+		if (cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[q].vaddr_page == vaddr_page &&
+			cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[q].valid)
+		{
+			if (found >= 0)
+			{
+				printf("ALREADY FOUND = %i\n", found);
+				int zz = found;
+				printf("%03i: ", zz);
+				printf("vaddr=%016llx\n", (long long)cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].vaddr_page);
+				printf("valid=%i\n", cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].valid);
+
+				zz = q;
+				printf("%03i: ", zz);
+				printf("vaddr=%016llx\n", (long long)cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].vaddr_page);
+				printf("valid=%i\n", cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].valid);
+
+				exit(1);
+			}
+			
+			found = q;
+		}
+	}
+*/
+
 	l3->host_load[x3] = NULL;
 	l3->host_store[x3] = NULL;
 	l3->phys_addr[x3] = 0;
@@ -1076,6 +1139,16 @@ static void DYNTRANS_INVALIDATE_TLB_ENTRY(struct cpu *cpu,
 		cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[
 		    l3->vaddr_to_tlbindex[x3] - 1].valid = 0;
 		l3->refcount --;
+	} else {/*
+		printf("APA: vaddr_page=%016llx l3->refcount = %i\n", (long long)vaddr_page, l3->refcount);
+		for (int zz = 0; zz < 128; ++zz)
+		{
+			printf("%03i: ", zz);
+			printf("vaddr=%016llx\n", (long long)cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].vaddr_page);
+			printf("valid=%i\n", cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[zz].valid);
+		}
+		
+		exit(1);*/
 	}
 	l3->vaddr_to_tlbindex[x3] = 0;
 
@@ -1154,7 +1227,30 @@ void DYNTRANS_INVALIDATE_TC(struct cpu *cpu, uint64_t addr, int flags)
 	/*  Quick case for _one_ virtual addresses: see note above.  */
 	if (flags & INVALIDATE_VADDR) {
 		/*  fatal("vaddr 0x%08x\n", (int)addr_page);  */
+
+#ifdef MODE32
 		DYNTRANS_INVALIDATE_TLB_ENTRY(cpu, addr_page, flags);
+#else
+		// In principle, there can be multiple entries, and they must
+		// be invalidated based on the bits that matter. For example,
+		// using vaddr 0x00A0000012345678 and 0x00B000001234567c
+		// actually point to the same physical page, and if then
+		// the virtual address 0x0000000012345000 is to be invalidated,
+		// then both of those should be invalidated!
+		int n = 0;
+		for (r=0; r<DYNTRANS_MAX_VPH_TLB_ENTRIES; r++)
+			if (cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid &&
+			    (cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].vaddr_page
+			    & cpu->vaddr_mask) == (addr_page & cpu->vaddr_mask)) {
+				DYNTRANS_INVALIDATE_TLB_ENTRY(cpu, cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].vaddr_page, flags);
+				n ++;
+			}
+
+		if (n > 1)
+			debug("[ DYNTRANS_INVALIDATE_TC: CACHE COLLISION vaddr 0x%016llx! ]\n",
+				(long long)addr_page);
+#endif
+
 		return;
 	}
 
@@ -1169,7 +1265,9 @@ void DYNTRANS_INVALIDATE_TC(struct cpu *cpu, uint64_t addr, int flags)
 				DYNTRANS_INVALIDATE_TLB_ENTRY(cpu, cpu->cd.
 				    DYNTRANS_ARCH.vph_tlb_entry[r].vaddr_page,
 				    0);
-				cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid=0;
+				// Not needed: DYNTRANS_INVALIDATE_TLB_ENTRY should
+				// set valid to 0 anyway.
+				// cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid=0;
 			}
 		}
 		return;
@@ -1182,7 +1280,9 @@ void DYNTRANS_INVALIDATE_TC(struct cpu *cpu, uint64_t addr, int flags)
 				DYNTRANS_INVALIDATE_TLB_ENTRY(cpu, cpu->cd.
 				    DYNTRANS_ARCH.vph_tlb_entry[r].vaddr_page,
 				    0);
-				cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid=0;
+				// Not needed: DYNTRANS_INVALIDATE_TLB_ENTRY should
+				// set valid to 0 anyway.
+				// cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid=0;
 			}
 		}
 		return;
@@ -1290,6 +1390,8 @@ void DYNTRANS_INVALIDATE_TC_CODE(struct cpu *cpu, uint64_t addr, int flags)
 				*physpage_entryp = ppp->next_ofs;
 		}
 #else
+		(void)prev_ppp;	// shut up compiler warning
+
 		/*
 		 *  Instead of removing the page from the code cache, each
 		 *  entry can be set to "to_be_translated". This is slow in
@@ -1398,8 +1500,8 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 	vaddr_page &= 0xffffffffULL;
 
 	if (paddr_page > 0xffffffffULL) {
-		fatal("update_translation_table(): v=0x%016"PRIx64", h=%p w=%i"
-		    " p=0x%016"PRIx64"\n", vaddr_page, host_page, writeflag,
+		fatal("update_translation_table(): v=0x%016" PRIx64", h=%p w=%i"
+		    " p=0x%016" PRIx64"\n", vaddr_page, host_page, writeflag,
 		    paddr_page);
 		exit(1);
 	}
@@ -1415,8 +1517,8 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 	struct DYNTRANS_L2_64_TABLE *l2;
 	struct DYNTRANS_L3_64_TABLE *l3;
 
-	/*  fatal("update_translation_table(): v=0x%016"PRIx64", h=%p w=%i"
-	    " p=0x%016"PRIx64"\n", (uint64_t)vaddr_page, host_page, writeflag,
+	/*  fatal("update_translation_table(): v=0x%016" PRIx64", h=%p w=%i"
+	    " p=0x%016" PRIx64"\n", (uint64_t)vaddr_page, host_page, writeflag,
 	    (uint64_t)paddr_page);  */
 #endif
 
@@ -1428,6 +1530,8 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 		useraccess = 1;
 	}
 
+	(void)useraccess;  // shut up compiler warning about unused var
+
 #ifdef DYNTRANS_M88K
 	/*  TODO  */
 	if (useraccess)
@@ -1508,7 +1612,8 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 			} else {
 				int i;
 				CHECK_ALLOCATION(l2 =
-				    cpu->cd.DYNTRANS_ARCH.l1_64[x1] = malloc(
+				    cpu->cd.DYNTRANS_ARCH.l1_64[x1] =
+				    (struct DYNTRANS_L2_64_TABLE *) malloc(
 				    sizeof(struct DYNTRANS_L2_64_TABLE)));
 				l2->refcount = 0;
 				for (i=0; i<(1 << DYNTRANS_L2N); i++)
@@ -1531,7 +1636,8 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 				    cpu->cd.DYNTRANS_ARCH.next_free_l3;
 				cpu->cd.DYNTRANS_ARCH.next_free_l3 = l3->next;
 			} else {
-				l3 = l2->l3[x2] = zeroed_alloc(sizeof(
+				l3 = l2->l3[x2] = (struct DYNTRANS_L3_64_TABLE *)
+				    zeroed_alloc(sizeof(
 				    struct DYNTRANS_L3_64_TABLE));
 			}
 			if (l3->refcount != 0) {
@@ -1628,6 +1734,10 @@ void DYNTRANS_UPDATE_TRANSLATION_TABLE(struct cpu *cpu, uint64_t vaddr_page,
 			l3->host_store[x3] = writeflag? host_page : NULL;
 			l3->phys_addr[x3] = paddr_page;
 		}
+		
+		/*  HM!  /2013-11-17  */
+		/*  Should this be here?  2014-08-02  */
+		//l3->phys_page[x3] = NULL;
 
 #ifdef BUGHUNT
 /*  Count how many pages are actually in use:  */
@@ -1675,17 +1785,17 @@ cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid);
 			if (curpc == (MODE_uint_t)
 			    cpu->machine->breakpoints.addr[i]) {
 				if (!cpu->machine->instruction_trace) {
-					int old_quiet_mode = quiet_mode;
+					int tmp_old_quiet_mode = quiet_mode;
 					quiet_mode = 0;
 					DISASSEMBLE(cpu, ib, 1, 0);
-					quiet_mode = old_quiet_mode;
+					quiet_mode = tmp_old_quiet_mode;
 				}
 #ifdef MODE32
-				fatal("BREAKPOINT: pc = 0x%"PRIx32"\n(The "
+				fatal("BREAKPOINT: pc = 0x%" PRIx32"\n(The "
 				    "instruction has not yet executed.)\n",
 				    (uint32_t)cpu->pc);
 #else
-				fatal("BREAKPOINT: pc = 0x%"PRIx64"\n(The "
+				fatal("BREAKPOINT: pc = 0x%" PRIx64"\n(The "
 				    "instruction has not yet executed.)\n",
 				    (uint64_t)cpu->pc);
 #endif
@@ -1695,7 +1805,7 @@ cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid);
 					    " slot! Not yet supported.\n");
 #endif
 				single_step_breakpoint = 1;
-				single_step = ENTER_SINGLE_STEPPING;
+				single_step = true;
 				goto stop_running_translated;
 			}
 	}
@@ -1712,7 +1822,7 @@ cpu->cd.DYNTRANS_ARCH.vph_tlb_entry[r].valid);
 	 */
 
 	/*  Make sure cur_physpage is in synch:  */
-	cpu->cd.DYNTRANS_ARCH.cur_physpage = (void *)
+	cpu->cd.DYNTRANS_ARCH.cur_physpage = (struct DYNTRANS_TC_PHYSPAGE *)
 	    cpu->cd.DYNTRANS_ARCH.cur_ic_page;
 
 	{
@@ -1840,25 +1950,35 @@ bad:	/*
 	if (cpu->translation_readahead)
 		return;
 
-	quiet_mode = 0;
-	fatal("to_be_translated(): TODO: unimplemented instruction");
 
-	if (cpu->machine->instruction_trace) {
-		if (cpu->is_32bit)
-			fatal(" at 0x%"PRIx32"\n", (uint32_t)cpu->pc);
-		else
-			fatal(" at 0x%"PRIx64"\n", (uint64_t)cpu->pc);
-	} else {
-		fatal(":\n");
+	/*
+	 *  Abort the emulation.  In principle, it may have been more accurate
+	 *  to treat invalid/unimplemented instructions as something that
+	 *  should cause e.g. an exception (depending on how the particular
+	 *  CPU works), but chances are that this is an unimplemented
+	 *  instruction that should be implemented for real.
+	 */
+
+	about_to_enter_single_step = true;
+
+	if (cpu->is_32bit)
+		debugmsg_cpu(cpu, SUBSYS_CPU, "", VERBOSITY_ERROR,
+		    "UNIMPLEMENTED instruction at 0x%" PRIx32, (uint32_t)cpu->pc);
+	else
+		debugmsg_cpu(cpu, SUBSYS_CPU, "", VERBOSITY_ERROR,
+		    "UNIMPLEMENTED instruction at 0x%" PRIx64, (uint64_t)cpu->pc);
+
+	if (!(single_step || cpu->machine->instruction_trace))
 		DISASSEMBLE(cpu, ib, 1, 0);
-	}
 
 	cpu->running = 0;
 
 	/*  Note: Single-stepping can jump here.  */
 stop_running_translated:
 
-	debugger_n_steps_left_before_interaction = 0;
+	// This makes sure that we stop in the debugger, even if e.g.
+	// "step 100" was executing and 50 steps remained.
+	debugger_reset();
 
 	ic = cpu->cd.DYNTRANS_ARCH.next_ic = &nothing_call;
 	cpu->cd.DYNTRANS_ARCH.next_ic ++;
