@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -47,12 +47,12 @@
 #include "machine.h"
 #include "memory.h"
 #include "misc.h"
-#include "mvmeprom.h"
 #include "settings.h"
 #include "symbol.h"
 
-#include "m8820x_pte.h"
-#include "m88k_dmt.h"
+#include "thirdparty/m8820x_pte.h"
+#include "thirdparty/m88k_dmt.h"
+#include "thirdparty/mvmeprom.h"
 
 #define DYNTRANS_32
 #define DYNTRANS_DELAYSLOT
@@ -62,18 +62,18 @@
 void m88k_pc_to_pointers(struct cpu *);
 void m88k_cpu_functioncall_trace(struct cpu *cpu, int n_args);
 
-static char *memop[4] = { ".d", "", ".h", ".b" };
+static const char *memop[4] = { ".d", "", ".h", ".b" };
 
 void m88k_irq_interrupt_assert(struct interrupt *interrupt);
 void m88k_irq_interrupt_deassert(struct interrupt *interrupt);
 
 
-static char *m88k_cr_names[] = M88K_CR_NAMES;
-static char *m88k_cr_197_names[] = M88K_CR_NAMES_197;
+static const char *m88k_cr_names[] = M88K_CR_NAMES;
+static const char *m88k_cr_197_names[] = M88K_CR_NAMES_197;
 
-static char *m88k_cr_name(struct cpu *cpu, int i)
+static const char *m88k_cr_name(struct cpu *cpu, int i)
 {
-	char **cr_names = m88k_cr_names;
+	const char **cr_names = m88k_cr_names;
 
 	/*  Hm. Is this really MVME197 specific? TODO  */
 	if (cpu->machine->machine_subtype == MACHINE_MVME88K_197)
@@ -125,16 +125,13 @@ int m88k_cpu_new(struct cpu *cpu, struct memory *mem,
 	cpu->translate_v2p = m88k_translate_v2p;
 
 	cpu->cd.m88k.cpu_type = cpu_type_defs[found];
-	cpu->name            = cpu->cd.m88k.cpu_type.name;
+	cpu->name            = strdup(cpu->cd.m88k.cpu_type.name);
 	cpu->is_32bit        = 1;
 	cpu->byte_order      = EMUL_BIG_ENDIAN;
 
 	cpu->instruction_has_delayslot = m88k_cpu_instruction_has_delayslot;
 
-	/*  Only show name and caches etc for CPU nr 0:  */
-	if (cpu_id == 0) {
-		debug("%s", cpu->name);
-	}
+	cpu->vaddr_mask = 0x00000000ffffffffULL;
 
 
 	/*
@@ -164,17 +161,17 @@ int m88k_cpu_new(struct cpu *cpu, struct memory *mem,
 
 	/*  Register the CPU interrupt pin:  */
 	{
-		struct interrupt template;
+		struct interrupt templ;
 		char name[50];
 		snprintf(name, sizeof(name), "%s", cpu->path);
 
-		memset(&template, 0, sizeof(template));
-		template.line = 0;
-		template.name = name;
-		template.extra = cpu;
-		template.interrupt_assert = m88k_irq_interrupt_assert;
-		template.interrupt_deassert = m88k_irq_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 0;
+		templ.name = name;
+		templ.extra = cpu;
+		templ.interrupt_assert = m88k_irq_interrupt_assert;
+		templ.interrupt_deassert = m88k_irq_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*  Set the Processor ID:  */
@@ -195,14 +192,12 @@ int m88k_cpu_new(struct cpu *cpu, struct memory *mem,
 /*
  *  m88k_cpu_dumpinfo():
  */
-void m88k_cpu_dumpinfo(struct cpu *cpu)
+void m88k_cpu_dumpinfo(struct cpu *cpu, bool verbose)
 {
-	/*  struct m88k_cpu_type_def *ct = &cpu->cd.m88k.cpu_type;  */
-
-	debug(", %s-endian",
+	debugmsg(SUBSYS_MACHINE, "cpu", VERBOSITY_INFO,
+	    "%s (%s-endian)",
+	    cpu->name,
 	    cpu->byte_order == EMUL_BIG_ENDIAN? "Big" : "Little");
-
-	debug("\n");
 }
 
 
@@ -278,7 +273,7 @@ void m88k_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 	if (gprs) {
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->pc, &offset);
-		debug("cpu%i:  pc  = 0x%08"PRIx32, x, (uint32_t)cpu->pc);
+		debug("cpu%i:  pc  = 0x%08" PRIx32, x, (uint32_t)cpu->pc);
 		debug("  <%s>\n", symbol != NULL? symbol : " no symbol ");
 
 		for (i=0; i<N_M88K_REGS; i++) {
@@ -287,7 +282,7 @@ void m88k_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 			if (i == 0)
 				debug("                  ");
 			else
-				debug("  r%-2i = 0x%08"PRIx32,
+				debug("  r%-2i = 0x%08" PRIx32,
 				    i, cpu->cd.m88k.r[i]);
 			if ((i % 4) == 3)
 				debug("\n");
@@ -304,7 +299,7 @@ void m88k_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		for (i=0; i<n_control_regs; i++) {
 			if ((i % 4) == 0)
 				debug("cpu%i:", x);
-			debug("  %4s=0x%08"PRIx32,
+			debug("  %4s=0x%08" PRIx32,
 			    m88k_cr_name(cpu, i), cpu->cd.m88k.cr[i]);
 			if ((i % 4) == 3)
 				debug("\n");
@@ -317,7 +312,7 @@ void m88k_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		for (i=0; i<n_fpu_control_regs; i++) {
 			if ((i % 4) == 0)
 				debug("cpu%i:", x);
-			debug("  %5s=0x%08"PRIx32,
+			debug("  %5s=0x%08" PRIx32,
 			    m88k_fcr_name(cpu, i), cpu->cd.m88k.fcr[i]);
 			if ((i % 4) == 3)
 				debug("\n");
@@ -330,71 +325,67 @@ void m88k_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
  *  m88k_cpu_tlbdump():
  *
  *  Called from the debugger to dump the TLB in a readable format.
- *  x is the cpu number to dump, or -1 to dump all CPUs.
  *
  *  If rawflag is nonzero, then the TLB contents isn't formated nicely,
  *  just dumped.
  */
-void m88k_cpu_tlbdump(struct machine *m, int x, int rawflag)
+void m88k_cpu_tlbdump(struct cpu* cpu, int rawflag)
 {
-	int cpu_nr, cmmu_nr, i;
+	int cpu_nr = -1, cmmu_nr, i;
 
-	for (cpu_nr = 0; cpu_nr < m->ncpus; cpu_nr++) {
-		struct cpu *cpu = m->cpus[cpu_nr];
+	for (i = 0; i < cpu->machine->ncpus; ++i)
+		if (cpu == cpu->machine->cpus[i])
+			cpu_nr = i;
 
-		if (x != -1 && cpu_nr != x)
+	for (cmmu_nr = 0; cmmu_nr < MAX_M8820X_CMMUS; cmmu_nr++) {
+		struct m8820x_cmmu *cmmu = cpu->cd.m88k.cmmu[cmmu_nr];
+		if (cmmu == NULL)
 			continue;
 
-		for (cmmu_nr = 0; cmmu_nr < MAX_M8820X_CMMUS; cmmu_nr++) {
-			struct m8820x_cmmu *cmmu = cpu->cd.m88k.cmmu[cmmu_nr];
-			if (cmmu == NULL)
-				continue;
+		printf("cpu%i: CMMU %i (%s)\n", cpu_nr, cmmu_nr,
+		    cmmu_nr & 1? "data" : "instruction");
 
-			printf("cpu%i: CMMU %i (%s)\n", cpu_nr, cmmu_nr,
-			    cmmu_nr & 1? "data" : "instruction");
+		/*  BATC:  */
+		for (i = 0; i < N_M88200_BATC_REGS; i++) {
+			uint32_t b = cmmu->batc[i];
+			printf("cpu%i: BATC[%2i]: ", cpu_nr, i);
+			printf("v=0x%08" PRIx32, b & 0xfff80000);
+			printf(", p=0x%08" PRIx32,
+			    (b << 13) & 0xfff80000);
+			printf(", %s %s %s %s %s %s\n",
+			    b & BATC_SO? "SP " : "!sp",
+			    b & BATC_WT? "WT " : "!wt",
+			    b & BATC_GLOBAL? "G  " : "!g ",
+			    b & BATC_INH? "CI " : "!ci",
+			    b & BATC_PROT? "WP " : "!wp",
+			    b & BATC_SO? "V " : "!v");
+		}
 
-			/*  BATC:  */
-			for (i = 0; i < N_M88200_BATC_REGS; i++) {
-				uint32_t b = cmmu->batc[i];
-				printf("cpu%i: BATC[%2i]: ", cpu_nr, i);
-				printf("v=0x%08"PRIx32, b & 0xfff80000);
-				printf(", p=0x%08"PRIx32,
-				    (b << 13) & 0xfff80000);
-				printf(", %s %s %s %s %s %s\n",
-				    b & BATC_SO? "SP " : "!sp",
-				    b & BATC_WT? "WT " : "!wt",
-				    b & BATC_GLOBAL? "G  " : "!g ",
-				    b & BATC_INH? "CI " : "!ci",
-				    b & BATC_PROT? "WP " : "!wp",
-				    b & BATC_SO? "V " : "!v");
-			}
+		/*  PATC:  */
+		for (i = 0; i < N_M88200_PATC_ENTRIES; i++) {
+			uint32_t v = cmmu->patc_v_and_control[i];
+			uint32_t p = cmmu->patc_p_and_supervisorbit[i];
 
-			/*  PATC:  */
-			for (i = 0; i < N_M88200_PATC_ENTRIES; i++) {
-				uint32_t v = cmmu->patc_v_and_control[i];
-				uint32_t p = cmmu->patc_p_and_supervisorbit[i];
+			printf("cpu%i: patc[%2i]: ", cpu_nr, i);
+			if (p & M8820X_PATC_SUPERVISOR_BIT)
+				printf("superv");
+			else
+				printf("user  ");
+			printf(" v=0x%08" PRIx32, v & 0xfffff000);
+			printf(", p=0x%08" PRIx32, p & 0xfffff000);
 
-				printf("cpu%i: patc[%2i]: ", cpu_nr, i);
-				if (p & M8820X_PATC_SUPERVISOR_BIT)
-					printf("superv");
-				else
-					printf("user  ");
-				printf(" v=0x%08"PRIx32, v & 0xfffff000);
-				printf(", p=0x%08"PRIx32, p & 0xfffff000);
+			printf("  %s %s %s %s %s %s %s",
+			    v & PG_U1?   "U1 " : "!u1",
+			    v & PG_U0?   "U0 " : "!u0",
+			    v & PG_SO?   "SP " : "!sp",
+			    v & PG_M?    "M "  : "!m",
+			    v & PG_U?    "U "  : "!u",
+			    v & PG_PROT? "WP " : "!wp",
+			    v & PG_V?    "V "  : "!v");
 
-				printf("  %s %s %s %s %s %s %s",
-				    v & PG_U1?   "U1 " : "!u1",
-				    v & PG_U0?   "U0 " : "!u0",
-				    v & PG_SO?   "SP " : "!sp",
-				    v & PG_M?    "M "  : "!m",
-				    v & PG_U?    "U "  : "!u",
-				    v & PG_PROT? "WP " : "!wp",
-				    v & PG_V?    "V "  : "!v");
-
-				if (i == cmmu->patc_update_index)
-					printf(" <--");
-				printf("\n");
-			}
+			if (i == cmmu->patc_update_index)
+				printf(" <--");
+			printf("\n");
 		}
 	}
 }
@@ -458,14 +449,18 @@ void m88k_ldcr(struct cpu *cpu, uint32_t *r32ptr, int cr)
 		 */
 		if (retval & DMT_VALID && !(retval & DMT_WRITE)) {
 			if (DMT_DREGBITS(retval) == M88K_ZERO_REG) {
-				fatal("DMT DREG = zero? Internal error.\n");
-				exit(1);
+				debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+				    "ldcr: DMT DREG = zero? Internal error.");
+				cpu->running = false;
+				return;
 			}
 		}
 		if (!!(cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_BO)
 		    != !!(retval & DMT_BO) && retval & DMT_VALID) {
-			fatal("DMT byte order not same as CPUs?\n");
-			exit(1);
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "ldcr: DMT byte order not same as CPUs?");
+			cpu->running = false;
+			return;
 		}
 
 		break;
@@ -479,15 +474,19 @@ void m88k_ldcr(struct cpu *cpu, uint32_t *r32ptr, int cr)
 		 *  should always be zero.
 		 */
 		if (retval & 3) {
-			fatal("DMAx not word-aligned? Internal error.\n");
-			exit(1);
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "ldcr: DMAx not word-aligned? Internal error.");
+			cpu->running = false;
+			return;
 		}
 
 		break;
 
-	default:fatal("m88k_ldcr: UNIMPLEMENTED cr = 0x%02x (%s)\n",
+	default:debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    "ldcr: UNIMPLEMENTED cr = 0x%02x (%s)\n",
 		    cr, m88k_cr_name(cpu, cr));
-		exit(1);
+		cpu->running = false;
+		return;
 	}
 
 	*r32ptr = retval;
@@ -506,26 +505,31 @@ void m88k_stcr(struct cpu *cpu, uint32_t value, int cr, int rte)
 
 	switch (cr) {
 
-	case M88K_CR_PSR:	/*  Processor Status Regoster  */
+	case M88K_CR_PSR:	/*  Processor Status Register  */
 		if ((cpu->byte_order == EMUL_LITTLE_ENDIAN
 		    && !(value & M88K_PSR_BO)) ||
 		    (cpu->byte_order == EMUL_BIG_ENDIAN
 		    && (value & M88K_PSR_BO))) {
-			fatal("TODO: attempt to change endianness by flipping"
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "stcr: TODO: attempt to change endianness by flipping"
 			    " the endianness bit in the PSR. How should this"
-			    " be handled? Aborting.\n");
-			exit(1);
+			    " be handled? Aborting.");
+			cpu->running = false;
+			return;
 		}
 
 		if (!rte && old & M88K_PSR_MODE && !(value & M88K_PSR_MODE))
-			fatal("[ m88k_stcr: WARNING! the PSR_MODE bit is being"
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_WARNING,
+			    "stcr: WARNING! the PSR_MODE bit is being"
 			    " cleared; this should be done using the RTE "
 			    "instruction only, according to the M88100 "
-			    "manual! Continuing anyway. ]\n");
+			    "manual! Continuing anyway.");
 
 		if (value & M88K_PSR_MXM) {
-			fatal("m88k_stcr: TODO: MXM support\n");
-			exit(1);
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "stcr: TODO: MXM support");
+			cpu->running = false;
+			return;
 		}
 
 		if ((old & M88K_PSR_MODE) != (value & M88K_PSR_MODE))
@@ -547,15 +551,19 @@ void m88k_stcr(struct cpu *cpu, uint32_t value, int cr, int rte)
 
 	case M88K_CR_SSBR:	/*  Shadow ScoreBoard Register  */
 		if (value & 1)
-			fatal("[ m88k_stcr: WARNING! bit 0 non-zero when"
-			    " writing to SSBR (?) ]\n");
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_WARNING,
+			    "stcr: WARNING! bit 0 non-zero when"
+			    " writing to SSBR (?)");
+
 		cpu->cd.m88k.cr[cr] = value;
 		break;
 
 	case M88K_CR_VBR:
 		if (value & 0x00000fff)
-			fatal("[ m88k_stcr: WARNING! bits 0..11 non-zero when"
-			    " writing to VBR (?) ]\n");
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_WARNING,
+			    "stcr: WARNING! bits 0..11 non-zero when"
+			    " writing to VBR (?)");
+
 		cpu->cd.m88k.cr[cr] = value;
 		break;
 
@@ -572,9 +580,11 @@ void m88k_stcr(struct cpu *cpu, uint32_t value, int cr, int rte)
 		cpu->cd.m88k.cr[cr] = value;
 		break;
 
-	default:fatal("m88k_stcr: UNIMPLEMENTED cr = 0x%02x (%s)\n",
+	default:debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    "stcr: UNIMPLEMENTED cr = 0x%02x (%s)",
 		    cr, m88k_cr_name(cpu, cr));
-		exit(1);
+		cpu->running = false;
+		return;
 	}
 }
 
@@ -586,18 +596,23 @@ void m88k_stcr(struct cpu *cpu, uint32_t value, int cr, int rte)
  */
 void m88k_fstcr(struct cpu *cpu, uint32_t value, int fcr)
 {
-#if 0
-	/*  TODO (?)  */
-	uint32_t old = cpu->cd.m88k.cr[fcr];
-
 	switch (fcr) {
-	default:fatal("m88k_fstcr: UNIMPLEMENTED fcr = 0x%02x (%s)\n",
-		    fcr, m88k_fcr_name(cpu, fcr));
-		exit(1);
+
+	case 0:
+	case 62:
+	case 63:
+		// Don't warn for writes with the value zero for these, for now.
+		if (value == 0)
+			break;
+
+		// fall through
+	
+	default:debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_WARNING,
+		    "WARNING: fstcr: write 0x%08x to UNIMPLEMENTED fcr = 0x%02x (%s)",
+		    value, fcr, m88k_fcr_name(cpu, fcr));
 	}
-#else
-	cpu->cd.m88k.cr[fcr] = value;
-#endif
+
+	cpu->cd.m88k.fcr[fcr] = value;
 }
 
 
@@ -633,8 +648,8 @@ static void m88k_memory_transaction_debug_dump(struct cpu *cpu, int n)
 		}
 		debug("bytebits=0x%x ]\n", DMT_ENBITS(dmt));
 
-		debug("[ DMD%i: 0x%08"PRIx32"; ", n, cpu->cd.m88k.dmd[n]);
-		debug("DMA%i: 0x%08"PRIx32" ]\n", n, cpu->cd.m88k.dma[n]);
+		debug("[ DMD%i: 0x%08" PRIx32"; ", n, cpu->cd.m88k.dmd[n]);
+		debug("DMA%i: 0x%08" PRIx32" ]\n", n, cpu->cd.m88k.dma[n]);
 	} else
 		debug("not valid ]\n");
 }
@@ -647,48 +662,100 @@ static void m88k_memory_transaction_debug_dump(struct cpu *cpu, int n)
  */
 void m88k_exception(struct cpu *cpu, int vector, int is_trap)
 {
-	int update_shadow_regs = 1;
+	bool update_shadow_regs = true;
+	bool implemented = false;
+
+	// TODO: Use a const char * instead or something like that, inside the switch...
 
-	debug("[ EXCEPTION 0x%03x: ", vector);
 	switch (vector) {
+
 	case M88K_EXCEPTION_RESET:
-		debug("RESET"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "RESET", VERBOSITY_INFO, "");
+		implemented = true;
+		break;
+
 	case M88K_EXCEPTION_INTERRUPT:
-		debug("INTERRUPT"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "INTERRUPT", VERBOSITY_INFO, "");
+		implemented = true;
+		break;
+
 	case M88K_EXCEPTION_INSTRUCTION_ACCESS:
-		debug("INSTRUCTION_ACCESS"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "INSTRUCTION_ACCESS", VERBOSITY_INFO, "");
+		implemented = true;
+		break;
+
 	case M88K_EXCEPTION_DATA_ACCESS:
-		debug("DATA_ACCESS"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "DATA_ACCESS", VERBOSITY_INFO, "");
+		implemented = true;
+		break;
+
 	case M88K_EXCEPTION_MISALIGNED_ACCESS:
-		debug("MISALIGNED_ACCESS"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "MISALIGNED_ACCESS", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_UNIMPLEMENTED_OPCODE:
-		debug("UNIMPLEMENTED_OPCODE"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "UNIMPLEMENTED_OPCODE", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_PRIVILEGE_VIOLATION:
-		debug("PRIVILEGE_VIOLATION"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "PRIVILEGE_VIOLATION", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_BOUNDS_CHECK_VIOLATION:
-		debug("BOUNDS_CHECK_VIOLATION"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "BOUNDS_CHECK_VIOLATION", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_ILLEGAL_INTEGER_DIVIDE:
-		debug("ILLEGAL_INTEGER_DIVIDE"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "ILLEGAL_INTEGER_DIVIDE", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_INTEGER_OVERFLOW:
-		debug("INTEGER_OVERFLOW"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "INTEGER_OVERFLOW", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_ERROR:
-		debug("ERROR"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "ERROR", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_SFU1_PRECISE:
-		debug("SFU1_PRECISE"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "SFU1_PRECISE", VERBOSITY_ERROR, "");
+		break;
+
 	case M88K_EXCEPTION_SFU1_IMPRECISE:
-		debug("SFU1_IMPRECISE"); break;
-	case 0x80:
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "SFU1_IMPRECISE", VERBOSITY_ERROR, "");
+		break;
+
+	case 0x80:	/*  up to OpenBSD 5.2  */
+	case 0x1c2:	/*  from OpenBSD 5.3 and forward...  */
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "syscall", VERBOSITY_INFO,
+		    "r13=%i", cpu->cd.m88k.r[13]);
 #if 0
 		fatal("[ syscall %i(", cpu->cd.m88k.r[13]);
 		m88k_cpu_functioncall_trace(cpu, 8);
 		fatal(") ]\n");
 #endif
-		debug("syscall, r13=%i", cpu->cd.m88k.r[13]); break;
+		break;
+
 	case MVMEPROM_VECTOR:
-		debug("MVMEPROM_VECTOR"); break;
-	default:debug("unknown"); break;
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "MVMEPROM_VECTOR", VERBOSITY_INFO, "");
+		break;
+
+	default:
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "unknown vector",
+		    vector < M88K_EXCEPTION_USER_TRAPS_START ? VERBOSITY_ERROR
+		    : VERBOSITY_INFO, "0x%03x", vector);
+	}
+
+	if (vector < M88K_EXCEPTION_USER_TRAPS_START && !implemented) {
+		debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "unimplemented",
+		    VERBOSITY_ERROR, "vector=0x%03x pc=0x%08x",
+			vector,
+			(int)cpu->pc);
+		for (int j = 0; j < cpu->machine->ncpus; ++j)
+			cpu->machine->cpus[j]->running = false;
+		m88k_pc_to_pointers(cpu);
+		return;
 	}
-	debug(" ]\n");
 
 	/*  Stuff common for all exceptions:  */
 	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFRZ) {
@@ -701,7 +768,7 @@ void m88k_exception(struct cpu *cpu, int vector, int is_trap)
 			fatal("[ SFRZ already set in PSR => ERROR ]\n");
 		}
 
-		update_shadow_regs = 0;
+		update_shadow_regs = false;
 	} else {
 		/*  Freeze shadow registers, and save the PSR:  */
 		cpu->cd.m88k.cr[M88K_CR_EPSR] = cpu->cd.m88k.cr[M88K_CR_PSR];
@@ -767,7 +834,9 @@ void m88k_exception(struct cpu *cpu, int vector, int is_trap)
 
 		case M88K_EXCEPTION_RESET:
 			fatal("[ m88k_exception: reset ]\n");
-			exit(1);
+			for (int j = 0; j < cpu->machine->ncpus; ++j)
+				cpu->machine->cpus[j]->running = false;
+			break;
 
 		case M88K_EXCEPTION_INTERRUPT:
 			/*  When returning with rte, we want to re-  */
@@ -809,9 +878,20 @@ void m88k_exception(struct cpu *cpu, int vector, int is_trap)
 			break;
 #endif
 
-		default:fatal("m88k_exception(): 0x%x: TODO\n", vector);
-			fflush(stdout);
-			exit(1);
+#if 0
+		// TODO: Read up more about SFU1 exceptions and how the
+		// control registers look when they happen.
+		case M88K_EXCEPTION_SFU1_PRECISE:
+			cpu->cd.m88k.cr[M88K_CR_SNIP] = 0; //?
+			cpu->cd.m88k.cr[M88K_CR_SFIP] = 0;
+			//cpu->cd.m88k.cr[M88K_CR_SNIP] -= 4;
+			//cpu->cd.m88k.cr[M88K_CR_SFIP] -= 4;
+			break;
+#endif
+
+		default:debugmsg_cpu(cpu, SUBSYS_EXCEPTION, "m88k",
+			    VERBOSITY_ERROR, "internal error");
+			cpu->running = false;
 		}
 	}
 
@@ -836,10 +916,10 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 {
 	int supervisor = cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE;
 	uint32_t iw;
-	char *symbol, *mnem = NULL;
+	const char *symbol, *mnem = NULL;
 	uint64_t offset;
 	uint32_t op26, op10, op11, d, s1, s2, w5, cr6, imm16;
-	int32_t d16, d26, simm16;
+	int32_t d16, d26;
 
 	if (running)
 		dumpaddr = cpu->pc;
@@ -852,7 +932,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	if (cpu->machine->ncpus > 1 && running)
 		debug("cpu%i:\t", cpu->cpu_id);
 
-	debug("%c%08"PRIx32": ",
+	debug("%c%08" PRIx32": ",
 	    cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE? 's' : 'u',
 	    (uint32_t) dumpaddr);
 
@@ -861,12 +941,14 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	else
 		iw = ib[3] + (ib[2]<<8) + (ib[1]<<16) + (ib[0]<<24);
 
-	debug("%08"PRIx32, (uint32_t) iw);
+	debug("%08" PRIx32, (uint32_t) iw);
 
 	if (running && cpu->delay_slot)
 		debug(" (d)");
+	else
+		debug("    ");
 
-	debug("\t");
+	cpu_print_pc_indicator_in_disassembly(cpu, running, dumpaddr);
 
 	op26   = (iw >> 26) & 0x3f;
 	op11   = (iw >> 11) & 0x1f;
@@ -875,7 +957,6 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	s1     = (iw >> 16) & 0x1f;
 	s2     =  iw        & 0x1f;
 	imm16  = iw & 0xffff;
-	simm16 = (int16_t) (iw & 0xffff);
 	w5     = (iw >>  5) & 0x1f;
 	cr6    = (iw >>  5) & 0x3f;
 	d16    = ((int16_t) (iw & 0xffff)) * 4;
@@ -915,23 +996,23 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 			if (symbol != NULL && supervisor)
 				debug("\t; [<%s>]", symbol);
 			else
-				debug("\t; [0x%08"PRIx32"]", tmpaddr);
+				debug("\t; [0x%08" PRIx32"]", tmpaddr);
 			if (op26 >= 0x08) {
 				/*  Store:  */
 				debug(" = ");
 				switch (op26 & 3) {
-				case 0:	debug("0x%016"PRIx64, (uint64_t)
+				case 0:	debug("0x%016" PRIx64, (uint64_t)
 					    ((((uint64_t) cpu->cd.m88k.r[d])
 					    << 32) + ((uint64_t)
 					    cpu->cd.m88k.r[d+1])) );
 					break;
-				case 1:	debug("0x%08"PRIx32,
+				case 1:	debug("0x%08" PRIx32,
 					    (uint32_t) cpu->cd.m88k.r[d]);
 					break;
-				case 2:	debug("0x%04"PRIx16,
+				case 2:	debug("0x%04" PRIx16,
 					    (uint16_t) cpu->cd.m88k.r[d]);
 					break;
-				case 3:	debug("0x%02"PRIx8,
+				case 3:	debug("0x%02" PRIx8,
 					    (uint8_t) cpu->cd.m88k.r[d]);
 					break;
 				}
@@ -968,7 +1049,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				if (symbol != NULL && supervisor)
 					debug("\t; [<%s>]", symbol);
 				else
-					debug("\t; [0x%08"PRIx32"]", tmpaddr);
+					debug("\t; [0x%08" PRIx32"]", tmpaddr);
 			}
 		}
 		debug("\n");
@@ -982,6 +1063,9 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 	case 0x15:	/*  xor.u   */
 	case 0x16:	/*  or      */
 	case 0x17:	/*  or.u    */
+		if (d == M88K_ZERO_REG)
+			debug("nop");
+
 		switch (op26) {
 		case 0x10:
 		case 0x11:	mnem = "and"; break;
@@ -992,8 +1076,13 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 		case 0x16:
 		case 0x17:	mnem = "or"; break;
 		}
-		debug("%s%s\t", mnem, op26 & 1? ".u" : "");
-		debug("r%i,r%i,0x%x", d, s1, imm16);
+
+		if (d != M88K_ZERO_REG || op26 != 0x16 || s1 != M88K_ZERO_REG) {
+			if (d == M88K_ZERO_REG)
+				debug("\t\t; weird nop encoding: ");
+			debug("%s%s\t", mnem, op26 & 1? ".u" : "");
+			debug("r%i,r%i,0x%x", d, s1, imm16);
+		}
 
 		if (op26 == 0x16 && d != M88K_ZERO_REG) {
 			/*
@@ -1024,7 +1113,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				if (symbol != NULL && supervisor)
 					debug("<%s>", symbol);
 				else
-					debug("0x%08"PRIx32, tmpaddr);
+					debug("0x%08" PRIx32, tmpaddr);
 			}
 		}
 
@@ -1054,17 +1143,19 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 
 	case 0x20:
 		if ((iw & 0x001ff81f) == 0x00004000) {
-			debug("ldcr\tr%i,%s\n", d,
-			    m88k_cr_name(cpu, cr6));
+			debug("ldcr\tr%i,%s", d, m88k_cr_name(cpu, cr6));
+			if (running)
+				debug("\t\t; %s = 0x%08x", m88k_cr_name(cpu, cr6), cpu->cd.m88k.cr[cr6]);
+			debug("\n");
 		} else if ((iw & 0x001ff81f) == 0x00004800) {
-			debug("fldcr\tr%i,%s\n", d,
-			    m88k_fcr_name(cpu, cr6));
+			debug("fldcr\tr%i,%s\n", d, m88k_fcr_name(cpu, cr6));
 		} else if ((iw & 0x03e0f800) == 0x00008000) {
-			debug("stcr\tr%i,%s", s1,
-			    m88k_cr_name(cpu, cr6));
+			debug("stcr\tr%i,%s", s1, m88k_cr_name(cpu, cr6));
 			if (s1 != s2)
 				debug("\t\t; NOTE: weird encoding: "
 				    "low 5 bits = 0x%02x", s2);
+			if (running)
+				debug("\t\t; r%i = 0x%08x", s1, cpu->cd.m88k.r[s1]);
 			debug("\n");
 		} else if ((iw & 0x03e0f800) == 0x00008800) {
 			debug("fstcr\tr%i,%s", s1,
@@ -1146,7 +1237,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 		debug("b%sr%s\t",
 		    op26 >= 0x32? "s" : "",
 		    op26 & 1? ".n" : "");
-		debug("0x%08"PRIx32, (uint32_t) (dumpaddr + d26));
+		debug("0x%08" PRIx32, (uint32_t) (dumpaddr + d26));
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    dumpaddr + d26, &offset);
 		if (symbol != NULL && supervisor)
@@ -1175,6 +1266,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 			case 0x1: debug("gt0"); break;
 			case 0x2: debug("eq0"); break;
 			case 0x3: debug("ge0"); break;
+			case 0x5: debug("not_maxneg_nor_zero"); break;
 			case 0x7: debug("not_maxneg"); break;
 			case 0x8: debug("maxneg"); break;
 			case 0xc: debug("lt0"); break;
@@ -1185,7 +1277,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 		} else {
 			debug("%i", d);
 		}
-		debug(",r%i,0x%08"PRIx32, s1, (uint32_t) (dumpaddr + d16));
+		debug(",r%i,0x%08" PRIx32, s1, (uint32_t) (dumpaddr + d16));
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    dumpaddr + d16, &offset);
 		if (symbol != NULL && supervisor)
@@ -1226,7 +1318,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				if (symbol != NULL && supervisor)
 					debug("\t; [<%s>]", symbol);
 				else
-					debug("\t; [0x%08"PRIx32"]", tmpaddr);
+					debug("\t; [0x%08" PRIx32"]", tmpaddr);
 			}
 
 			debug("\n");
@@ -1315,7 +1407,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				if (symbol != NULL && supervisor)
 					debug("\t; [<%s>]", symbol);
 				else
-					debug("\t; [0x%08"PRIx32"]", tmpaddr);
+					debug("\t; [0x%08" PRIx32"]", tmpaddr);
 			}
 
 			debug("\n");
@@ -1393,7 +1485,17 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 			case 0xa0: mnem = "mak"; break;
 			case 0xa8: mnem = "rot"; break;
 			}
-			debug("%s\tr%i,r%i,r%i\n", mnem, d, s1, s2);
+
+			if (((iw >> 8) & 0xff) == 0x58 && d == M88K_ZERO_REG)
+				debug("nop");
+
+			if (((iw >> 8) & 0xff) != 0x58 || d != M88K_ZERO_REG) {
+				if (d == M88K_ZERO_REG)
+					debug("\t\t; weird nop encoding: ");
+				debug("%s\tr%i,r%i,r%i", mnem, d, s1, s2);
+			}
+
+			debug("\n");
 			break;
 		case 0xc0:	/*  jmp  */
 		case 0xc4:	/*  jmp.n  */
@@ -1411,7 +1513,7 @@ int m88k_cpu_disassemble_instr(struct cpu *cpu, unsigned char *ib,
 				if (symbol != NULL && supervisor)
 					debug("<%s>", symbol);
 				else
-					debug("0x%08"PRIx32, tmpaddr);
+					debug("0x%08" PRIx32, tmpaddr);
 			}
 			debug("\n");
 			break;
