@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2007-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2007-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -35,16 +35,16 @@
 
 
 #define SYNCH_PC                {                                       \
-                int low_pc = ((size_t)ic - (size_t)cpu->cd.m88k.cur_ic_page) \
+                int low_pc_ = ((size_t)ic - (size_t)cpu->cd.m88k.cur_ic_page) \
                     / sizeof(struct m88k_instr_call);                   \
                 cpu->pc &= ~((M88K_IC_ENTRIES_PER_PAGE-1)               \
                     << M88K_INSTR_ALIGNMENT_SHIFT);                     \
-                cpu->pc += (low_pc << M88K_INSTR_ALIGNMENT_SHIFT);      \
+                cpu->pc += (low_pc_ << M88K_INSTR_ALIGNMENT_SHIFT);      \
         }
 
-#define	ABORT_EXECUTION	  {	cpu->cd.m88k.next_ic = &nothing_call;	\
-				cpu->running = 0;			\
-				debugger_n_steps_left_before_interaction = 0; }
+#define	ABORT_EXECUTION	  {	SYNCH_PC;				\
+				cpu->cd.m88k.next_ic = &nothing_call;	\
+				cpu->running = false; }
 
 
 /*
@@ -301,6 +301,7 @@ X(ff1)
 #define M88K_LOADSTORE_SCALEDNESS       32
 #define M88K_LOADSTORE_USR              64
 #define M88K_LOADSTORE_REGISTEROFFSET   128
+#define	M88K_LOADSTORE_NO_PC_SYNC	256
 
 
 /*
@@ -514,6 +515,10 @@ static void m88k_rot(struct cpu *cpu, struct m88k_instr_call *ic, int n)
 
 	reg(ic->arg[0]) = x;
 }
+X(rot_imm)
+{
+	m88k_rot(cpu, ic, ic->arg[2] & 0x1f);
+}
 X(rot)
 {
 	m88k_rot(cpu, ic, reg(ic->arg[2]) & 0x1f);
@@ -648,10 +653,12 @@ X(sub_imm)
  *  addu:   	d = s1 + s2
  *  addu_co:   	d = s1 + s2		carry out
  *  addu_ci:   	d = s1 + s2 + carry	carry in
+ *  addu_cio:   d = s1 + s2 + carry	carry in & carry out
  *  lda_reg_X:	same as addu, but s2 is scaled by 2, 4, or 8
  *  subu:   	d = s1 - s2
  *  subu_co:   	d = s1 - s2		carry/borrow out
  *  subu_ci:   	d = s1 - s2 - (carry? 0 : 1)	carry in
+ *  subu_cio:   d = s1 - s2 - (carry? 0 : 1)	carry in & carry/borrow out
  *  mul:    	d = s1 * s2
  *  divu:   	d = s1 / s2		(unsigned)
  *  div:    	d = s1 / s2		(signed)
@@ -733,6 +740,17 @@ X(addu_co)
 	if ((a >> 32) & 1)
 		cpu->cd.m88k.cr[M88K_CR_PSR] |= M88K_PSR_C;
 }
+X(addu_cio)
+{
+	uint64_t a = reg(ic->arg[1]), b = reg(ic->arg[2]);
+	a += b;
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_C)
+		a ++;
+	reg(ic->arg[0]) = a;
+	cpu->cd.m88k.cr[M88K_CR_PSR] &= ~M88K_PSR_C;
+	if ((a >> 32) & 1)
+		cpu->cd.m88k.cr[M88K_CR_PSR] |= M88K_PSR_C;
+}
 X(addu_ci)
 {
 	uint32_t result = reg(ic->arg[1]) + reg(ic->arg[2]);
@@ -740,10 +758,21 @@ X(addu_ci)
 		result ++;
 	reg(ic->arg[0]) = result;
 }
+X(subu_cio)
+{
+	uint64_t a = reg(ic->arg[1]), b = reg(ic->arg[2]) ^ 0xffffffff;
+	a += b;
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_C)
+		a ++;
+	reg(ic->arg[0]) = a;
+	cpu->cd.m88k.cr[M88K_CR_PSR] &= ~M88K_PSR_C;
+	if ((a >> 32) & 1)
+		cpu->cd.m88k.cr[M88K_CR_PSR] |= M88K_PSR_C;
+}
 X(subu_co)
 {
-	uint64_t a = reg(ic->arg[1]), b = reg(ic->arg[2]);
-	a -= b;
+	uint64_t a = reg(ic->arg[1]), b = reg(ic->arg[2]) ^ 0xffffffff;
+	a += b + 1;
 	reg(ic->arg[0]) = a;
 	cpu->cd.m88k.cr[M88K_CR_PSR] &= ~M88K_PSR_C;
 	if ((a >> 32) & 1)
@@ -751,9 +780,9 @@ X(subu_co)
 }
 X(subu_ci)
 {
-	uint32_t result = reg(ic->arg[1]) - reg(ic->arg[2]);
+	uint32_t result = reg(ic->arg[1]) + ~reg(ic->arg[2]);
 	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_C)
-		result --;
+		result ++;
 	reg(ic->arg[0]) = result;
 }
 
@@ -769,9 +798,12 @@ X(ldcr)
 {
 	SYNCH_PC;
 
-	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE)
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE) {
 		m88k_ldcr(cpu, (uint32_t *) (void *) ic->arg[0], ic->arg[1]);
-	else
+
+		if (!cpu->running)
+			ABORT_EXECUTION;
+	} else
 		m88k_exception(cpu, M88K_EXCEPTION_PRIVILEGE_VIOLATION, 0);
 }
 X(fldcr)
@@ -799,18 +831,24 @@ X(stcr)
 {
 	SYNCH_PC;
 
-	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE)
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE) {
 		m88k_stcr(cpu, reg(ic->arg[0]), ic->arg[1], 0);
-	else
+
+		if (!cpu->running)
+			ABORT_EXECUTION;
+	} else
 		m88k_exception(cpu, M88K_EXCEPTION_PRIVILEGE_VIOLATION, 0);
 }
 X(fstcr)
 {
 	SYNCH_PC;
 
-	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE || ic->arg[1] >= 62)
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE || ic->arg[1] >= 62) {
 		m88k_fstcr(cpu, reg(ic->arg[0]), ic->arg[1]);
-	else {
+
+		if (!cpu->running)
+			ABORT_EXECUTION;
+	} else {
 		/*  TODO: The manual says "floating point privilege
 		    violation", not just "privilege violation"!  */
 		m88k_exception(cpu, M88K_EXCEPTION_PRIVILEGE_VIOLATION, 0);
@@ -848,10 +886,33 @@ X(fadd_sss)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_S, 0);
+	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_S);
 
 	reg(ic->arg[0]) = d;
 }
+X(fadd_dss)
+{
+	struct ieee_float_value f1;
+	struct ieee_float_value f2;
+	uint64_t d;
+	uint32_t s2 = reg(ic->arg[2]);
+	uint32_t s1 = reg(ic->arg[1]);
+
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFD1) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FUNIMP;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
+	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
+
+	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D);
+
+	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
+	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
+}
 X(fadd_dsd)
 {
 	struct ieee_float_value f1;
@@ -871,7 +932,7 @@ X(fadd_dsd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -895,7 +956,7 @@ X(fadd_dds)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -920,11 +981,32 @@ X(fadd_ddd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f + f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
 }
+X(fsub_sss)
+{
+	struct ieee_float_value f1;
+	struct ieee_float_value f2;
+	uint32_t d;
+	uint32_t s2 = reg(ic->arg[2]);
+	uint32_t s1 = reg(ic->arg[1]);
+
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFD1) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FUNIMP;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
+	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
+
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_S);
+	reg(ic->arg[0]) = d;
+}
 X(fsub_sds)
 {
 	struct ieee_float_value f1;
@@ -944,7 +1026,7 @@ X(fsub_sds)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_S, 0);
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_S);
 	reg(ic->arg[0]) = d;
 }
 X(fsub_dss)
@@ -965,7 +1047,7 @@ X(fsub_dss)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -989,7 +1071,7 @@ X(fsub_dsd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1013,7 +1095,7 @@ X(fsub_dds)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1038,7 +1120,7 @@ X(fsub_ddd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f - f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1061,7 +1143,7 @@ X(fmul_sss)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_S, 0);
+	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_S);
 	reg(ic->arg[0]) = d;
 }
 X(fmul_dss)
@@ -1082,7 +1164,7 @@ X(fmul_dss)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1106,7 +1188,7 @@ X(fmul_dsd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1130,7 +1212,7 @@ X(fmul_dds)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
 
-	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1155,7 +1237,7 @@ X(fmul_ddd)
 	ieee_interpret_float_value(s1, &f1, IEEE_FMT_D);
 	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
 
-	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f * f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1185,10 +1267,40 @@ X(fdiv_sss)
 		return;
 	}
 
-	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_S);
 
 	reg(ic->arg[0]) = d;
 }
+X(fdiv_dss)
+{
+	struct ieee_float_value f1;
+	struct ieee_float_value f2;
+	uint64_t d;
+	uint32_t s1 = reg(ic->arg[1]);
+	uint32_t s2 = reg(ic->arg[2]);
+
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFD1) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FUNIMP;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
+	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
+
+	if (f2.f == 0) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FDVZ;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D);
+
+	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
+	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
+}
 X(fdiv_dsd)
 {
 	struct ieee_float_value f1;
@@ -1215,7 +1327,7 @@ X(fdiv_dsd)
 		return;
 	}
 
-	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1247,7 +1359,7 @@ X(fdiv_ddd)
 		return;
 	}
 
-	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D, 0);
+	d = ieee_store_float_value(f1.f / f2.f, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = d >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = d;	/*  and low word.  */
@@ -1271,17 +1383,22 @@ static uint32_t m88k_fcmp_common(struct ieee_float_value *f1,
 	/*  TODO: Implement all bits correctly, e.g. "in range" bits.  */
 	d = 0;
 	if (isnan(f1->f) || isnan(f2->f))
+		// This bit is set if the two numbers are not comparable.
 		d |= (1 << 0);
 	else {
+		// This bit is set if the two numbers are comparable.
 		d |= (1 << 1);
+
 		if (f1->f == f2->f)
 			d |= (1 << 2);
 		else
 			d |= (1 << 3);
+
 		if (f1->f > f2->f)
 			d |= (1 << 4);
 		else
 			d |= (1 << 5);
+
 		if (f1->f < f2->f)
 			d |= (1 << 6);
 		else
@@ -1290,6 +1407,43 @@ static uint32_t m88k_fcmp_common(struct ieee_float_value *f1,
 
 	return d;
 }
+X(fcmp_sss)
+{
+	struct ieee_float_value f1;
+	struct ieee_float_value f2;
+	uint32_t s2 = reg(ic->arg[2]);
+	uint32_t s1 = reg(ic->arg[1]);
+
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFD1) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FUNIMP;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
+	ieee_interpret_float_value(s2, &f2, IEEE_FMT_S);
+	reg(ic->arg[0]) = m88k_fcmp_common(&f1, &f2);
+}
+X(fcmp_ssd)
+{
+	struct ieee_float_value f1;
+	struct ieee_float_value f2;
+	uint64_t s2 = reg(ic->arg[2]);
+	uint32_t s1 = reg(ic->arg[1]);
+	s2 = (s2 << 32) + reg(ic->arg[2] + 4);
+
+	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_SFD1) {
+		SYNCH_PC;
+		cpu->cd.m88k.fcr[M88K_FPCR_FPECR] = M88K_FPECR_FUNIMP;
+		m88k_exception(cpu, M88K_EXCEPTION_SFU1_PRECISE, 0);
+		return;
+	}
+
+	ieee_interpret_float_value(s1, &f1, IEEE_FMT_S);
+	ieee_interpret_float_value(s2, &f2, IEEE_FMT_D);
+	reg(ic->arg[0]) = m88k_fcmp_common(&f1, &f2);
+}
 X(fcmp_sds)
 {
 	struct ieee_float_value f1;
@@ -1350,7 +1504,7 @@ X(flt_ss)
 		return;
 	}
 
-	reg(ic->arg[0]) = ieee_store_float_value((double)x, IEEE_FMT_S, 0);
+	reg(ic->arg[0]) = ieee_store_float_value((double)x, IEEE_FMT_S);
 }
 X(flt_ds)
 {
@@ -1364,7 +1518,7 @@ X(flt_ds)
 		return;
 	}
 
-	result = ieee_store_float_value((double)x, IEEE_FMT_D, 0);
+	result = ieee_store_float_value((double)x, IEEE_FMT_D);
 
 	reg(ic->arg[0]) = result >> 32;	/*  High 32-bit word,  */
 	reg(ic->arg[0] + 4) = result;	/*  and low word.  */
@@ -1426,8 +1580,12 @@ X(xcr)
 	if (cpu->cd.m88k.cr[M88K_CR_PSR] & M88K_PSR_MODE) {
 		tmp = reg(ic->arg[1]);
 		m88k_ldcr(cpu, &tmp2, ic->arg[2]);
-		m88k_stcr(cpu, tmp, ic->arg[2], 0);
-		reg(ic->arg[0]) = tmp2;
+		if (!cpu->running) {
+			ABORT_EXECUTION;
+		} else {
+			m88k_stcr(cpu, tmp, ic->arg[2], 0);
+			reg(ic->arg[0]) = tmp2;
+		}
 	} else
 		m88k_exception(cpu, M88K_EXCEPTION_PRIVILEGE_VIOLATION, 0);
 }
@@ -1448,12 +1606,14 @@ X(rte)
 	m88k_stcr(cpu, cpu->cd.m88k.cr[M88K_CR_EPSR], M88K_CR_PSR, 1);
 
 	if (cpu->cd.m88k.cr[M88K_CR_SNIP] & M88K_NIP_E) {
-		fatal("rte: NIP: TODO: single-step support\n");
+		debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    "rte: NIP: TODO: single-step support");
 		goto abort_dump;
 	}
 
 	if (cpu->cd.m88k.cr[M88K_CR_SFIP] & M88K_FIP_E) {
-		fatal("rte: TODO: FIP single-step support\n");
+		debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    "rte: TODO: FIP single-step support");
 		goto abort_dump;
 	}
 
@@ -1466,8 +1626,9 @@ X(rte)
 		if (!(cpu->cd.m88k.cr[M88K_CR_SFIP] & M88K_FIP_V)) {
 			if ((cpu->cd.m88k.cr[M88K_CR_SFIP] & M88K_FIP_ADDR)
 			    != (cpu->cd.m88k.cr[M88K_CR_SNIP] & M88K_NIP_ADDR) + 4) {
-				fatal("[ TODO: Neither FIP nor NIP has the "
-				    "Valid bit set?! ]\n");
+				debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    		    "rte: TODO: Neither FIP nor NIP has the "
+				    "Valid bit set?!");
 				goto abort_dump;
 			}
 
@@ -1494,16 +1655,18 @@ X(rte)
 		quick_pc_to_pointers(cpu);
 
 		if (cpu->pc != nip) {
-			fatal("NIP execution caused exception?! TODO\n");
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "NIP execution caused exception?! TODO");
 			goto abort_dump;
 		}
 
 		instr(to_be_translated)(cpu, cpu->cd.m88k.next_ic);
 
 		if ((cpu->pc & 0xfffff000) != (nip & 0xfffff000)) {
-			fatal("instruction in delay slot when returning via"
+			debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+			    "instruction in delay slot when returning via"
 			    " rte caused an exception?! nip=0x%08x, but pc "
-			    "changed to 0x%08x! TODO\n", nip, (int)cpu->pc);
+			    "changed to 0x%08x! TODO", nip, (int)cpu->pc);
 			goto abort_dump;
 		}
 
@@ -1513,14 +1676,15 @@ X(rte)
 		return;
 	}
 
-	/*  fatal("RTE: NIP=0x%08"PRIx32", FIP=0x%08"PRIx32"\n",
+	/*  fatal("RTE: NIP=0x%08" PRIx32", FIP=0x%08" PRIx32"\n",
 	    cpu->cd.m88k.cr[M88K_CR_SNIP], cpu->cd.m88k.cr[M88K_CR_SFIP]);  */
 
 	quick_pc_to_pointers(cpu);
 	return;
 
 abort_dump:
-	fatal("RTE failed. NIP=0x%08"PRIx32", FIP=0x%08"PRIx32"\n",
+	debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+	    "rte failed. NIP=0x%08" PRIx32", FIP=0x%08" PRIx32,
 	    cpu->cd.m88k.cr[M88K_CR_SNIP], cpu->cd.m88k.cr[M88K_CR_SFIP]);
 
 	ABORT_EXECUTION;
@@ -1531,94 +1695,104 @@ abort_dump:
  *  xmem_slow:  Unoptimized xmem (exchange register with memory)
  *
  *  arg[0] = copy of the instruction word
+ *
+ *  NOTE: The alternative would be to extend the load/store instructions
+ *  to also include xmem, and thus implement a "fast" version (where
+ *  the page is already accessible both readable and writable) and a
+ *  generic version. (The generic implementation would then not need to
+ *  have a NO_PC_SYNC flag, since they would always sync.)
  */
 X(xmem_slow)
 {
-	uint32_t iword = ic->arg[0], addr;
-	uint8_t tmp[4];
-	uint8_t data[4];
+	uint32_t iword = ic->arg[0];
+	uint32_t tmp;
 	int d      = (iword >> 21) & 0x1f;
 	int s1     = (iword >> 16) & 0x1f;
 	int s2     =  iword        & 0x1f;
 	int imm16  =  iword        & 0xffff;
+	int regofs = (iword & 0xf0000000) != 0;
 	int scaled = iword & 0x200;
 	int size   = iword & 0x400;
 	int user   = iword & 0x80;
+	uint32_t pc_before_memory_access;
+	void (*xmem_load)(struct cpu*, struct m88k_instr_call*);
+	void (*xmem_store)(struct cpu*, struct m88k_instr_call*);
+	struct m88k_instr_call call;
 
 	SYNCH_PC;
+	pc_before_memory_access = (uint32_t)cpu->pc;
 
-	if (user) {
-		fatal("xmem_slow: user: not yet (TODO)\n");
-		exit(1);
-	}
-
-	if ((iword & 0xf0000000) == 0) {
+	if (!regofs) {
 		/*  immediate offset:  */
-		addr = imm16;
 		scaled = 0;
 		size = (iword >> 26) & 1;
 		user = 0;
-	} else {
-		/*  register offset:  */
-		addr = cpu->cd.m88k.r[s2];
-		if (scaled && size)
-			addr *= sizeof(uint32_t);
 	}
 
-	addr += cpu->cd.m88k.r[s1];
+	tmp = cpu->cd.m88k.r[d];
 
-	if (size) {
-		uint32_t x = cpu->cd.m88k.r[d];
-		if (cpu->byte_order == EMUL_LITTLE_ENDIAN)
-			x = LE32_TO_HOST(x);
-		else
-			x = BE32_TO_HOST(x);
+	xmem_load = m88k_loadstore[ (size? 2 : 0)
+	    + (cpu->byte_order == EMUL_BIG_ENDIAN? M88K_LOADSTORE_ENDIANNESS : 0)
+	    + (scaled? M88K_LOADSTORE_SCALEDNESS : 0)
+	    + (user? M88K_LOADSTORE_USR : 0)
+	    + (regofs? M88K_LOADSTORE_REGISTEROFFSET : 0)
+	    + M88K_LOADSTORE_NO_PC_SYNC ];
 
-		{
-			uint32_t *p = (uint32_t *) tmp;
-			*p = x;
-		}
+	call.f = xmem_load;
+	call.arg[0] = (size_t) &cpu->cd.m88k.r[d];
+	call.arg[1] = (size_t) &cpu->cd.m88k.r[s1];
+	if (regofs)
+		call.arg[2] = (size_t) &cpu->cd.m88k.r[s2];
+	else
+		call.arg[2] = imm16;
 
-		if (addr & 3) {
-			m88k_exception(cpu,
-			    M88K_EXCEPTION_MISALIGNED_ACCESS, 0);
-			return;
-		}
-	} else
-		tmp[0] = cpu->cd.m88k.r[d];
+	if (d == M88K_ZERO_REG)
+		call.arg[0] = (size_t)&cpu->cd.m88k.zero_scratch;
 
-	if (!cpu->memory_rw(cpu, cpu->mem, addr, (uint8_t *) &data,
-	    size? 4 : 1, MEM_READ, CACHE_DATA)) {
-		/*  Exception.  */
+	xmem_load(cpu, &call);
 
-		fatal("XMEM exception: TODO: update the transaction"
-		    " registers!\n");
-		exit(1);
-		/*  return;  */
+	// If there was an exception in the load or store call, then the pc
+	// will have changed. Alternatively, one could check for changes in
+	// the transaction registers. (But handling the theoretical tiny risk
+	// that the pc was the same before and after, i.e. that the faulting
+	// instruction was the same as the exception handler, is not worth it.)
+	if ((uint32_t)cpu->pc != pc_before_memory_access) {
+		//printf("xmem_load exception, pc_before_memory_access = %08x\n", pc_before_memory_access);
+		cpu->cd.m88k.dmd[0] = tmp;
+		cpu->cd.m88k.dmt[0] |= DMT_LOCKBAR | DMT_WRITE;
+		cpu->cd.m88k.dmt[0] &= ~((0x1f) << DMT_DREGSHIFT);
+		cpu->cd.m88k.dmt[0] |= d << DMT_DREGSHIFT;
+		return;
 	}
 
-	if (!cpu->memory_rw(cpu, cpu->mem, addr, (uint8_t *) &tmp,
-	    size? 4 : 1, MEM_WRITE, CACHE_DATA)) {
-		/*  Exception.  */
+	xmem_store = m88k_loadstore[ (size? 2 : 0)
+	    + M88K_LOADSTORE_STORE
+	    + (cpu->byte_order == EMUL_BIG_ENDIAN? M88K_LOADSTORE_ENDIANNESS : 0)
+	    + (scaled? M88K_LOADSTORE_SCALEDNESS : 0)
+	    + (user? M88K_LOADSTORE_USR : 0)
+	    + (regofs? M88K_LOADSTORE_REGISTEROFFSET : 0)
+	    + M88K_LOADSTORE_NO_PC_SYNC ];
 
-		fatal("XMEM exception: TODO: update the transaction"
-		    " registers!\n");
-		exit(1);
-		/*  return;  */
+	call.f = xmem_store;
+	call.arg[0] = (size_t) &tmp;
+	// arg[1] and arg[2] are already set up from xmem_load call.
+
+	if (d == M88K_ZERO_REG)
+		tmp = 0;
+
+	xmem_store(cpu, &call);
+
+	// If there was an exception in xmem_store(), then return the d register
+	// to what it was before the xmem_load().
+	if ((uint32_t)cpu->pc != pc_before_memory_access) {
+		//printf("xmem_store exception, pc_before_memory_access = %08x\n", pc_before_memory_access);
+		// cpu->cd.m88k.r[d] = tmp;
+		cpu->cd.m88k.dmd[0] = tmp;
+		cpu->cd.m88k.dmt[0] |= DMT_LOCKBAR | DMT_WRITE;
+		cpu->cd.m88k.dmt[0] &= ~((0x1f) << DMT_DREGSHIFT);
+		cpu->cd.m88k.dmt[0] |= d << DMT_DREGSHIFT;
+		return;
 	}
-
-	if (size) {
-		uint32_t x;
-		uint32_t *p = (uint32_t *) data;
-		x = *p;
-
-		if (cpu->byte_order == EMUL_LITTLE_ENDIAN)
-			x = LE32_TO_HOST(x);
-		else
-			x = BE32_TO_HOST(x);
-		cpu->cd.m88k.r[d] = x;
-	} else
-		cpu->cd.m88k.r[d] = data[0];
 }
 
 
@@ -1635,10 +1809,17 @@ X(prom_call)
 	}
 
 	switch (cpu->machine->machine_type) {
+
+	case MACHINE_LUNA88K:
+		luna88kprom_emul(cpu);
+		break;
+
 	case MACHINE_MVME88K:
 		mvmeprom_emul(cpu);
 		break;
-	default:fatal("m88k prom_call: unimplemented machine type\n");
+
+	default:debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+		    "prom_call: unimplemented machine type");
 		ABORT_EXECUTION;
 	}
 
@@ -1714,9 +1895,8 @@ X(idle)
 
 	if (v == 0) {
 		SYNCH_PC;
-		usleep(50);
-		cpu->has_been_idling = 1;
-		cpu->n_translated_instrs += N_SAFE_DYNTRANS_LIMIT / 2;
+		cpu->wants_to_idle = true;
+		cpu->n_translated_instrs += N_DYNTRANS_IDLE_BREAK;
 		cpu->cd.m88k.next_ic = &nothing_call;
 	} else {
 		cpu->n_translated_instrs ++;
@@ -1756,9 +1936,8 @@ X(idle_with_tb1)
 
 	if (v == 0) {
 		SYNCH_PC;
-		usleep(50);
-		cpu->has_been_idling = 1;
-		cpu->n_translated_instrs += N_SAFE_DYNTRANS_LIMIT / 2;
+		cpu->wants_to_idle = true;
+		cpu->n_translated_instrs += N_DYNTRANS_IDLE_BREAK;
 		cpu->cd.m88k.next_ic = &nothing_call;
 	} else {
 		cpu->n_translated_instrs += 2;
@@ -1842,8 +2021,9 @@ X(end_of_page2)
 	if (cpu->delay_slot == NOT_DELAYED)
 		return;
 
-	fatal("end_of_page2: fatal error, we're in a delay slot\n");
-	exit(1);
+	debugmsg_cpu(cpu, SUBSYS_CPU, "m88k", VERBOSITY_ERROR,
+	    "end_of_page2: fatal error, we're in a delay slot");
+	ABORT_EXECUTION;
 }
 
 
@@ -1861,6 +2041,13 @@ X(end_of_page2)
  *  s00079d78: f020d8ff	tb1	1,r0,0xff
  *  s00079d7c: 15ac7320	ld	r13,r12,0x7320	; [<_sched_whichqs>]
  *  s00079d80: e84dfffe	bcnd	eq0,r13,0x00079d78	; <_sched_idle+0x158>
+ *
+ *  or
+ *
+ *  s00244840: f020d8ff	tb1	1,r0,0xff
+ *  s00244844: 15b80168	ld	r13,r24,0x168	; [<m88k_cpus+0x1b0>]
+ *  s00244848: ec4dfffe	bcnd.n	eq0,r13,0x00244840	; <sched_idle+0x100>
+ *  s0024484c: 5853fd80 (d)	or	r2,r19,0xfd80
  */
 void COMBINE(idle)(struct cpu *cpu, struct m88k_instr_call *ic, int low_addr)
 {
@@ -1888,6 +2075,24 @@ void COMBINE(idle)(struct cpu *cpu, struct m88k_instr_call *ic, int low_addr)
 		ic[-2].f = instr(idle_with_tb1);
 		return;
 	}
+
+	// NOTE: In this case, we are only checking the tb1, the ld, and
+	// the bcnd.n instruction. It actually executes the instruction after
+	// it: hopefully it is an 'or', but in principle it could be anything.
+	// Here we take a chance and hope that it is not something weird that
+	// would cause an exception. (It is unlikely that this particular loop
+	// construct, used for idle loops, would have something else after it.)
+	if (ic[0].f == instr(bcnd_n_eq0) &&
+	    ic[0].arg[2] == (size_t) low_addr - 8 &&
+	    ic[-2].f == instr(tb1) &&
+	    ic[-2].arg[1] == (size_t) &cpu->cd.m88k.r[M88K_ZERO_REG] &&
+	    ic[-1].f == instr(ld_u_4_be) &&
+	    ic[0].arg[0] == ic[-1].arg[0] &&
+	    ic[0].arg[0] != (size_t) &cpu->cd.m88k.r[M88K_ZERO_REG]) {
+		ic[-2].f = instr(idle_with_tb1);
+		// printf("NEW IDLE using bcnd.n!!!\n");
+		return;
+	}
 }
 
 
@@ -1908,7 +2113,7 @@ X(to_be_translated)
 	unsigned char *page;
 	unsigned char ib[4];
 	uint32_t op26, op10, op11, d, s1, s2, cr6, imm16;
-	int32_t d16, d26, simm16;
+	int32_t d16, d26; //, simm16;
 	int offset, shift;
 	int in_crosspage_delayslot = 0;
 	void (*samepage_function)(struct cpu *, struct m88k_instr_call *)=NULL;
@@ -1927,7 +2132,7 @@ X(to_be_translated)
 	addr = cpu->pc & ~((M88K_IC_ENTRIES_PER_PAGE-1)
 	    << M88K_INSTR_ALIGNMENT_SHIFT);
 	addr += (low_pc << M88K_INSTR_ALIGNMENT_SHIFT);
-	cpu->pc = (MODE_int_t)addr;
+	cpu->pc = (uint32_t)addr;
 	addr &= ~((1 << M88K_INSTR_ALIGNMENT_SHIFT) - 1);
 
 	/*  Read the instruction word from memory:  */
@@ -1970,8 +2175,10 @@ X(to_be_translated)
 	 *  scratch register instead).
 	 */
 	if (cpu->cd.m88k.r[M88K_ZERO_REG] != 0) {
-		fatal("INTERNAL ERROR! M88K_ZERO_REG != 0?\n");
-		exit(1);
+		debugmsg_cpu(cpu, SUBSYS_CPU, "to_be_translated", VERBOSITY_ERROR,
+		    "INTERNAL ERROR! M88K_ZERO_REG != 0");
+		cpu->running = false;
+		goto bad;
 	}
 
 	op26   = (iword >> 26) & 0x3f;
@@ -1981,7 +2188,7 @@ X(to_be_translated)
 	s1     = (iword >> 16) & 0x1f;
 	s2     =  iword        & 0x1f;
 	imm16  =  iword        & 0xffff;
-	simm16 = (int16_t) (iword & 0xffff);
+	// simm16 = (int16_t) (iword & 0xffff);
 	cr6    = (iword >>  5) & 0x3f;
 	d16    = ((int16_t) (iword & 0xffff)) * 4;
 	d26    = ((int32_t)((iword & 0x03ffffff) << 6)) >> 4;
@@ -2039,6 +2246,9 @@ X(to_be_translated)
 			    + (signedness? M88K_LOADSTORE_SIGNEDNESS:0)
 			    + (cpu->byte_order == EMUL_BIG_ENDIAN?
 			       M88K_LOADSTORE_ENDIANNESS : 0) ];
+
+			if (!store && d == 0)
+				ic->arg[0] = (size_t) &cpu->cd.m88k.zero_scratch;
 		}
 		break;
 
@@ -2196,6 +2406,7 @@ X(to_be_translated)
 			ic->arg[2] = (size_t) &cpu->cd.m88k.r[s2];
 			switch ((iword >> 5) & 0x3f) {
 			case 0x00:	ic->f = instr(fadd_sss); break;
+			case 0x01:	ic->f = instr(fadd_dss); break;
 			case 0x05:	ic->f = instr(fadd_dsd); break;
 			case 0x11:	ic->f = instr(fadd_dds); break;
 			case 0x15:	ic->f = instr(fadd_ddd); break;
@@ -2216,6 +2427,7 @@ X(to_be_translated)
 			ic->arg[1] = (size_t) &cpu->cd.m88k.r[s1];
 			ic->arg[2] = (size_t) &cpu->cd.m88k.r[s2];
 			switch ((iword >> 5) & 0x3f) {
+			case 0x00:	ic->f = instr(fsub_sss); break;
 			case 0x01:	ic->f = instr(fsub_dss); break;
 			case 0x05:	ic->f = instr(fsub_dsd); break;
 			case 0x10:	ic->f = instr(fsub_sds); break;
@@ -2238,6 +2450,8 @@ X(to_be_translated)
 			ic->arg[1] = (size_t) &cpu->cd.m88k.r[s1];
 			ic->arg[2] = (size_t) &cpu->cd.m88k.r[s2];
 			switch ((iword >> 5) & 0x3f) {
+			case 0x00:	ic->f = instr(fcmp_sss); break;
+			case 0x04:	ic->f = instr(fcmp_ssd); break;
 			case 0x10:	ic->f = instr(fcmp_sds); break;
 			case 0x14:	ic->f = instr(fcmp_sdd); break;
 			default:if (!cpu->translation_readahead)
@@ -2277,6 +2491,7 @@ X(to_be_translated)
 			ic->arg[2] = (size_t) &cpu->cd.m88k.r[s2];
 			switch ((iword >> 5) & 0x3f) {
 			case 0x00:	ic->f = instr(fdiv_sss); break;
+			case 0x01:	ic->f = instr(fdiv_dss); break;
 			case 0x05:	ic->f = instr(fdiv_dsd); break;
 			case 0x15:	ic->f = instr(fdiv_ddd); break;
 			default:if (!cpu->translation_readahead)
@@ -2394,7 +2609,8 @@ X(to_be_translated)
 		}
 
 		if ((iword & 0xffe0ffff) == 0xe840ffff ||
-		    (iword & 0xffe0ffff) == 0xe840fffe)
+		    (iword & 0xffe0ffff) == 0xe840fffe ||
+		    (iword & 0xffe0ffff) == 0xec40fffe)
 			cpu->cd.m88k.combination_check = COMBINE(idle);
 
 		break;
@@ -2407,6 +2623,7 @@ X(to_be_translated)
 		case 0x24:	/*  ext  */
 		case 0x26:	/*  extu  */
 		case 0x28:	/*  mak  */
+		case 0x2a:	/*  rot  */
 			ic->arg[0] = (size_t) &cpu->cd.m88k.r[d];
 			ic->arg[1] = (size_t) &cpu->cd.m88k.r[s1];
 			ic->arg[2] = iword & 0x3ff;
@@ -2435,6 +2652,7 @@ X(to_be_translated)
 			case 0x24: ic->f = instr(ext_imm); break;
 			case 0x26: ic->f = instr(extu_imm); break;
 			case 0x28: ic->f = instr(mak_imm); break;
+			case 0x2a: ic->f = instr(rot_imm); break;
 			}
 
 			if (d == M88K_ZERO_REG)
@@ -2525,8 +2743,8 @@ X(to_be_translated)
 				    + (user? M88K_LOADSTORE_USR : 0)
 				    + M88K_LOADSTORE_REGISTEROFFSET ];
 
-				if (op == 0 && d == M88K_ZERO_REG)
-					ic->f = instr(nop);
+				if (d == M88K_ZERO_REG && op == 0)
+					ic->arg[0] = (size_t)&cpu->cd.m88k.zero_scratch;
 
 				if (opsize == 3 && d == 31) {
 					fatal("m88k load/store of register "
@@ -2545,6 +2763,8 @@ X(to_be_translated)
 				} else {
 					ic->f = instr(addu);
 				}
+				if (d == M88K_ZERO_REG)
+					ic->f = instr(nop);
 			} else {
 				/*  xmem:  */
 				ic->f = instr(xmem_slow);
@@ -2562,9 +2782,11 @@ X(to_be_translated)
 		case 0x60:	/*  addu   */
 		case 0x61:	/*  addu.co  */
 		case 0x62:	/*  addu.ci  */
+		case 0x63:	/*  addu.cio */
 		case 0x64:	/*  subu   */
 		case 0x65:	/*  subu.co  */
 		case 0x66:	/*  subu.ci  */
+		case 0x67:	/*  subu.cio */
 		case 0x68:	/*  divu   */
 		case 0x6c:	/*  mul    */
 		case 0x70:	/*  add    */
@@ -2590,9 +2812,11 @@ X(to_be_translated)
 			case 0x60: ic->f = instr(addu);  break;
 			case 0x61: ic->f = instr(addu_co); break;
 			case 0x62: ic->f = instr(addu_ci); break;
+			case 0x63: ic->f = instr(addu_cio); break;
 			case 0x64: ic->f = instr(subu);  break;
 			case 0x65: ic->f = instr(subu_co); break;
 			case 0x66: ic->f = instr(subu_ci); break;
+			case 0x67: ic->f = instr(subu_cio); break;
 			case 0x68: ic->f = instr(divu);  break;
 			case 0x6c: ic->f = instr(mul);   break;
 			case 0x70: ic->f = instr(add);   break;
@@ -2623,10 +2847,13 @@ X(to_be_translated)
 			 */
 			if (d == M88K_ZERO_REG) {
 				int opc = (iword >> 8) & 0xff;
-				if (opc != 0x61 && opc != 0x63 &&
-				    opc != 0x65 && opc != 0x67 &&
-				    opc != 0x71 && opc != 0x73 &&
-				    opc != 0x75 && opc != 0x77)
+				if (opc != 0x61 /* addu.co */ && opc != 0x63 /* addu.cio */ &&
+				    opc != 0x65 /* subu.co */ && opc != 0x67 /* subu.cio */ &&
+				    opc != 0x71 /*  add.co */ && opc != 0x73 /*  add.cio */ &&
+				    opc != 0x75 /*  sub.co */ && opc != 0x77 /*  sub.cio */ &&
+				    opc != 0x68 /*  divu   */ && opc != 0x69 /* divu.d   */ &&
+				    opc != 0x6c /*  mul    */ && opc != 0x6d /* mulu.d   */ &&
+				    opc != 0x6e /*  muls   */ && opc != 0x78 /*  div     */)
 					ic->f = instr(nop);
 				else
 					ic->arg[0] = (size_t)
@@ -2689,7 +2916,7 @@ X(to_be_translated)
 				if (iword == 0xf400fc00)
 					ic->f = instr(rte);
 				else {
-					fatal("unimplemented rte variant: 0x%08"PRIx32"\n", iword);
+					fatal("unimplemented rte variant: 0x%08" PRIx32"\n", iword);
 					goto bad;
 				}
 				break;
