@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -52,22 +52,22 @@
 #include "symbol.h"
 
 
-static char *exception_names[] = EXCEPTION_NAMES;
+static const char *exception_names[] = EXCEPTION_NAMES;
 
-static char *hi6_names[] = HI6_NAMES;
-static char *regimm_names[] = REGIMM_NAMES;
-static char *special_names[] = SPECIAL_NAMES;
-static char *special_rot_names[] = SPECIAL_ROT_NAMES;
-static char *special2_names[] = SPECIAL2_NAMES;
-static char *mmi_names[] = MMI_NAMES;
-static char *mmi0_names[] = MMI0_NAMES;
-static char *mmi1_names[] = MMI1_NAMES;
-static char *mmi2_names[] = MMI2_NAMES;
-static char *mmi3_names[] = MMI3_NAMES;
-static char *special3_names[] = SPECIAL3_NAMES;
+static const char *hi6_names[] = HI6_NAMES;
+static const char *regimm_names[] = REGIMM_NAMES;
+static const char *special_names[] = SPECIAL_NAMES;
+static const char *special_rot_names[] = SPECIAL_ROT_NAMES;
+static const char *special2_names[] = SPECIAL2_NAMES;
+static const char *mmi_names[] = MMI_NAMES;
+static const char *mmi0_names[] = MMI0_NAMES;
+static const char *mmi1_names[] = MMI1_NAMES;
+static const char *mmi2_names[] = MMI2_NAMES;
+static const char *mmi3_names[] = MMI3_NAMES;
+static const char *special3_names[] = SPECIAL3_NAMES;
 
-static char *regnames[] = MIPS_REGISTER_NAMES;
-static char *cop0_names[] = COP0_NAMES;
+static const char *regnames[] = MIPS_REGISTER_NAMES;
+static const char *cop0_names[] = COP0_NAMES;
 
 
 #define DYNTRANS_DUALMODE_32
@@ -91,7 +91,6 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 {
 	int i, found, j, tags_size, n_cache_lines, size_per_cache_line;
 	struct mips_cpu_type_def cpu_type_defs[] = MIPS_CPU_TYPE_DEFS;
-	int64_t secondary_cache_size;
 	int x, linesize;
 
 	/*  Scan the cpu_type_defs list for this cpu type:  */
@@ -110,7 +109,7 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 	cpu->memory_rw                = mips_memory_rw;
 	cpu->cd.mips.cpu_type         = cpu_type_defs[found];
-	cpu->name                     = cpu->cd.mips.cpu_type.name;
+	cpu->name                     = strdup(cpu->cd.mips.cpu_type.name);
 	cpu->byte_order               = EMUL_LITTLE_ENDIAN;
 	cpu->cd.mips.gpr[MIPS_GPR_SP] = INITIAL_STACK_POINTER;
 
@@ -136,9 +135,6 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 	cpu->instruction_has_delayslot = mips_cpu_instruction_has_delayslot;
 
-	if (cpu_id == 0)
-		debug("%s", cpu->cd.mips.cpu_type.name);
-
 	/*
 	 *  CACHES:
 	 *
@@ -215,7 +211,7 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 		cpu->cd.mips.cache_mask[i] = cpu->cd.mips.cache_size[i] - 1;
 
-		CHECK_ALLOCATION(cpu->cd.mips.cache[i] =
+		CHECK_ALLOCATION(cpu->cd.mips.cache[i] = (unsigned char *)
 		    malloc(cpu->cd.mips.cache_size[i]));
 
 		n_cache_lines = cpu->cd.mips.cache_size[i] /
@@ -245,44 +241,18 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 		cpu->cd.mips.cache_last_paddr[i] = IMPOSSIBLE_PADDR;
 	}
 
-	/*
-	 *  Secondary cache:
-	 */
-	secondary_cache_size = 0;
-	if (cpu->cd.mips.cache_secondary)
-		secondary_cache_size = 1 << cpu->cd.mips.cache_secondary;
-	/*  TODO: linesize...  */
-
-	if (cpu_id == 0) {
-		debug(" (I+D = %i+%i KB",
-		    (int)(cpu->cd.mips.cache_size[CACHE_INSTRUCTION] / 1024),
-		    (int)(cpu->cd.mips.cache_size[CACHE_DATA] / 1024));
-
-		if (secondary_cache_size != 0) {
-			debug(", L2 = ");
-			if (secondary_cache_size >= 1048576)
-				debug("%i MB", (int)
-				    (secondary_cache_size / 1048576));
-			else
-				debug("%i KB", (int)
-				    (secondary_cache_size / 1024));
-		}
-
-		debug(")");
-	}
-
 	/*  Register the CPU's interrupts:  */
 	for (i=2; i<8; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char name[50];
 		snprintf(name, sizeof(name), "%s.%i", cpu->path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = 1 << (STATUS_IM_SHIFT + i);
-		template.name = name;
-		template.extra = cpu;
-		template.interrupt_assert = mips_cpu_interrupt_assert;
-		template.interrupt_deassert = mips_cpu_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 1 << (STATUS_IM_SHIFT + i);
+		templ.name = name;
+		templ.extra = cpu;
+		templ.interrupt_assert = mips_cpu_interrupt_assert;
+		templ.interrupt_deassert = mips_cpu_interrupt_deassert;
+		interrupt_handler_register(&templ);
 
 		if (i == 7)
 			INTERRUPT_CONNECT(name, cpu->cd.mips.irq_compare);
@@ -294,21 +264,33 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 	switch (cpu->cd.mips.cpu_type.mmu_model) {
 	case MMU3K:
+		cpu->vaddr_mask = 0x00000000ffffffffULL;
 		cpu->translate_v2p = translate_v2p_mmu3k;
 		break;
 	case MMU8K:
 		cpu->translate_v2p = translate_v2p_mmu8k;
 		break;
 	case MMU10K:
+		cpu->vaddr_mask = ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK_R10K | 0x1fff;
 		cpu->translate_v2p = translate_v2p_mmu10k;
 		break;
 	default:
+		if (cpu->is_32bit)
+			cpu->vaddr_mask = 0x00000000ffffffffULL;
+		else
+			cpu->vaddr_mask = ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK | 0x1fff;
+
 		if (cpu->cd.mips.cpu_type.rev == MIPS_R4100)
 			cpu->translate_v2p = translate_v2p_mmu4100;
 		else
 			cpu->translate_v2p = translate_v2p_generic;
 	}
 
+	// TODO: Consider moving to src/machine/machine_test.c?
+	// This is only here to make the floating point tests work out-of-the-box.
+	if (!cpu->is_32bit)
+		cpu->cd.mips.coproc[0]->reg[COP0_STATUS] |= STATUS_FR;
+
 	if (cpu->machine->prom_emulation) {
 		/*
 		 *  Default behaviour of jumping to 0xbfc00000 should be
@@ -341,15 +323,26 @@ int mips_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
  *
  *  Debug dump of MIPS-specific CPU data for specific CPU.
  */
-void mips_cpu_dumpinfo(struct cpu *cpu)
+void mips_cpu_dumpinfo(struct cpu *cpu, bool verbose)
 {
-	int iadd = DEBUG_INDENTATION;
 	struct mips_cpu_type_def *ct = &cpu->cd.mips.cpu_type;
+	int secondary_cache_size = 0;
+	if (cpu->cd.mips.cache_secondary)
+		secondary_cache_size = 1 << cpu->cd.mips.cache_secondary;
 
-	debug_indentation(iadd);
+	if (!verbose) {
+		debugmsg(SUBSYS_MACHINE, "cpu", VERBOSITY_INFO,
+		    "%s (%i-bit, I+D = %i+%i KB, L2 = %i KB)",
+		    cpu->name,
+		    cpu->is_32bit? 32 : 64,
+		    (int)(cpu->cd.mips.cache_size[CACHE_INSTRUCTION] / 1024),
+		    (int)(cpu->cd.mips.cache_size[CACHE_DATA] / 1024),
+		    (int)(secondary_cache_size / 1024));
 
-	debug("\n%i-bit %s-endian (MIPS",
-	    cpu->is_32bit? 32 : 64,
+		return;
+	}
+
+	debug("%s-endian (MIPS",
 	    cpu->byte_order == EMUL_BIG_ENDIAN? "Big" : "Little");
 
 	switch (ct->isa_level) {
@@ -405,8 +398,6 @@ void mips_cpu_dumpinfo(struct cpu *cpu)
 			debug(", direct-mapped");
 		debug("\n");
 	}
-
-	debug_indentation(-iadd);
 }
 
 
@@ -488,24 +479,24 @@ int mips_cpu_instruction_has_delayslot(struct cpu *cpu, unsigned char *ib)
  *  mips_cpu_tlbdump():
  *
  *  Called from the debugger to dump the TLB in a readable format.
- *  x is the cpu number to dump, or -1 to dump all CPUs.
  *
  *  If rawflag is nonzero, then the TLB contents isn't formated nicely,
  *  just dumped.
  */
-void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
+void mips_cpu_tlbdump(struct cpu *cpu, int rawflag)
 {
 	int i, j;
+	struct machine* m = cpu->machine;
 
 	/*  Raw output:  */
 	if (rawflag) {
 		for (i=0; i<m->ncpus; i++) {
+			if (cpu != m->cpus[i])
+				continue;
+
 			struct mips_coproc *cop0 =
 			    m->cpus[i]->cd.mips.coproc[0];
 
-			if (x >= 0 && i != x)
-				continue;
-
 			/*  Print index, random, and wired:  */
 			printf("cpu%i: (", i);
 
@@ -514,13 +505,13 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 				    (int) cop0->reg[COP0_INDEX],
 				    (int) cop0->reg[COP0_RANDOM]);
 			else
-				printf("index=0x%016"PRIx64
-				    " random=0x%016"PRIx64,
+				printf("index=0x%016" PRIx64
+				    " random=0x%016" PRIx64,
 				    (uint64_t) cop0->reg[COP0_INDEX],
 				    (uint64_t) cop0->reg[COP0_RANDOM]);
 
 			if (m->cpus[i]->cd.mips.cpu_type.isa_level >= 3)
-				printf(" wired=0x%"PRIx64,
+				printf(" wired=0x%" PRIx64,
 				    (uint64_t) cop0->reg[COP0_WIRED]);
 
 			printf(")\n");
@@ -529,22 +520,22 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 			    nr_of_tlb_entries; j++) {
 				if (m->cpus[i]->cd.mips.cpu_type.mmu_model ==
 				    MMU3K)
-					printf("%3i: hi=0x%08"PRIx32" lo=0x%08"
+					printf(" %02x: hi=0x%08" PRIx32" lo=0x%08"
 					    PRIx32"\n", j,
 					    (uint32_t) cop0->tlbs[j].hi,
 					    (uint32_t) cop0->tlbs[j].lo0);
 				else if (m->cpus[i]->is_32bit)
-					printf("%3i: hi=0x%08"PRIx32" mask=0x"
-					    "%08"PRIx32" lo0=0x%08"PRIx32
-					    " lo1=0x%08"PRIx32"\n", j,
+					printf(" %02x: hi=0x%08" PRIx32" mask=0x"
+					    "%08" PRIx32" lo0=0x%08" PRIx32
+					    " lo1=0x%08" PRIx32"\n", j,
 					    (uint32_t) cop0->tlbs[j].hi,
 					    (uint32_t) cop0->tlbs[j].mask,
 					    (uint32_t) cop0->tlbs[j].lo0,
 					    (uint32_t) cop0->tlbs[j].lo1);
 				else
-					printf("%3i: hi=0x%016"PRIx64" mask="
-					    "0x%016"PRIx64" lo0=0x%016"PRIx64
-					    " lo1=0x%016"PRIx64"\n", j,
+					printf(" %02x: hi=0x%016" PRIx64" mask="
+					    "0x%016" PRIx64" lo0=0x%016" PRIx64
+					    " lo1=0x%016" PRIx64"\n", j,
 					    (uint64_t) cop0->tlbs[j].hi,
 					    (uint64_t) cop0->tlbs[j].mask,
 					    (uint64_t) cop0->tlbs[j].lo0,
@@ -557,11 +548,12 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 
 	/*  Nicely formatted output:  */
 	for (i=0; i<m->ncpus; i++) {
+		if (cpu != m->cpus[i])
+			continue;
+
 		int pageshift = 12;
 		struct mips_coproc *cop0 = m->cpus[i]->cd.mips.coproc[0];
 
-		if (x >= 0 && i != x)
-			continue;
 
 		if (m->cpus[i]->cd.mips.cpu_type.rev == MIPS_R4100)
 			pageshift = 10;
@@ -579,7 +571,7 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 		default:printf("index=0x%x random=0x%x",
 			    (int) (cop0->reg[COP0_INDEX] & INDEX_MASK),
 			    (int) (cop0->reg[COP0_RANDOM] & RANDOM_MASK));
-			printf(" wired=0x%"PRIx64,
+			printf(" wired=0x%" PRIx64,
 			    (uint64_t) cop0->reg[COP0_WIRED]);
 		}
 
@@ -596,7 +588,7 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 			mask |= (1 << (pageshift+1)) - 1;
 			/*  here mask = e.g. 0x1fff for 4KB pages  */
 
-			printf("%3i: ", j);
+			printf(" %02x: ", j);
 
 			switch (m->cpus[i]->cd.mips.cpu_type.mmu_model) {
 			case MMU3K:
@@ -622,11 +614,11 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 				break;
 			default:switch (m->cpus[i]->cd.mips.cpu_type.mmu_model){
 				case MMU32:
-					printf("vaddr=0x%08"PRIx32" ",
+					printf("vaddr=0x%08" PRIx32" ",
 					    (uint32_t) (hi & ~mask));
 					break;
 				default:/*  R4x00, R1x000, MIPS64, etc.  */
-					printf("vaddr=%016"PRIx64" ",
+					printf("vaddr=%016" PRIx64" ",
 					    (uint64_t) (hi & ~mask));
 				}
 				if (hi & TLB_G)
@@ -644,7 +636,7 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 					paddr >>= ENTRYLO_PFN_SHIFT;
 					paddr <<= pageshift;
 					paddr &= ~(mask >> 1);
-					printf(" p0=0x%09"PRIx64" ",
+					printf(" p0=0x%09" PRIx64" ",
 					    (uint64_t) paddr);
 				}
 				printf(lo0 & ENTRYLO_D? "D" : " ");
@@ -656,7 +648,7 @@ void mips_cpu_tlbdump(struct machine *m, int x, int rawflag)
 					paddr >>= ENTRYLO_PFN_SHIFT;
 					paddr <<= pageshift;
 					paddr &= ~(mask >> 1);
-					printf(" p1=0x%09"PRIx64" ",
+					printf(" p1=0x%09" PRIx64" ",
 					    (uint64_t) paddr);
 				}
 				printf(lo1 & ENTRYLO_D? "D" : " ");
@@ -718,9 +710,9 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 		debug("cpu%i: ", cpu->cpu_id);
 
 	if (cpu->is_32bit)
-		debug("%08"PRIx32, (uint32_t)dumpaddr);
+		debug("%08" PRIx32, (uint32_t)dumpaddr);
 	else
-		debug("%016"PRIx64, (uint64_t)dumpaddr);
+		debug("%016" PRIx64, (uint64_t)dumpaddr);
 
 	memcpy(instr, originstr, sizeof(uint32_t));
 
@@ -740,8 +732,10 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 
 	if (running && cpu->delay_slot)
 		debug(" (d)");
+	else
+		debug("    ");
 
-	debug("\t");
+	cpu_print_pc_indicator_in_disassembly(cpu, running, dumpaddr);
 
 	/*
 	 *  Decode the instruction:
@@ -1008,9 +1002,9 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 		}
 
 		if (cpu->is_32bit)
-			debug("0x%08"PRIx32, (uint32_t)addr);
+			debug("0x%08" PRIx32, (uint32_t)addr);
 		else
-			debug("0x%016"PRIx64, (uint64_t)addr);
+			debug("0x%016" PRIx64, (uint64_t)addr);
 
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    addr, &offset);
@@ -1187,7 +1181,7 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 			    rt, imm, regnames[rs]);
 
 			if (running) {
-				debug("\t[0x%016"PRIx64" = %s]",
+				debug("\t[0x%016" PRIx64" = %s]",
 				    (uint64_t)(cpu->cd.mips.gpr[rs] + imm));
 				if (symbol != NULL)
 					debug(" = %s", symbol);
@@ -1212,10 +1206,10 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 			debug("\t[");
 
 			if (cpu->is_32bit)
-				debug("0x%08"PRIx32,
+				debug("0x%08" PRIx32,
 				    (uint32_t) (cpu->cd.mips.gpr[rs] + imm));
 			else
-				debug("0x%016"PRIx64,
+				debug("0x%016" PRIx64,
 				    (uint64_t) (cpu->cd.mips.gpr[rs] + imm));
 
 			if (symbol != NULL)
@@ -1239,9 +1233,9 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 		    addr, &offset);
 		debug("%s\t0x", hi6_names[hi6]);
 		if (cpu->is_32bit)
-			debug("%08"PRIx32, (uint32_t) addr);
+			debug("%08" PRIx32, (uint32_t) addr);
 		else
-			debug("%016"PRIx64, (uint64_t) addr);
+			debug("%016" PRIx64, (uint64_t) addr);
 		if (symbol != NULL)
 			debug("\t<%s>", symbol);
 		break;
@@ -1281,7 +1275,7 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 		if (cache_op==6)	debug("hit writeback");
 		if (cache_op==7)	debug("hit set virtual");
 		if (running)
-			debug(", addr 0x%016"PRIx64,
+			debug(", addr 0x%016" PRIx64,
 			    (uint64_t)(cpu->cd.mips.gpr[rt] + imm));
 		if (showtag)
 		debug(", taghi=%08lx lo=%08lx",
@@ -1462,9 +1456,23 @@ int mips_cpu_disassemble_instr(struct cpu *cpu, unsigned char *originstr,
 			addr = (dumpaddr + 4) + (imm << 2);
 
 			if (cpu->is_32bit)
-				debug("0x%08"PRIx32, (uint32_t) addr);
+				debug("0x%08" PRIx32, (uint32_t) addr);
 			else
-				debug("0x%016"PRIx64, (uint64_t) addr);
+				debug("0x%016" PRIx64, (uint64_t) addr);
+			break;
+
+		case REGIMM_TGEI:
+		case REGIMM_TGEIU:
+		case REGIMM_TLTI:
+		case REGIMM_TLTIU:
+		case REGIMM_TEQI:
+		case REGIMM_TNEI:
+			debug("%s\t%s,", regimm_names[regimm5], regnames[rs]);
+
+			if (cpu->is_32bit)
+				debug("0x%" PRIx32, (uint32_t) imm);
+			else
+				debug("0x%" PRIx64, (uint64_t) imm);
 			break;
 
 		case REGIMM_SYNCI:
@@ -1509,29 +1517,29 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		    cpu->pc, &offset);
 
 		if (bits32)
-			debug("cpu%i:  pc = %08"PRIx32,
+			debug("cpu%i:  pc = %08" PRIx32,
 			    cpu->cpu_id, (uint32_t) cpu->pc);
 		else if (bits128)
-			debug("cpu%i:  pc=%016"PRIx64,
+			debug("cpu%i:  pc=%016" PRIx64,
 			    cpu->cpu_id, (uint64_t) cpu->pc);
 		else
-			debug("cpu%i:    pc = 0x%016"PRIx64,
+			debug("cpu%i:    pc = 0x%016" PRIx64,
 			    cpu->cpu_id, (uint64_t) cpu->pc);
 
 		debug("    <%s>\n", symbol != NULL? symbol :
 		    " no symbol ");
 
 		if (bits32)
-			debug("cpu%i:  hi = %08"PRIx32"  lo = %08"PRIx32"\n",
+			debug("cpu%i:  hi = %08" PRIx32"  lo = %08" PRIx32"\n",
 			    cpu->cpu_id, (uint32_t) cpu->cd.mips.hi,
 			    (uint32_t) cpu->cd.mips.lo);
 		else if (bits128) {
-			debug("cpu%i:  hi=%016"PRIx64"%016"PRIx64"  lo="
-			    "%016"PRIx64"%016"PRIx64"\n", cpu->cpu_id,
+			debug("cpu%i:  hi=%016" PRIx64"%016" PRIx64"  lo="
+			    "%016" PRIx64"%016" PRIx64"\n", cpu->cpu_id,
 			    cpu->cd.mips.hi1, cpu->cd.mips.hi,
 			    cpu->cd.mips.lo1, cpu->cd.mips.lo);
 		} else {
-			debug("cpu%i:    hi = 0x%016"PRIx64"    lo = 0x%016"
+			debug("cpu%i:    hi = 0x%016" PRIx64"    lo = 0x%016"
 			    PRIx64"\n", cpu->cpu_id,
 			    (uint64_t) cpu->cd.mips.hi,
 			    (uint64_t) cpu->cd.mips.lo);
@@ -1548,7 +1556,7 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 					debug("                           "
 					    "          ");
 				else
-					debug(" %3s=%016"PRIx64"%016"PRIx64,
+					debug(" %3s=%016" PRIx64"%016" PRIx64,
 					    regnames[r], (uint64_t)
 					    cpu->cd.mips.gpr_quadhi[r],
 					    (uint64_t)cpu->cd.mips.gpr[r]);
@@ -1563,7 +1571,7 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 				if (i == MIPS_GPR_ZERO)
 					debug("               ");
 				else
-					debug(" %3s = %08"PRIx32, regnames[i],
+					debug(" %3s = %08" PRIx32, regnames[i],
 					    (uint32_t)cpu->cd.mips.gpr[i]);
 				if ((i & 3) == 3)
 					debug("\n");
@@ -1577,7 +1585,7 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 				if (r == MIPS_GPR_ZERO)
 					debug("                           ");
 				else
-					debug("   %3s = 0x%016"PRIx64,
+					debug("   %3s = 0x%016" PRIx64,
 					    regnames[r],
 					    (uint64_t)cpu->cd.mips.gpr[r]);
 				if ((i & 1) == 1)
@@ -1622,7 +1630,7 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 					    (int) cpu->cd.mips.coproc[
 					    coprocnr]->reg[i]);
 				else
-					debug(" = 0x%016"PRIx64, (uint64_t)
+					debug(" = 0x%016" PRIx64, (uint64_t)
 					    cpu->cd.mips.coproc[
 					    coprocnr]->reg[i]);
 			}
@@ -1640,10 +1648,10 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 			debug("cpu%i: ", cpu->cpu_id);
 			debug("config_select1 = 0x");
 			if (cpu->is_32bit)
-				debug("%08"PRIx32,
+				debug("%08" PRIx32,
 				    (uint32_t)cpu->cd.mips.cop0_config_select1);
 			else
-				debug("%016"PRIx64,
+				debug("%016" PRIx64,
 				    (uint64_t)cpu->cd.mips.cop0_config_select1);
 			debug("\n");
 		}
@@ -1673,7 +1681,7 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 	if (cpu->cd.mips.rmw) {
 		printf("cpu%i: Read-Modify-Write in progress, address "
-		    "0x%016"PRIx64"\n", cpu->cpu_id, cpu->cd.mips.rmw_addr);
+		    "0x%016" PRIx64"\n", cpu->cpu_id, cpu->cd.mips.rmw_addr);
 	}
 }
 
@@ -1686,12 +1694,12 @@ void mips_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
  */
 void mips_cpu_interrupt_assert(struct interrupt *interrupt)
 {
-	struct cpu *cpu = interrupt->extra;
+	struct cpu *cpu = (struct cpu *) interrupt->extra;
 	cpu->cd.mips.coproc[0]->reg[COP0_CAUSE] |= interrupt->line;
 }
 void mips_cpu_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct cpu *cpu = interrupt->extra;
+	struct cpu *cpu = (struct cpu *) interrupt->extra;
 	cpu->cd.mips.coproc[0]->reg[COP0_CAUSE] &= ~interrupt->line;
 }
 
@@ -1724,8 +1732,11 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 		 *  the instruction following the wait instruction be the one
 		 *  we continue at when the interrupt service routine returns.
 		 */
-		cpu->is_halted = 0;
+		cpu->is_halted = false;
 		cpu->pc += sizeof(uint32_t);
+
+		debugmsg_cpu(cpu, SUBSYS_CPU, "wakeup", VERBOSITY_DEBUG,
+		    "waking up from halted state");
 	}
 
 	if (!quiet_mode) {
@@ -1734,6 +1745,11 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 		char *symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->pc, &offset);
 
+
+		// TODO: build up full exception string...
+		//debugmsg_cpu(cpu, SUBSYS_EXCEPTION, exception_names[exccode], VERBOSITY_INFO, "todo");
+
+
 		debug("[ ");
 		if (cpu->machine->ncpus > 1)
 			debug("cpu%i: ", cpu->cpu_id);
@@ -1764,10 +1780,10 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 					    d, strbuf, sizeof(strbuf)));
 				} else {
 					if (cpu->is_32bit)
-						debug(" a%i=0x%"PRIx32, x,
+						debug(" a%i=0x%" PRIx32, x,
 						    (uint32_t)d);
 					else
-						debug(" a%i=0x%"PRIx64, x,
+						debug(" a%i=0x%" PRIx64, x,
 						    (uint64_t)d);
 				}
 			}
@@ -1781,13 +1797,13 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 			if (cpu->is_32bit)
 				debug(" vaddr=0x%08x", (int)vaddr);
 			else
-				debug(" vaddr=0x%016"PRIx64, (uint64_t)vaddr);
+				debug(" vaddr=0x%016" PRIx64, (uint64_t)vaddr);
 		}
 
 		if (cpu->is_32bit)
-			debug(" pc=0x%08"PRIx32" ", (uint32_t)cpu->pc);
+			debug(" pc=0x%08" PRIx32" ", (uint32_t)cpu->pc);
 		else
-			debug(" pc=0x%016"PRIx64" ", (uint64_t)cpu->pc);
+			debug(" pc=0x%016" PRIx64" ", (uint64_t)cpu->pc);
 
 		if (symbol != NULL)
 			debug("<%s> ]\n", symbol);
@@ -1799,19 +1815,23 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 		uint64_t offset;
 		char *symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->pc, &offset);
+
+		// TODO: debugmsg SUBSYS_CPU, or better: SUBSYS_EXCEPTION
+		// VERBOSITY_WARNING since this is a relatively sure sign of a bug.
+
 		fatal("[ ");
 		if (cpu->machine->ncpus > 1)
 			fatal("cpu%i: ", cpu->cpu_id);
 		fatal("warning: LOW reference: vaddr=");
 		if (cpu->is_32bit)
-			fatal("0x%08"PRIx32, (uint32_t) vaddr);
+			fatal("0x%08" PRIx32, (uint32_t) vaddr);
 		else
-			fatal("0x%016"PRIx64, (uint64_t) vaddr);
+			fatal("0x%016" PRIx64, (uint64_t) vaddr);
 		fatal(", exception %s, pc=", exception_names[exccode]);
 		if (cpu->is_32bit)
-			fatal("0x%08"PRIx32, (uint32_t) cpu->pc);
+			fatal("0x%08" PRIx32, (uint32_t) cpu->pc);
 		else
-			fatal("0x%016"PRIx64, (uint64_t)cpu->pc);
+			fatal("0x%016" PRIx64, (uint64_t)cpu->pc);
 		fatal(" <%s> ]\n", symbol? symbol : "(no symbol)");
 	}
 
@@ -1874,6 +1894,21 @@ void mips_cpu_exception(struct cpu *cpu, int exccode, int tlb, uint64_t vaddr,
 
 				/*  reg[COP0_PAGEMASK] = cpu->cd.mips.coproc[0]->tlbs[0].mask & PAGEMASK_MASK;  */
 
+				// Actually, the R10000 manual says:
+				// "When either a TLB refill, TLB invalid, or TLB modified exception occurs, the
+				//  EntryHi register is loaded with the virtual page number (VPN2) and the ASID of
+				//  the virtual address that did not have a matching TLB entry."
+				// Strange that it does not mention the 2 top bits (R).
+				//
+				// However, the MIPS64 manual says:
+				// "A TLB exception (TLB Refill, XTLB Refill, TLB Invalid, or TLB Modified) causes
+				//  the bits of the virtual address corresponding to the R and VPN2 fields to be
+				//  written into the EntryHi register."
+
+				if ((reg[COP0_ENTRYHI] & ENTRYHI_ASID) != (uint32_t)vaddr_asid)
+					fatal("[ huh? vaddr_asid 0x%02x not same as in ENTRYHI 0x%02x ]\n",
+						vaddr_asid, reg[COP0_ENTRYHI] & ENTRYHI_ASID);
+
 				if (cpu->cd.mips.cpu_type.mmu_model == MMU10K)
 					reg[COP0_ENTRYHI] = (vaddr & (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK_R10K)) | vaddr_asid;
 				else
