@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -46,10 +46,8 @@
 #include "timer.h"
 
 
-extern volatile int single_step;
-
-static char *cop0_names[] = COP0_NAMES;
-static char *regnames[] = MIPS_REGISTER_NAMES;
+static const char *cop0_names[] = COP0_NAMES;
+static const char *regnames[] = MIPS_REGISTER_NAMES;
 
 
 /*
@@ -326,6 +324,7 @@ static void initialize_cop1(struct cpu *cpu, struct mips_coproc *c)
 	case MIPS_4Kc:		/*  TODO: Is this the same as 5Kc?  */
 	case MIPS_5Kc:		other_stuff = COP1_REVISION_DOUBLE
 				    | COP1_REVISION_SINGLE;
+				// fall through
 	case MIPS_R5000:
 	case MIPS_RM5200:	fpu_rev = cpu->cd.mips.cpu_type.rev;
 				other_stuff |= 0x10;
@@ -357,14 +356,14 @@ struct mips_coproc *mips_coproc_new(struct cpu *cpu, int coproc_nr)
 {
 	struct mips_coproc *c;
 
-	CHECK_ALLOCATION(c = malloc(sizeof(struct mips_coproc)));
+	CHECK_ALLOCATION(c = (struct mips_coproc *) malloc(sizeof(struct mips_coproc)));
 	memset(c, 0, sizeof(struct mips_coproc));
 
 	c->coproc_nr = coproc_nr;
 
 	if (coproc_nr == 0) {
 		c->nr_of_tlbs = cpu->cd.mips.cpu_type.nr_of_tlb_entries;
-		c->tlbs = zeroed_alloc(c->nr_of_tlbs * sizeof(struct mips_tlb));
+		c->tlbs = (struct mips_tlb *) zeroed_alloc(c->nr_of_tlbs * sizeof(struct mips_tlb));
 
 		/*
 		 *  Start with nothing in the status register. This makes sure
@@ -373,11 +372,6 @@ struct mips_coproc *mips_coproc_new(struct cpu *cpu, int coproc_nr)
 		 */
 		c->reg[COP0_STATUS] = 0;
 
-		/*  For userland emulation, enable all four coprocessors:  */
-		if (cpu->machine->userland_emul)
-			c->reg[COP0_STATUS] |=
-			    ((uint32_t)0xf << STATUS_CU_SHIFT);
-
 		/*  Hm. Enable coprocessors 0 and 1 even if we're not just
 		    emulating userland? TODO: Think about this.  */
 		/*  if (cpu->machine->prom_emulation)  */
@@ -521,7 +515,7 @@ static void invalidate_asid(struct cpu *cpu, unsigned int asid)
 	struct mips_tlb *tlb = cp->tlbs;
 
 	if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
-		for (i=0; i<ntlbs; i++)
+		for (i = 0; i < ntlbs; i++)
 			if ((tlb[i].hi & R2K3K_ENTRYHI_ASID_MASK) == asid
 			    && (tlb[i].lo0 & R2K3K_ENTRYLO_V)
 			    && !(tlb[i].lo0 & R2K3K_ENTRYLO_G)) {
@@ -530,45 +524,45 @@ static void invalidate_asid(struct cpu *cpu, unsigned int asid)
 				    INVALIDATE_VADDR);
 			}
 	} else {
-		int non4kpages = 0;
-		uint64_t topbit = 1, fillmask = 0xffffff0000000000ULL;
-
-		if (cpu->is_32bit) {
-			topbit = 0x80000000;
-			fillmask = 0xffffffff00000000ULL;
-		} else if (cpu->cd.mips.cpu_type.mmu_model == MMU10K) {
-			topbit <<= 43;
-			fillmask <<= 4;
-		} else {
-			topbit <<= 39;
-		}
-
-		for (i=0; i<ntlbs; i++) {
-			if (tlb[i].mask != 0 && tlb[i].mask != 0x1800) {
-				non4kpages = 1;
+		for (i = 0; i < ntlbs; i++) {
+			if ((tlb[i].hi & ENTRYHI_ASID) != asid || (tlb[i].hi & TLB_G))
 				continue;
+
+			uint64_t mask = cp->tlbs[i].mask;
+			uint64_t pagesize = 0x1000;
+			uint64_t tmp = mask >> 13;
+			while ((tmp & 1)) {
+				tmp >>= 1;
+				pagesize <<= 1;
+			}
+			
+			uint64_t oldvaddr;
+
+			if (cpu->cd.mips.cpu_type.mmu_model == MMU10K) {
+				oldvaddr = cp->tlbs[i].hi &
+					    (ENTRYHI_VPN2_MASK_R10K | ENTRYHI_R_MASK);
+			} else if (cpu->is_32bit) {
+				/*  MIPS32 etc.:  */
+				oldvaddr = cp->tlbs[i].hi & ENTRYHI_VPN2_MASK;
+				oldvaddr = (int32_t)oldvaddr;
+			} else {
+				/*  Assume MMU4K  */
+				oldvaddr = cp->tlbs[i].hi &
+				    (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK);
 			}
 
-			if ((tlb[i].hi & ENTRYHI_ASID) == asid &&
-			    !(tlb[i].hi & TLB_G)) {
-				uint64_t vaddr0, vaddr1;
-				vaddr0 = cp->tlbs[i].hi & ~fillmask;
-				if (vaddr0 & topbit)
-					vaddr0 |= fillmask;
-				vaddr1 = vaddr0 | 0x1000;  /*  TODO: mask  */
+			mask |= 0x1fff;
+			oldvaddr &= ~mask;
 
-				if (tlb[i].lo0 & ENTRYLO_V)
-					cpu->invalidate_translation_caches(cpu,
-					    vaddr0, INVALIDATE_VADDR);
-				if (tlb[i].lo1 & ENTRYLO_V)
-					cpu->invalidate_translation_caches(cpu,
-					    vaddr1, INVALIDATE_VADDR);
-			}
-		}
-
-		if (non4kpages) {
-			cpu->invalidate_translation_caches(cpu,
-			    0, INVALIDATE_ALL);
+			// printf("pagesize = %016llx mask = %016llx\n", pagesize, mask);
+			
+			if (cp->tlbs[i].lo0 & ENTRYLO_V)
+				for (uint64_t ofs = 0; ofs < pagesize; ofs += 0x1000)
+					cpu->invalidate_translation_caches(cpu, oldvaddr + ofs, INVALIDATE_VADDR);
+			
+			if (cp->tlbs[i].lo1 & ENTRYLO_V)
+				for (uint64_t ofs = 0; ofs < pagesize; ofs += 0x1000)
+					cpu->invalidate_translation_caches(cpu, oldvaddr + ofs + pagesize, INVALIDATE_VADDR);
 		}
 	}
 }
@@ -595,6 +589,7 @@ void coproc_register_read(struct cpu *cpu,
 	if (cp->coproc_nr==0 && reg_nr==COP0_COUNT) {
 		/*  TODO: Increase count in a more meaningful way!  */
 		cp->reg[COP0_COUNT] = (int32_t) (cp->reg[COP0_COUNT] + 1);
+		cpu->cd.mips.count_register_read_count ++;
 		unimpl = 0;
 	}
 	if (cp->coproc_nr==0 && reg_nr==COP0_ENTRYHI)	unimpl = 0;
@@ -676,8 +671,7 @@ void coproc_register_write(struct cpu *cpu,
 			break;
 		case COP0_ENTRYLO0:
 			unimpl = 0;
-			if (cpu->cd.mips.cpu_type.mmu_model == MMU3K &&
-			    (tmp & 0xff)!=0) {
+			if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
 				/*  char *symbol;
 				    uint64_t offset;
 				    symbol = get_symbol_name(cpu->pc, &offset);
@@ -688,9 +682,19 @@ void coproc_register_write(struct cpu *cpu,
 				tmp &= (R2K3K_ENTRYLO_PFN_MASK |
 				    R2K3K_ENTRYLO_N | R2K3K_ENTRYLO_D |
 				    R2K3K_ENTRYLO_V | R2K3K_ENTRYLO_G);
-			} else if (cpu->cd.mips.cpu_type.mmu_model == MMU4K) {
+			} else {
+				// Both MIPS III (such as R4000) and MIPS32/MIPS64 (?)
 				tmp &= (ENTRYLO_PFN_MASK | ENTRYLO_C_MASK |
 				    ENTRYLO_D | ENTRYLO_V | ENTRYLO_G);
+
+				int c = (tmp & ENTRYLO_C_MASK) >> ENTRYLO_C_SHIFT;
+				switch (c) {
+				case 2:	// Uncached
+				case 3:	// Cached
+					break;
+				default:if (tmp & ENTRYLO_V)
+						debug("[ MIPS coproc_register_write: unimplemented c = %i ]\n", c);
+				}
 			}
 			break;
 		case COP0_BADVADDR:
@@ -699,7 +703,12 @@ void coproc_register_write(struct cpu *cpu,
 			break;
 		case COP0_ENTRYLO1:
 			unimpl = 0;
-			if (cpu->cd.mips.cpu_type.mmu_model == MMU4K) {
+			if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
+				fatal("Attempt to access ENTRYLO1 with MMU3K?\n");
+				exit(1);
+			}
+			// Both MIPS III (such as R4000) and MIPS32/MIPS64 (?)
+			if (cpu->cd.mips.cpu_type.mmu_model != MMU3K) {
 				tmp &= (ENTRYLO_PFN_MASK | ENTRYLO_C_MASK |
 				    ENTRYLO_D | ENTRYLO_V | ENTRYLO_G);
 			}
@@ -726,22 +735,32 @@ void coproc_register_write(struct cpu *cpu,
 			return;
 		case COP0_PAGEMASK:
 			tmp2 = tmp >> PAGEMASK_SHIFT;
-			if (tmp2 != 0x000 &&
-			    tmp2 != 0x003 &&
-			    tmp2 != 0x00f &&
-			    tmp2 != 0x03f &&
-			    tmp2 != 0x0ff &&
-			    tmp2 != 0x3ff &&
-			    tmp2 != 0xfff)
-				fatal("cpu%i: trying to write an invalid"
-				    " pagemask 0x%08lx to COP0_PAGEMASK\n",
-				    cpu->cpu_id, (long)tmp);
+			if (cpu->cd.mips.cpu_type.rev != MIPS_R4100) {
+				if (tmp2 != 0x000 &&
+				    tmp2 != 0x003 &&
+				    tmp2 != 0x00f &&
+				    tmp2 != 0x03f &&
+				    tmp2 != 0x0ff &&
+				    tmp2 != 0x3ff &&
+				    tmp2 != 0xfff &&
+				    tmp2 != 0x3fff &&
+				    tmp2 != 0xffff)
+					fatal("[ cpu%i: trying to write an invalid"
+					    " pagemask 0x%08lx to COP0_PAGEMASK ]\n",
+					    cpu->cpu_id, (long)tmp);
+				// Actually just 0xfff << shift for R10000.
+				if (cpu->cd.mips.cpu_type.mmu_model == MMU10K) {
+					tmp &= 0x0fff << PAGEMASK_SHIFT;
+				} else {
+					tmp &= 0xffff << PAGEMASK_SHIFT;
+				}
+			}
 			unimpl = 0;
 			break;
 		case COP0_WIRED:
 			if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
-				fatal("cpu%i: r2k/r3k wired register must "
-				    "always be 8\n", cpu->cpu_id);
+				fatal("[ cpu%i: r2k/r3k wired register must "
+				    "always be 8 ]\n", cpu->cpu_id);
 				tmp = 8;
 			}
 			cp->reg[COP0_RANDOM] = cp->nr_of_tlbs-1;
@@ -750,8 +769,8 @@ void coproc_register_write(struct cpu *cpu,
 			break;
 		case COP0_COUNT:
 			if (tmp != (uint64_t)(int64_t)(int32_t)tmp)
-				fatal("WARNING: trying to write a 64-bit value"
-				    " to the COUNT register!\n");
+				fatal("[ WARNING: trying to write a 64-bit value"
+				    " to the COUNT register! ]\n");
 			tmp = (int64_t)(int32_t)tmp;
 			unimpl = 0;
 			break;
@@ -769,10 +788,6 @@ void coproc_register_write(struct cpu *cpu,
 				else
 					hz = (double)cpu->machine->emulated_hz
 					    / (double)compare_diff;
-/*
- *  TODO: DON'T HARDCODE THIS!
- */
-hz = 100.0;
 
 				/*  Initialize or re-set the periodic timer:  */
 				if (hz > 0) {
@@ -794,8 +809,8 @@ hz = 100.0;
 			cp->reg[COP0_CAUSE] &= ~0x8000;
 
 			if (tmp != (uint64_t)(int64_t)(int32_t)tmp)
-				fatal("WARNING: trying to write a 64-bit value"
-				    " to the COMPARE register!\n");
+				fatal("[ WARNING: trying to write a 64-bit value"
+				    " to the COMPARE register! ]\n");
 
 			tmp = (int64_t)(int32_t)tmp;
 			cpu->cd.mips.compare_register_set = 1;
@@ -863,9 +878,9 @@ hz = 100.0;
 				switch (select) {
 				case 1:	cpu->cd.mips.cop0_config_select1 = tmp;
 					break;
-				default:fatal("coproc_register_write(): unimpl"
+				default:fatal("[ coproc_register_write(): unimpl"
 					    "emented config register select "
-					    "%i\n", select);
+					    "%i ]\n", select);
 					exit(1);
 				}
 				return;
@@ -882,6 +897,15 @@ hz = 100.0;
 			oldmode = cp->reg[COP0_STATUS];
 			tmp &= ~(1 << 21);	/*  bit 21 is read-only  */
 
+			/*
+			 *  When switching between 32-bit and 64-bit address
+			 *  spaces, invalidate all translations.
+			 */
+			if (!cpu->is_32bit &&
+			    (oldmode & (STATUS_KX | STATUS_SX | STATUS_UX)) !=
+			    (tmp & (STATUS_KX | STATUS_SX | STATUS_UX)))
+				cpu->invalidate_translation_caches(cpu, 0, INVALIDATE_ALL);
+
 			/*
 			 *  When isolating caches, invalidate all translations.
 			 *  During the isolation, a special hack in memory_rw.c
@@ -953,8 +977,8 @@ hz = 100.0;
 	}
 
 	if (unimpl) {
-		fatal("cpu%i: warning: write to unimplemented coproc%i "
-		    "register %i (%s), data = 0x%016llx\n", cpu->cpu_id,
+		fatal("[ cpu%i: warning: write to unimplemented coproc%i "
+		    "register %i (%s), data = 0x%016llx ]\n", cpu->cpu_id,
 		    cp->coproc_nr, reg_nr, cp->coproc_nr==0?
 		    cop0_names[reg_nr] : "?", (long long)tmp);
 
@@ -964,8 +988,8 @@ hz = 100.0;
 	}
 
 	if (readonly) {
-		fatal("cpu%i: warning: write to READONLY coproc%i register "
-		    "%i ignored\n", cpu->cpu_id, cp->coproc_nr, reg_nr);
+		fatal("[ cpu%i: warning: write to READONLY coproc%i register "
+		    "%i ignored ]\n", cpu->cpu_id, cp->coproc_nr, reg_nr);
 		return;
 	}
 
@@ -988,13 +1012,13 @@ static int mips_fmt_to_ieee_fmt[32] = {
 	IEEE_FMT_W, IEEE_FMT_L, /* PS (Paired Single) */ 0, 0,
 	0, 0, 0, 0,  0, 0, 0, 0  };
 
-static char *fmtname[32] = {
+static const char *fmtname[32] = {
 	 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",
 	 "8",  "9", "10", "11", "12", "13", "14", "15",
 	 "s",  "d", "18", "19",  "w",  "l", "ps", "23",
 	"24", "25", "26", "27", "28", "29", "30", "31" 	};
 
-static char *ccname[16] = {
+static const char *ccname[16] = {
 	"f",  "un",   "eq",  "ueq", "olt", "ult", "ole", "ule",
 	"sf", "ngle", "seq", "ngl", "lt",  "nge", "le",  "ngt"  };
 
@@ -1016,24 +1040,28 @@ static char *ccname[16] = {
  *
  *  Stores a float value (actually a double) in fmt format.
  */
-static void fpu_store_float_value(struct mips_coproc *cp, int fd,
+static void fpu_store_float_value(int fr_flag, struct mips_coproc *cp, int fd,
 	double nf, int fmt, int nan)
 {
 	int ieee_fmt = mips_fmt_to_ieee_fmt[fmt];
-	uint64_t r = ieee_store_float_value(nf, ieee_fmt, nan);
+	uint64_t r = ieee_store_float_value(nf, ieee_fmt);
 
 	/*
 	 *  TODO: This is for 32-bit mode. It has to be updated later
 	 *        for 64-bit coprocessor functionality!
 	 */
 	if (fmt == COP1_FMT_D || fmt == COP1_FMT_L) {
-		cp->reg[fd] = r & 0xffffffffULL;
-		cp->reg[(fd+1) & 31] = (r >> 32) & 0xffffffffULL;
+		if (fr_flag)
+			cp->reg[fd] = r;
+		else {
+			cp->reg[fd] = r & 0xffffffffULL;
+			cp->reg[(fd+1) & 31] = (r >> 32) & 0xffffffffULL;
 
-		if (cp->reg[fd] & 0x80000000ULL)
-			cp->reg[fd] |= 0xffffffff00000000ULL;
-		if (cp->reg[fd+1] & 0x80000000ULL)
-			cp->reg[fd+1] |= 0xffffffff00000000ULL;
+			if (cp->reg[fd] & 0x80000000ULL)
+				cp->reg[fd] |= 0xffffffff00000000ULL;
+			if (cp->reg[fd+1] & 0x80000000ULL)
+				cp->reg[fd+1] |= 0xffffffff00000000ULL;
+		}
 	} else {
 		cp->reg[fd] = r & 0xffffffffULL;
 
@@ -1060,46 +1088,52 @@ static int fpu_op(struct cpu *cpu, struct mips_coproc *cp, int op, int fmt,
 	int unordered, nan, ieee_fmt = mips_fmt_to_ieee_fmt[fmt];
 	uint64_t fs_v = 0;
 	double nf;
+	int fr = cpu->cd.mips.coproc[0]->reg[COP0_STATUS] & STATUS_FR ? 1 : 0;
+
+	// printf("op %x (fmt %i):\n", op, fmt);
 
 	if (fs >= 0) {
 		fs_v = cp->reg[fs];
-		/*  TODO: register-pair mode and plain
-		    register mode? "FR" bit?  */
-		if (fmt == COP1_FMT_D || fmt == COP1_FMT_L)
-			fs_v = (fs_v & 0xffffffffULL) +
-			    (cp->reg[(fs + 1) & 31] << 32);
+		if (fmt == COP1_FMT_D || fmt == COP1_FMT_L) {
+			if (!fr)
+				fs_v = (fs_v & 0xffffffffULL) +
+				    (cp->reg[(fs + 1) & 31] << 32);
+		}
+		// printf("    fs_v = 0x%016llx\n", (long long)fs_v);
 		ieee_interpret_float_value(fs_v, &float_value[0], ieee_fmt);
 	}
 	if (ft >= 0) {
 		uint64_t v = cp->reg[ft];
-		/*  TODO: register-pair mode and
-		    plain register mode? "FR" bit?  */
-		if (fmt == COP1_FMT_D || fmt == COP1_FMT_L)
-			v = (v & 0xffffffffULL) +
-			    (cp->reg[(ft + 1) & 31] << 32);
+		if (fmt == COP1_FMT_D || fmt == COP1_FMT_L) {
+			if (!fr)
+				v = (v & 0xffffffffULL) +
+				    (cp->reg[(ft + 1) & 31] << 32);
+		}
+		// printf("    ft_v = 0x%016llx\n", (long long)v);
 		ieee_interpret_float_value(v, &float_value[1], ieee_fmt);
 	}
 
+
 	switch (op) {
 	case FPU_OP_ADD:
 		nf = float_value[0].f + float_value[1].f;
 		/*  debug("  add: %f + %f = %f\n",
 		    float_value[0].f, float_value[1].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan || float_value[1].nan);
 		break;
 	case FPU_OP_SUB:
 		nf = float_value[0].f - float_value[1].f;
 		/*  debug("  sub: %f - %f = %f\n",
 		    float_value[0].f, float_value[1].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan || float_value[1].nan);
 		break;
 	case FPU_OP_MUL:
 		nf = float_value[0].f * float_value[1].f;
 		/*  debug("  mul: %f * %f = %f\n",
 		    float_value[0].f, float_value[1].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan || float_value[1].nan);
 		break;
 	case FPU_OP_DIV:
@@ -1107,13 +1141,15 @@ static int fpu_op(struct cpu *cpu, struct mips_coproc *cp, int op, int fmt,
 		if (fabs(float_value[1].f) > 0.00000000001)
 			nf = float_value[0].f / float_value[1].f;
 		else {
-			fatal("DIV by zero !!!!\n");
+			fatal("DIV by zero !!!! TODO\n");
 			nf = 0.0;	/*  TODO  */
 			nan = 1;
+			// mips_cpu_exception(cpu, EXCEPTION_FPE, 0, 0, 1, 0, 0, 0);
+			return 0;
 		}
 		/*  debug("  div: %f / %f = %f\n",
 		    float_value[0].f, float_value[1].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt, nan);
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt, nan);
 		break;
 	case FPU_OP_SQRT:
 		nan = float_value[0].nan;
@@ -1126,24 +1162,24 @@ static int fpu_op(struct cpu *cpu, struct mips_coproc *cp, int op, int fmt,
 			nan = 1;
 		}
 		/*  debug("  sqrt: %f => %f\n", float_value[0].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt, nan);
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt, nan);
 		break;
 	case FPU_OP_ABS:
 		nf = fabs(float_value[0].f);
 		/*  debug("  abs: %f => %f\n", float_value[0].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan);
 		break;
 	case FPU_OP_NEG:
 		nf = - float_value[0].f;
 		/*  debug("  neg: %f => %f\n", float_value[0].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan);
 		break;
 	case FPU_OP_CVT:
 		nf = float_value[0].f;
 		/*  debug("  mov: %f => %f\n", float_value[0].f, nf);  */
-		fpu_store_float_value(cp, fd, nf, output_fmt,
+		fpu_store_float_value(fr, cp, fd, nf, output_fmt,
 		    float_value[0].nan);
 		break;
 	case FPU_OP_MOV:
@@ -1153,12 +1189,16 @@ static int fpu_op(struct cpu *cpu, struct mips_coproc *cp, int op, int fmt,
 		 *		for 64-bit coprocessor stuff.
 		 */
 		if (output_fmt == COP1_FMT_D || output_fmt == COP1_FMT_L) {
-			cp->reg[fd] = fs_v & 0xffffffffULL;
-			cp->reg[(fd+1) & 31] = (fs_v >> 32) & 0xffffffffULL;
-			if (cp->reg[fd] & 0x80000000ULL)
-				cp->reg[fd] |= 0xffffffff00000000ULL;
-			if (cp->reg[fd+1] & 0x80000000ULL)
-				cp->reg[fd+1] |= 0xffffffff00000000ULL;
+			if (fr)
+				cp->reg[fd] = fs_v;
+			else {
+				cp->reg[fd] = fs_v & 0xffffffffULL;
+				cp->reg[(fd+1) & 31] = (fs_v >> 32) & 0xffffffffULL;
+				if (cp->reg[fd] & 0x80000000ULL)
+					cp->reg[fd] |= 0xffffffff00000000ULL;
+				if (cp->reg[fd+1] & 0x80000000ULL)
+					cp->reg[fd+1] |= 0xffffffff00000000ULL;
+			}
 		} else {
 			cp->reg[fd] = fs_v & 0xffffffffULL;
 			if (cp->reg[fd] & 0x80000000ULL)
@@ -1244,7 +1284,7 @@ static int fpu_function(struct cpu *cpu, struct mips_coproc *cp,
 	/*  bc1f, bc1t, bc1fl, bc1tl:  */
 	if ((function & 0x03e00000) == 0x01000000) {
 		int nd, tf, imm;
-		char *instr_mnem;
+		const char *instr_mnem;
 
 		/*  cc are bits 20..18:  */
 		cc = (function >> 18) & 7;
@@ -1262,7 +1302,7 @@ static int fpu_function(struct cpu *cpu, struct mips_coproc *cp,
 
 		if (cpu->machine->instruction_trace || unassemble_only)
 			debug("%s\t%i,0x%016llx\n", instr_mnem, cc,
-			    (long long) (cpu->pc + (imm << 2)));
+			    (long long) (cpu->pc + 4 + (imm << 2)));
 		if (unassemble_only)
 			return 1;
 
@@ -1493,8 +1533,12 @@ void coproc_tlbpr(struct cpu *cpu, int readflag)
 		} else {
 			/*  R4000:  */
 			i = cp->reg[COP0_INDEX] & INDEX_MASK;
-			if (i >= cp->nr_of_tlbs)	/*  TODO:  exception  */
+			if (i >= cp->nr_of_tlbs) {
+				/*  TODO:  exception?  */
+				fatal("[ warning: tlbr from index %i (too "
+				    "high) ]\n", i);
 				return;
+			}
 
 			cp->reg[COP0_PAGEMASK] = cp->tlbs[i].mask;
 			cp->reg[COP0_ENTRYHI]  = cp->tlbs[i].hi;
@@ -1542,6 +1586,7 @@ void coproc_tlbpr(struct cpu *cpu, int readflag)
 		else
 			xmask = ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK;
 		vpn2 = cp->reg[COP0_ENTRYHI] & xmask;
+
 		found = -1;
 		for (i=0; i<cp->nr_of_tlbs; i++) {
 			int gbit = cp->tlbs[i].hi & TLB_G;
@@ -1561,6 +1606,7 @@ void coproc_tlbpr(struct cpu *cpu, int readflag)
 			}
 		}
 	}
+
 	if (found == -1)
 		cp->reg[COP0_INDEX] = INDEX_P;
 	else {
@@ -1622,19 +1668,19 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 	/*  Debug dump of the previous entry at that index:  */
 	fatal("{ old TLB entry at index %02x:", index);
 	if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
-		fatal(" hi=%08"PRIx32, (uint32_t)cp->tlbs[index].hi);
-		fatal(" lo=%08"PRIx32, (uint32_t)cp->tlbs[index].lo0);
+		fatal(" hi=%08" PRIx32, (uint32_t)cp->tlbs[index].hi);
+		fatal(" lo=%08" PRIx32, (uint32_t)cp->tlbs[index].lo0);
 	} else {
 		if (cpu->is_32bit) {
-			fatal(" mask=%08"PRIx32,(uint32_t)cp->tlbs[index].mask);
-			fatal(" hi=%08"PRIx32, (uint32_t)cp->tlbs[index].hi);
-			fatal(" lo0=%08"PRIx32, (uint32_t)cp->tlbs[index].lo0);
-			fatal(" lo1=%08"PRIx32, (uint32_t)cp->tlbs[index].lo1);
+			fatal(" mask=%08" PRIx32,(uint32_t)cp->tlbs[index].mask);
+			fatal(" hi=%08" PRIx32, (uint32_t)cp->tlbs[index].hi);
+			fatal(" lo0=%08" PRIx32, (uint32_t)cp->tlbs[index].lo0);
+			fatal(" lo1=%08" PRIx32, (uint32_t)cp->tlbs[index].lo1);
 		} else {
-			fatal(" mask=%016"PRIx64, cp->tlbs[index].mask);
-			fatal(" hi=%016"PRIx64, cp->tlbs[index].hi);
-			fatal(" lo0=%016"PRIx64, cp->tlbs[index].lo0);
-			fatal(" lo1=%016"PRIx64, cp->tlbs[index].lo1);
+			fatal(" mask=%016" PRIx64, cp->tlbs[index].mask);
+			fatal(" hi=%016" PRIx64, cp->tlbs[index].hi);
+			fatal(" lo0=%016" PRIx64, cp->tlbs[index].lo0);
+			fatal(" lo1=%016" PRIx64, cp->tlbs[index].lo1);
 		}
 	}
 	fatal(" }\n");
@@ -1667,9 +1713,6 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 	default:if (cpu->cd.mips.cpu_type.mmu_model == MMU10K) {
 			oldvaddr = cp->tlbs[index].hi &
 			    (ENTRYHI_VPN2_MASK_R10K | ENTRYHI_R_MASK);
-			/*  44 addressable bits:  */
-			if (oldvaddr & 0x80000000000ULL)
-				oldvaddr |= 0x3ffff00000000000ULL;
 		} else if (cpu->is_32bit) {
 			/*  MIPS32 etc.:  */
 			oldvaddr = cp->tlbs[index].hi & ENTRYHI_VPN2_MASK;
@@ -1678,20 +1721,42 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 			/*  Assume MMU4K  */
 			oldvaddr = cp->tlbs[index].hi &
 			    (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK);
-			/*  40 addressable bits:  */
-			if (oldvaddr & 0x8000000000ULL)
-				oldvaddr |= 0x3fffff0000000000ULL;
 		}
 
-		/*
-		 *  TODO: non-4KB page sizes!
-		 */
-		if (cp->tlbs[index].lo0 & ENTRYLO_V)
-			cpu->invalidate_translation_caches(cpu, oldvaddr,
-			    INVALIDATE_VADDR);
-		if (cp->tlbs[index].lo1 & ENTRYLO_V)
-			cpu->invalidate_translation_caches(cpu, oldvaddr|0x1000,
-			    INVALIDATE_VADDR);
+		{
+			// This is the "dual page" mask:
+			uint64_t dpmask = cp->tlbs[index].mask;
+			if (cpu->cd.mips.cpu_type.rev == MIPS_R4100) {
+				dpmask |= 0x07ff;
+			} else {
+				dpmask |= 0x1fff;
+			}
+
+			if (dpmask == 0x7ff) {
+				if (cp->tlbs[index].lo0 & ENTRYLO_V ||
+				    cp->tlbs[index].lo1 & ENTRYLO_V) {
+					fatal("1KB pages don't work with dyntrans.\n");
+					exit(1);
+				}
+			}
+
+			// Mask for bytes within a single page:
+			uint64_t mask = dpmask >> 1;
+
+			uint64_t pagesize = mask + 1;
+			
+			oldvaddr &= ~dpmask;
+
+			// printf("pagesize = %016llx mask = %016llx\n", pagesize, mask);
+			
+			if (cp->tlbs[index].lo0 & ENTRYLO_V)
+				for (uint64_t ofs = 0; ofs < pagesize; ofs += 0x1000)
+					cpu->invalidate_translation_caches(cpu, oldvaddr + ofs, INVALIDATE_VADDR);
+			
+			if (cp->tlbs[index].lo1 & ENTRYLO_V)
+				for (uint64_t ofs = 0; ofs < pagesize; ofs += 0x1000)
+					cpu->invalidate_translation_caches(cpu, oldvaddr + ofs + pagesize, INVALIDATE_VADDR);
+		}
 	}
 
 #if 0
@@ -1741,7 +1806,6 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 	if (cpu->cd.mips.cpu_type.mmu_model == MMU3K) {
 		uint32_t vaddr, paddr;
 		int wf = cp->reg[COP0_ENTRYLO0] & R2K3K_ENTRYLO_D? 1 : 0;
-		unsigned char *memblock = NULL;
 
 		cp->tlbs[index].hi = cp->reg[COP0_ENTRYHI];
 		cp->tlbs[index].lo0 = cp->reg[COP0_ENTRYLO0];
@@ -1749,8 +1813,6 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 		vaddr =  cp->reg[COP0_ENTRYHI] & R2K3K_ENTRYHI_VPN_MASK;
 		paddr = cp->reg[COP0_ENTRYLO0] & R2K3K_ENTRYLO_PFN_MASK;
 
-		memblock = memory_paddr_to_hostaddr(cpu->mem, paddr, 0);
-
 		/*  Invalidate any code translation, if we are writing
 		    a Dirty page to the TLB:  */
 		if (wf) {
@@ -1771,17 +1833,15 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 
 		/*  If we have a memblock (host page) for the physical
 		    page, then add a translation for it immediately:  */
-		if (memblock != NULL &&
-		    cp->reg[COP0_ENTRYLO0] & R2K3K_ENTRYLO_V)
-			cpu->update_translation_table(cpu, vaddr, memblock,
-			    wf, paddr);
+		if (cp->reg[COP0_ENTRYLO0] & R2K3K_ENTRYLO_V) {
+		    	unsigned char *memblock = memory_paddr_to_hostaddr(cpu->mem, paddr, 0);
+			if (memblock != NULL)
+				cpu->update_translation_table(cpu, vaddr, memblock, wf, paddr);
+		}
 	} else {
-		/*  R4000 etc.:  */
-		unsigned char *memblock = NULL;
-		int pfn_shift = 12, vpn_shift = 12;
-		int wf0, wf1, mask;
-		uint64_t vaddr0, vaddr1, paddr0, paddr1, ptmp;
-		uint64_t psize;
+		/*  R4000, R10000, VR41xx etc.:  */
+		int wf0, wf1;
+		uint64_t vaddr0, vaddr1, ptmp, psize;
 
 		cp->tlbs[index].mask = cp->reg[COP0_PAGEMASK];
 		cp->tlbs[index].hi   = cp->reg[COP0_ENTRYHI];
@@ -1791,64 +1851,57 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 		wf0 = cp->tlbs[index].lo0 & ENTRYLO_D;
 		wf1 = cp->tlbs[index].lo1 & ENTRYLO_D;
 
-		mask = cp->reg[COP0_PAGEMASK];
+		// This is the "dual page" mask:
+		uint64_t dpmask = cp->reg[COP0_PAGEMASK];
 		if (cpu->cd.mips.cpu_type.rev == MIPS_R4100) {
-			pfn_shift = 10;
-			mask |= 0x07ff;
+			dpmask |= 0x07ff;
 		} else {
-			mask |= 0x1fff;
+			dpmask |= 0x1fff;
 		}
-		switch (mask) {
-		case 0x00007ff:
+
+		if (dpmask == 0x7ff) {
 			if (cp->tlbs[index].lo0 & ENTRYLO_V ||
 			    cp->tlbs[index].lo1 & ENTRYLO_V) {
 				fatal("1KB pages don't work with dyntrans.\n");
 				exit(1);
 			}
-			vpn_shift = 10;
-			break;
-		case 0x0001fff:	break;
-		case 0x0007fff:	vpn_shift = 14; break;
-		case 0x001ffff:	vpn_shift = 16; break;
-		case 0x007ffff:	vpn_shift = 18; break;
-		case 0x01fffff:	vpn_shift = 20; break;
-		case 0x07fffff:	vpn_shift = 22; break;
-		case 0x1ffffff:	vpn_shift = 24; break;
-		case 0x7ffffff:	vpn_shift = 26; break;
-		default:fatal("Unimplemented MASK = 0x%016x\n", mask);
-			exit(1);
 		}
 
-		paddr0 = ((cp->tlbs[index].lo0 & ENTRYLO_PFN_MASK)
-		    >> ENTRYLO_PFN_SHIFT) << pfn_shift
-		    >> vpn_shift << vpn_shift;
-		paddr1 = ((cp->tlbs[index].lo1 & ENTRYLO_PFN_MASK)
-		    >> ENTRYLO_PFN_SHIFT) << pfn_shift
-		    >> vpn_shift << vpn_shift;
+		// Mask for bytes within a single page:
+		uint64_t mask = dpmask >> 1;
+
+		// printf("mask = %016llx\n", (long long)mask);
+
+		uint64_t pfn0 = (cp->tlbs[index].lo0 & ENTRYLO_PFN_MASK) >> ENTRYLO_PFN_SHIFT;
+		uint64_t pfn1 = (cp->tlbs[index].lo1 & ENTRYLO_PFN_MASK) >> ENTRYLO_PFN_SHIFT;
+
+		uint64_t paddr0, paddr1;
+
+		if (cpu->cd.mips.cpu_type.rev == MIPS_R4100) {
+			paddr0 = (pfn0 << 10) & ~mask;
+			paddr1 = (pfn1 << 10) & ~mask;
+		} else {
+			paddr0 = (pfn0 << 12) & ~mask;
+			paddr1 = (pfn1 << 12) & ~mask;
+		}
 
 		if (cpu->cd.mips.cpu_type.mmu_model == MMU10K) {
-			vaddr0 = cp->tlbs[index].hi &
-			    (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK_R10K);
-			/*  44 addressable bits:  */
-			if (vaddr0 & 0x80000000000ULL)
-				vaddr0 |= 0x3ffff00000000000ULL;
+			vaddr0 = cp->tlbs[index].hi & (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK_R10K);
 		} else if (cpu->is_32bit) {
 			/*  MIPS32 etc.:  */
 			vaddr0 = cp->tlbs[index].hi & ENTRYHI_VPN2_MASK;
 			vaddr0 = (int32_t)vaddr0;
 		} else {
 			/*  Assume MMU4K  */
-			vaddr0 = cp->tlbs[index].hi &
-			    (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK);
-			/*  40 addressable bits:  */
-			if (vaddr0 & 0x8000000000ULL)
-				vaddr0 |= 0x3fffff0000000000ULL;
+			vaddr0 = cp->tlbs[index].hi & (ENTRYHI_R_MASK | ENTRYHI_VPN2_MASK);
 		}
 
-		vaddr1 = vaddr0 | (1 << vpn_shift);
+		psize = mask + 1;
 
-		g_bit = (cp->reg[COP0_ENTRYLO0] &
-		    cp->reg[COP0_ENTRYLO1]) & ENTRYLO_G;
+		vaddr0 &= ~dpmask;
+		vaddr1 = vaddr0 | psize;
+
+		g_bit = (cp->reg[COP0_ENTRYLO0] & cp->reg[COP0_ENTRYLO1]) & ENTRYLO_G;
 
 		if (cpu->cd.mips.cpu_type.rev == MIPS_R4100) {
 			/*  NOTE: The VR4131 (and possibly others) don't have
@@ -1865,14 +1918,16 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 
 		/*
 		 *  Invalidate any code translations, if we are writing Dirty
-		 *  pages to the TLB:  (TODO: 4KB hardcoded... ugly)
+		 *  pages to the TLB:
 		 */
-		psize = 1 << pfn_shift;
-		for (ptmp = 0; ptmp < psize; ptmp += 0x1000) {
-			if (wf0)
+
+		if (wf0) {
+			for (ptmp = 0; ptmp < psize; ptmp += 0x1000)
 				cpu->invalidate_code_translation(cpu,
 				    paddr0 + ptmp, INVALIDATE_PADDR);
-			if (wf1)
+		}
+		if (wf1) {
+			for (ptmp = 0; ptmp < psize; ptmp += 0x1000)
 				cpu->invalidate_code_translation(cpu,
 				    paddr1 + ptmp, INVALIDATE_PADDR);
 		}
@@ -1887,14 +1942,18 @@ void coproc_tlbwri(struct cpu *cpu, int randomflag)
 		 *             be too expensive to add e.g. 16MB pages like
 		 *             this.
 		 */
-		memblock = memory_paddr_to_hostaddr(cpu->mem, paddr0, 0);
-		if (memblock != NULL && cp->reg[COP0_ENTRYLO0] & ENTRYLO_V)
-			cpu->update_translation_table(cpu, vaddr0, memblock,
-			    wf0, paddr0);
-		memblock = memory_paddr_to_hostaddr(cpu->mem, paddr1, 0);
-		if (memblock != NULL && cp->reg[COP0_ENTRYLO1] & ENTRYLO_V)
-			cpu->update_translation_table(cpu, vaddr1, memblock,
-			    wf1, paddr1);
+		if (psize == 0x1000) {
+			if (cp->reg[COP0_ENTRYLO0] & ENTRYLO_V) {
+				unsigned char *memblock = memory_paddr_to_hostaddr(cpu->mem, paddr0, 0);
+				if (memblock != NULL)
+					cpu->update_translation_table(cpu, vaddr0, memblock, wf0, paddr0);
+			}
+			if (cp->reg[COP0_ENTRYLO1] & ENTRYLO_V) {
+				unsigned char *memblock = memory_paddr_to_hostaddr(cpu->mem, paddr1, 0);
+				if (memblock != NULL)
+					cpu->update_translation_table(cpu, vaddr1, memblock, wf1, paddr1);
+			}
+		}
 
 		/*  Set new last_written_tlb_index hint:  */
 		cpu->cd.mips.last_written_tlb_index = index;
@@ -1987,6 +2046,13 @@ void coproc_function(struct cpu *cpu, struct mips_coproc *cp, int cpnr,
 
 	if (cpnr < 2 && (((function & 0x03e007f8) == (COPz_MTCz << 21))
 	              || ((function & 0x03e007f8) == (COPz_DMTCz << 21)))) {
+		tmpvalue = cpu->cd.mips.gpr[rt];
+		if (copz == COPz_MTCz) {
+			/*  Sign-extend:  */
+			tmpvalue &= 0xffffffffULL;
+			if (tmpvalue & 0x80000000ULL)
+				tmpvalue |= 0xffffffff00000000ULL;
+		}
 		if (unassemble_only) {
 			debug("%s%i\t%s,", copz==COPz_DMTCz? "dmtc" : "mtc",
 			    cpnr, regnames[rt]);
@@ -1996,16 +2062,10 @@ void coproc_function(struct cpu *cpu, struct mips_coproc *cp, int cpnr,
 				debug("r%i", rd);
 			if (function & 7)
 				debug(",%i", (int)(function & 7));
+			debug(" [%016llx]", (long long)tmpvalue);
 			debug("\n");
 			return;
 		}
-		tmpvalue = cpu->cd.mips.gpr[rt];
-		if (copz == COPz_MTCz) {
-			/*  Sign-extend:  */
-			tmpvalue &= 0xffffffffULL;
-			if (tmpvalue & 0x80000000ULL)
-				tmpvalue |= 0xffffffff00000000ULL;
-		}
 		coproc_register_write(cpu, cpu->cd.mips.coproc[cpnr], rd,
 		    &tmpvalue, copz == COPz_DMTCz, function & 7);
 		return;
@@ -2219,8 +2279,8 @@ void coproc_function(struct cpu *cpu, struct mips_coproc *cp, int cpnr,
 		return;
 	}
 
-	fatal("cpu%i: UNIMPLEMENTED coproc%i function %08"PRIx32" "
-	    "(pc = %016"PRIx64")\n", cpu->cpu_id, cp->coproc_nr,
+	fatal("cpu%i: UNIMPLEMENTED coproc%i function %08" PRIx32" "
+	    "(pc = %016" PRIx64")\n", cpu->cpu_id, cp->coproc_nr,
 	    (uint32_t)function, cpu->pc);
 
 	mips_cpu_exception(cpu, EXCEPTION_CPU, 0, 0, cp->coproc_nr, 0, 0, 0);
