@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -28,10 +28,11 @@
  *  PowerPC/POWER CPU emulation.
  */
 
+#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <ctype.h>
+#include <unistd.h>
 
 #include "cpu.h"
 #include "devices.h"
@@ -41,14 +42,13 @@
 #include "misc.h"
 #include "of.h"
 #include "opcodes_ppc.h"
-#include "ppc_bat.h"
-#include "ppc_pte.h"
-#include "ppc_spr.h"
 #include "ppc_spr_strings.h"
 #include "settings.h"
 #include "symbol.h"
-#include "useremul.h"
 
+#include "thirdparty/ppc_bat.h"
+#include "thirdparty/ppc_pte.h"
+#include "thirdparty/ppc_spr.h"
 
 #define	DYNTRANS_DUALMODE_32
 #include "tmp_ppc_head.c"
@@ -72,7 +72,6 @@ void ppc_irq_interrupt_deassert(struct interrupt *interrupt);
 int ppc_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 	int cpu_id, char *cpu_type_name)
 {
-	int any_cache = 0;
 	int i, found;
 	struct ppc_cpu_type_def cpu_type_defs[] = PPC_CPU_TYPE_DEFS;
 
@@ -92,7 +91,7 @@ int ppc_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 	cpu->memory_rw = ppc_memory_rw;
 
 	cpu->cd.ppc.cpu_type = cpu_type_defs[found];
-	cpu->name            = cpu->cd.ppc.cpu_type.name;
+	cpu->name            = strdup(cpu->cd.ppc.cpu_type.name);
 	cpu->byte_order      = EMUL_BIG_ENDIAN;
 	cpu->cd.ppc.mode     = MODE_PPC;	/*  TODO  */
 
@@ -138,30 +137,6 @@ int ppc_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 	cpu->translate_v2p = ppc_translate_v2p;
 
-	/*  Only show name and caches etc for CPU nr 0 (in SMP machines):  */
-	if (cpu_id == 0) {
-		debug("%s", cpu->cd.ppc.cpu_type.name);
-
-		if (cpu->cd.ppc.cpu_type.icache_shift != 0)
-			any_cache = 1;
-		if (cpu->cd.ppc.cpu_type.dcache_shift != 0)
-			any_cache = 1;
-		if (cpu->cd.ppc.cpu_type.l2cache_shift != 0)
-			any_cache = 1;
-
-		if (any_cache) {
-			debug(" (I+D = %i+%i KB",
-			    (int)(1 << (cpu->cd.ppc.cpu_type.icache_shift-10)),
-			    (int)(1 << (cpu->cd.ppc.cpu_type.dcache_shift-10)));
-			if (cpu->cd.ppc.cpu_type.l2cache_shift != 0) {
-				debug(", L2 = %i KB",
-				    (int)(1 << (cpu->cd.ppc.cpu_type.
-				    l2cache_shift-10)));
-			}
-			debug(")");
-		}
-	}
-
 	cpu->cd.ppc.spr[SPR_PIR] = cpu_id;
 
 	/*  Some default stack pointer value.  TODO: move this?  */
@@ -221,18 +196,21 @@ int ppc_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 
 	/*  Register the CPU as an interrupt handler:  */
 	{
-		struct interrupt template;
+		struct interrupt templ;
 		char name[150];
 		snprintf(name, sizeof(name), "%s", cpu->path);
-		memset(&template, 0, sizeof(template));
-		template.line = 0;
-		template.name = name;
-		template.extra = cpu;
-		template.interrupt_assert = ppc_irq_interrupt_assert;
-		template.interrupt_deassert = ppc_irq_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 0;
+		templ.name = name;
+		templ.extra = cpu;
+		templ.interrupt_assert = ppc_irq_interrupt_assert;
+		templ.interrupt_deassert = ppc_irq_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
+	if (cpu->is_32bit)
+		cpu->vaddr_mask = 0x00000000ffffffffULL;
+
 	return 1;
 }
 
@@ -262,35 +240,22 @@ void ppc_cpu_list_available_types(void)
 /*
  *  ppc_cpu_dumpinfo():
  */
-void ppc_cpu_dumpinfo(struct cpu *cpu)
+void ppc_cpu_dumpinfo(struct cpu *cpu, bool verbose)
 {
 	struct ppc_cpu_type_def *ct = &cpu->cd.ppc.cpu_type;
+	int l2kb = 0;
+	if (ct->l2cache_shift)
+		l2kb = (1 << ct->l2cache_shift) / 1024;
 
-	debug(" (%i-bit ", cpu->cd.ppc.bits);
-
-	switch (cpu->cd.ppc.mode) {
-	case MODE_PPC:
-		debug("PPC");
-		break;
-	case MODE_POWER:
-		debug("POWER");
-		break;
-	default:
-		debug("_INTERNAL ERROR_");
-	}
-
-	debug(", I+D = %i+%i KB",
+	debugmsg(SUBSYS_MACHINE, "cpu", VERBOSITY_INFO,
+	    "%s (%i-bit, %s, I+D = %i+%i KB, L2 = %i %cB)",
+	    cpu->name,
+	    cpu->cd.ppc.bits,
+	    cpu->cd.ppc.mode == MODE_PPC ? "PPC" : "POWER",
 	    (1 << ct->icache_shift) / 1024,
-	    (1 << ct->dcache_shift) / 1024);
-
-	if (ct->l2cache_shift) {
-		int kb = (1 << ct->l2cache_shift) / 1024;
-		debug(", L2 = %i %cB",
-		    kb >= 1024? kb / 1024 : kb,
-		    kb >= 1024? 'M' : 'K');
-	}
-
-	debug(")\n");
+	    (1 << ct->dcache_shift) / 1024,
+	    l2kb >= 1024? l2kb / 1024 : l2kb,
+	    l2kb >= 1024? 'M' : 'K');
 }
 
 
@@ -362,7 +327,7 @@ void ppc_exception(struct cpu *cpu, int exception_nr)
 		cpu->cd.ppc.spr[SPR_SRR1] = (cpu->cd.ppc.msr & 0x87c0ffff);
 
 	if (!quiet_mode)
-		fatal("[ PPC Exception 0x%x; pc=0x%"PRIx64" ]\n",
+		fatal("[ PPC Exception 0x%x; pc=0x%" PRIx64" ]\n",
 		    exception_nr, cpu->pc);
 
 	/*  Disable External Interrupts, Recoverable Interrupt Mode,
@@ -402,17 +367,17 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 		debug("cpu%i: pc  = 0x", x);
 		if (bits32)
-			debug("%08"PRIx32, (uint32_t)cpu->pc);
+			debug("%08" PRIx32, (uint32_t)cpu->pc);
 		else
-			debug("%016"PRIx64, (uint64_t)cpu->pc);
+			debug("%016" PRIx64, (uint64_t)cpu->pc);
 		debug("  <%s>\n", symbol != NULL? symbol : " no symbol ");
 
 		debug("cpu%i: lr  = 0x", x);
 		if (bits32)
-			debug("%08"PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_LR]);
+			debug("%08" PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_LR]);
 		else
-			debug("%016"PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_LR]);
-		debug("  cr  = 0x%08"PRIx32, (uint32_t)cpu->cd.ppc.cr);
+			debug("%016" PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_LR]);
+		debug("  cr  = 0x%08" PRIx32, (uint32_t)cpu->cd.ppc.cr);
 
 		if (bits32)
 			debug("  ");
@@ -420,15 +385,15 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 			debug("\ncpu%i: ", x);
 		debug("ctr = 0x", x);
 		if (bits32)
-			debug("%08"PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_CTR]);
+			debug("%08" PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_CTR]);
 		else
-			debug("%016"PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_CTR]);
+			debug("%016" PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_CTR]);
 
 		debug("  xer = 0x", x);
 		if (bits32)
-			debug("%08"PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_XER]);
+			debug("%08" PRIx32, (uint32_t)cpu->cd.ppc.spr[SPR_XER]);
 		else
-			debug("%016"PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_XER]);
+			debug("%016" PRIx64, (uint64_t)cpu->cd.ppc.spr[SPR_XER]);
 
 		debug("\n");
 
@@ -437,7 +402,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 			for (i=0; i<PPC_NGPRS; i++) {
 				if ((i % 4) == 0)
 					debug("cpu%i:", x);
-				debug(" r%02i = 0x%08"PRIx32" ", i,
+				debug(" r%02i = 0x%08" PRIx32" ", i,
 				    (uint32_t) cpu->cd.ppc.gpr[i]);
 				if ((i % 4) == 3)
 					debug("\n");
@@ -448,7 +413,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 				int r = (i >> 1) + ((i & 1) << 4);
 				if ((i % 2) == 0)
 					debug("cpu%i:", x);
-				debug(" r%02i = 0x%016"PRIx64" ", r,
+				debug(" r%02i = 0x%016" PRIx64" ", r,
 				    (uint64_t) cpu->cd.ppc.gpr[r]);
 				if ((i % 2) == 1)
 					debug("\n");
@@ -457,13 +422,13 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 		/*  Other special registers:  */
 		if (bits32) {
-			debug("cpu%i: srr0 = 0x%08"PRIx32
-			    " srr1 = 0x%08"PRIx32"\n", x,
+			debug("cpu%i: srr0 = 0x%08" PRIx32
+			    " srr1 = 0x%08" PRIx32"\n", x,
 			    (uint32_t) cpu->cd.ppc.spr[SPR_SRR0],
 			    (uint32_t) cpu->cd.ppc.spr[SPR_SRR1]);
 		} else {
-			debug("cpu%i: srr0 = 0x%016"PRIx64
-			    "  srr1 = 0x%016"PRIx64"\n", x,
+			debug("cpu%i: srr0 = 0x%016" PRIx64
+			    "  srr1 = 0x%016" PRIx64"\n", x,
 			    (uint64_t) cpu->cd.ppc.spr[SPR_SRR0],
 			    (uint64_t) cpu->cd.ppc.spr[SPR_SRR1]);
 		}
@@ -471,25 +436,25 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		debug("cpu%i: msr = ", x);
 		reg_access_msr(cpu, &tmp, 0, 0);
 		if (bits32)
-			debug("0x%08"PRIx32, (uint32_t) tmp);
+			debug("0x%08" PRIx32, (uint32_t) tmp);
 		else
-			debug("0x%016"PRIx64, (uint64_t) tmp);
+			debug("0x%016" PRIx64, (uint64_t) tmp);
 
-		debug("  tb  = 0x%08"PRIx32"%08"PRIx32"\n",
+		debug("  tb  = 0x%08" PRIx32"%08" PRIx32"\n",
 		    (uint32_t) cpu->cd.ppc.spr[SPR_TBU],
 		    (uint32_t) cpu->cd.ppc.spr[SPR_TBL]);
 
-		debug("cpu%i: dec = 0x%08"PRIx32,
+		debug("cpu%i: dec = 0x%08" PRIx32,
 		    x, (uint32_t) cpu->cd.ppc.spr[SPR_DEC]);
 		if (!bits32)
-			debug("  hdec = 0x%08"PRIx32"\n",
+			debug("  hdec = 0x%08" PRIx32"\n",
 			    (uint32_t) cpu->cd.ppc.spr[SPR_HDEC]);
 
 		debug("\n");
 	}
 
 	if (coprocs & 1) {
-		debug("cpu%i: fpscr = 0x%08"PRIx32"\n",
+		debug("cpu%i: fpscr = 0x%08" PRIx32"\n",
 		    x, (uint32_t) cpu->cd.ppc.fpscr);
 
 		/*  TODO: show floating-point values :-)  */
@@ -499,7 +464,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		for (i=0; i<PPC_NFPRS; i++) {
 			if ((i % 2) == 0)
 				debug("cpu%i:", x);
-			debug(" f%02i = 0x%016"PRIx64" ", i,
+			debug(" f%02i = 0x%016" PRIx64" ", i,
 			    (uint64_t) cpu->cd.ppc.fpr[i]);
 			if ((i % 2) == 1)
 				debug("\n");
@@ -507,7 +472,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 	}
 
 	if (coprocs & 2) {
-		debug("cpu%i:  sdr1 = 0x%"PRIx64"\n", x,
+		debug("cpu%i:  sdr1 = 0x%" PRIx64"\n", x,
 		    (uint64_t) cpu->cd.ppc.spr[SPR_SDR1]);
 		if (cpu->cd.ppc.cpu_type.flags & PPC_601)
 			debug("cpu%i:  PPC601-style, TODO!\n");
@@ -518,8 +483,8 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 				uint32_t lower = cpu->cd.ppc.spr[spr+1];
 				uint32_t len = (((upper & BAT_BL) << 15)
 				    | 0x1ffff) + 1;
-				debug("cpu%i:  %sbat%i: u=0x%08"PRIx32
-				    " l=0x%08"PRIx32" ",
+				debug("cpu%i:  %sbat%i: u=0x%08" PRIx32
+				    " l=0x%08" PRIx32" ",
 				    x, i<4? "i" : "d", i&3, upper, lower);
 				if (!(upper & BAT_V)) {
 					debug(" (not valid)\n");
@@ -556,7 +521,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 			uint32_t s = cpu->cd.ppc.sr[i];
 
 			debug("cpu%i:", x);
-			debug("  sr%-2i = 0x%08"PRIx32, i, s);
+			debug("  sr%-2i = 0x%08" PRIx32, i, s);
 
 			s &= (SR_TYPE | SR_SUKEY | SR_PRKEY | SR_NOEXEC);
 			if (s != 0) {
@@ -594,7 +559,7 @@ void ppc_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
  *
  *  Not currently used for PPC.
  */
-void ppc_cpu_tlbdump(struct machine *m, int x, int rawflag)
+void ppc_cpu_tlbdump(struct cpu* cpu, int rawflag)
 {
 }
 
@@ -634,12 +599,12 @@ void ppc_irq_interrupt_deassert(struct interrupt *interrupt)
 int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 	int running, uint64_t dumpaddr)
 {
-	int hi6, xo, lev, rt, rs, ra, rb, imm, sh, me, rc, l_bit, oe_bit;
+	int hi6, xo, lev, rt, rs, ra, rb, imm, sh, me, rc, l_bit; //, oe_bit;
 	int spr, aa_bit, lk_bit, bf, bh, bi, bo, mb, nb, bt, ba, bb, fpreg;
 	int bfa, to, load, wlen, no_rb = 0;
 	uint64_t offset, addr;
 	uint32_t iword;
-	char *symbol, *mnem = "ERROR";
+	const char *symbol, *mnem = "ERROR";
 	int power = cpu->cd.ppc.mode == MODE_POWER;
 
 	if (running)
@@ -654,15 +619,18 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		debug("cpu%i: ", cpu->cpu_id);
 
 	if (cpu->cd.ppc.bits == 32)
-		debug("%08"PRIx32, (uint32_t) dumpaddr);
+		debug("%08" PRIx32, (uint32_t) dumpaddr);
 	else
-		debug("%016"PRIx64, (uint64_t) dumpaddr);
+		debug("%016" PRIx64, (uint64_t) dumpaddr);
 
 	/*  NOTE: Fixed to big-endian.  */
 	iword = (instr[0] << 24) + (instr[1] << 16) + (instr[2] << 8)
 	    + instr[3];
 
-	debug(": %08"PRIx32"\t", iword);
+	debug(": %08" PRIx32, iword);
+
+	cpu_print_pc_indicator_in_disassembly(cpu, running, dumpaddr);
+
 
 	/*
 	 *  Decode the instruction:
@@ -763,9 +731,9 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		if (cpu->cd.ppc.bits == 32)
 			addr &= 0xffffffff;
 		if (cpu->cd.ppc.bits == 32)
-			debug("0x%"PRIx32, (uint32_t) addr);
+			debug("0x%" PRIx32, (uint32_t) addr);
 		else
-			debug("0x%"PRIx64, (uint64_t) addr);
+			debug("0x%" PRIx64, (uint64_t) addr);
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    addr, &offset);
 		if (symbol != NULL)
@@ -796,9 +764,9 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		if (cpu->cd.ppc.bits == 32)
 			addr &= 0xffffffff;
 		if (cpu->cd.ppc.bits == 32)
-			debug("\t0x%"PRIx32, (uint32_t) addr);
+			debug("\t0x%" PRIx32, (uint32_t) addr);
 		else
-			debug("\t0x%"PRIx64, (uint64_t) addr);
+			debug("\t0x%" PRIx64, (uint64_t) addr);
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    addr, &offset);
 		if (symbol != NULL)
@@ -1087,7 +1055,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			if (symbol != NULL)
 				debug(" \t<%s", symbol);
 			else
-				debug(" \t<0x%"PRIx64, (uint64_t) addr);
+				debug(" \t<0x%" PRIx64, (uint64_t) addr);
 			if (wlen > 0 && !fpreg /* && !reverse */) {
 				/*  TODO  */
 			}
@@ -1097,7 +1065,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		case PPC_31_NEGO:
 			rt = (iword >> 21) & 31;
 			ra = (iword >> 16) & 31;
-			oe_bit = (iword >> 10) & 1;
+			// oe_bit = (iword >> 10) & 1;
 			rc = iword & 1;
 			switch (xo) {
 			case PPC_31_NEG:  mnem = "neg"; break;
@@ -1120,7 +1088,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		case PPC_31_ADDZEO:
 			rt = (iword >> 21) & 31;
 			ra = (iword >> 16) & 31;
-			oe_bit = (iword >> 10) & 1;
+			// oe_bit = (iword >> 10) & 1;
 			rc = iword & 1;
 			switch (xo) {
 			case PPC_31_ADDZE:
@@ -1179,7 +1147,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			rt = (iword >> 21) & 31;
 			ra = (iword >> 16) & 31;
 			rb = (iword >> 11) & 31;
-			oe_bit = (iword >> 10) & 1;
+			// oe_bit = (iword >> 10) & 1;
 			rc = iword & 1;
 			switch (xo) {
 			case PPC_31_ADDC:
@@ -1258,10 +1226,10 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			    ppc_spr_names[spr]==NULL? "?" : ppc_spr_names[spr]);
 			if (running) {
 				if (cpu->cd.ppc.bits == 32)
-					debug(": 0x%"PRIx32, (uint32_t)
+					debug(": 0x%" PRIx32, (uint32_t)
 					    cpu->cd.ppc.spr[spr]);
 				else
-					debug(": 0x%"PRIx64, (uint64_t)
+					debug(": 0x%" PRIx64, (uint64_t)
 					    cpu->cd.ppc.spr[spr]);
 			}
 			debug(">");
@@ -1392,7 +1360,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			rt = (iword >> 21) & 31;
 			ra = (iword >> 16) & 31;
 			rb = (iword >> 11) & 31;
-			oe_bit = (iword >> 10) & 1;
+			// oe_bit = (iword >> 10) & 1;
 			rc = iword & 1;
 			switch (xo) {
 			case PPC_31_DIVWU:  mnem = "divwu"; break;
@@ -1418,10 +1386,10 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			    ppc_spr_names[spr]==NULL? "?" : ppc_spr_names[spr]);
 			if (running) {
 				if (cpu->cd.ppc.bits == 32)
-					debug(": 0x%"PRIx32, (uint32_t)
+					debug(": 0x%" PRIx32, (uint32_t)
 					    cpu->cd.ppc.gpr[rs]);
 				else
-					debug(": 0x%"PRIx64, (uint64_t)
+					debug(": 0x%" PRIx64, (uint64_t)
 					    cpu->cd.ppc.gpr[rs]);
 			}
 			debug(">");
@@ -1574,7 +1542,7 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 		if (symbol != NULL)
 			debug(" \t<%s", symbol);
 		else
-			debug(" \t<0x%"PRIx64, (uint64_t) addr);
+			debug(" \t<0x%" PRIx64, (uint64_t) addr);
 		if (wlen > 0 && load && wlen > 0) {
 			unsigned char tw[8];
 			uint64_t tdata = 0;
@@ -1598,12 +1566,12 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 					if (symbol != NULL)
 						debug("%s", symbol);
 					else
-						debug("0x%"PRIx64,
+						debug("0x%" PRIx64,
 						    (uint64_t) tdata);
 				} else {
 					/*  TODO: if load==2, then this is
 					    a _signed_ load.  */
-					debug("0x%"PRIx64, (uint64_t) tdata);
+					debug("0x%" PRIx64, (uint64_t) tdata);
 				}
 			} else
 				debug(": unreadable");
@@ -1621,12 +1589,12 @@ int ppc_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 				if (symbol != NULL)
 					debug("%s", symbol);
 				else
-					debug("0x%"PRIx64, (uint64_t) tdata);
+					debug("0x%" PRIx64, (uint64_t) tdata);
 			} else {
 				if (tdata > -256 && tdata < 256)
 					debug("%i", (int)tdata);
 				else
-					debug("0x%"PRIx64, (uint64_t) tdata);
+					debug("0x%" PRIx64, (uint64_t) tdata);
 			}
 		}
 		debug(">");
@@ -1818,7 +1786,7 @@ static void debug_spr_usage(uint64_t pc, int spr)
 			break;
 		} else
 			fatal("[ using UNIMPLEMENTED spr %i (%s), pc = "
-			    "0x%"PRIx64" ]\n", spr, ppc_spr_names[spr] == NULL?
+			    "0x%" PRIx64" ]\n", spr, ppc_spr_names[spr] == NULL?
 			    "UNKNOWN" : ppc_spr_names[spr], (uint64_t) pc);
 	}
 
