@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -893,6 +893,21 @@ X(fmr)
 }
 
 
+/*
+ *  fabs:  Floating-point Absulute Value
+ *
+ *  arg[0] = ptr to frb
+ *  arg[1] = ptr to frt
+ */
+X(fabs)
+{
+	uint64_t v;
+	CHECK_FOR_FPU_EXCEPTION;
+	v = *(uint64_t *)ic->arg[0];
+	*(uint64_t *)ic->arg[1] = v & 0x7fffffffffffffffULL;
+}
+
+
 /*
  *  fneg:  Floating-point Negate
  *
@@ -971,8 +986,7 @@ X(frsp)
 	/*  TODO: Signaling vs Quiet NaN  */
 	cpu->cd.ppc.fpscr &= ~(PPC_FPSCR_FPCC | PPC_FPSCR_VXNAN);
 	cpu->cd.ppc.fpscr |= (c << PPC_FPSCR_FPCC_SHIFT);
-	(*(uint64_t *)ic->arg[1]) =
-	    ieee_store_float_value(fl, IEEE_FMT_D, frb.nan);
+	(*(uint64_t *)ic->arg[1]) = ieee_store_float_value(fl, IEEE_FMT_D);
 }
 
 
@@ -996,8 +1010,9 @@ X(fctiwz)
 		else if (frb.f <= -2147483648.0)
 			res = 0x80000000;
 		else
-			res = frb.f;
+			res = (int32_t) frb.f;
 	}
+
 	*(uint64_t *)ic->arg[1] = (uint32_t)res;
 }
 
@@ -1014,17 +1029,14 @@ X(fmul)
 	struct ieee_float_value fra;
 	struct ieee_float_value frc;
 	double result = 0.0;
-	int c, nan = 0;
+	int c;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[2], &frc, IEEE_FMT_D);
-	if (fra.nan || frc.nan)
-		nan = 1;
-	else
-		result = fra.f * frc.f;
-	if (nan)
+	result = fra.f * frc.f;
+	if (isnan(result))
 		c = 1;
 	else {
 		if (result < 0.0)
@@ -1039,7 +1051,7 @@ X(fmul)
 	cpu->cd.ppc.fpscr |= (c << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[0]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 X(fmuls)
 {
@@ -1063,18 +1075,15 @@ X(fmadd)
 	struct ieee_float_value frb;
 	struct ieee_float_value frc;
 	double result = 0.0;
-	int nan = 0, cc;
+	int cc;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(cpu->cd.ppc.fpr[b], &frb, IEEE_FMT_D);
 	ieee_interpret_float_value(cpu->cd.ppc.fpr[c], &frc, IEEE_FMT_D);
-	if (fra.nan || frb.nan || frc.nan)
-		nan = 1;
-	else
-		result = fra.f * frc.f + frb.f;
-	if (nan)
+	result = fra.f * frc.f + frb.f;
+	if (isnan(result))
 		cc = 1;
 	else {
 		if (result < 0.0)
@@ -1089,7 +1098,7 @@ X(fmadd)
 	cpu->cd.ppc.fpscr |= (cc << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[0]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 
 
@@ -1108,18 +1117,15 @@ X(fmsub)
 	struct ieee_float_value frb;
 	struct ieee_float_value frc;
 	double result = 0.0;
-	int nan = 0, cc;
+	int cc;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(cpu->cd.ppc.fpr[b], &frb, IEEE_FMT_D);
 	ieee_interpret_float_value(cpu->cd.ppc.fpr[c], &frc, IEEE_FMT_D);
-	if (fra.nan || frb.nan || frc.nan)
-		nan = 1;
-	else
-		result = fra.f * frc.f - frb.f;
-	if (nan)
+	result = fra.f * frc.f - frb.f;
+	if (isnan(result))
 		cc = 1;
 	else {
 		if (result < 0.0)
@@ -1134,7 +1140,7 @@ X(fmsub)
 	cpu->cd.ppc.fpscr |= (cc << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[0]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 
 
@@ -1150,17 +1156,14 @@ X(fadd)
 	struct ieee_float_value fra;
 	struct ieee_float_value frb;
 	double result = 0.0;
-	int nan = 0, c;
+	int c;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[0], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &frb, IEEE_FMT_D);
-	if (fra.nan || frb.nan)
-		nan = 1;
-	else
-		result = fra.f + frb.f;
-	if (nan)
+	result = fra.f + frb.f;
+	if (isnan(result))
 		c = 1;
 	else {
 		if (result < 0.0)
@@ -1175,7 +1178,7 @@ X(fadd)
 	cpu->cd.ppc.fpscr |= (c << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[2]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 X(fadds)
 {
@@ -1187,17 +1190,14 @@ X(fsub)
 	struct ieee_float_value fra;
 	struct ieee_float_value frb;
 	double result = 0.0;
-	int nan = 0, c;
+	int c;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[0], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &frb, IEEE_FMT_D);
-	if (fra.nan || frb.nan)
-		nan = 1;
-	else
-		result = fra.f - frb.f;
-	if (nan)
+	result = fra.f - frb.f;
+	if (isnan(result))
 		c = 1;
 	else {
 		if (result < 0.0)
@@ -1212,7 +1212,7 @@ X(fsub)
 	cpu->cd.ppc.fpscr |= (c << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[2]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 X(fsubs)
 {
@@ -1224,17 +1224,14 @@ X(fdiv)
 	struct ieee_float_value fra;
 	struct ieee_float_value frb;
 	double result = 0.0;
-	int nan = 0, c;
+	int c;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[0], &fra, IEEE_FMT_D);
 	ieee_interpret_float_value(*(uint64_t *)ic->arg[1], &frb, IEEE_FMT_D);
-	if (fra.nan || frb.nan || frb.f == 0)
-		nan = 1;
-	else
-		result = fra.f / frb.f;
-	if (nan)
+	result = fra.f / frb.f;
+	if (isnan(result))
 		c = 1;
 	else {
 		if (result < 0.0)
@@ -1249,7 +1246,7 @@ X(fdiv)
 	cpu->cd.ppc.fpscr |= (c << PPC_FPSCR_FPCC_SHIFT);
 
 	(*(uint64_t *)ic->arg[2]) =
-	    ieee_store_float_value(result, IEEE_FMT_D, nan);
+	    ieee_store_float_value(result, IEEE_FMT_D);
 }
 X(fdivs)
 {
@@ -1273,6 +1270,7 @@ X(llsc)
 	switch (xo) {
 	case PPC_31_LDARX:
 		len = 8;
+		// fall through
 	case PPC_31_LWARX:
 		load = 1;
 		break;
@@ -1722,6 +1720,15 @@ X(mtspr) {
 	/*  TODO: Check permission  */
 	reg(ic->arg[1]) = reg(ic->arg[0]);
 }
+X(mtspr_sprg2) {
+	if (cpu->cd.ppc.bits == 32) {
+		// Ignore for now. FreeBSD/powerpc seems to write 0xffffffe0
+		// here, and read it back. If it is non-zero, it assumes a 64-bit
+		// cpu.
+	} else {
+		reg(ic->arg[1]) = reg(ic->arg[0]);
+	}
+}
 X(mtlr) {
 	cpu->cd.ppc.spr[SPR_LR] = reg(ic->arg[0]);
 }
@@ -2341,7 +2348,7 @@ X(lfs)
 		ieee_interpret_float_value(*(uint64_t *)ic->arg[0],
 		    &val, IEEE_FMT_S);
 		(*(uint64_t *)ic->arg[0]) =
-		    ieee_store_float_value(val.f, IEEE_FMT_D, val.nan);
+		    ieee_store_float_value(val.f, IEEE_FMT_D);
 	}
 }
 X(lfsx)
@@ -2371,7 +2378,7 @@ X(lfsx)
 		ieee_interpret_float_value(*(uint64_t *)ic->arg[0],
 		    &val, IEEE_FMT_S);
 		(*(uint64_t *)ic->arg[0]) =
-		    ieee_store_float_value(val.f, IEEE_FMT_D, val.nan);
+		    ieee_store_float_value(val.f, IEEE_FMT_D);
 	}
 }
 X(lfd)
@@ -2400,14 +2407,14 @@ X(lfdx)
 }
 X(stfs)
 {
-	uint64_t *old_arg0 = (void *)ic->arg[0];
+	uint64_t *old_arg0 = (uint64_t *) ic->arg[0];
 	struct ieee_float_value val;
 	uint64_t tmp_val;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*old_arg0, &val, IEEE_FMT_D);
-	tmp_val = ieee_store_float_value(val.f, IEEE_FMT_S, val.nan);
+	tmp_val = ieee_store_float_value(val.f, IEEE_FMT_S);
 
 	ic->arg[0] = (size_t)&tmp_val;
 
@@ -2423,14 +2430,14 @@ X(stfs)
 }
 X(stfsx)
 {
-	uint64_t *old_arg0 = (void *)ic->arg[0];
+	uint64_t *old_arg0 = (uint64_t *)ic->arg[0];
 	struct ieee_float_value val;
 	uint64_t tmp_val;
 
 	CHECK_FOR_FPU_EXCEPTION;
 
 	ieee_interpret_float_value(*old_arg0, &val, IEEE_FMT_D);
-	tmp_val = ieee_store_float_value(val.f, IEEE_FMT_S, val.nan);
+	tmp_val = ieee_store_float_value(val.f, IEEE_FMT_S);
 
 	ic->arg[0] = (size_t)&tmp_val;
 
@@ -2590,36 +2597,23 @@ X(sc)
 }
 
 
-/*
- *  user_syscall:  Userland syscall.
- *
- *  arg[0] = syscall "level" (usually 0)
- */
-X(user_syscall)
-{
-	useremul_syscall(cpu, ic->arg[0]);
-
-	if (!cpu->running) {
-		cpu->n_translated_instrs --;
-		cpu->cd.ppc.next_ic = &nothing_call;
-	}
-}
-
-
 /*
  *  openfirmware:
  */
 X(openfirmware)
 {
 	of_emul(cpu);
-	if (cpu->running == 0) {
-		cpu->n_translated_instrs --;
-		cpu->cd.ppc.next_ic = &nothing_call;
-	}
+
 	cpu->pc = cpu->cd.ppc.spr[SPR_LR];
 	if (cpu->machine->show_trace_tree)
 		cpu_functioncall_trace_return(cpu);
+
 	quick_pc_to_pointers(cpu);
+
+	if (!cpu->running) {
+		cpu->n_translated_instrs --;
+		cpu->cd.ppc.next_ic = &nothing_call;
+	}
 }
 
 
@@ -2768,7 +2762,7 @@ X(to_be_translated)
 		} else {
 			if (!cpu->translation_readahead)
 				fatal("[ TODO: Unimplemented ALTIVEC, iword"
-				    " = 0x%08"PRIx32"x ]\n", iword);
+				    " = 0x%08" PRIx32"x ]\n", iword);
 			goto bad;
 		}
 		break;
@@ -3003,9 +2997,7 @@ X(to_be_translated)
 	case PPC_HI6_SC:
 		ic->arg[0] = (iword >> 5) & 0x7f;
 		ic->arg[1] = (addr & 0xfff) + 4;
-		if (cpu->machine->userland_emul != NULL)
-			ic->f = instr(user_syscall);
-		else if (iword == 0x44ee0002) {
+		if (iword == 0x44ee0002) {
 			/*  Special case/magic hack for OpenFirmware emul:  */
 			ic->f = instr(openfirmware);
 		} else
@@ -3269,6 +3261,9 @@ X(to_be_translated)
 			ic->arg[0] = (size_t)(&cpu->cd.ppc.gpr[rt]);
 			ic->arg[1] = (size_t)(&cpu->cd.ppc.spr[spr]);
 			switch (spr) {
+			// Reuse SPR_TB* for TBR_TB*:
+			case TBR_TBL: ic->f = instr(mftb); break;
+			case TBR_TBU: ic->f = instr(mftbu); break;
 			case SPR_PMC1:	ic->f = instr(mfspr_pmc1); break;
 			default:	ic->f = instr(mfspr);
 			}
@@ -3287,6 +3282,9 @@ X(to_be_translated)
 			case SPR_CTR:
 				ic->f = instr(mtctr);
 				break;
+			case SPR_SPRG2:
+				ic->f = instr(mtspr_sprg2);
+				break;
 			default:ic->f = instr(mtspr);
 			}
 			break;
@@ -3863,11 +3861,13 @@ X(to_be_translated)
 			case PPC_63_FRSP:
 			case PPC_63_FCTIWZ:
 			case PPC_63_FNEG:
+			case PPC_63_FABS:
 			case PPC_63_FMR:
 				switch (xo) {
 				case PPC_63_FRSP:   ic->f = instr(frsp); break;
 				case PPC_63_FCTIWZ: ic->f = instr(fctiwz);break;
 				case PPC_63_FNEG:   ic->f = instr(fneg); break;
+				case PPC_63_FABS:   ic->f = instr(fabs); break;
 				case PPC_63_FMR:    ic->f = instr(fmr); break;
 				}
 				ic->arg[0] = (size_t)(&cpu->cd.ppc.fpr[rb]);
@@ -3897,7 +3897,7 @@ X(to_be_translated)
 
 
 #define	DYNTRANS_TO_BE_TRANSLATED_TAIL
-#include "cpu_dyntrans.c" 
+#include "cpu_dyntrans.c"
 #undef	DYNTRANS_TO_BE_TRANSLATED_TAIL
 }
 
