@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -62,9 +62,12 @@ void LS_GENERIC_N(struct cpu *cpu, struct ppc_instr_call *ic)
 
 #ifndef LS_B
 	if ((addr & 0xfff) + LS_SIZE-1 > 0xfff) {
-		fatal("PPC LOAD/STORE misalignment across page boundary: TODO"
+		debugmsg_cpu(cpu, SUBSYS_CPU, "ppc", VERBOSITY_ERROR,
+		    "PPC LOAD/STORE misalignment across page boundary: TODO"
 		    " (addr=0x%08x, LS_SIZE=%i)\n", (int)addr, LS_SIZE);
-		exit(1);
+		cpu->running = false;
+		cpu->cd.ppc.next_ic = &nothing_call;
+		return;
 	}
 #endif
 
@@ -72,6 +75,8 @@ void LS_GENERIC_N(struct cpu *cpu, struct ppc_instr_call *ic)
 	if (!cpu->memory_rw(cpu, cpu->mem, addr, data, sizeof(data),
 	    MEM_READ, CACHE_DATA)) {
 		/*  Exception.  */
+		if (!cpu->running)
+			cpu->cd.ppc.next_ic = &nothing_call;
 		return;
 	}
 #ifdef LS_B
@@ -156,6 +161,8 @@ void LS_GENERIC_N(struct cpu *cpu, struct ppc_instr_call *ic)
 	if (!cpu->memory_rw(cpu, cpu->mem, addr, data, sizeof(data),
 	    MEM_WRITE, CACHE_DATA)) {
 		/*  Exception.  */
+		if (!cpu->running)
+			cpu->cd.ppc.next_ic = &nothing_call;
 		return;
 	}
 #endif
@@ -169,6 +176,16 @@ void LS_GENERIC_N(struct cpu *cpu, struct ppc_instr_call *ic)
 
 void LS_N(struct cpu *cpu, struct ppc_instr_call *ic)
 {
+#ifndef MODE32
+	/************************************************************/
+	/* For now, 64-bit access is done using the slow fallback.  */
+	if (!cpu->is_32bit) {
+		LS_GENERIC_N(cpu, ic);
+		return;
+	}
+#endif
+
+
 #ifdef MODE32
 	uint32_t addr =
 #else
@@ -202,17 +219,6 @@ void LS_N(struct cpu *cpu, struct ppc_instr_call *ic)
 	}
 #endif
 
-
-#ifndef MODE32
-/*******************************************/
-if (!cpu->is_32bit) {
-LS_GENERIC_N(cpu, ic);
-return;
-}
-/*******************************************/
-#endif
-
-
 	if (page == NULL) {
 		LS_GENERIC_N(cpu, ic);
 		return;
