@@ -1,12 +1,12 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
  *
  *  1. Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
- *  2. Redistributions in binary form must reproduce the above copyright  
+ *  2. Redistributions in binary form must reproduce the above copyright
  *     notice, this list of conditions and the following disclaimer in the 
  *     documentation and/or other materials provided with the distribution.
  *  3. The name of the author may not be used to endorse or promote products
@@ -49,8 +49,8 @@
 #include "settings.h"
 #include "symbol.h"
 
-#include "sh4_exception.h"
-#include "sh4_mmu.h"
+#include "thirdparty/sh4_exception.h"
+#include "thirdparty/sh4_mmu.h"
 
 
 #define DYNTRANS_32
@@ -112,11 +112,6 @@ int sh_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 	cpu->invalidate_code_translation =
 	    sh_invalidate_code_translation;
 
-	/*  Only show name and caches etc for CPU nr 0 (in SMP machines):  */
-	if (cpu_id == 0) {
-		debug("%s", cpu->name);
-	}
-
 	/*  Initial value of FPSCR (according to the SH4 manual):  */
 	cpu->cd.sh.fpscr = 0x00040001;
 
@@ -132,7 +127,9 @@ int sh_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 	CPU_SETTINGS_ADD_REGISTER32("sr", cpu->cd.sh.sr);
 	CPU_SETTINGS_ADD_REGISTER32("pr", cpu->cd.sh.pr);
 	CPU_SETTINGS_ADD_REGISTER32("vbr", cpu->cd.sh.vbr);
+	CPU_SETTINGS_ADD_REGISTER32("sgr", cpu->cd.sh.sgr);
 	CPU_SETTINGS_ADD_REGISTER32("gbr", cpu->cd.sh.gbr);
+	CPU_SETTINGS_ADD_REGISTER32("dbr", cpu->cd.sh.dbr);
 	CPU_SETTINGS_ADD_REGISTER32("macl", cpu->cd.sh.macl);
 	CPU_SETTINGS_ADD_REGISTER32("mach", cpu->cd.sh.mach);
 	CPU_SETTINGS_ADD_REGISTER32("expevt", cpu->cd.sh.expevt);
@@ -175,32 +172,32 @@ int sh_cpu_new(struct cpu *cpu, struct memory *mem, struct machine *machine,
 	/*  Register the CPU's interrupts:  */
 	if (cpu->cd.sh.cpu_type.arch == 4) {
 		for (i=SH_INTEVT_NMI; i<0x1000; i+=0x20) {
-			struct interrupt template;
+			struct interrupt templ;
 			char name[100];
 			snprintf(name, sizeof(name), "%s.irq[0x%x]",
 			    cpu->path, i);
-			memset(&template, 0, sizeof(template));
-			template.line = i;
-			template.name = name;
-			template.extra = cpu;
-			template.interrupt_assert = sh_cpu_interrupt_assert;
-			template.interrupt_deassert = sh_cpu_interrupt_deassert;
-			interrupt_handler_register(&template);
+			memset(&templ, 0, sizeof(templ));
+			templ.line = i;
+			templ.name = name;
+			templ.extra = cpu;
+			templ.interrupt_assert = sh_cpu_interrupt_assert;
+			templ.interrupt_deassert = sh_cpu_interrupt_deassert;
+			interrupt_handler_register(&templ);
 		}
 	} else {
-		struct interrupt template;
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = cpu->path;
-		template.extra = cpu;
-		template.interrupt_assert = sh3_cpu_interrupt_assert;
-		template.interrupt_deassert = sh3_cpu_interrupt_deassert;
-		interrupt_handler_register(&template);
+		struct interrupt templ;
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = cpu->path;
+		templ.extra = cpu;
+		templ.interrupt_assert = sh3_cpu_interrupt_assert;
+		templ.interrupt_deassert = sh3_cpu_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*  SH4-specific memory mapped registers, TLBs, caches, etc:  */
 	if (cpu->cd.sh.cpu_type.arch == 4) {
-		cpu->cd.sh.pcic_pcibus = device_add(machine, "sh4");
+		cpu->cd.sh.pcic_pcibus = (struct pci_data *) device_add(machine, "sh4");
 
 		/*
 		 *  Interrupt Controller initial values, according to the
@@ -301,7 +298,7 @@ void sh3_cpu_interrupt_deassert(struct interrupt *interrupt)
  */
 void sh_cpu_interrupt_assert(struct interrupt *interrupt)
 {
-	struct cpu *cpu = interrupt->extra;
+	struct cpu *cpu = (struct cpu *) interrupt->extra;
 	unsigned int irq_nr = interrupt->line;
 	unsigned int index = irq_nr / 0x20;
 	unsigned int prio;
@@ -328,7 +325,7 @@ void sh_cpu_interrupt_assert(struct interrupt *interrupt)
  */
 void sh_cpu_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct cpu *cpu = interrupt->extra;
+	struct cpu *cpu = (struct cpu *) interrupt->extra;
 	int irq_nr = interrupt->line;
 	int index = irq_nr / 0x20;
 
@@ -378,9 +375,11 @@ void sh_cpu_list_available_types(void)
 /*
  *  sh_cpu_dumpinfo():
  */
-void sh_cpu_dumpinfo(struct cpu *cpu)
+void sh_cpu_dumpinfo(struct cpu *cpu, bool verbose)
 {
-	debug(" (%s-endian)\n",
+	debugmsg(SUBSYS_MACHINE, "cpu", VERBOSITY_INFO,
+	    "%s (%s-endian)",
+	    cpu->name,
 	    cpu->byte_order == EMUL_BIG_ENDIAN? "Big" : "Little");
 }
 
@@ -393,7 +392,7 @@ void sh_cpu_dumpinfo(struct cpu *cpu)
 int sh_cpu_instruction_has_delayslot(struct cpu *cpu, unsigned char *ib)
 {
 	uint16_t iword = *((uint16_t *)&ib[0]);
-	int hi4, lo4, lo8;
+	int hi4, lo8; // , lo4
 
 	if (!cpu->is_32bit)
 		return 0;
@@ -403,7 +402,9 @@ int sh_cpu_instruction_has_delayslot(struct cpu *cpu, unsigned char *ib)
 	else
 		iword = LE16_TO_HOST(iword);
 
-	hi4 = iword >> 12; lo4 = iword & 15; lo8 = iword & 255;
+	hi4 = iword >> 12;
+	lo8 = iword & 255;
+	// lo4 = iword & 15;
 
         switch (hi4) {
 	case 0x0:
@@ -458,10 +459,10 @@ void sh_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->pc, &offset);
 
-		debug("cpu%i: pc  = 0x%08"PRIx32, x, (uint32_t)cpu->pc);
+		debug("cpu%i: pc  = 0x%08" PRIx32, x, (uint32_t)cpu->pc);
 		debug("  <%s>\n", symbol != NULL? symbol : " no symbol ");
 
-		debug("cpu%i: sr  = 0x%08"PRIx32"  (%s, %s, %s, %s, %s, %s,"
+		debug("cpu%i: sr  = 0x%08" PRIx32"  (%s, %s, %s, %s, %s, %s,"
 		    " imask=0x%x, %s, %s)\n", x, (int32_t)cpu->cd.sh.sr,
 		    (cpu->cd.sh.sr & SH_SR_MD)? "MD" : "!md",
 		    (cpu->cd.sh.sr & SH_SR_RB)? "RB" : "!rb",
@@ -475,11 +476,11 @@ void sh_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 		symbol = get_symbol_name(&cpu->machine->symbol_context,
 		    cpu->cd.sh.pr, &offset);
-		debug("cpu%i: pr  = 0x%08"PRIx32, x, (uint32_t)cpu->cd.sh.pr);
+		debug("cpu%i: pr  = 0x%08" PRIx32, x, (uint32_t)cpu->cd.sh.pr);
 		debug("  <%s>\n", symbol != NULL? symbol : " no symbol ");
 
-		debug("cpu%i: mach = 0x%08"PRIx32"  macl = 0x%08"PRIx32
-		    "  gbr = 0x%08"PRIx32"\n", x, (uint32_t)cpu->cd.sh.mach,
+		debug("cpu%i: mach = 0x%08" PRIx32"  macl = 0x%08" PRIx32
+		    "  gbr = 0x%08" PRIx32"\n", x, (uint32_t)cpu->cd.sh.mach,
 		    (uint32_t)cpu->cd.sh.macl, (uint32_t)cpu->cd.sh.gbr);
 
 		for (i=0; i<SH_N_GPRS; i++) {
@@ -493,7 +494,7 @@ void sh_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 	if (coprocs & 1) {
 		/*  Floating point:  */
-		debug("cpu%i: fpscr = 0x%08"PRIx32" (%s,%s,%s)  fpul = 0x%08"
+		debug("cpu%i: fpscr = 0x%08" PRIx32" (%s,%s,%s)  fpul = 0x%08"
 		    PRIx32"\n", x, cpu->cd.sh.fpscr,
 		    cpu->cd.sh.fpscr & SH_FPSCR_PR? "PR" : "!pr",
 		    cpu->cd.sh.fpscr & SH_FPSCR_SZ? "SZ" : "!sz",
@@ -519,12 +520,13 @@ void sh_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
 
 	if (coprocs & 2) {
 		/*  System registers, etc:  */
-		debug("cpu%i: vbr = 0x%08"PRIx32"  sgr = 0x%08"PRIx32
-		    "\n", x, cpu->cd.sh.vbr, cpu->cd.sh.sgr);
-		debug("cpu%i: spc = 0x%08"PRIx32"  ssr = 0x%08"PRIx32"\n",
+		debug("cpu%i: vbr = 0x%08" PRIx32"  sgr = 0x%08" PRIx32
+		    "  dbr = 0x%08" PRIx32"\n", x, cpu->cd.sh.vbr, cpu->cd.sh.sgr,
+		    cpu->cd.sh.dbr);
+		debug("cpu%i: spc = 0x%08" PRIx32"  ssr = 0x%08" PRIx32"\n",
 		    x, cpu->cd.sh.spc, cpu->cd.sh.ssr);
-		debug("cpu%i: expevt = 0x%"PRIx32"  intevt = 0x%"PRIx32
-		    "  tra = 0x%"PRIx32"\n", x, cpu->cd.sh.expevt,
+		debug("cpu%i: expevt = 0x%" PRIx32"  intevt = 0x%" PRIx32
+		    "  tra = 0x%" PRIx32"\n", x, cpu->cd.sh.expevt,
 		    cpu->cd.sh.intevt, cpu->cd.sh.tra);
 
 		for (i=0; i<SH_N_GPRS_BANKED; i++) {
@@ -543,29 +545,27 @@ void sh_cpu_register_dump(struct cpu *cpu, int gprs, int coprocs)
  *  sh_cpu_tlbdump():
  *
  *  Called from the debugger to dump the TLB in a readable format.
- *  x is the cpu number to dump, or -1 to dump all CPUs.
  *
  *  If rawflag is nonzero, then the TLB contents isn't formated nicely,
  *  just dumped.
  */
-void sh_cpu_tlbdump(struct machine *m, int x, int rawflag)
+void sh_cpu_tlbdump(struct cpu* cpu, int rawflag)
 {
 	int i, j;
+	struct machine* m = cpu->machine;
 
 	for (j=0; j<m->ncpus; j++) {
-		struct cpu *cpu = m->cpus[j];
-
-		if (x >= 0 && j != x)
+		if (cpu != m->cpus[j])
 			continue;
 
 		for (i=0; i<SH_N_ITLB_ENTRIES; i++)
-			printf("cpu%i: itlb_hi_%-2i = 0x%08"PRIx32"  "
-			    "itlb_lo_%-2i = 0x%08"PRIx32"\n", j, i,
+			printf("cpu%i: itlb_hi_%-2i = 0x%08" PRIx32"  "
+			    "itlb_lo_%-2i = 0x%08" PRIx32"\n", j, i,
 			    (uint32_t) cpu->cd.sh.itlb_hi[i], i,
 			    (uint32_t) cpu->cd.sh.itlb_lo[i]);
 		for (i=0; i<SH_N_UTLB_ENTRIES; i++)
-			printf("cpu%i: utlb_hi_%-2i = 0x%08"PRIx32"  "
-			    "utlb_lo_%-2i = 0x%08"PRIx32"\n", j, i,
+			printf("cpu%i: utlb_hi_%-2i = 0x%08" PRIx32"  "
+			    "utlb_lo_%-2i = 0x%08" PRIx32"\n", j, i,
 			    (uint32_t) cpu->cd.sh.utlb_hi[i], i,
 			    (uint32_t) cpu->cd.sh.utlb_lo[i]);
 	}
@@ -634,9 +634,9 @@ void sh_exception(struct cpu *cpu, int expevt, int intevt, uint32_t vaddr)
 		else
 			debug("[ exception 0x%03x", expevt);
 
-		debug(", pc=0x%08"PRIx32" ", (uint32_t)cpu->pc);
+		debug(", pc=0x%08" PRIx32" ", (uint32_t)cpu->pc);
 		if (intevt == 0)
-			debug("vaddr=0x%08"PRIx32" ", vaddr);
+			debug("vaddr=0x%08" PRIx32" ", vaddr);
 
 		debug(" ]\n");
 	}
@@ -658,7 +658,7 @@ void sh_exception(struct cpu *cpu, int expevt, int intevt, uint32_t vaddr)
 		 *  where execution resumes when the interrupt service routine
 		 *  returns.
 		 */
-		cpu->is_halted = 0;
+		cpu->is_halted = false;
 		cpu->pc += sizeof(uint16_t);
 	}
 
@@ -706,6 +706,7 @@ void sh_exception(struct cpu *cpu, int expevt, int intevt, uint32_t vaddr)
 	case EXPEVT_TLB_MISS_LD:
 	case EXPEVT_TLB_MISS_ST:
 		cpu->pc = vbr + 0x400;
+		// fall through
 	case EXPEVT_TLB_PROT_LD:
 	case EXPEVT_TLB_PROT_ST:
 	case EXPEVT_TLB_MOD:
@@ -732,7 +733,7 @@ void sh_exception(struct cpu *cpu, int expevt, int intevt, uint32_t vaddr)
 		 *  these are not very common.
 		 */
 #if 1
-		printf("\nRESERVED SuperH instruction at spc=%08"PRIx32"\n",
+		printf("\nRESERVED SuperH instruction at spc=%08" PRIx32"\n",
 		    cpu->cd.sh.spc);
 		exit(1);
 #else
@@ -773,7 +774,6 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 	char *symbol;
 	uint64_t offset, addr;
 	uint16_t iword;
-	int hi4, lo4, lo8, r8, r4;
 
 	if (running)
 		dumpaddr = cpu->pc;
@@ -786,16 +786,19 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 	if (cpu->machine->ncpus > 1 && running)
 		debug("cpu%i: ", cpu->cpu_id);
 
-	debug("%08"PRIx32, (uint32_t) dumpaddr);
+	debug("%08" PRIx32, (uint32_t) dumpaddr);
 
 	if (cpu->byte_order == EMUL_BIG_ENDIAN)
 		iword = (instr[0] << 8) + instr[1];
 	else
 		iword = (instr[1] << 8) + instr[0];
 
-	debug(":  %04x %s\t", iword, cpu->delay_slot? "(d)" : "");
-	hi4 = iword >> 12; lo4 = iword & 15; lo8 = iword & 255;
-	r8 = (iword >> 8) & 15; r4 = (iword >> 4) & 15;
+	debug(":  %04x %s", iword, cpu->delay_slot? "(d)" : "   ");
+
+	cpu_print_pc_indicator_in_disassembly(cpu, running, dumpaddr);
+
+	const int hi4 = iword >> 12, lo4 = iword & 15, lo8 = iword & 255;
+	int r8 = (iword >> 8) & 15, r4 = (iword >> 4) & 15;
 
 
 	/*
@@ -816,8 +819,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			else if (lo4 == 0x6)
 				debug("mov.l\tr%i,@(r0,r%i)", r4, r8);
 			if (running) {
-				uint32_t addr = cpu->cd.sh.r[0] +
-				    cpu->cd.sh.r[r8];
+				addr = cpu->cd.sh.r[0] + cpu->cd.sh.r[r8];
 				debug("\t; r0+r%i = ", r8);
 				symbol = get_symbol_name(
 				    &cpu->machine->symbol_context,
@@ -825,7 +827,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 				if (symbol != NULL)
 					debug("<%s>", symbol);
 				else
-					debug("0x%08"PRIx32, addr);
+					debug("0x%08" PRIx32, addr);
 			}
 			debug("\n");
 		} else if (lo4 == 0x7)
@@ -846,8 +848,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			else if (lo4 == 0xe)
 				debug("mov.l\t@(r0,r%i),r%i", r4, r8);
 			if (running) {
-				uint32_t addr = cpu->cd.sh.r[0] +
-				    cpu->cd.sh.r[r4];
+				addr = cpu->cd.sh.r[0] + cpu->cd.sh.r[r4];
 				debug("\t; r0+r%i = ", r4);
 				symbol = get_symbol_name(
 				    &cpu->machine->symbol_context,
@@ -855,7 +856,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 				if (symbol != NULL)
 					debug("<%s>", symbol);
 				else
-					debug("0x%08"PRIx32, addr);
+					debug("0x%08" PRIx32, addr);
 			}
 			debug("\n");
 		} else if (lo8 == 0x12)
@@ -918,50 +919,54 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 	case 0x1:
 		debug("mov.l\tr%i,@(%i,r%i)", r4, lo4 * 4, r8);
 		if (running) {
-			uint32_t addr = cpu->cd.sh.r[r8] + lo4 * 4;
+			addr = cpu->cd.sh.r[r8] + lo4 * 4;
 			debug("\t; r%i+%i = ", r8, lo4 * 4);
 			symbol = get_symbol_name(&cpu->machine->symbol_context,
 			    addr, &offset);
 			if (symbol != NULL)
 				debug("<%s>", symbol);
 			else
-				debug("0x%08"PRIx32, addr);
+				debug("0x%08" PRIx32, addr);
 		}
 		debug("\n");
 		break;
 	case 0x2:
 		if (lo4 == 0x0)
-			debug("mov.b\tr%i,@r%i\n", r4, r8);
+			debug("mov.b\tr%i,@r%i", r4, r8);
 		else if (lo4 == 0x1)
-			debug("mov.w\tr%i,@r%i\n", r4, r8);
+			debug("mov.w\tr%i,@r%i", r4, r8);
 		else if (lo4 == 0x2)
-			debug("mov.l\tr%i,@r%i\n", r4, r8);
+			debug("mov.l\tr%i,@r%i", r4, r8);
 		else if (lo4 == 0x4)
-			debug("mov.b\tr%i,@-r%i\n", r4, r8);
+			debug("mov.b\tr%i,@-r%i", r4, r8);
 		else if (lo4 == 0x5)
-			debug("mov.w\tr%i,@-r%i\n", r4, r8);
+			debug("mov.w\tr%i,@-r%i", r4, r8);
 		else if (lo4 == 0x6)
-			debug("mov.l\tr%i,@-r%i\n", r4, r8);
+			debug("mov.l\tr%i,@-r%i", r4, r8);
 		else if (lo4 == 0x7)
-			debug("div0s\tr%i,r%i\n", r4, r8);
+			debug("div0s\tr%i,r%i", r4, r8);
 		else if (lo4 == 0x8)
-			debug("tst\tr%i,r%i\n", r4, r8);
+			debug("tst\tr%i,r%i", r4, r8);
 		else if (lo4 == 0x9)
-			debug("and\tr%i,r%i\n", r4, r8);
+			debug("and\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xa)
-			debug("xor\tr%i,r%i\n", r4, r8);
+			debug("xor\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xb)
-			debug("or\tr%i,r%i\n", r4, r8);
+			debug("or\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xc)
-			debug("cmp/str\tr%i,r%i\n", r4, r8);
+			debug("cmp/str\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xd)
-			debug("xtrct\tr%i,r%i\n", r4, r8);
+			debug("xtrct\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xe)
-			debug("mulu.w\tr%i,r%i\n", r4, r8);
+			debug("mulu.w\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xf)
-			debug("muls.w\tr%i,r%i\n", r4, r8);
+			debug("muls.w\tr%i,r%i", r4, r8);
 		else
-			debug("UNIMPLEMENTED hi4=0x%x, lo8=0x%02x\n", hi4, lo8);
+			debug("UNIMPLEMENTED hi4=0x%x, lo8=0x%02x", hi4, lo8);
+		if (running && lo4 <= 6) {
+			debug("\t; r%i = 0x%08" PRIx32, r8, cpu->cd.sh.r[r8]);
+		}
+		debug("\n");
 		break;
 	case 0x3:
 		if (lo4 == 0x0)
@@ -1122,46 +1127,54 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 	case 0x5:
 		debug("mov.l\t@(%i,r%i),r%i", lo4 * 4, r4, r8);
 		if (running) {
-			debug("\t; r%i+%i = 0x%08"PRIx32, r4, lo4 * 4,
-			    cpu->cd.sh.r[r4] + lo4 * 4);
+			addr = cpu->cd.sh.r[r4] + lo4 * 4;
+			symbol = get_symbol_name(&cpu->machine->symbol_context, addr, &offset);
+			if (symbol != NULL)
+				debug("\t; r%i+%i <%s>", r4, lo4 * 4, symbol);
+			else
+				debug("\t; r%i+%i = 0x%08" PRIx32, r4, lo4 * 4, (int)addr);
 		}
 		debug("\n");
 		break;
 	case 0x6:
 		if (lo4 == 0x0)
-			debug("mov.b\t@r%i,r%i\n", r4, r8);
+			debug("mov.b\t@r%i,r%i", r4, r8);
 		else if (lo4 == 0x1)
-			debug("mov.w\t@r%i,r%i\n", r4, r8);
+			debug("mov.w\t@r%i,r%i", r4, r8);
 		else if (lo4 == 0x2)
-			debug("mov.l\t@r%i,r%i\n", r4, r8);
+			debug("mov.l\t@r%i,r%i", r4, r8);
 		else if (lo4 == 0x3)
-			debug("mov\tr%i,r%i\n", r4, r8);
+			debug("mov\tr%i,r%i", r4, r8);
 		else if (lo4 == 0x4)
-			debug("mov.b\t@r%i+,r%i\n", r4, r8);
+			debug("mov.b\t@r%i+,r%i", r4, r8);
 		else if (lo4 == 0x5)
-			debug("mov.w\t@r%i+,r%i\n", r4, r8);
+			debug("mov.w\t@r%i+,r%i", r4, r8);
 		else if (lo4 == 0x6)
-			debug("mov.l\t@r%i+,r%i\n", r4, r8);
+			debug("mov.l\t@r%i+,r%i", r4, r8);
 		else if (lo4 == 0x7)
-			debug("not\tr%i,r%i\n", r4, r8);
+			debug("not\tr%i,r%i", r4, r8);
 		else if (lo4 == 0x8)
-			debug("swap.b\tr%i,r%i\n", r4, r8);
+			debug("swap.b\tr%i,r%i", r4, r8);
 		else if (lo4 == 0x9)
-			debug("swap.w\tr%i,r%i\n", r4, r8);
+			debug("swap.w\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xa)
-			debug("negc\tr%i,r%i\n", r4, r8);
+			debug("negc\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xb)
-			debug("neg\tr%i,r%i\n", r4, r8);
+			debug("neg\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xc)
-			debug("extu.b\tr%i,r%i\n", r4, r8);
+			debug("extu.b\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xd)
-			debug("extu.w\tr%i,r%i\n", r4, r8);
+			debug("extu.w\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xe)
-			debug("exts.b\tr%i,r%i\n", r4, r8);
+			debug("exts.b\tr%i,r%i", r4, r8);
 		else if (lo4 == 0xf)
-			debug("exts.w\tr%i,r%i\n", r4, r8);
+			debug("exts.w\tr%i,r%i", r4, r8);
 		else
-			debug("UNIMPLEMENTED hi4=0x%x, lo8=0x%02x\n", hi4, lo8);
+			debug("UNIMPLEMENTED hi4=0x%x, lo8=0x%02x", hi4, lo8);
+		if (running && lo4 < 8 && (lo4 & 3) < 3) {
+			debug("\t; r%i = 0x%08" PRIx32, r4, cpu->cd.sh.r[r4]);
+		}
+		debug("\n");
 		break;
 	case 0x7:
 		debug("add\t#%i,r%i\n", (int8_t)lo8, r8);
@@ -1173,7 +1186,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			else if (r8 == 0x4)
 				debug("mov.b\t@(%i,r%i),r0", lo4, r4);
 			if (running) {
-				debug("\t; r%i+%i = 0x%08"PRIx32, r4, lo4,
+				debug("\t; r%i+%i = 0x%08" PRIx32, r4, lo4,
 				    cpu->cd.sh.r[r4] + lo4);
 			}
 			debug("\n");
@@ -1183,7 +1196,7 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			else if (r8 == 0x5)
 				debug("mov.w\t@(%i,r%i),r0", lo4 * 2, r4);
 			if (running) {
-				debug("\t; r%i+%i = 0x%08"PRIx32, r4, lo4 * 2,
+				debug("\t; r%i+%i = 0x%08" PRIx32, r4, lo4 * 2,
 				    cpu->cd.sh.r[r4] + lo4 * 2);
 			}
 			debug("\n");
@@ -1293,49 +1306,49 @@ int sh_cpu_disassemble_instr(struct cpu *cpu, unsigned char *instr,
 			    cpu->cd.sh.fpscr & SH_FPSCR_PR? "d" : "f", r4,
 			    cpu->cd.sh.fpscr & SH_FPSCR_PR? "d" : "f", r8);
 		else if (lo4 == 0x6) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r8 & 1)? "xd" : "dr";
 				r8 &= ~1;
 			}
 			debug("fmov\t@(r0,r%i),%s%i\n", r4, n, r8);
 		} else if (lo4 == 0x7) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r4 & 1)? "xd" : "dr";
 				r4 &= ~1;
 			}
 			debug("fmov\t%s%i,@(r0,r%i)\n", n, r4, r8);
 		} else if (lo4 == 0x8) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r8 & 1)? "xd" : "dr";
 				r8 &= ~1;
 			}
 			debug("fmov\t@r%i,%s%i\n", r4, n, r8);
 		} else if (lo4 == 0x9) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r8 & 1)? "xd" : "dr";
 				r8 &= ~1;
 			}
 			debug("fmov\t@r%i+,%s%i\n", r4, n, r8);
 		} else if (lo4 == 0xa) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r4 & 1)? "xd" : "dr";
 				r4 &= ~1;
 			}
 			debug("fmov\t%s%i,@r%i\n", n, r4, r8);
 		} else if (lo4 == 0xb) {
-			char *n = "fr";
+			const char *n = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n = (r4 & 1)? "xd" : "dr";
 				r4 &= ~1;
 			}
 			debug("fmov\t%s%i,@-r%i\n", n, r4, r8);
 		} else if (lo4 == 0xc) {
-			char *n1 = "fr", *n2 = "fr";
+			const char *n1 = "fr", *n2 = "fr";
 			if (cpu->cd.sh.fpscr & SH_FPSCR_SZ) {
 				n1 = (r4 & 1)? "xd" : "dr";
 				n2 = (r8 & 1)? "xd" : "dr";
