@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -39,7 +39,8 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "armreg.h"
+#include "thirdparty/armreg.h"
+
 
 extern int quiet_mode;
 
@@ -102,8 +103,6 @@ static int arm_check_access(struct cpu *cpu, int ap, int dav, int user)
 /*
  *  arm_translate_v2p_mmu():
  *
- *  Don't call this function if userland_emul is non-NULL, or cpu is NULL.
- *
  *  Return values:
  *	0  Failure
  *	1  Success, the page is readable only
@@ -206,6 +205,7 @@ int arm_translate_v2p_mmu(struct cpu *cpu, uint64_t vaddr64,
 				break;
 			}
 			/*  NOTE: Fall-through for XScale!  */
+			// fall through
 		case 2:	/*  4KB page:  */
 			ap3 = (d2 >> 10) & 3;
 			ap2 = (d2 >>  8) & 3;
