@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -71,9 +71,7 @@ int MEMORY_RW(struct cpu *cpu, struct memory *mem, uint64_t vaddr,
 	const int offset_mask = 0xfff;
 #endif
 
-#ifndef MEM_USERLAND
 	int ok = 2;
-#endif
 	uint64_t paddr;
 	int cache, no_exceptions, offset;
 	unsigned char *memblock;
@@ -83,13 +81,6 @@ int MEMORY_RW(struct cpu *cpu, struct memory *mem, uint64_t vaddr,
 	cache = misc_flags & CACHE_FLAGS_MASK;
 
 
-#ifdef MEM_USERLAND
-#ifdef MEM_ALPHA
-	paddr = vaddr;
-#else
-	paddr = vaddr & 0x7fffffff;
-#endif
-#else	/*  !MEM_USERLAND  */
 	if (misc_flags & PHYSICAL || cpu->translate_v2p == NULL) {
 		paddr = vaddr;
 	} else {
@@ -108,10 +99,136 @@ int MEMORY_RW(struct cpu *cpu, struct memory *mem, uint64_t vaddr,
 			return MEMORY_ACCESS_FAILED;
 	}
 
-#endif	/*  !MEM_USERLAND  */
+#if 0
+	/*
+	 *  For quick-and-dirty debugging of 32-bit code, typically
+	 *  unknown ROM code:  Write out a summary of accessed memory ranges.
+	 */
+	if (cpu->running && paddr < 0x00300000)
+	{
+		static int swriteflag = -1, lastswriteflag;
+		static int32_t start = -1, stop = -1;
+		static int32_t laststart = -1, laststop = -1;
 
+		if (start == -1)
+		{
+			start = stop = paddr;
+			swriteflag = writeflag;
+		}
+		else
+		{
+			if (paddr == stop + len && writeflag == swriteflag)
+			{
+				stop = paddr;
+			}
+			else
+			{
+				if (start != laststart || stop != laststop || swriteflag != lastswriteflag)
+				{
+					printf("%s  %08x -- %08x  %i-bit\n",
+						swriteflag ? "WRITE" : "READ ",
+						(int)start, (int)(stop + len - 1), (int)len*8);
+				}
+
+				laststart = start; laststop = stop; lastswriteflag = swriteflag;
+
+				start = stop = paddr;
+				swriteflag = writeflag;
+			}
+		}
+
+		// Make sure the access is not put into quick lookup tables:
+		ok |= MEMORY_NOT_FULL_PAGE;
+	}
+#endif
+
+#if 0
+	/*
+	 *  For quick-and-dirty test to see which memory addresses are in use,
+	 *  making a "waterfall spectrum"...
+	 */
+	if (cpu->running && (paddr < 0x200000 ||
+		(paddr >= 0x0c000000 && paddr < 0x0d000000)))
+	{
+		int64_t xsize = 1920, ysize = 1080;
+		static int y = 0;
+		static int count = 0;
+		static uint8_t* buf = NULL;
+
+		if (buf == NULL)
+		{
+			buf = (uint8_t*)malloc(xsize * ysize*3);
+			memset(buf, 0xff, xsize * ysize*3);
+		}
+
+		uint64_t paddr2 = paddr - 0x0c000000;
+		
+		if (paddr < 0x200000)
+			paddr2 = paddr + 16*1048576;
+
+		int64_t max = 16*1048576 + (2*1048576);
+
+		int64_t x = xsize * paddr2 / max;
+
+		if (x >= 0 && x < xsize)
+		{
+			if (writeflag)
+			{
+				int c = buf[(x+y*xsize)*3+0];
+				if (c == 255)
+					c = 128;
+				else if (c > 0)
+				{
+					c = c - 8;
+				}
+
+				buf[(x+y*xsize)*3+1] = c;
+				buf[(x+y*xsize)*3+2] = c;
+			}
+			else
+			{
+				int c = buf[(x+y*xsize)*3+1];
+				if (c == 255)
+					c = 128;
+				else if (c > 0)
+				{
+					c = c - 8;
+				}
+
+				buf[(x+y*xsize)*3+0] = c;
+				buf[(x+y*xsize)*3+2] = c;
+			}
+		}
+		
+		for (int meg = 0; meg < 18; ++meg)
+			buf[((xsize * (meg * 0x100000LL) / max) +y*xsize)*3+1] = 64;
+
+		count ++;
+		if (count >= 8192)
+		{
+			count = 0;
+			y ++;
+			
+			if (y >= ysize)
+			{
+				static int n = 0;
+				char name[40];
+				snprintf(name, sizeof(name), "memory_%05i.ppm", n++);
+				FILE* f = fopen(name, "w");
+				printf("writing out %s\n", name);
+				fprintf(f, "P6\n%i %i\n255\n", (int)xsize, (int)ysize);
+				fwrite((char*)buf, 1, (int)(xsize*ysize*3), f);
+				fclose(f);
+				memset(buf, 0xff, xsize*ysize);
+				y = 0;
+			}
+		}
+		
+		// Make sure the access is not put into quick lookup tables:
+		ok |= MEMORY_NOT_FULL_PAGE;
+	}
+#endif
 
-#ifndef MEM_USERLAND
 	/*
 	 *  Memory mapped device?
 	 *
@@ -219,11 +336,16 @@ not just the device in question.
 
 				res = 0;
 				if (!no_exceptions || (mem->devices[i].flags &
-				    DM_READS_HAVE_NO_SIDE_EFFECTS))
+				    DM_READS_HAVE_NO_SIDE_EFFECTS)) {
+					bool running_before_device_access = cpu->running;
 					res = mem->devices[i].f(cpu, mem, paddr,
 					    data, len, writeflag,
 					    mem->devices[i].extra);
 
+					if (running_before_device_access && !cpu->running)
+						return MEMORY_ACCESS_FAILED;
+				}
+
 				if (res == 0)
 					res = -1;
 
@@ -306,8 +428,9 @@ not just the device in question.
 #endif /* MIPS */
 		{
 			if (paddr >= mem->physical_max && !no_exceptions)
-				memory_warn_about_unimplemented_addr
-				    (cpu, mem, writeflag, paddr, data, len);
+				if (memory_warn_about_unimplemented_addr
+				    (cpu, mem, writeflag, paddr, len))
+					return MEMORY_ACCESS_FAILED;
 
 			if (writeflag == MEM_READ) {
 				/*  Return all zeroes? (Or 0xff? TODO)  */
@@ -344,8 +467,6 @@ not just the device in question.
 		}
 	}
 
-#endif	/*  ifndef MEM_USERLAND  */
-
 
 	/*
 	 *  Uncached access:
@@ -374,18 +495,12 @@ not just the device in question.
 	    && (cpu->cd.mips.cpu_type.mmu_model != MMU3K ||
             !(cpu->cd.mips.coproc[0]->reg[COP0_STATUS] & MIPS1_ISOL_CACHES))
 #endif
-#ifndef MEM_USERLAND
 	    && !(ok & MEMORY_NOT_FULL_PAGE)
-#endif
 	    && !no_exceptions)
 		cpu->update_translation_table(cpu, vaddr & ~offset_mask,
 		    memblock, (misc_flags & MEMORY_USER_ACCESS) |
-#if !defined(MEM_USERLAND)
 		    (cache == CACHE_INSTRUCTION?
 			(writeflag == MEM_WRITE? 1 : 0) : ok - 1),
-#else
-		    (writeflag == MEM_WRITE? 1 : 0),
-#endif
 		    paddr & ~offset_mask);
 
 	/*
@@ -394,9 +509,7 @@ not just the device in question.
 	 */
 
 	if ((writeflag == MEM_WRITE
-#if !defined(MEM_USERLAND)
 	    || (ok == 2 && cache == CACHE_DATA)
-#endif
 	    ) && cpu->invalidate_code_translation != NULL)
 		cpu->invalidate_code_translation(cpu, paddr, INVALIDATE_PADDR);
 
