@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -56,7 +56,9 @@
 extern int extra_argc;
 extern char **extra_argv;
 extern struct settings *global_settings;
+
 extern int quiet_mode;
+extern int verbose;
 
 
 /*
@@ -65,16 +67,18 @@ extern int quiet_mode;
  *  TODO: Some of these should be moved to some other place!
  */
 
-volatile int single_step = NOT_SINGLE_STEPPING;
-volatile int exit_debugger;
-int force_debugger_at_exit = 0;
-
+bool single_step = false;
+bool about_to_enter_single_step = false;
+bool debugger_enter_at_end_of_run = false;
 volatile int single_step_breakpoint = 0;
-int debugger_n_steps_left_before_interaction = 0;
 
-int old_instruction_trace = 0;
+extern bool emul_shutdown;
+
 int old_quiet_mode = 0;
-int old_show_trace_tree = 0;
+
+static int debugger_n_steps_left_before_interaction = 0;
+static volatile bool exit_debugger;
+static volatile bool exit_debugger_to_continue_single_stepping;
 
 
 /*
@@ -83,8 +87,6 @@ int old_show_trace_tree = 0;
 
 static volatile int ctrl_c;
 
-static struct emul *debugger_emul;
-
 /*  Currently focused CPU, machine, and emulation:  */
 int debugger_cur_cpu;
 int debugger_cur_machine;
@@ -133,18 +135,19 @@ void debugger_activate(int x)
 {
 	ctrl_c = 1;
 
-	if (single_step != NOT_SINGLE_STEPPING) {
-		/*  Already in the debugger. Do nothing.  */
-		int i;
-		for (i=0; i<MAX_CMD_BUFLEN; i++)
-			console_makeavail(MAIN_CONSOLE, '\b');
-		console_makeavail(MAIN_CONSOLE, ' ');
-		console_makeavail(MAIN_CONSOLE, '\n');
+	if (single_step) {
+		/*  Already in the debugger. Just abort the current line.  */
+		console_makeavail(MAIN_CONSOLE, 3);
 		printf("^C");
 		fflush(stdout);
 	} else {
-		/*  Enter the single step debugger.  */
-		single_step = ENTER_SINGLE_STEPPING;
+		/*
+		 * Enter the single step debugger.
+		 *
+		 * NOTE: Setting single_step = true directly is not wise,
+		 * since it may be in the middle of executing.
+		 */
+		about_to_enter_single_step = true;
 
 		/*  Discard any chars in the input queue:  */
 		while (console_charavail(MAIN_CONSOLE))
@@ -166,9 +169,9 @@ static void show_breakpoint(struct machine *m, int i)
 {
 	printf("%3i: 0x", i);
 	if (m->cpus[0]->is_32bit)
-		printf("%08"PRIx32, (uint32_t) m->breakpoints.addr[i]);
+		printf("%08" PRIx32, (uint32_t) m->breakpoints.addr[i]);
 	else
-		printf("%016"PRIx64, (uint64_t) m->breakpoints.addr[i]);
+		printf("%016" PRIx64, (uint64_t) m->breakpoints.addr[i]);
 	if (m->breakpoints.string[i] != NULL)
 		printf(" (%s)", m->breakpoints.string[i]);
 	printf("\n");
@@ -196,7 +199,7 @@ void debugger_assignment(struct machine *m, char *cmd)
 	uint64_t tmp;
 	uint64_t old_pc = m->cpus[0]->pc;	/*  TODO: multiple cpus?  */
 
-	CHECK_ALLOCATION(left = malloc(MAX_CMD_BUFLEN));
+	CHECK_ALLOCATION(left = (char *) malloc(MAX_CMD_BUFLEN));
 	strlcpy(left, cmd, MAX_CMD_BUFLEN);
 	right = strchr(left, '=');
 	if (right == NULL) {
@@ -356,7 +359,11 @@ static char *debugger_readline(void)
 	char *cmd = last_cmd[last_cmd_index];
 
 	cmd_len = 0; cmd[0] = '\0';
+
+	color_prompt();
 	printf("GXemul> ");
+	color_normal();
+
 	fflush(stdout);
 
 	ch = '\0';
@@ -402,6 +409,15 @@ static char *debugger_readline(void)
 				printf("\b");
 				cursor_pos --;
 			}
+		} else if (ch == 3) {
+			/*  CTRL-C: Abort the current line.  */
+			cmd_len = 0; cmd[0] = '\0'; cursor_pos = 0;
+			color_normal();
+			printf("\n");
+			color_prompt();
+			printf("GXemul> ");
+			color_normal();
+			fflush(stdout);
 		} else if (ch == 5) {
 			/*  CTRL-E: End of line.  */
 			while (cursor_pos < cmd_len) {
@@ -463,7 +479,80 @@ static char *debugger_readline(void)
 					cursor_pos = cmd_len;
 				}
 			} while (0);
-		} else if (ch >= ' ' && cmd_len < MAX_CMD_BUFLEN-1) {
+		} else if (ch == 20) {
+			/*  CTRL-T: Print SIGINFO-like info.  */
+			color_normal();
+			printf("^T\n");
+
+			struct cpu* cpu = debugger_machine->cpus[debugger_cur_cpu];
+
+			char info[1000];
+			snprintf(info, sizeof(info), "ninstrs=%lli, pc=", (long long) cpu->ninstrs);
+
+			if (cpu->is_32bit)
+				snprintf(info + strlen(info), sizeof(info) - strlen(info),
+				    "0x%08" PRIx32, (uint32_t) cpu->pc);
+			else
+				snprintf(info + strlen(info), sizeof(info) - strlen(info),
+				    "0x%016" PRIx64, (uint64_t) cpu->pc);
+
+			uint64_t offset;
+			int n_args;
+			char* pc_symbol = get_symbol_name_and_n_args(&debugger_machine->symbol_context, cpu->pc, &offset, &n_args);
+			if (pc_symbol != NULL)
+				snprintf(info + strlen(info), sizeof(info) - strlen(info),
+				    " <%s>", pc_symbol);
+
+			if (!cpu->running)
+				snprintf(info + strlen(info), sizeof(info) - strlen(info),
+				    ", stopped");
+
+			debugmsg_cpu(cpu, SUBSYS_CPU, "", VERBOSITY_INFO, "%s", info);
+
+			color_prompt();
+			printf("GXemul> ");
+			color_normal();
+
+			// Reprint all chars, then backspace if cursor_pos is less than cmd_len.
+			for (i = 0; i < cmd_len; ++i)
+				printf("%c", cmd[i]);
+			for (i = 0; i < cmd_len - cursor_pos; ++i)
+				printf("\b");
+
+			fflush(stdout);
+		} else if (ch == 23) {
+			/*  CTRL-W: Backspace a full word.  */
+
+			while (cursor_pos >= 1 && cmd[cursor_pos-1] == ' ') {
+				/*  Backspace.  */
+				cursor_pos --;
+				cmd_len --;
+				memmove(cmd + cursor_pos, cmd + cursor_pos + 1,
+				    cmd_len);
+				cmd[cmd_len] = '\0';
+				printf("\b");
+				for (i=cursor_pos; i<cmd_len; i++)
+					printf("%c", cmd[i]);
+				printf(" \b");
+				for (i=cursor_pos; i<cmd_len; i++)
+					printf("\b");
+			}
+
+			while (cursor_pos >= 1 && cmd[cursor_pos-1] != ' ') {
+				/*  Backspace.  */
+				cursor_pos --;
+				cmd_len --;
+				memmove(cmd + cursor_pos, cmd + cursor_pos + 1,
+				    cmd_len);
+				cmd[cmd_len] = '\0';
+				printf("\b");
+				for (i=cursor_pos; i<cmd_len; i++)
+					printf("%c", cmd[i]);
+				printf(" \b");
+				for (i=cursor_pos; i<cmd_len; i++)
+					printf("\b");
+			}
+		} else if (ch >= ' ' && ch != 127 && cmd_len < MAX_CMD_BUFLEN-1) {
 			/*  Visible character:  */
 			memmove(cmd + cursor_pos + 1, cmd + cursor_pos,
 			    cmd_len - cursor_pos);
@@ -538,7 +627,9 @@ static char *debugger_readline(void)
 				}
 				if (j != 0)
 					printf("\n");
+				color_prompt();
 				printf("GXemul> ");
+				color_normal();
 				for (i=0; i<cmd_len; i++)
 					printf("%c", cmd[i]);
 			}
@@ -636,7 +727,8 @@ void debugger(void)
 	/*  Stop timers while interacting with the user:  */
 	timer_stop();
 
-	exit_debugger = 0;
+	exit_debugger = false;
+	exit_debugger_to_continue_single_stepping = false;
 
 	while (!exit_debugger) {
 		/*  Read a line from the terminal:  */
@@ -668,7 +760,7 @@ void debugger(void)
 		debugger_execute_cmd(cmd, cmd_len);
 
 		/*  Special hack for the "step" command:  */
-		if (exit_debugger == -1)
+		if (exit_debugger_to_continue_single_stepping)
 			return;
 	}
 
@@ -682,9 +774,7 @@ void debugger(void)
 		debugger_machine->cpus[i]->ninstrs_since_gettimeofday = 0;
 	}
 
-	single_step = NOT_SINGLE_STEPPING;
-	debugger_machine->instruction_trace = old_instruction_trace;
-	debugger_machine->show_trace_tree = old_show_trace_tree;
+	single_step = false;
 	quiet_mode = old_quiet_mode;
 }
 
@@ -726,7 +816,7 @@ void debugger_init(struct emul *emul)
 	debugger_cur_machine = 0;
 
 	for (i=0; i<N_PREVIOUS_CMDS; i++) {
-		CHECK_ALLOCATION(last_cmd[i] = malloc(MAX_CMD_BUFLEN));
+		CHECK_ALLOCATION(last_cmd[i] = (char *) malloc(MAX_CMD_BUFLEN));
 		last_cmd[i][0] = '\0';
 	}
 
