@@ -97,7 +97,7 @@ void isa_interrupt_common(struct bus_isa_data *d, int old_isa_assert)
  */
 void isa_interrupt_assert(struct interrupt *interrupt)
 {
-	struct bus_isa_data *d = interrupt->extra;
+	struct bus_isa_data *d = (struct bus_isa_data *) interrupt->extra;
 	int old_isa_assert, line = interrupt->line;
 	int mask = 1 << (line & 7);
 
@@ -119,7 +119,7 @@ void isa_interrupt_assert(struct interrupt *interrupt)
  */
 void isa_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct bus_isa_data *d = interrupt->extra;
+	struct bus_isa_data *d = (struct bus_isa_data *) interrupt->extra;
 	int line = interrupt->line, mask = 1 << (line & 7);
 	int old_irr1 = d->pic1->irr, old_isa_assert;
 
@@ -181,7 +181,7 @@ struct bus_isa_data *bus_isa_init(struct machine *machine,
 	int i, tmp_handle, kbd_in_use;
 	int lptbase = 0x378;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct bus_isa_data)));
+	CHECK_ALLOCATION(d = (struct bus_isa_data *) malloc(sizeof(struct bus_isa_data)));
 	memset(d, 0, sizeof(struct bus_isa_data));
 
 	d->isa_portbase = isa_portbase;
@@ -194,17 +194,17 @@ struct bus_isa_data *bus_isa_init(struct machine *machine,
 
 		/*  Register the 16 possible ISA interrupts:  */
 		for (i=0; i<16; i++) {
-			struct interrupt template;
+			struct interrupt templ;
 			char name[300];
 			snprintf(name, sizeof(name),
 			    "%s.isa.%i", interrupt_base_path, i);
-			memset(&template, 0, sizeof(template));
-			template.line = i;
-			template.name = name;
-			template.extra = d;
-			template.interrupt_assert = isa_interrupt_assert;
-			template.interrupt_deassert = isa_interrupt_deassert;
-			interrupt_handler_register(&template);
+			memset(&templ, 0, sizeof(templ));
+			templ.line = i;
+			templ.name = name;
+			templ.extra = d;
+			templ.interrupt_assert = isa_interrupt_assert;
+			templ.interrupt_deassert = isa_interrupt_deassert;
+			interrupt_handler_register(&templ);
 		}
 	}
 
@@ -219,7 +219,7 @@ struct bus_isa_data *bus_isa_init(struct machine *machine,
 	if (!(bus_isa_flags & BUS_ISA_EXTERNAL_PIC)) {
 		snprintf(tmpstr, sizeof(tmpstr), "8259 irq=%s addr=0x%llx",
 		    interrupt_base_path, (long long)(isa_portbase + 0x20));
-		d->pic1 = machine->isa_pic_data.pic1 =
+		d->pic1 = machine->isa_pic_data.pic1 = (struct pic8259_data *)
 		    device_add(machine, tmpstr);
 		d->ptr_to_pending_timer_interrupts =
 		    machine->isa_pic_data.pending_timer_interrupts;
@@ -231,7 +231,7 @@ struct bus_isa_data *bus_isa_init(struct machine *machine,
 			snprintf(tmpstr, sizeof(tmpstr),
 			    "8259 irq=%s.isa.2 addr=0x%llx",
 			    interrupt_base_path,(long long)(isa_portbase+0xa0));
-			d->pic2 = machine->isa_pic_data.pic2 =
+			d->pic2 = machine->isa_pic_data.pic2 = (struct pic8259_data *)
 			    device_add(machine, tmpstr);
 		}
 	} else {
