@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -48,7 +48,7 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "i8253reg.h"
+#include "thirdparty/i8253reg.h"
 
 
 /*  #define debug fatal  */
@@ -76,7 +76,7 @@ struct pit8253_data {
 
 static void timer0_tick(struct timer *t, void *extra)
 {
-	struct pit8253_data *d = extra;
+	struct pit8253_data *d = (struct pit8253_data *) extra;
 	d->pending_interrupts_timer0 ++;
 
 	/*  printf("%i ", d->pending_interrupts_timer0); fflush(stdout);  */
@@ -85,7 +85,7 @@ static void timer0_tick(struct timer *t, void *extra)
 
 DEVICE_TICK(8253)
 {
-	struct pit8253_data *d = extra;
+	struct pit8253_data *d = (struct pit8253_data *) extra;
 
 	if (!d->in_use)
 		return;
@@ -123,9 +123,9 @@ DEVICE_ACCESS(8253)
 				d->counter[relative_addr] &= 0x00ff;
 				d->counter[relative_addr] |= ((idata&0xff)<<8);
 				if (d->counter[relative_addr] != 0)
-					d->hz[relative_addr] =
+					d->hz[relative_addr] = (int) (
 					    I8253_TIMER_FREQ / (float)
-					    d->counter[relative_addr] + 0.5;
+					    d->counter[relative_addr] + 0.5);
 				else
 					d->hz[relative_addr] = 0;
 				debug("[ 8253: counter %i set to %i (%i Hz) "
@@ -194,6 +194,7 @@ DEVICE_ACCESS(8253)
 					break;
 				case I8253_TIMER_16BIT:
 					debug("LSB+");
+					// fall through
 				case I8253_TIMER_MSB:
 					debug("MSB ");
 				}
@@ -231,7 +232,7 @@ DEVINIT(8253)
 {
 	struct pit8253_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pit8253_data)));
+	CHECK_ALLOCATION(d = (struct pit8253_data *) malloc(sizeof(struct pit8253_data)));
 	memset(d, 0, sizeof(struct pit8253_data));
 
 	d->in_use = devinit->in_use;
