@@ -39,7 +39,7 @@
 #include "misc.h"
 #include "x11.h"
 
-#include "bt459.h"
+#include "thirdparty/bt459.h"
 
 
 #ifdef WITH_X11
@@ -109,7 +109,9 @@ struct bt459_data {
 static void bt459_update_X_cursor(struct cpu *cpu, struct bt459_data *d)
 {
 	int i, x,y, xmax=0, ymax=0;
+#ifdef WITH_X11
 	int bw_only = 1;
+#endif
 
 	/*  First, let's calculate the size of the cursor:  */
 	for (y=0; y<64; y++)
@@ -124,8 +126,10 @@ static void bt459_update_X_cursor(struct cpu *cpu, struct bt459_data *d)
 				int color = (data >> (6-2*i)) & 3;
 				if (color != 0)
 					xmax = x + i;
+#ifdef WITH_X11
 				if (color != 0 && color != 3)
 					bw_only = 0;
+#endif
 			}
 		}
 
@@ -244,7 +248,7 @@ static void bt459_update_cursor_position(struct bt459_data *d,
 
 DEVICE_TICK(bt459)
 {
-	struct bt459_data *d = extra;
+	struct bt459_data *d = (struct bt459_data *) extra;
 	int old_cursor_on = d->cursor_on;
 
 	if (d->need_to_update_cursor_shape) {
@@ -282,11 +286,12 @@ DEVICE_TICK(bt459)
 
 DEVICE_ACCESS(bt459_irq)
 {
-	struct bt459_data *d = extra;
-	uint64_t idata = 0, odata = 0;
+	struct bt459_data *d = (struct bt459_data *) extra;
+	// uint64_t idata = 0;
+	uint64_t odata = 0;
 
-	if (writeflag == MEM_WRITE)
-		idata = memory_readmax64(cpu, data, len);
+//	if (writeflag == MEM_WRITE)
+//		idata = memory_readmax64(cpu, data, len);
 
 #ifdef BT459_DEBUG
 	fatal("[ bt459: IRQ ack ]\n");
@@ -305,7 +310,7 @@ DEVICE_ACCESS(bt459_irq)
 
 DEVICE_ACCESS(bt459)
 {
-	struct bt459_data *d = extra;
+	struct bt459_data *d = (struct bt459_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int btaddr, old_cursor_on = d->cursor_on, modified;
 
@@ -521,11 +526,11 @@ DEVICE_ACCESS(bt459)
  */
 void dev_bt459_init(struct machine *machine, struct memory *mem,
 	uint64_t baseaddr, uint64_t baseaddr_irq, struct vfb_data *vfb_data,
-	int planes, char *irq_path, int type)
+	int planes, const char *irq_path, int type)
 {
 	struct bt459_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct bt459_data)));
+	CHECK_ALLOCATION(d = (struct bt459_data *) malloc(sizeof(struct bt459_data)));
 	memset(d, 0, sizeof(struct bt459_data));
 
 	INTERRUPT_CONNECT(irq_path, d->irq);
