@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -57,7 +57,7 @@ struct cons_data {
 
 DEVICE_TICK(cons)
 {
-	struct cons_data *d = extra;
+	struct cons_data *d = (struct cons_data *) extra;
 
 	if (console_charavail(d->console_handle))
 		INTERRUPT_ASSERT(d->irq);
@@ -68,17 +68,16 @@ DEVICE_TICK(cons)
 
 DEVICE_ACCESS(cons)
 {
-	struct cons_data *d = extra;
+	struct cons_data *d = (struct cons_data *) extra;
 	unsigned int i;
 
 	/*  Exit the emulator:  */
 	if (relative_addr == DEV_CONS_HALT) {
 		/*  cpu->running = 0;
-		    cpu->machine->exit_without_entering_debugger = 1;
 		    return 1;  */
 		/*  TODO: this doesn't work yet. for now, let's
 		    simply use exit()  */
-		exit(1);
+		exit(0);
 	}
 
 	if (writeflag == MEM_WRITE) {
@@ -116,13 +115,13 @@ DEVINIT(cons)
 	char *name3;
 	size_t nlen;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct cons_data)));
+	CHECK_ALLOCATION(d = (struct cons_data *) malloc(sizeof(struct cons_data)));
 	memset(d, 0, sizeof(struct cons_data));
 
 	nlen = strlen(devinit->name) + 10;
 	if (devinit->name2 != NULL)
 		nlen += strlen(devinit->name2) + 10;
-	CHECK_ALLOCATION(name3 = malloc(nlen));
+	CHECK_ALLOCATION(name3 = (char *) malloc(nlen));
 	if (devinit->name2 != NULL && devinit->name2[0])
 		snprintf(name3, nlen, "%s [%s]", devinit->name, devinit->name2);
 	else
