@@ -40,7 +40,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "cpc700reg.h"
+#include "thirdparty/cpc700reg.h"
 
 
 struct cpc700_data {
@@ -55,14 +55,14 @@ struct cpc700_data {
 
 void cpc700_interrupt_assert(struct interrupt *interrupt)
 {
-	struct cpc700_data *d = interrupt->extra;
+	struct cpc700_data *d = (struct cpc700_data *) interrupt->extra;
 	d->sr |= interrupt->line;
 	if (d->sr & d->er)
 		INTERRUPT_ASSERT(d->ppc_irq);
 }
 void cpc700_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct cpc700_data *d = interrupt->extra;
+	struct cpc700_data *d = (struct cpc700_data *) interrupt->extra;
 	d->sr &= ~interrupt->line;
 	if (!(d->sr & d->er))
 		INTERRUPT_DEASSERT(d->ppc_irq);
@@ -78,7 +78,7 @@ DEVICE_ACCESS(cpc700_pci)
 {
 	uint64_t idata = 0, odata = 0;
 	int bus, dev, func, reg;
-	struct cpc700_data *d = extra;
+	struct cpc700_data *d = (struct cpc700_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len|MEM_PCI_LITTLE_ENDIAN);
@@ -109,7 +109,7 @@ DEVICE_ACCESS(cpc700_pci)
  */
 DEVICE_ACCESS(cpc700_int)
 {
-	struct cpc700_data *d = extra;
+	struct cpc700_data *d = (struct cpc700_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -187,7 +187,7 @@ DEVINIT(cpc700)
 	char tmp[300];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct cpc700_data)));
+	CHECK_ALLOCATION(d = (struct cpc700_data *) malloc(sizeof(struct cpc700_data)));
 	memset(d, 0, sizeof(struct cpc700_data));
 
 	/*  Connect to the CPU's interrupt pin:  */
@@ -195,17 +195,17 @@ DEVINIT(cpc700)
 
 	/*  Register 32 CPC700 interrupts:  */
 	for (i=0; i<32; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char n[300];
 		snprintf(n, sizeof(n), "%s.cpc700.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = 1 << i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = cpc700_interrupt_assert;
-		template.interrupt_deassert = cpc700_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 1 << i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = cpc700_interrupt_assert;
+		templ.interrupt_deassert = cpc700_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*  Register a PCI bus:  */
