@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -39,21 +39,18 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "dc7085.h"
+#include "thirdparty/dc7085.h"
 
 
 #define	DC_TICK_SHIFT		14
 
-#define	MAX_QUEUE_LEN		32768
+#define	MAX_QUEUE_LEN		1024
 
 struct dc_data {
 	struct dc7085regs	regs;
 
 	int			console_handle;
 
-	/*  For slow_serial_interrupts_hack_for_linux:  */
-	int			just_transmitted_something;
-
 	unsigned char		rx_queue_char[MAX_QUEUE_LEN];
 	char			rx_queue_lineno[MAX_QUEUE_LEN];
 	int			cur_rx_queue_pos_write;
@@ -68,6 +65,25 @@ struct dc_data {
 };
 
 
+/*
+ *  Check if space is available in the receive queue.
+ */
+int space_available_in_queue(void *e, int line_no)
+{
+	struct dc_data *d = (struct dc_data *) e;
+	int entries_in_use = d->cur_rx_queue_pos_write -
+	    d->cur_rx_queue_pos_read;
+
+	while (entries_in_use < 0)
+		entries_in_use += MAX_QUEUE_LEN;
+
+	if (entries_in_use < MAX_QUEUE_LEN - 20)
+		return 1;
+
+	return 0;
+}
+
+
 /*
  *  Add a character to the receive queue.
  */
@@ -101,20 +117,9 @@ DEVICE_TICK(dc7085)
 	 *  If a key is available from the keyboard, add it to the rx queue.
 	 *  If other bits are set, an interrupt might need to be caused.
 	 */
-	struct dc_data *d = extra;
+	struct dc_data *d = (struct dc_data *) extra;
 	int avail;
 
-	if (cpu->machine->slow_serial_interrupts_hack_for_linux) {
-		/*
-		 *  Special hack to prevent Linux from Oopsing. (This makes
-		 *  interrupts not come as fast as possible.)
-		 */
-		if (d->just_transmitted_something) {
-			d->just_transmitted_something --;
-			return;
-		}
-	}
-
 	d->regs.dc_csr &= ~CSR_RDONE;
 
 	if ((d->regs.dc_csr & CSR_MSE) && !(d->regs.dc_csr & CSR_TRDY)) {
@@ -159,7 +164,7 @@ DEVICE_TICK(dc7085)
 
 DEVICE_ACCESS(dc7085)
 {
-	struct dc_data *d = extra;
+	struct dc_data *d = (struct dc_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	size_t i;
 
@@ -222,8 +227,6 @@ DEVICE_ACCESS(dc7085)
 
 			d->regs.dc_csr &= ~CSR_RDONE;
 			INTERRUPT_DEASSERT(d->irq);
-
-			d->just_transmitted_something = 4;
 		}
 		break;
 
@@ -253,8 +256,6 @@ DEVICE_ACCESS(dc7085)
 
 			d->regs.dc_csr &= ~CSR_TRDY;
 			INTERRUPT_DEASSERT(d->irq);
-
-			d->just_transmitted_something = 4;
 		} else {
 			/*  read:  */
 			d->regs.dc_msr_tdr |= MSR_DSR2 | MSR_CD2 |
@@ -300,7 +301,7 @@ int dev_dc7085_init(struct machine *machine, struct memory *mem,
 {
 	struct dc_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct dc_data)));
+	CHECK_ALLOCATION(d = (struct dc_data *) malloc(sizeof(struct dc_data)));
 	memset(d, 0, sizeof(struct dc_data));
 
 	INTERRUPT_CONNECT(irq_path, d->irq);
@@ -311,7 +312,10 @@ int dev_dc7085_init(struct machine *machine, struct memory *mem,
 
 	d->console_handle = console_start_slave(machine, "DC7085", 1);
 
-	lk201_init(&d->lk201, use_fb, add_to_rx_queue, d->console_handle, d);
+	lk201_init(&d->lk201, use_fb,
+	    space_available_in_queue,
+	    add_to_rx_queue,
+	    d->console_handle, d);
 
 	memory_device_register(mem, "dc7085", baseaddr, DEV_DC7085_LENGTH,
 	    dev_dc7085_access, d, DM_DEFAULT, NULL);
