@@ -67,14 +67,14 @@ struct dec5800_data {
 
 void dec5800_interrupt_assert(struct interrupt *interrupt)
 {
-	struct dec5800_data *d = interrupt->extra;
+	struct dec5800_data *d = (struct dec5800_data *) interrupt->extra;
 	d->csr |= (1 << interrupt->line);
 	if (d->csr & 0x10000000)
 		INTERRUPT_ASSERT(d->cpu_irq);
 }
 void dec5800_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct dec5800_data *d = interrupt->extra;
+	struct dec5800_data *d = (struct dec5800_data *) interrupt->extra;
 	d->csr &= ~(1 << interrupt->line);
 	if (!(d->csr & 0x10000000))
 		INTERRUPT_DEASSERT(d->cpu_irq);
@@ -83,7 +83,7 @@ void dec5800_interrupt_deassert(struct interrupt *interrupt)
 
 DEVICE_TICK(dec5800)
 {
-	struct dec5800_data *d = extra;
+	struct dec5800_data *d = (struct dec5800_data *) extra;
 
 	/*  Timer interrupts?  */
 	if (d->csr & 0x8000) {
@@ -100,7 +100,7 @@ DEVICE_TICK(dec5800)
 DEVICE_ACCESS(dec5800_vectors)
 {
 	uint64_t idata = 0, odata = 0;
-	struct dec5800_data *d = extra;
+	struct dec5800_data *d = (struct dec5800_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
@@ -129,7 +129,7 @@ DEVICE_ACCESS(dec5800_vectors)
 DEVICE_ACCESS(dec5800)
 {
 	uint64_t idata = 0, odata = 0;
-	struct dec5800_data *d = extra;
+	struct dec5800_data *d = (struct dec5800_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
@@ -178,7 +178,7 @@ DEVINIT(dec5800)
 	char tmpstr[200];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct dec5800_data)));
+	CHECK_ALLOCATION(d = (struct dec5800_data *) malloc(sizeof(struct dec5800_data)));
 	memset(d, 0, sizeof(struct dec5800_data));
 
 	snprintf(tmpstr, sizeof(tmpstr), "%s.2", devinit->interrupt_path);
@@ -190,16 +190,16 @@ DEVINIT(dec5800)
 	/*  Register 32 CSR interrupts, corresponding to bits in the CSR:  */
 	for (i=0; i<32; i++) {
 		char n[200];
-		struct interrupt template;
+		struct interrupt templ;
 		snprintf(n, sizeof(n), "%s.dec5800.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = dec5800_interrupt_assert;
-		template.interrupt_deassert = dec5800_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = dec5800_interrupt_assert;
+		templ.interrupt_deassert = dec5800_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	memory_device_register(devinit->machine->memory, "dec5800",
@@ -218,7 +218,7 @@ DEVINIT(dec5800)
 /*****************************************************************************/
 
 
-#include "bireg.h"
+#include "thirdparty/bireg.h"
 
 /*  16 slots, 0x2000 bytes each  */
 #define DEV_DECBI_LENGTH		0x20000
@@ -232,7 +232,7 @@ DEVICE_ACCESS(decbi)
 {
 	uint64_t idata = 0, odata = 0;
 	int node_nr;
-	struct decbi_data *d = extra;
+	struct decbi_data *d = (struct decbi_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
@@ -314,7 +314,7 @@ DEVINIT(decbi)
 {
 	struct decbi_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct decbi_data)));
+	CHECK_ALLOCATION(d = (struct decbi_data *) malloc(sizeof(struct decbi_data)));
 	memset(d, 0, sizeof(struct decbi_data));
 
 	memory_device_register(devinit->machine->memory, "decbi",
@@ -390,7 +390,7 @@ void dev_deccca_init(struct memory *mem, uint64_t baseaddr)
 {
 	struct deccca_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct deccca_data)));
+	CHECK_ALLOCATION(d = (struct deccca_data *) malloc(sizeof(struct deccca_data)));
 	memset(d, 0, sizeof(struct deccca_data));
 
 	memory_device_register(mem, "deccca", baseaddr, DEV_DECCCA_LENGTH,
@@ -405,7 +405,7 @@ void dev_deccca_init(struct memory *mem, uint64_t baseaddr)
  *  DEC 5800 XMI (this has to do with SMP...)
  */
 
-#include "xmireg.h"
+#include "thirdparty/xmireg.h"
 
 struct decxmi_data {
 	uint32_t		reg_0xc[NNODEXMI];
@@ -419,7 +419,7 @@ DEVICE_ACCESS(decxmi)
 {
 	uint64_t idata = 0, odata = 0;
 	int node_nr;
-	struct decxmi_data *d = extra;
+	struct decxmi_data *d = (struct decxmi_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
@@ -502,7 +502,7 @@ void dev_decxmi_init(struct memory *mem, uint64_t baseaddr)
 {
 	struct decxmi_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct decxmi_data)));
+	CHECK_ALLOCATION(d = (struct decxmi_data *) malloc(sizeof(struct decxmi_data)));
 	memset(d, 0, sizeof(struct decxmi_data));
 
 	memory_device_register(mem, "decxmi", baseaddr, DEV_DECXMI_LENGTH,
