@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -139,7 +139,7 @@ void dev_fb_resize(struct vfb_data *d, int new_xsize, int new_ysize)
 	new_bytes_per_line = new_xsize * d->bit_depth / 8;
 	size = new_ysize * new_bytes_per_line;
 
-	CHECK_ALLOCATION(new_framebuffer = malloc(size));
+	CHECK_ALLOCATION(new_framebuffer = (unsigned char *) malloc(size));
 
 	/*  Copy the old framebuffer to the new:  */
 	if (d->framebuffer != NULL) {
@@ -180,8 +180,11 @@ void dev_fb_resize(struct vfb_data *d, int new_xsize, int new_ysize)
 
 #ifdef WITH_X11
 	if (d->fb_window != NULL) {
-		x11_fb_resize(d->fb_window, new_xsize, new_ysize);
-		x11_set_standard_properties(d->fb_window, d->title);
+		x11_fb_resize(d->fb_window, d->x11_xsize, d->x11_ysize);
+		if (d->fb_window->name != NULL)
+			free(d->fb_window->name);
+		d->fb_window->name = strdup(d->title);
+		x11_set_standard_properties(d->fb_window);
 	}
 #endif
 }
@@ -236,7 +239,7 @@ void framebuffer_blockcopyfill(struct vfb_data *d, int fillflag, int fill_r,
 	int from_x, int from_y)
 {
 	int x, y;
-	long from_ofs, dest_ofs, linelen;
+	size_t from_ofs, dest_ofs, linelen;
 
 	if (fillflag)
 		debug("framebuffer_blockcopyfill(FILL, %i,%i, %i,%i, "
@@ -262,7 +265,7 @@ void framebuffer_blockcopyfill(struct vfb_data *d, int fillflag, int fill_r,
 				    d->framebuffer + dest_ofs;
 
 				if (d->bit_depth == 24) {
-					for (x=0; x<linelen && x <
+					for (x=0; x<(ssize_t)linelen && x <
 					    (int) sizeof(buf); x += 3) {
 						buf[x] = fill_r;
 						buf[x+1] = fill_g;
@@ -404,7 +407,7 @@ void (*redraw[2 * 4 * 2])(struct vfb_data *, int, int) = {
 
 DEVICE_TICK(fb)
 {
-	struct vfb_data *d = extra;
+	struct vfb_data *d = (struct vfb_data *) extra;
 #ifdef WITH_X11
 	int need_to_flush_x11 = 0;
 	int need_to_redraw_cursor = 0;
@@ -529,8 +532,9 @@ DEVICE_TICK(fb)
 	if (d->update_x2 != -1) {
 #ifdef WITH_X11
 		int y;
+		int addr, addr2;
 #endif
-		int addr, addr2, q = d->vfb_scaledown;
+		int q = d->vfb_scaledown;
 
 		if (d->update_x1 >= d->visible_xsize)
 			d->update_x1 = d->visible_xsize - 1;
@@ -550,12 +554,12 @@ DEVICE_TICK(fb)
 		d->update_y1 = d->update_y1 / q * q;
 		d->update_y2 = d->update_y2 / q * q;
 
+#ifdef WITH_X11
 		addr  = d->update_y1 * d->bytes_per_line +
 		    d->update_x1 * d->bit_depth / 8;
 		addr2 = d->update_y1 * d->bytes_per_line +
 		    d->update_x2 * d->bit_depth / 8;
 
-#ifdef WITH_X11
 		for (y=d->update_y1; y<=d->update_y2; y+=q) {
 			d->redraw_func(d, addr, addr2 - addr);
 			addr  += d->bytes_per_line * q;
@@ -604,7 +608,7 @@ DEVICE_TICK(fb)
 
 DEVICE_ACCESS(fb)
 {
-	struct vfb_data *d = extra;
+	struct vfb_data *d = (struct vfb_data *) extra;
 	size_t i;
 
 #ifdef FB_DEBUG
@@ -729,12 +733,14 @@ DEVICE_ACCESS(fb)
  *  VFB_HPC is like generic, but the color encoding is done as on HPCmips
  *  and Dreamcast.
  *
+ *  VFB_REVERSEBITS is like generic, but with bit order reversed.
+ *
  *  If bit_depth = -15 (note the minus sign), then a special hack is used for
  *  the Playstation Portable's 5-bit R, 5-bit G, 5-bit B.
  */
 struct vfb_data *dev_fb_init(struct machine *machine, struct memory *mem,
 	uint64_t baseaddr, int vfb_type, int visible_xsize, int visible_ysize,
-	int xsize, int ysize, int bit_depth, char *name)
+	int xsize, int ysize, int bit_depth, const char *name)
 {
 	struct vfb_data *d;
 	size_t size, nlen;
@@ -742,7 +748,7 @@ struct vfb_data *dev_fb_init(struct machine *machine, struct memory *mem,
 	int reverse_start = 0;
 	char *name2;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct vfb_data)));
+	CHECK_ALLOCATION(d = (struct vfb_data *) malloc(sizeof(struct vfb_data)));
 	memset(d, 0, sizeof(struct vfb_data));
 
 	if (vfb_type & VFB_REVERSE_START) {
@@ -805,7 +811,7 @@ struct vfb_data *dev_fb_init(struct machine *machine, struct memory *mem,
 	d->bytes_per_line = d->xsize * d->bit_depth / 8;
 	size = d->ysize * d->bytes_per_line;
 
-	CHECK_ALLOCATION(d->framebuffer = malloc(size));
+	CHECK_ALLOCATION(d->framebuffer = (unsigned char *) malloc(size));
 
 	/*  Clear the framebuffer (all black pixels):  */
 	d->framebuffer_size = size;
@@ -848,7 +854,7 @@ struct vfb_data *dev_fb_init(struct machine *machine, struct memory *mem,
 		d->fb_window = NULL;
 
 	nlen = strlen(name) + 10;
-	CHECK_ALLOCATION(name2 = malloc(nlen));
+	CHECK_ALLOCATION(name2 = (char *) malloc(nlen));
 
 	snprintf(name2, nlen, "fb [%s]", name);
 
