@@ -57,7 +57,7 @@ struct fdc_data {
 
 DEVICE_ACCESS(fdc)
 {
-	struct fdc_data *d = extra;
+	struct fdc_data *d = (struct fdc_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	size_t i;
 
@@ -91,7 +91,7 @@ DEVINIT(fdc)
 {
 	struct fdc_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct fdc_data)));
+	CHECK_ALLOCATION(d = (struct fdc_data *) malloc(sizeof(struct fdc_data)));
 	memset(d, 0, sizeof(struct fdc_data));
 
 	INTERRUPT_CONNECT(devinit->interrupt_path, d->irq);
