@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2010  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -45,7 +45,7 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "gtreg.h"
+#include "thirdparty/gtreg.h"
 
 
 #define	TICK_SHIFT		14
@@ -79,14 +79,14 @@ struct gt_data {
  */
 static void timer_tick(struct timer *timer, void *extra)
 {
-	struct gt_data *d = extra;
+	struct gt_data *d = (struct gt_data *) extra;
 	d->pending_timer0_interrupts ++;
 }
 
 
 DEVICE_TICK(gt)
 {
-	struct gt_data *d = extra;
+	struct gt_data *d = (struct gt_data *) extra;
 	if (d->pending_timer0_interrupts > 0)
 		INTERRUPT_ASSERT(d->timer0_irq);
 }
@@ -94,7 +94,7 @@ DEVICE_TICK(gt)
 
 DEVICE_ACCESS(gt)
 {
-	struct gt_data *d = extra;
+	struct gt_data *d = (struct gt_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int bus, dev, func, reg;
 	size_t i;
@@ -173,8 +173,12 @@ DEVICE_ACCESS(gt)
 
 	case GT_PCI0_CFG_ADDR:
 		if (cpu->byte_order != EMUL_LITTLE_ENDIAN) {
-			fatal("[ gt: TODO: big endian PCI access ]\n");
-			exit(1);
+			// Hack:
+			if (len == sizeof(uint32_t))
+			{
+				idata = data[0] + (data[1] << 8)
+					+ (data[2] << 16) + (data[3] << 24);
+			}
 		}
 		bus_pci_decompose_1(idata, &bus, &dev, &func, &reg);
 		bus_pci_setaddr(cpu, d->pci_data, bus, dev, func, reg);
@@ -182,8 +186,12 @@ DEVICE_ACCESS(gt)
 
 	case GT_PCI0_CFG_DATA:
 		if (cpu->byte_order != EMUL_LITTLE_ENDIAN) {
-			fatal("[ gt: TODO: big endian PCI access ]\n");
-			exit(1);
+			// Hack:
+			if (len == sizeof(uint32_t))
+			{
+				idata = data[0] + (data[1] << 8)
+					+ (data[2] << 16) + (data[3] << 24);
+			}
 		}
 		bus_pci_data_access(cpu, d->pci_data, writeflag == MEM_READ?
 		    &odata : &idata, len, writeflag);
@@ -215,15 +223,15 @@ DEVICE_ACCESS(gt)
  *  is added to the bus, then a pointer to the bus is returned.
  */
 struct pci_data *dev_gt_init(struct machine *machine, struct memory *mem,
-	uint64_t baseaddr, char *timer_irq_path, char *isa_irq_path, int type)
+	uint64_t baseaddr, const char *timer_irq_path, const char *isa_irq_path, int type)
 {
 	struct gt_data *d;
 	uint64_t pci_portbase = 0, pci_membase = 0;
 	uint64_t isa_portbase = 0, isa_membase = 0;
 	uint64_t pci_io_offset = 0, pci_mem_offset = 0;
-	char *gt_name = "NO";
+	const char *gt_name = "NO";
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct gt_data)));
+	CHECK_ALLOCATION(d = (struct gt_data *) malloc(sizeof(struct gt_data)));
 	memset(d, 0, sizeof(struct gt_data));
 
 	INTERRUPT_CONNECT(timer_irq_path, d->timer0_irq);
