@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -50,8 +50,8 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "i80321reg.h"
-#include "iopi2creg.h"
+#include "thirdparty/i80321reg.h"
+#include "thirdparty/iopi2creg.h"
 
 
 #define	TICK_SHIFT		15
@@ -101,10 +101,10 @@ static void i80321_deassert(struct i80321_data *d, uint32_t linemask)
  *  Called whenever an i80321 interrupt is asserted/deasserted.
  */
 void i80321_interrupt_assert(struct interrupt *interrupt)
-{ i80321_assert(interrupt->extra, interrupt->line); }
+{ i80321_assert((struct i80321_data *)interrupt->extra, interrupt->line); }
 void i80321_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct i80321_data *d = interrupt->extra;
+	struct i80321_data *d = (struct i80321_data *) interrupt->extra;
 
 	/*  Ack. timer interrupts:  */
 	if (interrupt->line == 1 << 9 &&
@@ -118,14 +118,14 @@ void i80321_interrupt_deassert(struct interrupt *interrupt)
 /*  TMR0 ticks, called d->hz times per second.  */
 static void tmr0_tick(struct timer *t, void *extra)
 {
-	struct i80321_data *d = extra;
+	struct i80321_data *d = (struct i80321_data *) extra;
 	d->pending_tmr0_interrupts ++;
 }
 
 
 DEVICE_TICK(i80321)
 {
-	struct i80321_data *d = extra;
+	struct i80321_data *d = (struct i80321_data *) extra;
 
 	if (cpu->cd.arm.tmr0 & TMRx_ENABLE && d->pending_tmr0_interrupts > 0) {
 		i80321_assert(d, 1 << 9);
@@ -139,9 +139,9 @@ DEVICE_TICK(i80321)
         
 DEVICE_ACCESS(i80321)
 {
-	struct i80321_data *d = extra;
+	struct i80321_data *d = (struct i80321_data *) extra;
 	uint64_t idata = 0, odata = 0;
-	char *n = NULL;
+	const char *n = NULL;
 	int bus, dev, func, reg;
 
 	if (writeflag == MEM_WRITE)
@@ -198,8 +198,7 @@ DEVICE_ACCESS(i80321)
 		if (writeflag == MEM_WRITE && idata == 0x30) {
 			int j;
 			for (j=0; j<cpu->machine->ncpus; j++)
-				cpu->machine->cpus[j]->running = 0;
-			cpu->machine->exit_without_entering_debugger = 1;
+				cpu->machine->cpus[j]->running = false;
 		}
 		break;
 	case VERDE_ATU_BASE + ATU_ATUIMR:
@@ -413,7 +412,7 @@ DEVINIT(i80321)
 	    machine->bootstrap_cpu];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct i80321_data)));
+	CHECK_ALLOCATION(d = (struct i80321_data *) malloc(sizeof(struct i80321_data)));
 	memset(d, 0, sizeof(struct i80321_data));
 
 	/*  Connect to the CPU interrupt pin:  */
@@ -421,26 +420,26 @@ DEVINIT(i80321)
 
 	/*  Register 32 i80321 interrupts:  */
 	for (i=0; i<32; i++) {
-		struct interrupt template;
-		char tmpstr[300];
-		snprintf(tmpstr, sizeof(tmpstr), "%s.i80321.%i",
+		struct interrupt templ;
+		char tmpstr2[300];
+		snprintf(tmpstr2, sizeof(tmpstr2), "%s.i80321.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = 1 << i;
-		template.name = tmpstr;
-		template.extra = d;
-		template.interrupt_assert = i80321_interrupt_assert;
-		template.interrupt_deassert = i80321_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 1 << i;
+		templ.name = tmpstr2;
+		templ.extra = d;
+		templ.interrupt_assert = i80321_interrupt_assert;
+		templ.interrupt_deassert = i80321_interrupt_deassert;
+		interrupt_handler_register(&templ);
 
 		/*
 		 *  Connect the CPU's TMR0 and TMR1 interrupts to these
 		 *  i80321 timer interrupts (nr 9 and 10):
 		 */
 		if (i == 9)
-			INTERRUPT_CONNECT(tmpstr, cpu->cd.arm.tmr0_irq);
+			INTERRUPT_CONNECT(tmpstr2, cpu->cd.arm.tmr0_irq);
 		if (i == 10)
-			INTERRUPT_CONNECT(tmpstr, cpu->cd.arm.tmr1_irq);
+			INTERRUPT_CONNECT(tmpstr2, cpu->cd.arm.tmr1_irq);
 	}
 
 	d->status = &cpu->cd.arm.i80321_isrc;
