@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -54,8 +54,8 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "jazz_r4030_dma.h"
-#include "pica.h"
+#include "thirdparty/jazz_r4030_dma.h"
+#include "thirdparty/pica.h"
 
 
 #define	DEV_JAZZ_LENGTH			0x280
@@ -112,7 +112,7 @@ void reassert_isa_interrupts(struct jazz_data *d)
 
 void jazz_interrupt_assert(struct interrupt *interrupt)
 {
-	struct jazz_data *d = interrupt->extra;
+	struct jazz_data *d = (struct jazz_data *) interrupt->extra;
 	d->int_asserted |= (1 << interrupt->line);
 
 	if (d->int_asserted & 0x7fff)
@@ -122,7 +122,7 @@ void jazz_interrupt_assert(struct interrupt *interrupt)
 }
 void jazz_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct jazz_data *d = interrupt->extra;
+	struct jazz_data *d = (struct jazz_data *) interrupt->extra;
 	d->int_asserted &= ~(1 << interrupt->line);
 
 	if (!(d->int_asserted & 0x7fff))
@@ -132,13 +132,13 @@ void jazz_interrupt_deassert(struct interrupt *interrupt)
 }
 void jazz_isa_interrupt_assert(struct interrupt *interrupt)
 {
-	struct jazz_data *d = interrupt->extra;
+	struct jazz_data *d = (struct jazz_data *) interrupt->extra;
 	d->isa_int_asserted |= (1 << interrupt->line);
 	reassert_isa_interrupts(d);
 }
 void jazz_isa_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct jazz_data *d = interrupt->extra;
+	struct jazz_data *d = (struct jazz_data *) interrupt->extra;
 	d->isa_int_asserted &= ~(1 << interrupt->line);
 	reassert_isa_interrupts(d);
 }
@@ -153,7 +153,7 @@ size_t dev_jazz_dma_controller(void *dma_controller_data,
 	struct jazz_data *d = (struct jazz_data *) dma_controller_data;
 	struct cpu *cpu = d->cpu;
 	int i, enab_writeflag;
-	int res, ncpy;
+	int ncpy;
 	uint32_t dma_addr;
 	unsigned char tr[sizeof(uint32_t)];
 	uint32_t phys_addr;
@@ -186,8 +186,7 @@ size_t dev_jazz_dma_controller(void *dma_controller_data,
 	dma_addr = d->dma0_addr;
 	i = 0;
 	while (dma_addr < d->dma0_addr + d->dma0_count && i < (int32_t)len) {
-
-		res = cpu->memory_rw(cpu, cpu->mem,
+		/* int res = */  cpu->memory_rw(cpu, cpu->mem,
 		    d->dma_translation_table_base + (dma_addr >> 12) * 8,
 		    tr, sizeof(tr), 0, PHYSICAL | NO_EXCEPTIONS);
 
@@ -210,7 +209,7 @@ size_t dev_jazz_dma_controller(void *dma_controller_data,
 		if ((phys_addr & 255) == 0 && i + 255 <= (int32_t)len)
 			ncpy = 255;
 
-		res = cpu->memory_rw(cpu, cpu->mem, phys_addr,
+		/* int res = */  cpu->memory_rw(cpu, cpu->mem, phys_addr,
 		    &data[i], ncpy, writeflag, PHYSICAL | NO_EXCEPTIONS);
 
 		dma_addr += ncpy;
@@ -226,14 +225,14 @@ size_t dev_jazz_dma_controller(void *dma_controller_data,
 
 static void timer_tick(struct timer *t, void *extra)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	d->pending_timer_interrupts ++;
 }
 
 
 DEVICE_TICK(jazz)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 
 	/*  Used by NetBSD/arc and OpenBSD/arc:  */
 	if (d->interval_start > 0 && d->interval > 0
@@ -266,14 +265,13 @@ DEVICE_TICK(jazz)
 
 DEVICE_ACCESS(jazz)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	uint64_t idata = 0, odata = 0;
-	int regnr;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
 
-	regnr = relative_addr / sizeof(uint32_t);
+	// int regnr = relative_addr / sizeof(uint32_t);
 
 	switch (relative_addr) {
 	case R4030_SYS_CONFIG:
@@ -420,14 +418,13 @@ printf("R4030_SYS_ISA_VECTOR: w=%i\n", writeflag);
 
 DEVICE_ACCESS(jazz_led)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	uint64_t idata = 0, odata = 0;
-	int regnr;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
 
-	regnr = relative_addr / sizeof(uint32_t);
+	// int regnr = relative_addr / sizeof(uint32_t);
 
 	switch (relative_addr) {
 	case 0:
@@ -466,7 +463,7 @@ DEVICE_ACCESS(jazz_led)
  */
 DEVICE_ACCESS(jazz_a0)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -518,7 +515,7 @@ DEVICE_ACCESS(jazz_a0)
  */
 DEVICE_ACCESS(jazz_20)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -570,7 +567,7 @@ DEVICE_ACCESS(jazz_20)
  */
 DEVICE_ACCESS(jazz_jazzio)
 {
-	struct jazz_data *d = extra;
+	struct jazz_data *d = (struct jazz_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int i, v;
 
@@ -623,7 +620,7 @@ DEVINIT(jazz)
 	char tmpstr[300];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct jazz_data)));
+	CHECK_ALLOCATION(d = (struct jazz_data *) malloc(sizeof(struct jazz_data)));
 	memset(d, 0, sizeof(struct jazz_data));
 
 	d->cpu = devinit->machine->cpus[0];	/*  TODO  */
@@ -637,30 +634,30 @@ DEVINIT(jazz)
 	 *  machine[y].cpu[z].jazz.isa.%i	(ISA)
 	 */
 	for (i=0; i<16; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char n[300];
 		snprintf(n, sizeof(n), "%s.jazz.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = jazz_interrupt_assert;
-		template.interrupt_deassert = jazz_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = jazz_interrupt_assert;
+		templ.interrupt_deassert = jazz_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 	for (i=0; i<16; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char n[300];
 		snprintf(n, sizeof(n), "%s.jazz.isa.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = jazz_isa_interrupt_assert;
-		template.interrupt_deassert = jazz_isa_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = jazz_isa_interrupt_assert;
+		templ.interrupt_deassert = jazz_isa_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*  Connect to MIPS CPU interrupt lines:  */
