@@ -43,7 +43,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "dec_kn01.h"
+#include "thirdparty/dec_kn01.h"
 
 
 struct kn01_data {
@@ -77,7 +77,7 @@ struct vdac_data {
 
 DEVICE_ACCESS(kn01)
 {
-	struct kn01_data *d = extra;
+	struct kn01_data *d = (struct kn01_data *) extra;
 	int csr;
 
 	if (writeflag == MEM_WRITE) {
@@ -235,7 +235,7 @@ void dev_vdac_init(struct memory *mem, uint64_t baseaddr,
 {
 	struct vdac_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct vdac_data)));
+	CHECK_ALLOCATION(d = (struct vdac_data *) malloc(sizeof(struct vdac_data)));
 	memset(d, 0, sizeof(struct vdac_data));
 
 	d->rgb_palette   = rgb_palette;
@@ -253,7 +253,7 @@ void dev_kn01_init(struct memory *mem, uint64_t baseaddr, int color_fb)
 {
 	struct kn01_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct kn01_data)));
+	CHECK_ALLOCATION(d = (struct kn01_data *) malloc(sizeof(struct kn01_data)));
 	memset(d, 0, sizeof(struct kn01_data));
 
 	d->color_fb = color_fb;
