@@ -41,8 +41,8 @@
 #include "memory.h"
 #include "misc.h"
 
+#include "thirdparty/dec_kn02.h"
 
-#include "dec_kn02.h"
 
 #define	DEV_KN02_LENGTH		0x1000
 
@@ -65,7 +65,7 @@ struct kn02_data {
  */
 void kn02_interrupt_assert(struct interrupt *interrupt)
 {
-	struct kn02_data *d = interrupt->extra;
+	struct kn02_data *d = (struct kn02_data *) interrupt->extra;
 	d->csr[0] |= interrupt->line;
 	if (d->csr[0] & d->csr[2] && !d->int_asserted) {
 		d->int_asserted = 1;
@@ -74,7 +74,7 @@ void kn02_interrupt_assert(struct interrupt *interrupt)
 }
 void kn02_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct kn02_data *d = interrupt->extra;
+	struct kn02_data *d = (struct kn02_data *) interrupt->extra;
 	d->csr[0] &= ~interrupt->line;
 	if (!(d->csr[0] & d->csr[2]) && d->int_asserted) {
 		d->int_asserted = 0;
@@ -85,7 +85,7 @@ void kn02_interrupt_deassert(struct interrupt *interrupt)
 
 DEVICE_ACCESS(kn02)
 {
-	struct kn02_data *d = extra;
+	struct kn02_data *d = (struct kn02_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -150,7 +150,7 @@ DEVINIT(kn02)
 	uint32_t csr;
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct kn02_data)));
+	CHECK_ALLOCATION(d = (struct kn02_data *) malloc(sizeof(struct kn02_data)));
 	memset(d, 0, sizeof(struct kn02_data));
 
 	/*  Connect the KN02 to a specific MIPS CPU interrupt line:  */
@@ -158,17 +158,17 @@ DEVINIT(kn02)
 
 	/*  Register the 8 possible TurboChannel interrupts:  */
 	for (i=0; i<8; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char tmpstr[300];
 		snprintf(tmpstr, sizeof(tmpstr), "%s.kn02.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = 1 << i;
-		template.name = tmpstr;
-		template.extra = d;
-		template.interrupt_assert = kn02_interrupt_assert;
-		template.interrupt_deassert = kn02_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = 1 << i;
+		templ.name = tmpstr;
+		templ.extra = d;
+		templ.interrupt_assert = kn02_interrupt_assert;
+		templ.interrupt_deassert = kn02_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*
