@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -62,19 +62,18 @@
 #include "misc.h"
 #include "net.h"
 
-#include "if_lereg.h"
+#include "thirdparty/if_lereg.h"
 
 
 #define	LE_TICK_SHIFT		14
 
-/*  #define LE_DEBUG  */
-/*  #define debug fatal  */
 
 extern int quiet_mode;
 
-#define	LE_MODE_LOOP		4
-#define	LE_MODE_DTX		2
-#define	LE_MODE_DRX		1
+#define	LE_MODE_PROM		0x8000
+#define	LE_MODE_LOOP		0x0004
+#define	LE_MODE_DTX		0x0002
+#define	LE_MODE_DRX		0x0001
 
 
 #define	N_REGISTERS		4
@@ -83,6 +82,10 @@ extern int quiet_mode;
 
 
 struct le_data {
+	struct nic_data	nic;
+
+	int		subsys;		/*  for debugmsg  */
+
 	struct interrupt irq;
 	int		irq_asserted;
 
@@ -95,19 +98,22 @@ struct le_data {
 	int		reg_select;
 	uint16_t	reg[N_REGISTERS];
 
+	int		is_bigendian;
+
 	unsigned char	*sram;
 
 	/*  Initialization block:  */
 	uint32_t	init_block_addr;
 
 	uint16_t	mode;
-	uint64_t	padr;	/*  MAC address  */
-	uint64_t	ladrf;
+	uint16_t	ladrf[4];
 	uint32_t	rdra;	/*  receive descriptor ring address  */
 	int		rlen;	/*  nr of rx descriptors  */
 	uint32_t	tdra;	/*  transmit descriptor ring address  */
 	int		tlen;	/*  nr ot tx descriptors  */
 
+	int		allmulti;/* receive all multicast packets */
+
 	/*  Current rx and tx descriptor indices:  */
 	int		rxp;
 	int		txp;
@@ -129,10 +135,12 @@ struct le_data {
  */
 static uint64_t le_read_16bit(struct le_data *d, int addr)
 {
-	/*  TODO: This is for little endian only  */
-	int x = d->sram[addr & (SRAM_SIZE-1)] +
+	if (d->is_bigendian)
+		return (d->sram[addr & (SRAM_SIZE-1)] << 8) +
+			(d->sram[(addr+1) & (SRAM_SIZE-1)]);
+
+	return d->sram[addr & (SRAM_SIZE-1)] +
 	       (d->sram[(addr+1) & (SRAM_SIZE-1)] << 8);
-	return x;
 }
 
 
@@ -143,9 +151,17 @@ static uint64_t le_read_16bit(struct le_data *d, int addr)
  */
 static void le_write_16bit(struct le_data *d, int addr, uint16_t x)
 {
-	/*  TODO: This is for little endian only  */
-	d->sram[addr & (SRAM_SIZE-1)] = x & 0xff;
-	d->sram[(addr+1) & (SRAM_SIZE-1)] = (x >> 8) & 0xff;
+	if (d->is_bigendian)
+	{
+		d->sram[addr & (SRAM_SIZE-1)] = (x >> 8) & 0xff;
+		d->sram[(addr+1) & (SRAM_SIZE-1)] = x & 0xff;
+	}
+	else
+	{
+		d->sram[addr & (SRAM_SIZE-1)] = x & 0xff;
+		d->sram[(addr+1) & (SRAM_SIZE-1)] = (x >> 8) & 0xff;
+	}
+
 }
 
 
@@ -157,21 +173,48 @@ static void le_write_16bit(struct le_data *d, int addr, uint16_t x)
  */
 static void le_chip_init(struct le_data *d)
 {
+	uint16_t tmp;
+	uint8_t macaddr[6];
+
 	d->init_block_addr = (d->reg[1] & 0xffff) + ((d->reg[2] & 0xff) << 16);
 	if (d->init_block_addr & 1)
-		fatal("[ le: WARNING! initialization block address "
-		    "not word aligned? ]\n");
+		debugmsg(d->subsys, "initialization block", VERBOSITY_WARNING,
+		    "WARNING! address not word aligned?");
 
-	debug("[ le: d->init_block_addr = 0x%06x ]\n", d->init_block_addr);
+	debugmsg(d->subsys, "initialization block", VERBOSITY_DEBUG,
+	    "init_block_addr = 0x%06x", d->init_block_addr);
 
 	d->mode = le_read_16bit(d, d->init_block_addr + 0);
-	d->padr = le_read_16bit(d, d->init_block_addr + 2);
-	d->padr += (le_read_16bit(d, d->init_block_addr + 4) << 16);
-	d->padr += (le_read_16bit(d, d->init_block_addr + 6) << 32);
-	d->ladrf = le_read_16bit(d, d->init_block_addr + 8);
-	d->ladrf += (le_read_16bit(d, d->init_block_addr + 10) << 16);
-	d->ladrf += (le_read_16bit(d, d->init_block_addr + 12) << 32);
-	d->ladrf += (le_read_16bit(d, d->init_block_addr + 14) << 48);
+
+	/*
+	 * The MAC address is packed into the PADR field as 3 little-endian
+	 * 16-bit words.
+	 */
+	tmp = le_read_16bit(d, d->init_block_addr + 2);
+	macaddr[0] = (uint8_t)(tmp);
+	macaddr[1] = (uint8_t)(tmp >> 8);
+	tmp = le_read_16bit(d, d->init_block_addr + 4);
+	macaddr[2] = (uint8_t)(tmp);
+	macaddr[3] = (uint8_t)(tmp >> 8);
+	tmp = le_read_16bit(d, d->init_block_addr + 6);
+	macaddr[4] = (uint8_t)(tmp);
+	macaddr[5] = (uint8_t)(tmp >> 8);
+	memcpy(d->nic.mac_address, macaddr, sizeof(d->nic.mac_address));
+
+	/*
+	 * The muticast address filter is packed into the LADRF field
+	 * as 4 little-endian 16-bit words.
+	 */
+	d->ladrf[0] = le_read_16bit(d, d->init_block_addr + 8);
+	d->ladrf[1] = le_read_16bit(d, d->init_block_addr + 10);
+	d->ladrf[2] = le_read_16bit(d, d->init_block_addr + 12);
+	d->ladrf[3] = le_read_16bit(d, d->init_block_addr + 14);
+	if (d->ladrf[0] == 0xffff && d->ladrf[1] == 0xffff &&
+	    d->ladrf[2] == 0xffff && d->ladrf[3] == 0xffff)
+		d->allmulti = 1;
+	else
+		d->allmulti = 0;
+
 	d->rdra = le_read_16bit(d, d->init_block_addr + 16);
 	d->rdra += ((le_read_16bit(d, d->init_block_addr + 18) & 0xff) << 16);
 	d->rlen = 1 << ((le_read_16bit(d, d->init_block_addr + 18) >> 13) & 7);
@@ -179,13 +222,16 @@ static void le_chip_init(struct le_data *d)
 	d->tdra += ((le_read_16bit(d, d->init_block_addr + 22) & 0xff) << 16);
 	d->tlen = 1 << ((le_read_16bit(d, d->init_block_addr + 22) >> 13) & 7);
 
-	debug("[ le: DEBUG: mode              %04x ]\n", d->mode);
-	debug("[ le: DEBUG: padr  %016llx ]\n", (long long)d->padr);
-	debug("[ le: DEBUG: ladrf %016llx ]\n", (long long)d->ladrf);
-	debug("[ le: DEBUG: rdra            %06llx ]\n", d->rdra);
-	debug("[ le: DEBUG: rlen               %3i ]\n", d->rlen);
-	debug("[ le: DEBUG: tdra            %06llx ]\n", d->tdra);
-	debug("[ le: DEBUG: tlen               %3i ]\n", d->tlen);
+	debug("[ le: DEBUG: mode                 %04x ]\n", d->mode);
+	debug("[ le: DEBUG: padr    %02x:%02x:%02x:%02x:%02x:%02x ]\n",
+	      macaddr[0], macaddr[1], macaddr[2],
+	      macaddr[3], macaddr[4], macaddr[5]);
+	debug("[ le: DEBUG: ladrf %04x:%04x:%04x:%04x ]\n",
+	      d->ladrf[0], d->ladrf[1], d->ladrf[2], d->ladrf[3]);
+	debug("[ le: DEBUG: rdra               %06llx ]\n", d->rdra);
+	debug("[ le: DEBUG: rlen                  %3i ]\n", d->rlen);
+	debug("[ le: DEBUG: tdra               %06llx ]\n", d->tdra);
+	debug("[ le: DEBUG: tlen                  %3i ]\n", d->tlen);
 
 	/*  Set TXON and RXON, unless they are disabled by 'mode':  */
 	if (d->mode & LE_MODE_DTX)
@@ -198,6 +244,9 @@ static void le_chip_init(struct le_data *d)
 	else
 		d->reg[0] |= LE_RXON;
 
+	/*  Initialize promiscuous mode.  */
+	d->nic.promiscuous_mode = (d->mode & LE_MODE_PROM) ? 1 : 0;
+
 	/*  Go to the start of the descriptor rings:  */
 	d->rxp = d->txp = 0;
 
@@ -261,22 +310,25 @@ static void le_tx(struct net *net, struct le_data *d)
 		if ((tx_descr[2] & 0xf000) != 0xf000)
 			return;
 		if (buflen < 12 || buflen > 1900) {
-			fatal("[ le_tx(): buflen = %i ]\n", buflen);
+			debugmsg(d->subsys, "tx", VERBOSITY_WARNING, "buflen = %i", buflen);
 			return;
 		}
 
-		debug("[ le_tx(): descr %3i DUMP: 0x%04x 0x%04x 0x%04x 0x%04x "
-		    "=> addr=0x%06x, len=%i bytes, STP=%i ENP=%i ]\n", d->txp,
+		debugmsg(d->subsys, "tx", VERBOSITY_DEBUG,
+		    "descr %3i DUMP: 0x%04x 0x%04x 0x%04x 0x%04x "
+		    "=> addr=0x%06x, len=%i bytes, STP=%i ENP=%i", d->txp,
 		    tx_descr[0], tx_descr[1], tx_descr[2], tx_descr[3],
 		    bufaddr, buflen, stp, enp);
 
 		if (d->tx_packet == NULL && !stp) {
-			fatal("[ le_tx(): !stp but tx_packet == NULL ]\n");
+			debugmsg(d->subsys, "tx", VERBOSITY_WARNING,
+			    "!stp but tx_packet == NULL");
 			return;
 		}
 
 		if (d->tx_packet != NULL && stp) {
-			fatal("[ le_tx(): stp but tx_packet != NULL ]\n");
+			debugmsg(d->subsys, "tx", VERBOSITY_WARNING,
+			    "stp but tx_packet != NULL");
 			free(d->tx_packet);
 			d->tx_packet = NULL;
 			d->tx_packet_len = 0;
@@ -288,10 +340,10 @@ static void le_tx(struct net *net, struct le_data *d)
 		/*  Start of a new packet:  */
 		if (stp) {
 			d->tx_packet_len = buflen;
-			CHECK_ALLOCATION(d->tx_packet = malloc(buflen));
+			CHECK_ALLOCATION(d->tx_packet = (unsigned char *) malloc(buflen));
 		} else {
 			d->tx_packet_len += buflen;
-			CHECK_ALLOCATION(d->tx_packet =
+			CHECK_ALLOCATION(d->tx_packet = (unsigned char *)
 			    realloc(d->tx_packet, d->tx_packet_len));
 		}
 
@@ -308,7 +360,8 @@ static void le_tx(struct net *net, struct le_data *d)
 		 *  the packet.
 		 */
 		if (enp) {
-			net_ethernet_tx(net, d, d->tx_packet, d->tx_packet_len);
+			net_ethernet_tx(net, &d->nic, d->tx_packet,
+			    d->tx_packet_len);
 
 			free(d->tx_packet);
 			d->tx_packet = NULL;
@@ -332,7 +385,7 @@ static void le_tx(struct net *net, struct le_data *d)
 	} while (d->txp != start_txp);
 
 	/*  We are here if all descriptors were taken care of.  */
-	fatal("[ le_tx(): all TX descriptors used up? ]\n");
+	debugmsg(d->subsys, "tx", VERBOSITY_WARNING, "all descriptors used up?");
 }
 
 
@@ -371,12 +424,13 @@ static void le_rx(struct net *net, struct le_data *d)
 		if ((rx_descr[2] & 0xf000) != 0xf000)
 			return;
 		if (buflen < 12 || buflen > 1900) {
-			fatal("[ le_rx(): buflen = %i ]\n", buflen);
+			debugmsg(d->subsys, "rx", VERBOSITY_WARNING, "buflen = %i", buflen);
 			return;
 		}
 
-		debug("[ le_rx(): descr %3i DUMP: 0x%04x 0x%04x 0x%04x 0x%04x "
-		    "=> addr=0x%06x, len=%i bytes ]\n", d->rxp,
+		debugmsg(d->subsys, "rx", VERBOSITY_DEBUG,
+		    "descr %3i DUMP: 0x%04x 0x%04x 0x%04x 0x%04x "
+		    "=> addr=0x%06x, len=%i bytes", d->rxp,
 		    rx_descr[0], rx_descr[1], rx_descr[2], rx_descr[3],
 		    bufaddr, buflen);
 
@@ -441,7 +495,63 @@ static void le_rx(struct net *net, struct le_data *d)
 	} while (d->rxp != start_rxp);
 
 	/*  We are here if all descriptors were taken care of.  */
-	fatal("[ le_rx(): all RX descriptors used up? ]\n");
+	debugmsg(d->subsys, "rx", VERBOSITY_WARNING, "all descriptors used up?");
+}
+
+
+/*
+ *  le_rx_drop_packet():
+ *
+ *  Implement the logic to determine if we should drop a packet
+ *  before passing it to the guest.  Returns 1 if the packet was
+ *  dropped.
+ */
+static int
+le_rx_drop_packet(struct net *net, struct le_data *d)
+{
+	/* Only implement filtering if using a tap device. */
+	if (net->tapdev == NULL)
+		return 0;
+
+	/*
+	 * The network layer has already checked for our MAC address
+	 * or promiscuous mode.  We just need to check the multicast
+	 * filter or broadcast.
+	 */
+
+	/* If the packet is not multicast, we know it should be received. */
+	if (! net_ether_multicast(d->rx_packet))
+		return 0;
+
+	/*
+	 * Optimization -- if the guest has set all of the filter
+	 * bits, then we can skip additional checks.
+	 */
+	if (d->allmulti)
+		return 0;
+
+	/* Check for broadcast. */
+	if (net_ether_broadcast(d->rx_packet))
+		return 0;
+
+	/*
+	 * Check the multicast address filter.  We pass the address
+	 * through the little-endian Ethernet CRC generator.  The
+	 * high-order 6 bits are the index into the 64-bit filter.
+	 * The upper 2 bits select the 16-bit filter word, and the
+	 * remaining 4 select the bit in the word.
+	 */
+	uint32_t crc = net_ether_crc32_le(d->rx_packet, 6);
+	crc >>= 26;
+	if (d->ladrf[crc >> 4] & (1 << (crc & 0xf)))
+		return 0;
+
+	/* Not for us; drop the packet. */
+	free(d->rx_packet);
+	d->rx_packet = NULL;
+	d->rx_packet_len = 0;
+
+	return 1;
 }
 
 
@@ -454,16 +564,6 @@ static void le_register_fix(struct net *net, struct le_data *d)
 	if (d->reg[0] & LE_INIT)
 		le_chip_init(d);
 
-#ifdef LE_DEBUG
-	{
-		static int x = 1234;
-		if (x != d->reg[0]) {
-			debug("[ le reg[0] = 0x%04x ]\n", d->reg[0]);
-			x = d->reg[0];
-		}
-	}
-#endif
-
 	/*
 	 *  If the receiver is on:
 	 *  If there is a current rx_packet, try to receive it into the
@@ -481,9 +581,12 @@ static void le_register_fix(struct net *net, struct le_data *d)
 				break;
 
 			if (d->rx_packet == NULL &&
-			    net_ethernet_rx_avail(net, d))
-				net_ethernet_rx(net, d,
+			    net_ethernet_rx_avail(net, &d->nic)) {
+				net_ethernet_rx(net, &d->nic,
 				    &d->rx_packet, &d->rx_packet_len);
+				if (le_rx_drop_packet(net, d))
+					continue;
+			}
 		} while (d->rx_packet != NULL);
 	}
 
@@ -523,7 +626,7 @@ static void le_register_fix(struct net *net, struct le_data *d)
 
 DEVICE_TICK(le)
 {
-	struct le_data *d = extra;
+	struct le_data *d = (struct le_data *) extra;
 	int new_assert;
 
 	le_register_fix(cpu->machine->emul->net, d);
@@ -573,8 +676,8 @@ void le_register_write(struct le_data *d, int r, uint32_t x)
 		}
 		if (x & LE_INIT) {
 			if (!(d->reg[r] & LE_STOP))
-				fatal("[ le: attempt to INIT before"
-				    " STOPped! ]\n");
+				debugmsg(d->subsys, "register write", VERBOSITY_WARNING,
+				    "attempt to INIT before STOPped!");
 			d->reg[r] |= LE_INIT;
 			d->reg[r] &= ~LE_STOP;
 		}
@@ -598,7 +701,7 @@ void le_register_write(struct le_data *d, int r, uint32_t x)
 
 DEVICE_ACCESS(le_sram)
 {
-	struct le_data *d = extra;
+	struct le_data *d = (struct le_data *) extra;
 	size_t i;
 	int retval;
 
@@ -643,7 +746,7 @@ DEVICE_ACCESS(le_sram)
 
 DEVICE_ACCESS(le)
 {
-	struct le_data *d = extra;
+	struct le_data *d = (struct le_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int retval = 1;
 	size_t i;
@@ -687,9 +790,10 @@ DEVICE_ACCESS(le)
 	case 0:
 		if (writeflag==MEM_READ) {
 			odata = d->reg[d->reg_select];
-			if (!quiet_mode)
-				debug("[ le: read from register 0x%02x: 0x"
-				    "%02x ]\n", d->reg_select, (int)odata);
+			debugmsg(d->subsys, "register", VERBOSITY_DEBUG,
+			    "READ from register 0x%02x: 0x"
+			    "%02x", d->reg_select, (int)odata);
+
 			/*
 			 *  A read from csr1..3 should return "undefined"
 			 *  result if the stop bit is set.  However, Ultrix
@@ -697,9 +801,9 @@ DEVICE_ACCESS(le)
 			 *  a warning here.
 			 */
 		} else {
-			if (!quiet_mode)
-				debug("[ le: write to register 0x%02x: 0x"
-				    "%02x ]\n", d->reg_select, (int)idata);
+			debugmsg(d->subsys, "register", VERBOSITY_DEBUG,
+			    "WRITE to register 0x%02x: 0x%02x",
+			    d->reg_select, (int)idata);
 			/*
 			 *  A write to from csr1..3 when the stop bit is
 			 *  set should be ignored. However, Ultrix writes
@@ -714,37 +818,37 @@ DEVICE_ACCESS(le)
 	case 4:
 		if (writeflag==MEM_READ) {
 			odata = d->reg_select;
-			if (!quiet_mode)
-				debug("[ le: read from register select: "
-				    "0x%02x ]\n", (int)odata);
+			debugmsg(d->subsys, "register select", VERBOSITY_DEBUG,
+			    "READ: 0x%02x", (int)odata);
 		} else {
-			if (!quiet_mode)
-				debug("[ le: write to register select: "
-				    "0x%02x ]\n", (int)idata);
+			debugmsg(d->subsys, "register select", VERBOSITY_DEBUG,
+			     "WRITE: 0x%02x", (int)idata);
 			d->reg_select = idata & (N_REGISTERS - 1);
+
 			if (idata >= N_REGISTERS)
-				fatal("[ le: WARNING! register select %i "
-				    "(max is %i) ]\n", idata, N_REGISTERS - 1);
+				debugmsg(d->subsys, "register select", VERBOSITY_WARNING,
+				    "WARNING! register %i (max is %i)",
+				    idata, N_REGISTERS - 1);
 		}
 		break;
 
 	default:
 		if (writeflag==MEM_READ) {
-			fatal("[ le: read from UNIMPLEMENTED addr 0x%06x ]\n",
+			debugmsg(d->subsys, "access", VERBOSITY_WARNING,
+			    "read from UNIMPLEMENTED addr 0x%06x",
 			    (int)relative_addr);
 		} else {
-			fatal("[ le: write to UNIMPLEMENTED addr 0x%06x: "
-			    "0x%08x ]\n", (int)relative_addr, (int)idata);
+			debugmsg(d->subsys, "access", VERBOSITY_WARNING,
+			    "write to UNIMPLEMENTED addr 0x%06x: "
+			    "0x%08x", (int)relative_addr, (int)idata);
 		}
 	}
 
 do_return:
 	if (writeflag == MEM_READ) {
 		memory_writemax64(cpu, data, len, odata);
-#ifdef LE_DEBUG
-		fatal("[ le: read from addr 0x%06x: 0x%08x ]\n",
-		    relative_addr, odata);
-#endif
+		debugmsg(d->subsys, "access", VERBOSITY_DEBUG,
+		    "read from addr 0x%06x: 0x%08x", relative_addr, odata);
 	}
 
 	dev_le_tick(cpu, extra);
@@ -757,18 +861,22 @@ do_return:
  *  dev_le_init():
  */
 void dev_le_init(struct machine *machine, struct memory *mem, uint64_t baseaddr,
-	uint64_t buf_start, uint64_t buf_end, char *irq_path, int len)
+	uint64_t buf_start, uint64_t buf_end, const char *irq_path, int len)
 {
 	char *name2;
 	size_t nlen = 55;
 	struct le_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct le_data)));
+	CHECK_ALLOCATION(d = (struct le_data *) malloc(sizeof(struct le_data)));
 	memset(d, 0, sizeof(struct le_data));
 
+	d->subsys = debugmsg_register_subsystem("le");
+
+	d->is_bigendian = machine->cpus[0]->byte_order == EMUL_BIG_ENDIAN;
+
 	INTERRUPT_CONNECT(irq_path, d->irq);
 
-	CHECK_ALLOCATION(d->sram = malloc(SRAM_SIZE));
+	CHECK_ALLOCATION(d->sram = (unsigned char *) malloc(SRAM_SIZE));
 	memset(d->sram, 0, SRAM_SIZE);
 
 	/*  TODO:  Are these actually used yet?  */
@@ -804,7 +912,7 @@ void dev_le_init(struct machine *machine, struct memory *mem, uint64_t baseaddr,
 	    DM_DYNTRANS_OK | DM_DYNTRANS_WRITE_OK
 	    | DM_READS_HAVE_NO_SIDE_EFFECTS, d->sram);
 
-	CHECK_ALLOCATION(name2 = malloc(nlen));
+	CHECK_ALLOCATION(name2 = (char *) malloc(nlen));
 	snprintf(name2, nlen, "le [%02x:%02x:%02x:%02x:%02x:%02x]",
 	    d->rom[0], d->rom[1], d->rom[2], d->rom[3], d->rom[4], d->rom[5]);
 
@@ -813,6 +921,7 @@ void dev_le_init(struct machine *machine, struct memory *mem, uint64_t baseaddr,
 
 	machine_add_tickfunction(machine, dev_le_tick, d, LE_TICK_SHIFT);
 
-	net_add_nic(machine->emul->net, d, &d->rom[0]);
+	memcpy(d->nic.mac_address, &d->rom[0], sizeof(d->nic.mac_address));
+	net_add_nic(machine->emul->net, &d->nic);
 }
 
