@@ -41,7 +41,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "maltareg.h"
+#include "thirdparty/maltareg.h"
 
 
 #define	DEV_MALTA_LCD_LENGTH		0x80
@@ -58,7 +58,7 @@ struct malta_lcd_data {
 
 DEVICE_TICK(malta_lcd)
 { 
-	struct malta_lcd_data *d = extra;
+	struct malta_lcd_data *d = (struct malta_lcd_data *) extra;
 	int i;
 
 	if (d->display_modified == 0)
@@ -82,7 +82,7 @@ DEVICE_TICK(malta_lcd)
 
 DEVICE_ACCESS(malta_lcd)
 {
-	struct malta_lcd_data *d = extra;
+	struct malta_lcd_data *d = (struct malta_lcd_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int pos;
 
@@ -132,7 +132,7 @@ DEVINIT(malta_lcd)
 {
 	struct malta_lcd_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct malta_lcd_data)));
+	CHECK_ALLOCATION(d = (struct malta_lcd_data *) malloc(sizeof(struct malta_lcd_data)));
 	memset(d, 0, sizeof(struct malta_lcd_data));
 
 	d->base_addr = devinit->addr;
