@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -49,7 +49,7 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "mc146818reg.h"
+#include "thirdparty/mc146818reg.h"
 
 
 #define	to_bcd(x)	( ((x)/10) * 16 + ((x)%10) )
@@ -114,7 +114,7 @@ static void timer_tick(struct timer *timer, void *extra)
 
 DEVICE_TICK(mc146818)
 {
-	struct mc_data *d = extra;
+	struct mc_data *d = (struct mc_data *) extra;
 	int pti = d->pending_timer_interrupts;
 
 	if ((d->reg[MC_REGB * 4] & MC_REGB_PIE) && pti > 0) {
@@ -151,7 +151,7 @@ DEVICE_TICK(mc146818)
  */
 DEVICE_ACCESS(mc146818_jazz)
 {
-	struct mc_data *d = extra;
+	struct mc_data *d = (struct mc_data *) extra;
 
 #ifdef MC146818_DEBUG
 	if (writeflag == MEM_WRITE) {
@@ -243,8 +243,11 @@ static void mc146818_update_time(struct mc_data *d)
 		break;
 	case MC146818_DEC:
 		/*
-		 *  DECstations must have 72 or 73 in the
-		 *  Year field, or Ultrix screems.  (Weird.)
+		 *  DECstations must have 72 or 73 in the Year field, or
+		 *  Ultrix screems. This is also mentioned in NetBSD's
+		 *  mcclock.c.  Unfortunately, this means that NetBSD cannot
+		 *  know what year it actually is, so it has to deduce it
+		 *  from the filesystem.
 		 */
 		d->reg[4 * MC_YEAR] = 72;
 
@@ -276,7 +279,7 @@ static void mc146818_update_time(struct mc_data *d)
 
 DEVICE_ACCESS(mc146818)
 {
-	struct mc_data *d = extra;
+	struct mc_data *d = (struct mc_data *) extra;
 	struct tm *tmp;
 	time_t timet;
 	size_t i;
@@ -475,9 +478,7 @@ DEVICE_ACCESS(mc146818)
 				/*  Used on SGI to power off the machine.  */
 				fatal("[ md146818: power off ]\n");
 				for (j=0; j<cpu->machine->ncpus; j++)
-					cpu->machine->cpus[j]->running = 0;
-				cpu->machine->
-				    exit_without_entering_debugger = 1;
+					cpu->machine->cpus[j]->running = false;
 			}
 			break;
 		default:
@@ -518,6 +519,7 @@ DEVICE_ACCESS(mc146818)
 					break;
 				}
 			}
+			// fall through
 		case 4 * MC_MIN:
 		case 4 * MC_HOUR:
 		case 4 * MC_DOW:
@@ -591,7 +593,7 @@ void dev_mc146818_init(struct machine *machine, struct memory *mem,
 	int i, dev_len;
 	struct mc_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct mc_data)));
+	CHECK_ALLOCATION(d = (struct mc_data *) malloc(sizeof(struct mc_data)));
 	memset(d, 0, sizeof(struct mc_data));
 
 	d->access_style  = access_style;
