@@ -43,8 +43,7 @@
 #include "memory.h"
 #include "misc.h"
 
-
-#include "mk48txxreg.h"
+#include "thirdparty/mk48txxreg.h"
 
 
 #define	MK48TXX_LEN		MK48T08_CLKSZ
@@ -76,7 +75,7 @@ void mk48txx_update_regs(struct mk48txx_data *d)
 
 DEVICE_ACCESS(mk48txx)
 {
-	struct mk48txx_data *d = extra;
+	struct mk48txx_data *d = (struct mk48txx_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -131,7 +130,7 @@ DEVINIT(mk48txx)
 {
 	struct mk48txx_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct mk48txx_data)));
+	CHECK_ALLOCATION(d = (struct mk48txx_data *) malloc(sizeof(struct mk48txx_data)));
 	memset(d, 0, sizeof(struct mk48txx_data));
 
 	mk48txx_update_regs(d);
