@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -64,12 +64,9 @@ struct mp_data {
 };
 
 
-extern int single_step;
-
-
 DEVICE_ACCESS(mp)
 {
-	struct mp_data *d = extra;
+	struct mp_data *d = (struct mp_data *) extra;
 	int i, which_cpu;
 	uint64_t idata = 0, odata = 0;
 
@@ -95,7 +92,7 @@ DEVICE_ACCESS(mp)
 	case DEV_MP_STARTUPCPU:
 		which_cpu = idata;
 		d->cpus[which_cpu]->pc = d->startup_addr;
-		switch (cpu->machine->arch) {
+		switch (cpu->cpu_family->arch) {
 		case ARCH_MIPS:
 			d->cpus[which_cpu]->cd.mips.gpr[MIPS_GPR_SP] =
 			    d->stack_addr;
@@ -108,7 +105,7 @@ DEVICE_ACCESS(mp)
 			    " arch yet!\n");
 			exit(1);
 		}
-		d->cpus[which_cpu]->running = 1;
+		d->cpus[which_cpu]->running = true;
 		/*  debug("[ dev_mp: starting up cpu%i at 0x%llx ]\n", 
 		    which_cpu, (long long)d->startup_addr);  */
 		break;
@@ -129,7 +126,7 @@ DEVICE_ACCESS(mp)
 
 		for (i=0; i<cpu->machine->ncpus; i++)
 			if (i != which_cpu)
-				d->cpus[i]->running = 0;
+				d->cpus[i]->running = false;
 		break;
 
 	case DEV_MP_UNPAUSE_CPU:
@@ -137,7 +134,7 @@ DEVICE_ACCESS(mp)
 		which_cpu = idata;
 
 		if (which_cpu >= 0 && which_cpu <cpu->machine->ncpus)
-			d->cpus[which_cpu]->running = 1;
+			d->cpus[which_cpu]->running = true;
 		break;
 
 	case DEV_MP_STARTUPSTACK:
@@ -166,7 +163,7 @@ DEVICE_ACCESS(mp)
 		 *  It is actually located at machine->memory_offset_in_mb
 		 *  but that is only used for SGI emulation so far.)
 		 */
-		odata = cpu->machine->physical_ram_in_mb * 1048576;
+		odata = cpu->machine->physical_ram_in_mb * 1048576ULL;
 		break;
 
 	case DEV_MP_IPI_ONE:
@@ -192,7 +189,7 @@ DEVICE_ACCESS(mp)
 				send_it = 1;
 			if (send_it) {
 				d->n_pending_ipis[i] ++;
-				CHECK_ALLOCATION(d->ipi[i] = realloc(d->ipi[i],
+				CHECK_ALLOCATION(d->ipi[i] = (int *) realloc(d->ipi[i],
 				    d->n_pending_ipis[i] * sizeof(int)));
 
 				/*  Add the IPI last in the array:  */
@@ -256,7 +253,7 @@ DEVINIT(mp)
 	struct mp_data *d;
 	int n, i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct mp_data)));
+	CHECK_ALLOCATION(d = (struct mp_data *) malloc(sizeof(struct mp_data)));
 	memset(d, 0, sizeof(struct mp_data));
 
 	d->cpus = devinit->machine->cpus;
@@ -266,7 +263,8 @@ DEVINIT(mp)
 	n = devinit->machine->ncpus;
 
 	/*  Connect to all CPUs' IPI pins:  */
-	CHECK_ALLOCATION(d->ipi_irq = malloc(n * sizeof(struct interrupt)));
+	CHECK_ALLOCATION(d->ipi_irq = (struct interrupt *)
+	    malloc(n * sizeof(struct interrupt)));
 
 	for (i=0; i<n; i++) {
 		char tmpstr[200];
@@ -275,10 +273,10 @@ DEVINIT(mp)
 		INTERRUPT_CONNECT(tmpstr, d->ipi_irq[i]);
 	}
 
-	CHECK_ALLOCATION(d->n_pending_ipis = malloc(n * sizeof(int)));
+	CHECK_ALLOCATION(d->n_pending_ipis = (int *) malloc(n * sizeof(int)));
 	memset(d->n_pending_ipis, 0, sizeof(int) * n);
 
-	CHECK_ALLOCATION(d->ipi = malloc(n * sizeof(int *)));
+	CHECK_ALLOCATION(d->ipi = (int **) malloc(n * sizeof(int *)));
 	memset(d->ipi, 0, sizeof(int *) * n);
 
 	memory_device_register(devinit->machine->memory, devinit->name,
