@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2007-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2007-2014  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -52,9 +52,9 @@
 #include "memory.h"
 #include "misc.h"
 
+#include "thirdparty/mvme187.h"
+#include "thirdparty/mvme_pcctworeg.h"
 
-#include "mvme187.h"
-#include "mvme_pcctworeg.h"
 
 #define	INTERRUPT_LEVEL_MASK	0x07
 
@@ -62,7 +62,7 @@
 #define DEV_PCC2_TICK_SHIFT	14
 
 
-#define	debug fatal
+// #define	debug fatal
 
 struct pcc2_data {
 	struct interrupt	cpu_irq;
@@ -81,7 +81,7 @@ struct pcc2_data {
 static uint32_t read32(unsigned char *p)
 {
 	uint32_t x = 0;
-	int i;
+	size_t i;
 
 	for (i=0; i<sizeof(x); i++) {
 		x <<= 8;
@@ -94,7 +94,7 @@ static uint32_t read32(unsigned char *p)
 
 static void write32(unsigned char *p, uint32_t x)
 {
-	int i;
+	size_t i;
 
 	for (i=0; i<sizeof(x); i++) {
 		p[sizeof(x) - 1 - i] = x;
@@ -251,6 +251,10 @@ static void reassert_interrupts(struct pcc2_data *d)
 	    (d->pcctwo_reg[PCCTWO_MASK] & INTERRUPT_LEVEL_MASK))
 		assert = 1;
 
+	debug("[ pcc2: IPL = %i, MASK = %i => %s ]\n", d->pcctwo_reg[PCCTWO_IPL]& INTERRUPT_LEVEL_MASK,
+		d->pcctwo_reg[PCCTWO_MASK] & INTERRUPT_LEVEL_MASK,
+		assert ? "ASSERT" : "no assert");
+
 	/*  ... but only allow interrupts if Master Interrupt Enable is on:  */
 	if (!(d->pcctwo_reg[PCCTWO_GENCTL] & PCC2_MIEN))
 		assert = 0;
@@ -264,7 +268,7 @@ static void reassert_interrupts(struct pcc2_data *d)
 
 void pcctwo_interrupt_common(struct interrupt *interrupt, int assert)
 {
-	struct pcc2_data *d = interrupt->extra;
+	struct pcc2_data *d = (struct pcc2_data *) interrupt->extra;
 
 	switch (interrupt->line) {
 
@@ -310,7 +314,7 @@ static void pcctwo_interrupt_deassert(struct interrupt *interrupt)
 
 DEVICE_TICK(pcc2)
 {
-	struct pcc2_data *d = extra;
+	struct pcc2_data *d = (struct pcc2_data *) extra;
 	reassert_interrupts(d);
 }
 
@@ -318,7 +322,7 @@ DEVICE_TICK(pcc2)
 DEVICE_ACCESS(pcc2)
 {
 	uint64_t idata = 0;
-	struct pcc2_data *d = extra;
+	struct pcc2_data *d = (struct pcc2_data *) extra;
 
 	/*  0xfff42000..0xfff42fff, but only 0x40 unique registers:  */
 	relative_addr %= PCC2_SIZE;
@@ -568,13 +572,15 @@ DEVICE_ACCESS(pcc2)
 DEVICE_ACCESS(mvme187_iack)
 {
 	uint64_t odata = 0;
-	struct pcc2_data *d = extra;
+	struct pcc2_data *d = (struct pcc2_data *) extra;
 
 	if (writeflag == MEM_WRITE) {
 		fatal("[ pcc2: write to mvme187_iack? ]\n");
 	} else {
 		odata = d->pcctwo_reg[PCCTWO_VECBASE] + d->cur_int_vec[
 		    (relative_addr >> 2) & INTERRUPT_LEVEL_MASK];
+		debug("[ pcc2: mvme187_iack level %i => vector 0x%02x ]\n",
+		    (int)(relative_addr >> 2), (int)odata);
 
 		memory_writemax64(cpu, data, len, odata);
 	}
@@ -588,7 +594,7 @@ DEVINIT(pcc2)
 	struct pcc2_data *d;
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pcc2_data)));
+	CHECK_ALLOCATION(d = (struct pcc2_data *) malloc(sizeof(struct pcc2_data)));
 	memset(d, 0, sizeof(struct pcc2_data));
 
 	/*
@@ -610,19 +616,19 @@ DEVINIT(pcc2)
 	 *  Register the 16 PCC2 interrupt vectors:
 	 */
 	for (i=0; i<16; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char n[300];
 		snprintf(n, sizeof(n), "%s.pcc2.%i",
 		    devinit->interrupt_path, i);
 
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = pcctwo_interrupt_assert;
-		template.interrupt_deassert = pcctwo_interrupt_deassert;
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = pcctwo_interrupt_assert;
+		templ.interrupt_deassert = pcctwo_interrupt_deassert;
 
-		interrupt_handler_register(&template);
+		interrupt_handler_register(&templ);
 	}
 
 	memory_device_register(devinit->machine->memory, "pcc2",
