@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2019  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -56,7 +56,7 @@ struct pccmos_data {
 
 DEVICE_ACCESS(pccmos)
 {
-	struct pccmos_data *d = extra;
+	struct pccmos_data *d = (struct pccmos_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	unsigned char b = 0;
 	int r = 1;
@@ -114,7 +114,7 @@ DEVINIT(pccmos)
 	int type = MC146818_PC_CMOS, len = DEV_PCCMOS_LENGTH;
 	struct pccmos_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pccmos_data)));
+	CHECK_ALLOCATION(d = (struct pccmos_data *) malloc(sizeof(struct pccmos_data)));
 	memset(d, 0, sizeof(struct pccmos_data));
 
 	switch (devinit->machine->machine_type) {
@@ -135,9 +135,7 @@ DEVINIT(pccmos)
 		/*  Malta etc.  */
 		type = MC146818_ALGOR;
 		break;
-	case MACHINE_QEMU_MIPS:
 	case MACHINE_COBALT:
-	case MACHINE_BEBOX:
 	case MACHINE_PREP:
 	case MACHINE_MVMEPPC:
 	case MACHINE_ALPHA:
