@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -44,8 +44,8 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "i82365reg.h"
-#include "pcmciareg.h"
+#include "thirdparty/i82365reg.h"
+#include "thirdparty/pcmciareg.h"
 
 
 /*  #define debug fatal  */
@@ -61,9 +61,9 @@ struct pcic_data {
 DEVICE_ACCESS(pcic_cis)
 {
 	/*  struct pcic_data *d = (struct pcic_data *) extra;  */
-	uint64_t idata = 0, odata = 0;
+	uint64_t odata = 0;
 
-	idata = memory_readmax64(cpu, data, len);
+	// uint64_t idata = memory_readmax64(cpu, data, len);
 
 {
 #if 0
@@ -140,7 +140,7 @@ DEVICE_ACCESS(pcic_cis)
 
 DEVICE_ACCESS(pcic)
 {
-	struct pcic_data *d = extra;
+	struct pcic_data *d = (struct pcic_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int controller_nr, socket_nr;
 
@@ -216,7 +216,7 @@ DEVINIT(pcic)
 	char tmpstr[200];
 	struct pcic_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pcic_data)));
+	CHECK_ALLOCATION(d = (struct pcic_data *) malloc(sizeof(struct pcic_data)));
 	memset(d, 0, sizeof(struct pcic_data));
 
 	INTERRUPT_CONNECT(devinit->interrupt_path, d->irq);
