@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -31,9 +31,11 @@
  *
  *  Quick source of good info: http://my.execpc.com/~geezer/osd/kbd/kbd.txt
  *
- *
- *  TODOs:
- *	Finish the rewrite for 8242.
+ *  Some scancode listings for various types:
+ *	http://www.win.tue.nl/~aeb/linux/kbd/scancodes-1.html
+ *	http://www.computer-engineering.org/ps2keyboard/scancodes3.html
+ *	https://www.win.tue.nl/~aeb/linux/kbd/scancodes-10.html#scancodesets
+ *	http://nixdoc.net/man-pages/irix/man7/pckeyboard.7.html
  */
 
 #include <stdio.h>
@@ -47,14 +49,17 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "kbdreg.h"
+#include "thirdparty/kbdreg.h"
 
 
-/*  #define PCKBC_DEBUG  */
-/*  #define debug fatal  */
+// #define PCKBC_DEBUG
+// #define debug fatal
 
 
-#define	MAX_8042_QUEUELEN	256
+// Should be 256 to emulate a real 8042? Having a larger value allows
+// pasting of larger text chunks which is useful in the emulator.
+// #define	MAX_8042_QUEUELEN	256
+#define	MAX_8042_QUEUELEN	32768
 
 #define	PC_DATA			0
 #define	PC_CMD			0
@@ -63,12 +68,17 @@
 #define	PS2_TXBUF		0
 #define	PS2_RXBUF		1
 #define	PS2_CONTROL		2
+#define         PS2_CONTROL_TXINTEN   (1 << 2)
+#define         PS2_CONTROL_RXINTEN   (1 << 3)
 #define	PS2_STATUS		3
+#define         PS2_STATUS_TXEMPTY    (1 << 3)
+#define         PS2_STATUS_RXFULL     (1 << 4)
 
-#define	PS2	100
+#define	PS2	100	// internal offset
 
 #define	PCKBC_TICKSHIFT		15
 
+
 struct pckbc_data {
 	int		console_handle;
 	int		in_use;
@@ -83,18 +93,20 @@ struct pckbc_data {
 
 	/*  TODO: one of these for each port?  */
 	int		clocksignal;
-	int		rx_int_enable;
-	int		tx_int_enable;
+	int		rx_int_enable[2];
+	int		tx_int_enable[2];
 
-	int		keyscanning_enabled;
+	int		scanning_enabled[2];
 	int		translation_table;
-	int		state;
+	int		state[2];
 	int		cmdbyte;
 	int		output_byte;
 	int		last_scancode;
 
 	unsigned	key_queue[2][MAX_8042_QUEUELEN];
 	int		head[2], tail[2];
+
+	int		mouse_buttons;
 };
 
 #define	STATE_NORMAL			0
@@ -136,167 +148,22 @@ int pckbc_get_code(struct pckbc_data *d, int port)
 		fatal("[ pckbc: queue empty, port %i! ]\n", port);
 	else
 		d->tail[port] = (d->tail[port]+1) % MAX_8042_QUEUELEN;
+
 	return d->key_queue[port][d->tail[port]];
 }
 
 
 /*
- *  ascii_to_scancodes_type3():
+ *  ascii_to_scancodes_type1():
  *
  *  Conversion from ASCII codes to default (US) keyboard scancodes.
- *  (See http://www.computer-engineering.org/ps2keyboard/scancodes3.html)
  */
-static void ascii_to_pc_scancodes_type3(int a, struct pckbc_data *d)
+static void ascii_to_pc_scancodes_type1(int a, struct pckbc_data *d)
 {
 	int old_head;
 	int p = 0;	/*  port  */
 	int shift = 0, ctrl = 0;
 
-	if (a >= 'A' && a <= 'Z') { a += 32; shift = 1; }
-	if ((a >= 1 && a <= 26) && (a!='\n' && a!='\t' && a!='\b' && a!='\r'))
-		{ a += 96; ctrl = 1; }
-	if (a=='!')  {	a = '1'; shift = 1; }
-	if (a=='@')  {	a = '2'; shift = 1; }
-	if (a=='#')  {	a = '3'; shift = 1; }
-	if (a=='$')  {	a = '4'; shift = 1; }
-	if (a=='%')  {	a = '5'; shift = 1; }
-	if (a=='^')  {	a = '6'; shift = 1; }
-	if (a=='&')  {	a = '7'; shift = 1; }
-	if (a=='*')  {	a = '8'; shift = 1; }
-	if (a=='(')  {	a = '9'; shift = 1; }
-	if (a==')')  {	a = '0'; shift = 1; }
-	if (a=='_')  {	a = '-'; shift = 1; }
-	if (a=='+')  {	a = '='; shift = 1; }
-	if (a=='{')  {	a = '['; shift = 1; }
-	if (a=='}')  {	a = ']'; shift = 1; }
-	if (a==':')  {	a = ';'; shift = 1; }
-	if (a=='"')  {	a = '\''; shift = 1; }
-	if (a=='|')  {	a = '\\'; shift = 1; }
-	if (a=='<')  {	a = ','; shift = 1; }
-	if (a=='>')  {	a = '.'; shift = 1; }
-	if (a=='?')  {	a = '/'; shift = 1; }
-
-	if (shift)
-		pckbc_add_code(d, 0x12, p);
-	if (ctrl)
-		pckbc_add_code(d, 0x11, p);
-
-	/*
-	 *  Note: The ugly hack used to add release codes for all of these
-	 *  keys is as follows:  we remember how much of the kbd buf that
-	 *  is in use here, before we add any scancode. After we've added
-	 *  one or more scancodes (ie an optional shift + another key)
-	 *  then we add 0xf0 + the last scancode _if_ the kbd buf was altered.
-	 */
-
-	old_head = d->head[p];
-
-	if (a==27)	pckbc_add_code(d, 0x08, p);
-
-	if (a=='1')	pckbc_add_code(d, 0x16, p);
-	if (a=='2')	pckbc_add_code(d, 0x1e, p);
-	if (a=='3')	pckbc_add_code(d, 0x26, p);
-	if (a=='4')	pckbc_add_code(d, 0x25, p);
-	if (a=='5')	pckbc_add_code(d, 0x2e, p);
-	if (a=='6')	pckbc_add_code(d, 0x36, p);
-	if (a=='7')	pckbc_add_code(d, 0x3d, p);
-	if (a=='8')	pckbc_add_code(d, 0x3e, p);
-	if (a=='9')	pckbc_add_code(d, 0x46, p);
-	if (a=='0')	pckbc_add_code(d, 0x45, p);
-	if (a=='-')	pckbc_add_code(d, 0x4e, p);
-	if (a=='=')	pckbc_add_code(d, 0x55, p);
-
-	if (a=='\b')	pckbc_add_code(d, 0x29, p);
-
-	if (a=='\t')	pckbc_add_code(d, 0x0d, p);
-	if (a=='q')	pckbc_add_code(d, 0x15, p);
-	if (a=='w')	pckbc_add_code(d, 0x1d, p);
-	if (a=='e')	pckbc_add_code(d, 0x24, p);
-	if (a=='r')	pckbc_add_code(d, 0x2d, p);
-	if (a=='t')	pckbc_add_code(d, 0x2c, p);
-	if (a=='y')	pckbc_add_code(d, 0x35, p);
-	if (a=='u')	pckbc_add_code(d, 0x3c, p);
-	if (a=='i')	pckbc_add_code(d, 0x43, p);
-	if (a=='o')	pckbc_add_code(d, 0x44, p);
-	if (a=='p')	pckbc_add_code(d, 0x4d, p);
-
-	if (a=='[')	pckbc_add_code(d, 0x54, p);
-	if (a==']')	pckbc_add_code(d, 0x5b, p);
-
-	if (a=='\n' || a=='\r')	pckbc_add_code(d, 0x5a, p);
-
-	if (a=='a')	pckbc_add_code(d, 0x1c, p);
-	if (a=='s')	pckbc_add_code(d, 0x1b, p);
-	if (a=='d')	pckbc_add_code(d, 0x23, p);
-	if (a=='f')	pckbc_add_code(d, 0x2b, p);
-	if (a=='g')	pckbc_add_code(d, 0x34, p);
-	if (a=='h')	pckbc_add_code(d, 0x33, p);
-	if (a=='j')	pckbc_add_code(d, 0x3b, p);
-	if (a=='k')	pckbc_add_code(d, 0x42, p);
-	if (a=='l')	pckbc_add_code(d, 0x4b, p);
-
-	if (a==';')	pckbc_add_code(d, 0x4c, p);
-	if (a=='\'')	pckbc_add_code(d, 0x52, p);
-/*	if (a=='~')	pckbc_add_code(d, 0x29, p);  ?  */
-	if (a=='\\')	pckbc_add_code(d, 0x5c, p);
-
-	if (a=='z')	pckbc_add_code(d, 0x1a, p);
-	if (a=='x')	pckbc_add_code(d, 0x22, p);
-	if (a=='c')	pckbc_add_code(d, 0x21, p);
-	if (a=='v')	pckbc_add_code(d, 0x2a, p);
-	if (a=='b')	pckbc_add_code(d, 0x32, p);
-	if (a=='n')	pckbc_add_code(d, 0x31, p);
-	if (a=='m')	pckbc_add_code(d, 0x3a, p);
-
-	if (a==',')	pckbc_add_code(d, 0x41, p);
-	if (a=='.')	pckbc_add_code(d, 0x49, p);
-	if (a=='/')	pckbc_add_code(d, 0x4a, p);
-
-	if (a==' ')	pckbc_add_code(d, 0x29, p);
-
-	/*  Add release code, if a key was pressed:  */
-	if (d->head[p] != old_head) {
-		int code = d->key_queue[p][d->head[p]];
-		pckbc_add_code(d, 0xf0, p);
-		pckbc_add_code(d, code, p);
-	}
-
-	/*  Release shift and ctrl:  */
-	if (shift) {
-		pckbc_add_code(d, 0xf0, p);
-		pckbc_add_code(d, 0x12, p);
-	}
-	if (ctrl) {
-		pckbc_add_code(d, 0xf0, p);
-		pckbc_add_code(d, 0x11, p);
-	}
-}
-
-
-/*
- *  ascii_to_scancodes_type2():
- *
- *  Conversion from ASCII codes to default (US) keyboard scancodes.
- *  (See http://www.win.tue.nl/~aeb/linux/kbd/scancodes-1.html)
- *
- *  NOTE/TODO: This seems to be type 2, not type 1.
- */
-static void ascii_to_pc_scancodes_type2(int a, struct pckbc_data *d)
-{
-	int old_head;
-	int p = 0;	/*  port  */
-	int shift = 0, ctrl = 0;
-
-	if (d->translation_table == 3) {
-		ascii_to_pc_scancodes_type3(a, d);
-		return;
-	}
-
-	if (d->translation_table != 2) {
-		fatal("[ ascii_to_pc_scancodes: unimplemented type! ]\n");
-		return;
-	}
-
 	if (a >= 'A' && a <= 'Z') { a += 32; shift = 1; }
 	if ((a >= 1 && a <= 26) && (a!='\n' && a!='\t' && a!='\b' && a!='\r'))
 		{ a += 96; ctrl = 1; }
@@ -436,48 +303,381 @@ static void ascii_to_pc_scancodes_type2(int a, struct pckbc_data *d)
 }
 
 
-DEVICE_TICK(pckbc)
+/*
+ *  ascii_to_scancodes_type2():
+ *
+ *  Conversion from ASCII codes to default (US) keyboard scancodes.
+ */
+static void ascii_to_pc_scancodes_type2(int a, struct pckbc_data *d)
 {
-	struct pckbc_data *d = extra;
-	int port_nr, ch, ints_enabled;
+	int old_head;
+	int p = 0;	/*  port  */
+	int shift = 0, ctrl = 0;
 
-	if (d->in_use && console_charavail(d->console_handle)) {
-		ch = console_readchar(d->console_handle);
-		if (ch >= 0)
-			ascii_to_pc_scancodes_type2(ch, d);
+	if (a >= 'A' && a <= 'Z') { a += 32; shift = 1; }
+	if ((a >= 1 && a <= 26) && (a!='\n' && a!='\t' && a!='\b' && a!='\r'))
+		{ a += 96; ctrl = 1; }
+	if (a=='!')  {	a = '1'; shift = 1; }
+	if (a=='@')  {	a = '2'; shift = 1; }
+	if (a=='#')  {	a = '3'; shift = 1; }
+	if (a=='$')  {	a = '4'; shift = 1; }
+	if (a=='%')  {	a = '5'; shift = 1; }
+	if (a=='^')  {	a = '6'; shift = 1; }
+	if (a=='&')  {	a = '7'; shift = 1; }
+	if (a=='*')  {	a = '8'; shift = 1; }
+	if (a=='(')  {	a = '9'; shift = 1; }
+	if (a==')')  {	a = '0'; shift = 1; }
+	if (a=='_')  {	a = '-'; shift = 1; }
+	if (a=='+')  {	a = '='; shift = 1; }
+	if (a=='{')  {	a = '['; shift = 1; }
+	if (a=='}')  {	a = ']'; shift = 1; }
+	if (a==':')  {	a = ';'; shift = 1; }
+	if (a=='"')  {	a = '\''; shift = 1; }
+	if (a=='|')  {	a = '\\'; shift = 1; }
+	if (a=='<')  {	a = ','; shift = 1; }
+	if (a=='>')  {	a = '.'; shift = 1; }
+	if (a=='?')  {	a = '/'; shift = 1; }
+	if (a=='~')  {	a = '`'; shift = 1; }
+
+	if (shift)
+		pckbc_add_code(d, 0x12, p);
+	if (ctrl)
+		pckbc_add_code(d, 0x14, p);
+
+	/*
+	 *  Note: The ugly hack used to add release codes for all of these
+	 *  keys is as follows:  we remember how much of the kbd buf that
+	 *  is in use here, before we add any scancode. After we've added
+	 *  one or more scancodes (ie an optional shift + another key)
+	 *  then we add 0xf0 + the last scancode _if_ the kbd buf was altered.
+	 */
+
+	old_head = d->head[p];
+
+	if (a==27)	pckbc_add_code(d, 0x76, p);
+
+	if (a=='1')	pckbc_add_code(d, 0x16, p);
+	if (a=='2')	pckbc_add_code(d, 0x1e, p);
+	if (a=='3')	pckbc_add_code(d, 0x26, p);
+	if (a=='4')	pckbc_add_code(d, 0x25, p);
+	if (a=='5')	pckbc_add_code(d, 0x2e, p);
+	if (a=='6')	pckbc_add_code(d, 0x36, p);
+	if (a=='7')	pckbc_add_code(d, 0x3d, p);
+	if (a=='8')	pckbc_add_code(d, 0x3e, p);
+	if (a=='9')	pckbc_add_code(d, 0x46, p);
+	if (a=='0')	pckbc_add_code(d, 0x45, p);
+	if (a=='-')	pckbc_add_code(d, 0x4e, p);
+	if (a=='=')	pckbc_add_code(d, 0x55, p);
+
+	if (a=='\b')	pckbc_add_code(d, 0x66, p);
+
+	if (a=='\t')	pckbc_add_code(d, 0x0d, p);
+	if (a=='q')	pckbc_add_code(d, 0x15, p);
+	if (a=='w')	pckbc_add_code(d, 0x1d, p);
+	if (a=='e')	pckbc_add_code(d, 0x24, p);
+	if (a=='r')	pckbc_add_code(d, 0x2d, p);
+	if (a=='t')	pckbc_add_code(d, 0x2c, p);
+	if (a=='y')	pckbc_add_code(d, 0x35, p);
+	if (a=='u')	pckbc_add_code(d, 0x3c, p);
+	if (a=='i')	pckbc_add_code(d, 0x43, p);
+	if (a=='o')	pckbc_add_code(d, 0x44, p);
+	if (a=='p')	pckbc_add_code(d, 0x4d, p);
+
+	if (a=='[')	pckbc_add_code(d, 0x54, p);
+	if (a==']')	pckbc_add_code(d, 0x5b, p);
+
+	if (a=='\n' || a=='\r')	pckbc_add_code(d, 0x5a, p);
+
+	if (a=='a')	pckbc_add_code(d, 0x1c, p);
+	if (a=='s')	pckbc_add_code(d, 0x1b, p);
+	if (a=='d')	pckbc_add_code(d, 0x23, p);
+	if (a=='f')	pckbc_add_code(d, 0x2b, p);
+	if (a=='g')	pckbc_add_code(d, 0x34, p);
+	if (a=='h')	pckbc_add_code(d, 0x33, p);
+	if (a=='j')	pckbc_add_code(d, 0x3b, p);
+	if (a=='k')	pckbc_add_code(d, 0x42, p);
+	if (a=='l')	pckbc_add_code(d, 0x4b, p);
+
+	if (a==';')	pckbc_add_code(d, 0x4c, p);
+	if (a=='\'')	pckbc_add_code(d, 0x52, p);
+	if (a=='`')	pckbc_add_code(d, 0x0e, p);	// ?
+	if (a=='\\')	pckbc_add_code(d, 0x5d, p);	// 0x5c?
+
+	if (a=='z')	pckbc_add_code(d, 0x1a, p);
+	if (a=='x')	pckbc_add_code(d, 0x22, p);
+	if (a=='c')	pckbc_add_code(d, 0x21, p);
+	if (a=='v')	pckbc_add_code(d, 0x2a, p);
+	if (a=='b')	pckbc_add_code(d, 0x32, p);
+	if (a=='n')	pckbc_add_code(d, 0x31, p);
+	if (a=='m')	pckbc_add_code(d, 0x3a, p);
+
+	if (a==',')	pckbc_add_code(d, 0x41, p);
+	if (a=='.')	pckbc_add_code(d, 0x49, p);
+	if (a=='/')	pckbc_add_code(d, 0x4a, p);
+
+	if (a==' ')	pckbc_add_code(d, 0x29, p);
+
+	/*  Add release code, if a key was pressed:  */
+	if (d->head[p] != old_head) {
+		int code = d->key_queue[p][d->head[p]];
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, code, p);
 	}
 
-	ints_enabled = d->rx_int_enable;
+	/*  Release shift and ctrl:  */
+	if (shift) {
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, 0x12, p);
+	}
+	if (ctrl) {
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, 0x14, p);
+	}
+}
 
-	/*  TODO: mouse movements?  */
 
-	if (d->cmdbyte & KC8_KDISABLE)
-		ints_enabled = 0;
+/*
+ *  ascii_to_scancodes_type3():
+ *
+ *  Conversion from ASCII codes to default (US) keyboard scancodes.
+ */
+static void ascii_to_pc_scancodes_type3(int a, struct pckbc_data *d)
+{
+	int old_head;
+	int p = 0;	/*  port  */
+	int shift = 0, ctrl = 0;
 
-	for (port_nr=0; port_nr<2; port_nr++) {
-		/*  Cause receive interrupt, if there's something in the
-		    receive buffer: (Otherwise deassert the interrupt.)  */
+	if (a >= 'A' && a <= 'Z') { a += 32; shift = 1; }
+	if ((a >= 1 && a <= 26) && (a!='\n' && a!='\t' && a!='\b' && a!='\r'))
+		{ a += 96; ctrl = 1; }
+	if (a=='!')  {	a = '1'; shift = 1; }
+	if (a=='@')  {	a = '2'; shift = 1; }
+	if (a=='#')  {	a = '3'; shift = 1; }
+	if (a=='$')  {	a = '4'; shift = 1; }
+	if (a=='%')  {	a = '5'; shift = 1; }
+	if (a=='^')  {	a = '6'; shift = 1; }
+	if (a=='&')  {	a = '7'; shift = 1; }
+	if (a=='*')  {	a = '8'; shift = 1; }
+	if (a=='(')  {	a = '9'; shift = 1; }
+	if (a==')')  {	a = '0'; shift = 1; }
+	if (a=='_')  {	a = '-'; shift = 1; }
+	if (a=='+')  {	a = '='; shift = 1; }
+	if (a=='{')  {	a = '['; shift = 1; }
+	if (a=='}')  {	a = ']'; shift = 1; }
+	if (a==':')  {	a = ';'; shift = 1; }
+	if (a=='"')  {	a = '\''; shift = 1; }
+	if (a=='|')  {	a = '\\'; shift = 1; }
+	if (a=='<')  {	a = ','; shift = 1; }
+	if (a=='>')  {	a = '.'; shift = 1; }
+	if (a=='?')  {	a = '/'; shift = 1; }
+	if (a=='~')  {	a = '`'; shift = 1; }
+
+	if (shift)
+		pckbc_add_code(d, 0x12, p);
+	if (ctrl)
+		pckbc_add_code(d, 0x11, p);
+
+	/*
+	 *  Note: The ugly hack used to add release codes for all of these
+	 *  keys is as follows:  we remember how much of the kbd buf that
+	 *  is in use here, before we add any scancode. After we've added
+	 *  one or more scancodes (ie an optional shift + another key)
+	 *  then we add 0xf0 + the last scancode _if_ the kbd buf was altered.
+	 */
+
+	old_head = d->head[p];
+
+	if (a==27)	pckbc_add_code(d, 0x08, p);
+
+	if (a=='1')	pckbc_add_code(d, 0x16, p);
+	if (a=='2')	pckbc_add_code(d, 0x1e, p);
+	if (a=='3')	pckbc_add_code(d, 0x26, p);
+	if (a=='4')	pckbc_add_code(d, 0x25, p);
+	if (a=='5')	pckbc_add_code(d, 0x2e, p);
+	if (a=='6')	pckbc_add_code(d, 0x36, p);
+	if (a=='7')	pckbc_add_code(d, 0x3d, p);
+	if (a=='8')	pckbc_add_code(d, 0x3e, p);
+	if (a=='9')	pckbc_add_code(d, 0x46, p);
+	if (a=='0')	pckbc_add_code(d, 0x45, p);
+	if (a=='-')	pckbc_add_code(d, 0x4e, p);
+	if (a=='=')	pckbc_add_code(d, 0x55, p);
+
+	if (a=='\b')	pckbc_add_code(d, 0x66, p);
+
+	if (a=='\t')	pckbc_add_code(d, 0x0d, p);
+	if (a=='q')	pckbc_add_code(d, 0x15, p);
+	if (a=='w')	pckbc_add_code(d, 0x1d, p);
+	if (a=='e')	pckbc_add_code(d, 0x24, p);
+	if (a=='r')	pckbc_add_code(d, 0x2d, p);
+	if (a=='t')	pckbc_add_code(d, 0x2c, p);
+	if (a=='y')	pckbc_add_code(d, 0x35, p);
+	if (a=='u')	pckbc_add_code(d, 0x3c, p);
+	if (a=='i')	pckbc_add_code(d, 0x43, p);
+	if (a=='o')	pckbc_add_code(d, 0x44, p);
+	if (a=='p')	pckbc_add_code(d, 0x4d, p);
+
+	if (a=='[')	pckbc_add_code(d, 0x54, p);
+	if (a==']')	pckbc_add_code(d, 0x5b, p);
+
+	if (a=='\n' || a=='\r')	pckbc_add_code(d, 0x5a, p);
+
+	if (a=='a')	pckbc_add_code(d, 0x1c, p);
+	if (a=='s')	pckbc_add_code(d, 0x1b, p);
+	if (a=='d')	pckbc_add_code(d, 0x23, p);
+	if (a=='f')	pckbc_add_code(d, 0x2b, p);
+	if (a=='g')	pckbc_add_code(d, 0x34, p);
+	if (a=='h')	pckbc_add_code(d, 0x33, p);
+	if (a=='j')	pckbc_add_code(d, 0x3b, p);
+	if (a=='k')	pckbc_add_code(d, 0x42, p);
+	if (a=='l')	pckbc_add_code(d, 0x4b, p);
+
+	if (a==';')	pckbc_add_code(d, 0x4c, p);
+	if (a=='\'')	pckbc_add_code(d, 0x52, p);
+	if (a=='`')	pckbc_add_code(d, 0x0e, p);	// ?
+	if (a=='\\')	pckbc_add_code(d, 0x5d, p);	// 0x5c?
+
+	if (a=='z')	pckbc_add_code(d, 0x1a, p);
+	if (a=='x')	pckbc_add_code(d, 0x22, p);
+	if (a=='c')	pckbc_add_code(d, 0x21, p);
+	if (a=='v')	pckbc_add_code(d, 0x2a, p);
+	if (a=='b')	pckbc_add_code(d, 0x32, p);
+	if (a=='n')	pckbc_add_code(d, 0x31, p);
+	if (a=='m')	pckbc_add_code(d, 0x3a, p);
+
+	if (a==',')	pckbc_add_code(d, 0x41, p);
+	if (a=='.')	pckbc_add_code(d, 0x49, p);
+	if (a=='/')	pckbc_add_code(d, 0x4a, p);
+
+	if (a==' ')	pckbc_add_code(d, 0x29, p);
+
+	/*  Add release code, if a key was pressed:  */
+	if (d->head[p] != old_head) {
+		int code = d->key_queue[p][d->head[p]];
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, code, p);
+	}
+
+	/*  Release shift and ctrl:  */
+	if (shift) {
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, 0x12, p);
+	}
+	if (ctrl) {
+		pckbc_add_code(d, 0xf0, p);
+		pckbc_add_code(d, 0x11, p);
+	}
+}
+
+
+void pckbc_reassert_interrupts(struct pckbc_data *d)
+{
+	int ints_enabled = 1;
+
+	// if (d->cmdbyte & KC8_KDISABLE)
+	//	ints_enabled = 0;
+
+	for (int port_nr = 0; port_nr < 2; port_nr++) {
+		/*
+		 *  Cause receive interrupt, if there's something in the
+		 *  receive buffer: (Otherwise deassert the interrupt.)
+		 */
+		if (d->head[port_nr] != d->tail[port_nr] && ints_enabled && d->rx_int_enable[port_nr]) {
+			if (!d->currently_asserted[port_nr]) {
+				// fatal("[ pckbc: interrupt port %i ]\n", port_nr);
+				if (port_nr == 0)
+					INTERRUPT_ASSERT(d->irq_keyboard);
+				else
+					INTERRUPT_ASSERT(d->irq_mouse);
+			}
 
-		if (d->head[port_nr] != d->tail[port_nr] && ints_enabled) {
-			debug("[ pckbc: interrupt port %i ]\n", port_nr);
-			if (port_nr == 0)
-				INTERRUPT_ASSERT(d->irq_keyboard);
-			else
-				INTERRUPT_ASSERT(d->irq_mouse);
 			d->currently_asserted[port_nr] = 1;
 		} else {
 			if (d->currently_asserted[port_nr]) {
+				// fatal("[ pckbc: DEASSERT interrupt port %i ]\n", port_nr);
 				if (port_nr == 0)
 					INTERRUPT_DEASSERT(d->irq_keyboard);
 				else
 					INTERRUPT_DEASSERT(d->irq_mouse);
 			}
+
 			d->currently_asserted[port_nr] = 0;
 		}
 	}
 }
 
 
+DEVICE_TICK(pckbc)
+{
+	struct pckbc_data *d = (struct pckbc_data *) extra;
+	int ch;
+
+	// Keyboard input:
+	if (d->in_use) {
+		while (console_charavail(d->console_handle)) {
+			ch = console_readchar(d->console_handle);
+			if (ch >= 0) {
+				switch (d->translation_table) {
+				case 1:	ascii_to_pc_scancodes_type1(ch, d);
+					break;
+				case 2:	ascii_to_pc_scancodes_type2(ch, d);
+					break;
+				case 3:	ascii_to_pc_scancodes_type3(ch, d);
+					break;
+				default:debugmsg(SUBSYS_DEVICE, "pckbc", VERBOSITY_WARNING,
+					    "unimplemented translation table type %i",
+					    d->translation_table);
+				}
+			}
+		}
+	}
+
+	// Mouse input:
+	/*  Don't do mouse updates if we're running in serial console mode:  */
+	if (cpu->machine->x11_md.in_use && d->state[1] == STATE_NORMAL && d->scanning_enabled[1]) {
+		int xdelta, ydelta, mouse_buttons, mouse_fb_nr;
+		console_getmouse(&xdelta, &ydelta, &mouse_buttons, &mouse_fb_nr);
+
+		ydelta = 0 - ydelta;
+
+		const int m = 100;
+
+		if (xdelta > m)
+			xdelta = m;
+		if (xdelta < -m)
+			xdelta = -m;
+		if (ydelta > m)
+			ydelta = m;
+		if (ydelta < -m)
+			ydelta = -m;
+
+		/*  Only send update if there is an actual diff.  */
+		if (xdelta != 0 || ydelta != 0 || d->mouse_buttons != mouse_buttons) {
+			d->mouse_buttons = mouse_buttons;
+
+			// See "The default protocol" at
+			// https://www.win.tue.nl/~aeb/linux/kbd/scancodes-13.html
+			uint8_t b1 = 0x08;
+			uint8_t b2 = xdelta;
+			uint8_t b3 = ydelta;
+
+			b1 |= (mouse_buttons & 4) >> 2;	// Left
+			b1 |= (mouse_buttons & 2) << 1;	// Middle
+			b1 |= (mouse_buttons & 1) << 1;	// Right
+			b1 |= ((xdelta >> 8) & 1) << 4;
+			b1 |= ((ydelta >> 8) & 1) << 5;
+
+			// printf("x=%i y=%i b1=%02x %02x %02x\n", xdelta, ydelta, b1,b2,b3);
+
+			pckbc_add_code(d, b1, 1);
+			pckbc_add_code(d, b2, 1);
+			pckbc_add_code(d, b3, 1);
+		}
+	}
+
+	pckbc_reassert_interrupts(d);
+}
+
+
 /*
  *  dev_pckbc_command():
  *
@@ -485,58 +685,76 @@ DEVICE_TICK(pckbc)
  */
 static void dev_pckbc_command(struct pckbc_data *d, int port_nr)
 {
-	int cmd = d->reg[PC_CMD];
+	int cmd = d->type == PCKBC_8242 ? d->reg[PS2_TXBUF] : d->reg[PC_CMD];
 
-	if (d->type == PCKBC_8242)
-		cmd = d->reg[PS2_TXBUF];
+#if 0
+	// Mouse port debugging:
+	if (port_nr == 1)
+		fatal("[ pckbc: (port %i) command 0x%02x ]\n", port_nr, cmd);
+#endif
 
-	if (d->state == STATE_WAITING_FOR_TRANSLTABLE) {
+	if (d->state[port_nr] == STATE_WAITING_FOR_TRANSLTABLE) {
 		debug("[ pckbc: (port %i) switching to translation table "
 		    "0x%02x ]\n", port_nr, cmd);
+
 		switch (cmd) {
-		case 2:
+		case 0:	// TODO. "Writing 0xf0 followed by 0 queries the mode,
+			//  resulting in a scancode byte 43, 41 or 3f from the
+			//  keyboard." according to
+			// https://www.win.tue.nl/~aeb/linux/kbd/scancodes-10.html#scancodesets
+			fatal("[ pckbc: TODO: return current translation table ]\n");
+			break;
+		case 1:	// Fall-through.
+		case 2:	// Fall-through.
 		case 3:	d->translation_table = cmd;
 			break;
 		default:fatal("[ pckbc: (port %i) translation table "
 			    "0x%02x is NOT YET IMPLEMENTED ]\n",
 			    port_nr, cmd);
 		}
+
+		// Hack for OpenBSD/cats and NetBSD/cats:
+		if (d->type != PCKBC_8242) {
+			fatal("[ pckbc: TODO: hack for non-8242. figure out how to deal with this ]\n");
+			d->translation_table = 1;
+		}
+
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_NORMAL;
+		d->state[port_nr] = STATE_NORMAL;
 		return;
 	}
 
-	if (d->state == STATE_WAITING_FOR_RATE) {
-		debug("[ pckbc: (port %i) received Typematic Rate data: "
+	if (d->state[port_nr] == STATE_WAITING_FOR_RATE) {
+		debug("[ pckbc: (port %i) received Typematic/Sample Rate data: "
 		    "0x%02x ]\n", port_nr, cmd);
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_NORMAL;
+		d->state[port_nr] = STATE_NORMAL;
 		return;
 	}
 
-	if (d->state == STATE_WAITING_FOR_ONEKEY_MB) {
+	if (d->state[port_nr] == STATE_WAITING_FOR_ONEKEY_MB) {
 		debug("[ pckbc: (port %i) received One-key make/break data: "
 		    "0x%02x ]\n", port_nr, cmd);
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_NORMAL;
+		d->state[port_nr] = STATE_NORMAL;
 		return;
 	}
 
-	if (d->state == STATE_WAITING_FOR_AUX) {
+	if (d->state[port_nr] == STATE_WAITING_FOR_AUX) {
 		debug("[ pckbc: (port %i) received aux data: "
 		    "0x%02x ]\n", port_nr, cmd);
 		/*  Echo back.  */
 		pckbc_add_code(d, cmd, port_nr);
-		d->state = STATE_NORMAL;
+		d->state[port_nr] = STATE_NORMAL;
 		return;
 	}
 
-	if (d->state == STATE_WAITING_FOR_AUX_OUT) {
+	if (d->state[port_nr] == STATE_WAITING_FOR_AUX_OUT) {
 		debug("[ pckbc: (port %i) received aux out data: "
 		    "0x%02x ]\n", port_nr, cmd);
 		/*  Echo back.  */
 		pckbc_add_code(d, cmd, port_nr);
-		d->state = STATE_NORMAL;
+		d->state[port_nr] = STATE_NORMAL;
 		return;
 	}
 
@@ -555,18 +773,22 @@ static void dev_pckbc_command(struct pckbc_data *d, int port_nr)
 		pckbc_add_code(d, KBR_ACK, port_nr);
 		break;
 
+	case KBC_ECHO:
+		pckbc_add_code(d, KBR_ECHO, port_nr);
+		break;
+
 	case KBC_SETTABLE:
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_WAITING_FOR_TRANSLTABLE;
+		d->state[port_nr] = STATE_WAITING_FOR_TRANSLTABLE;
 		break;
 
 	case KBC_ENABLE:
-		d->keyscanning_enabled = 1;
+		d->scanning_enabled[port_nr] = 1;
 		pckbc_add_code(d, KBR_ACK, port_nr);
 		break;
 
 	case KBC_DISABLE:
-		d->keyscanning_enabled = 0;
+		d->scanning_enabled[port_nr] = 0;
 		pckbc_add_code(d, KBR_ACK, port_nr);
 		break;
 
@@ -575,16 +797,27 @@ static void dev_pckbc_command(struct pckbc_data *d, int port_nr)
 		break;
 
 	case KBC_GETID:
-		/*  Get keyboard ID.  NOTE/TODO: Ugly hardcoded answer.  */
+		/*  Get keyboard/mouse ID.  NOTE/TODO: Ugly hardcoded answer.  */
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		pckbc_add_code(d, 0xab, port_nr);
-		pckbc_add_code(d, 0x41, port_nr);
+		
+		if (port_nr == 0) {
+			// Keyboard:
+			pckbc_add_code(d, 0xab, port_nr);
+			pckbc_add_code(d, 0x41, port_nr);
+		} else {
+			// Mouse:
+			pckbc_add_code(d, 0x00, port_nr);
+			pckbc_add_code(d, 0x00, port_nr);
+		}
 		break;
 
 	case KBC_TYPEMATIC:
-		/*  Set typematic (auto-repeat) delay/speed:  */
+		/*
+		 *  Keyboard: Set typematic (auto-repeat) delay/speed.
+		 *  PS/2 mouse: Set sample ratre.
+		 */
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_WAITING_FOR_RATE;
+		d->state[port_nr] = STATE_WAITING_FOR_RATE;
 		break;
 
 	case KBC_ALLKEYS_TMB:
@@ -595,30 +828,38 @@ static void dev_pckbc_command(struct pckbc_data *d, int port_nr)
 	case KBC_ONEKEY_MB:
 		/*  "Make one key typematic/make/break"  */
 		pckbc_add_code(d, KBR_ACK, port_nr);
-		d->state = STATE_WAITING_FOR_ONEKEY_MB;
+		d->state[port_nr] = STATE_WAITING_FOR_ONEKEY_MB;
 		break;
 
 	case KBC_RESET:
 		pckbc_add_code(d, KBR_ACK, port_nr);
 		pckbc_add_code(d, KBR_RSTDONE, port_nr);
+
+		if (port_nr == 1) {
+			// Mouse (based on NetBSD's pckbport/pms.c) sends
+			// mouse ID after RSTDONE:
+			pckbc_add_code(d, 0x00, port_nr);
+		}
+
 		/*
 		 *  Disable interrupts during reset, or Linux 2.6
 		 *  prints warnings about spurious interrupts.
 		 */
-		d->rx_int_enable = 0;
+		// d->rx_int_enable = 0;
+		d->scanning_enabled[port_nr] = 0;
 		break;
 
 	default:
-		fatal("[ pckbc: (port %i) UNIMPLEMENTED 8048 command"
-		    " 0x%02x ]\n", port_nr, cmd);
-		exit(1);
+		pckbc_add_code(d, KBR_RESEND, port_nr); // Error
+		debugmsg(SUBSYS_DEVICE, "pckbc", VERBOSITY_WARNING,
+		    "UNIMPLEMENTED command 0x%02x (port %i)", cmd, port_nr);
 	}
 }
 
 
 DEVICE_ACCESS(pckbc)
 {
-	struct pckbc_data *d = extra;
+	struct pckbc_data *d = (struct pckbc_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int port_nr = 0;
 	size_t i;
@@ -679,14 +920,14 @@ if (x&1)
 
 	case 0:		/*  data  */
 		if (writeflag==MEM_READ) {
-			switch (d->state) {
+			switch (d->state[port_nr]) {
 			case STATE_RDCMDBYTE:
 				odata = d->cmdbyte;
-				d->state = STATE_NORMAL;
+				d->state[port_nr] = STATE_NORMAL;
 				break;
 			case STATE_RDOUTPUT:
 				odata = d->output_byte;
-				d->state = STATE_NORMAL;
+				d->state[port_nr] = STATE_NORMAL;
 				break;
 			default:if (d->head[0] != d->tail[0]) {
 					odata = pckbc_get_code(d, 0);
@@ -699,21 +940,23 @@ if (x&1)
 			/*  debug("[ pckbc: read from DATA: 0x%02x ]\n",
 			    (int)odata);  */
 		} else {
-			debug("[ pckbc: write to DATA:");
-			for (i=0; i<len; i++)
-				debug(" %02x", data[i]);
-			debug(" ]\n");
-
-			switch (d->state) {
+			if (port_nr == 1) {
+				fatal("[ pckbc: write to DATA:");
+				for (i=0; i<len; i++)
+					fatal(" %02x", data[i]);
+				fatal(" ]\n");
+			}
+			
+			switch (d->state[port_nr]) {
 			case STATE_LDCMDBYTE:
 				d->cmdbyte = idata;
-				d->rx_int_enable = d->cmdbyte &
+				d->rx_int_enable[port_nr] = d->cmdbyte &
 				    (KC8_KENABLE | KC8_MENABLE) ? 1 : 0;
-				d->state = STATE_NORMAL;
+				d->state[port_nr] = STATE_NORMAL;
 				break;
 			case STATE_LDOUTPUT:
 				d->output_byte = idata;
-				d->state = STATE_NORMAL;
+				d->state[port_nr] = STATE_NORMAL;
 				break;
 			default:d->reg[relative_addr] = idata;
 				dev_pckbc_command(d, port_nr);
@@ -722,17 +965,15 @@ if (x&1)
 		break;
 	case 1:		/*  control  */
 		if (writeflag==MEM_READ) {
-			dev_pckbc_tick(cpu, d);
-
 			odata = 0;
 
 			/*  "Data in buffer" bit  */
 			if (d->head[0] != d->tail[0] ||
-			    d->state == STATE_RDCMDBYTE ||
-			    d->state == STATE_RDOUTPUT)
+			    d->state[port_nr] == STATE_RDCMDBYTE ||
+			    d->state[port_nr] == STATE_RDOUTPUT)
 				odata |= KBS_DIB;
 
-			if (d->state == STATE_RDCMDBYTE)
+			if (d->state[port_nr] == STATE_RDCMDBYTE)
 				odata |= KBS_OCMD;
 
 			odata |= KBS_NOSEC;
@@ -752,10 +993,10 @@ if (x&1)
 				    these. NetBSD sends these.  */
 				break;
 			case K_RDCMDBYTE:
-				d->state = STATE_RDCMDBYTE;
+				d->state[port_nr] = STATE_RDCMDBYTE;
 				break;
 			case K_LDCMDBYTE:
-				d->state = STATE_LDCMDBYTE;
+				d->state[port_nr] = STATE_LDCMDBYTE;
 				break;
 			case 0xa7:
 				d->cmdbyte |= KC8_MDISABLE;
@@ -779,24 +1020,24 @@ if (x&1)
 				d->cmdbyte &= ~KC8_KDISABLE;
 				break;
 			case 0xd0:
-				d->state = STATE_RDOUTPUT;
+				d->state[port_nr] = STATE_RDOUTPUT;
 				break;
 			case 0xd1:
-				d->state = STATE_LDOUTPUT;
+				d->state[port_nr] = STATE_LDOUTPUT;
 				break;
 			case 0xd3:	/*  write to auxiliary device
 					    output buffer  */
 				debug("[ pckbc: CONTROL 0xd3, TODO ]\n");
-				d->state = STATE_WAITING_FOR_AUX_OUT;
+				d->state[port_nr] = STATE_WAITING_FOR_AUX_OUT;
 				break;
 			case 0xd4:	/*  write to auxiliary port  */
 				debug("[ pckbc: CONTROL 0xd4, TODO ]\n");
-				d->state = STATE_WAITING_FOR_AUX;
+				d->state[port_nr] = STATE_WAITING_FOR_AUX;
 				break;
 			default:
 				fatal("[ pckbc: unknown CONTROL 0x%x ]\n",
 				    (int)idata);
-				d->state = STATE_NORMAL;
+				d->state[port_nr] = STATE_NORMAL;
 			}
 		}
 		break;
@@ -824,7 +1065,7 @@ if (x&1)
 		if (writeflag==MEM_READ) {
 			/*  TODO: What should be returned if no data 
 			    is available?  */
-			odata = random() & 0xff;
+			odata = 0;
 			if (d->head[port_nr] != d->tail[port_nr])
 				odata = pckbc_get_code(d, port_nr);
 			debug("[ pckbc: read from port %i, PS2_RXBUF: "
@@ -843,23 +1084,30 @@ if (x&1)
 			debug("[ pckbc: write to port %i, PS2_CONTROL:"
 			    " 0x%llx ]\n", port_nr, (long long)idata);
 			d->clocksignal = (idata & 0x10) ? 1 : 0;
-			d->rx_int_enable = (idata & 0x08) ? 1 : 0;
-			d->tx_int_enable = (idata & 0x04) ? 1 : 0;
+			d->rx_int_enable[port_nr] = (idata & PS2_CONTROL_RXINTEN) ? 1 : 0;
+			d->tx_int_enable[port_nr] = (idata & PS2_CONTROL_TXINTEN) ? 1 : 0;
+
+			// HACK/TODO: NetBSD/sgimips' X11 seems to work when
+			// interrupts are enabled, but it seems to not turn
+			// them on. Having this here makes it work, but it not
+			// the correct solution.
+			if (port_nr == 1) {
+				d->rx_int_enable[port_nr] = 1;
+			}
 		}
 		break;
 
 	case PS2 + PS2_STATUS:
 		if (writeflag==MEM_READ) {
-			/* 0x08 = transmit buffer empty  */
-			odata = d->clocksignal + 0x08;
+			odata = d->clocksignal + PS2_STATUS_TXEMPTY;
 
 			if (d->head[port_nr] != d->tail[port_nr]) {
-				/*  0x10 = receicer data available (?)  */
-				odata |= 0x10;
+				/*  receiced data available  */
+				odata |= PS2_STATUS_RXFULL;
 			}
 
-			debug("[ pckbc: read from port %i, PS2_STATUS: "
-			    "0x%llx ]\n", port_nr, (long long)odata);
+			//debug("[ pckbc: read from port %i, PS2_STATUS: "
+			//    "0x%llx ]\n", port_nr, (long long)odata);
 		} else {
 			debug("[ pckbc: write to port %i, PS2_STATUS: "
 			    "0x%llx ]\n", port_nr, (long long)idata);
@@ -882,15 +1130,16 @@ if (x&1)
 	}
 
 	/*  SGI? TODO: fix  */
+#if 0
 	if (len == 8)
 		odata |= (odata << 8) | (odata << 16) | (odata << 24) |
 		    (odata << 32) | (odata << 40) | (odata << 48) |
 		    (odata << 56);
-
+#endif
 	if (writeflag == MEM_READ)
 		memory_writemax64(cpu, data, len, odata);
 
-	dev_pckbc_tick(cpu, d);
+	pckbc_reassert_interrupts(d);
 
 	return 1;
 }
@@ -908,12 +1157,16 @@ int dev_pckbc_init(struct machine *machine, struct memory *mem,
 	struct pckbc_data *d;
 	int len = DEV_PCKBC_LENGTH;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pckbc_data)));
+	CHECK_ALLOCATION(d = (struct pckbc_data *) malloc(sizeof(struct pckbc_data)));
 	memset(d, 0, sizeof(struct pckbc_data));
 
-	if (type == PCKBC_8242)
-		len = 0x18;
+	d->translation_table = 2;
 
+	if (type == PCKBC_8242) {
+		len = 0x40;
+		d->translation_table = 3;
+	}
+	
 	if (type == PCKBC_JAZZ) {
 		type = PCKBC_8042;
 		len = DEV_PCKBC_LENGTH + 0x60;
@@ -925,10 +1178,14 @@ int dev_pckbc_init(struct machine *machine, struct memory *mem,
 	d->type              = type;
 	d->in_use            = in_use;
 	d->pc_style_flag     = pc_style_flag;
-	d->translation_table = 2;
-	d->rx_int_enable     = 1;
+	d->rx_int_enable[0]  = 1;
+	d->rx_int_enable[1]  = 0;
 	d->output_byte       = 0x02;	/*  A20 enable on PCs  */
 
+	// Default is ENABLE keyboard but DISABLE mouse port:
+	d->scanning_enabled[0] = 1;
+	d->scanning_enabled[1] = 0;
+
 	d->console_handle = console_start_slave_inputonly(
 	    machine, "pckbc", d->in_use);
 
