@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -34,6 +34,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "cpu.h"
 #include "devices.h"
 #include "machine.h"
 #include "memory.h"
@@ -64,7 +65,7 @@ struct pmagja_data {
 
 DEVICE_ACCESS(pmagja)
 {
-	struct pmagja_data *d = extra;
+	struct pmagja_data *d = (struct pmagja_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	size_t i, res = 1;
 
@@ -134,13 +135,14 @@ for (i=0; i<len; i++)
 		    (long long)relative_addr, (long long)idata);
 }
 #endif
+
+#if 0
 			int x, y, ofs;
 
 			ofs = (relative_addr - 0x100000) * 2;
 			y = ofs / XSIZE;
 			x = ofs - y * XSIZE;
 
-#if 0
 			if (writeflag == MEM_WRITE) {
 				int ix;
 				for (ix=0; ix<len*2; ix++) {
@@ -194,16 +196,16 @@ for (i=0; i<len; i++)
 
 
 void dev_pmagja_init(struct machine *machine, struct memory *mem,
-	uint64_t baseaddr, char *irq_path)
+	uint64_t baseaddr, const char *irq_path)
 {
 	struct pmagja_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pmagja_data)));
+	CHECK_ALLOCATION(d = (struct pmagja_data *) malloc(sizeof(struct pmagja_data)));
 	memset(d, 0, sizeof(struct pmagja_data));
 
 	INTERRUPT_CONNECT(irq_path, d->irq);
 
-	d->fb_mem = memory_new(XSIZE * YSIZE * 3, machine->arch);
+	d->fb_mem = memory_new(XSIZE * YSIZE * 3);
 	if (d->fb_mem == NULL) {
 		fprintf(stderr, "dev_pmagja_init(): out of memory (1)\n");
 		exit(1);
