@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -40,7 +40,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "pmppc.h"
+#include "thirdparty/pmppc.h"
 
 
 struct pmppc_data {
@@ -53,7 +53,7 @@ struct pmppc_data {
 
 DEVICE_ACCESS(pmppc_board)
 {
-	struct pmppc_data *d = extra;
+	struct pmppc_data *d = (struct pmppc_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -88,8 +88,6 @@ DEVICE_ACCESS(pmppc_board)
 			if (d->reset_reg == PMPPC_RESET_SEQ_STEP1 &&
 			    idata == PMPPC_RESET_SEQ_STEP2) {
 				cpu->running = 0;
-				cpu->machine->
-				    exit_without_entering_debugger = 1;
 			}
 			d->reset_reg = idata;
 		}
@@ -117,7 +115,7 @@ DEVINIT(pmppc)
 	struct pmppc_data *d;
 	struct memory *mem = devinit->machine->memory;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct pmppc_data)));
+	CHECK_ALLOCATION(d = (struct pmppc_data *) malloc(sizeof(struct pmppc_data)));
 	memset(d, 0, sizeof(struct pmppc_data));
 
 	/*
