@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -62,7 +62,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "sccreg.h"
+#include "thirdparty/sccreg.h"
 
 
 #define	SCC_TICK_SHIFT		14
@@ -75,7 +75,8 @@
 
 
 struct scc_data {
-	int		irq_nr;
+	struct interrupt	irq;
+
 	int		use_fb;
 	int		console_handle;
 
@@ -96,6 +97,26 @@ struct scc_data {
 };
 
 
+/*
+ *  Check if space is available in the receive queue.
+ */
+int dev_scc_space_available_in_queue(void *e, int portnr)
+{
+	struct scc_data *d = (struct scc_data *) e;
+
+	int entries_in_use = d->cur_rx_queue_pos_write[portnr] -
+	    d->cur_rx_queue_pos_read[portnr];
+
+	while (entries_in_use < 0)
+		entries_in_use += MAX_QUEUE_LEN;
+
+	if (entries_in_use < MAX_QUEUE_LEN - 20)
+		return 1;
+
+	return 0;
+}
+
+
 /*
  *  dev_scc_add_to_rx_queue():
  *
@@ -149,17 +170,18 @@ static unsigned char rx_nextchar(struct scc_data *d, int portnr)
 
 DEVICE_TICK(scc)
 {
-	struct scc_data *d = extra;
+	struct scc_data *d = (struct scc_data *) extra;
 	int i;
 
 	/*  Add keystrokes to the rx queue:  */
-	if (d->use_fb == 0 && d->scc_nr == 1) {
-		if (console_charavail(d->console_handle))
-			dev_scc_add_to_rx_queue(extra, console_readchar(
+	if (d->scc_nr == 1) {
+		if (d->use_fb == 0) {
+			if (console_charavail(d->console_handle))
+				dev_scc_add_to_rx_queue(extra, console_readchar(
 			    d->console_handle), 2);
+		} else if (d->use_fb == 1)
+			lk201_tick(cpu->machine, &d->lk201);
 	}
-	if (d->use_fb == 1 && d->scc_nr == 1)
-		lk201_tick(cpu->machine, &d->lk201);
 
 	for (i=0; i<N_SCC_PORTS; i++) {
 		d->scc_register_r[i * N_SCC_REGS + SCC_RR0] |= SCC_RR0_TX_EMPTY;
@@ -185,9 +207,7 @@ DEVICE_TICK(scc)
 				    & SCC_RR3_TX_IP_A ||
 				    d->scc_register_r[i * N_SCC_REGS + SCC_RR3]
 				    & SCC_RR3_TX_IP_B) {
-fatal("TODO: legacy rewrite!\n");
-abort();
-//					cpu_interrupt(cpu, d->irq_nr);
+					INTERRUPT_ASSERT(d->irq);
 				}
 			}
 
@@ -208,9 +228,7 @@ abort();
 				    & SCC_RR3_RX_IP_A ||
 				    d->scc_register_r[i * N_SCC_REGS + SCC_RR3]
 				    & SCC_RR3_RX_IP_B) {
-fatal("TODO: legacy rewrite!\n");
-abort();
-//					cpu_interrupt(cpu, d->irq_nr);
+					INTERRUPT_ASSERT(d->irq);
 				}
 			}
 
@@ -231,14 +249,9 @@ abort();
 				if (d->scc_register_r[i * N_SCC_REGS + SCC_RR3]
 				    & SCC_RR3_EXT_IP_A ||
 				    d->scc_register_r[i * N_SCC_REGS + SCC_RR3]
-				    & SCC_RR3_EXT_IP_B)
-{
-fatal("TODO: legacy rewrite!\n");
-abort();
-//					cpu_interrupt(cpu, d->irq_nr);
-/*  TODO: huh?  */
-//cpu_interrupt(cpu, 8 + 0x02000000);
-}
+				    & SCC_RR3_EXT_IP_B) {
+					INTERRUPT_ASSERT(d->irq);
+				}
 			}
 		}
 	}
@@ -305,7 +318,7 @@ int dev_scc_dma_func(struct cpu *cpu, void *extra, uint64_t addr,
 
 DEVICE_ACCESS(scc)
 {
-	struct scc_data *d = extra;
+	struct scc_data *d = (struct scc_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int port;
 	int ultrix_mode = 0;
@@ -355,9 +368,7 @@ DEVICE_ACCESS(scc)
 				d->scc_register_r[port * N_SCC_REGS +
 				    SCC_RR3] = 0;
 
-fatal("TODO: legacy rewrite!\n");
-abort();
-//				cpu_interrupt_ack(cpu, d->irq_nr);
+				INTERRUPT_DEASSERT(d->irq);
 			}
 
 #ifdef SCC_DEBUG
@@ -405,9 +416,7 @@ abort();
 			/*  TODO:  perhaps only clear the RX part of RR3?  */
 			d->scc_register_r[N_SCC_REGS + SCC_RR3] = 0;
 
-fatal("TODO: legacy rewrite!\n");
-abort();
-//			cpu_interrupt_ack(cpu, d->irq_nr);
+			INTERRUPT_DEASSERT(d->irq);
 
 //			debug("[ scc: (port %i) read from 0x%08lx: 0x%02x ]\n",
 //			    port, (long)relative_addr, (int)odata);
@@ -470,20 +479,24 @@ abort();
  *	addmul = 1 in most cases, 8 on SGI?
  */
 void *dev_scc_init(struct machine *machine, struct memory *mem,
-	uint64_t baseaddr, int irq_nr, int use_fb, int scc_nr, int addrmul)
+	uint64_t baseaddr, char* irq_path, int use_fb, int scc_nr, int addrmul)
 {
 	struct scc_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct scc_data)));
+	CHECK_ALLOCATION(d = (struct scc_data *) malloc(sizeof(struct scc_data)));
 	memset(d, 0, sizeof(struct scc_data));
 
-	d->irq_nr  = irq_nr;
+	INTERRUPT_CONNECT(irq_path, d->irq);
 	d->scc_nr  = scc_nr;
 	d->use_fb  = use_fb;
 	d->addrmul = addrmul;
 	d->console_handle = console_start_slave(machine, "SCC", 1);
 
-	lk201_init(&d->lk201, use_fb, dev_scc_add_to_rx_queue,
+	d->scc_register_r[SCC_RR0] |= SCC_RR0_TX_UNDERRUN;
+
+	lk201_init(&d->lk201, use_fb,
+	    dev_scc_space_available_in_queue,
+	    dev_scc_add_to_rx_queue,
 	    d->console_handle, d);
 
 	memory_device_register(mem, "scc", baseaddr, DEV_SCC_LENGTH,
