@@ -38,7 +38,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "imcreg.h"
+#include "thirdparty/imcreg.h"
 
 
 #define	SGI_IP22_TICK_SHIFT		14
@@ -46,8 +46,9 @@
 
 DEVICE_TICK(sgi_ip22)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 
+	// TODO: Document this!
 	if (d->reg[0x38 / 4] != 0)
 		d->reg[0x38 / 4] --;
 }
@@ -60,7 +61,7 @@ DEVICE_TICK(sgi_ip22)
  */
 DEVICE_ACCESS(sgi_ip22_imc)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int regnr;
 
@@ -164,7 +165,7 @@ DEVICE_ACCESS(sgi_ip22_imc)
  */
 DEVICE_ACCESS(sgi_ip22_unknown)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -207,7 +208,7 @@ DEVICE_ACCESS(sgi_ip22_unknown)
  */
 DEVICE_ACCESS(sgi_ip22_unknown2)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int regnr;
 
@@ -241,7 +242,7 @@ DEVICE_ACCESS(sgi_ip22_unknown2)
 
 DEVICE_ACCESS(sgi_ip22_sysid)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -275,7 +276,7 @@ DEVICE_ACCESS(sgi_ip22_sysid)
 
 DEVICE_ACCESS(sgi_ip22)
 {
-	struct sgi_ip22_data *d = extra;
+	struct sgi_ip22_data *d = (struct sgi_ip22_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int regnr;
 
@@ -410,7 +411,7 @@ struct sgi_ip22_data *dev_sgi_ip22_init(struct machine *machine,
 {
 	struct sgi_ip22_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct sgi_ip22_data)));
+	CHECK_ALLOCATION(d = (struct sgi_ip22_data *) malloc(sizeof(struct sgi_ip22_data)));
 	memset(d, 0, sizeof(struct sgi_ip22_data));
 
 	d->guiness_flag = guiness_flag;
