@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2018  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -25,7 +25,7 @@
  *  SUCH DAMAGE.
  *   
  *
- *  COMMENT: SGI IP30 stuff
+ *  COMMENT: SGI IP30 ("Octane") stuff
  *
  *  NOTE/TODO: This is just comprised of hardcoded guesses so far. (Ugly.)
  */
@@ -67,7 +67,7 @@ struct sgi_ip30_data {
 
 DEVICE_TICK(sgi_ip30)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 
 	d->reg_0x20000 += 1000;
 
@@ -83,7 +83,7 @@ abort();
 
 DEVICE_ACCESS(sgi_ip30)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -186,7 +186,7 @@ abort();
 
 DEVICE_ACCESS(sgi_ip30_2)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -236,7 +236,7 @@ DEVICE_ACCESS(sgi_ip30_2)
 
 DEVICE_ACCESS(sgi_ip30_3)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -300,7 +300,7 @@ DEVICE_ACCESS(sgi_ip30_3)
 
 DEVICE_ACCESS(sgi_ip30_4)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -343,7 +343,7 @@ DEVICE_ACCESS(sgi_ip30_4)
 
 DEVICE_ACCESS(sgi_ip30_5)
 {
-	struct sgi_ip30_data *d = extra;
+	struct sgi_ip30_data *d = (struct sgi_ip30_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	idata = memory_readmax64(cpu, data, len);
@@ -380,7 +380,7 @@ DEVINIT(sgi_ip30)
 {
 	struct sgi_ip30_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct sgi_ip30_data)));
+	CHECK_ALLOCATION(d = (struct sgi_ip30_data *) malloc(sizeof(struct sgi_ip30_data)));
 	memset(d, 0, sizeof(struct sgi_ip30_data));
 
 	memory_device_register(devinit->machine->memory, "sgi_ip30_1",
