@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2006-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2006-2020  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -29,10 +29,10 @@
  *
  *  TODO: Among other things:
  *
- *	x)  Interrupt masks (msk register stuff).
+ *	x)  Interrupt masks (msk register stuff). Are these really correct?
  *	x)  BSC (Bus state controller).
- *	x)  DMA
- *	x)  UBC
+ *	x)  DMA: Right now there's a hack for Dreamcast emulation
+ *	x)  UBC (User Break Controller)
  *	x)  ...
  */
 
@@ -49,19 +49,19 @@
 #include "machine.h"
 #include "memory.h"
 #include "misc.h"
+#include "sh4_dmacreg.h"
 #include "timer.h"
 
-#include "sh4_bscreg.h"
-#include "sh4_cache.h"
-#include "sh4_dmacreg.h"
-#include "sh4_exception.h"
-#include "sh4_intcreg.h"
-#include "sh4_mmu.h"
-#include "sh4_pcicreg.h"
-#include "sh4_rtcreg.h"
-#include "sh4_scifreg.h"
-#include "sh4_scireg.h"
-#include "sh4_tmureg.h"
+#include "thirdparty/sh4_bscreg.h"
+#include "thirdparty/sh4_cache.h"
+#include "thirdparty/sh4_exception.h"
+#include "thirdparty/sh4_intcreg.h"
+#include "thirdparty/sh4_mmu.h"
+#include "thirdparty/sh4_pcicreg.h"
+#include "thirdparty/sh4_rtcreg.h"
+#include "thirdparty/sh4_scifreg.h"
+#include "thirdparty/sh4_scireg.h"
+#include "thirdparty/sh4_tmureg.h"
 
 
 #define	SH4_REG_BASE		0xff000000
@@ -79,17 +79,13 @@
 #define	SCIF_TX_FIFO_SIZE	16
 #define	SCIF_DELAYED_TX_VALUE	2	/*  2 to be safe, 1 = fast but buggy  */
 
-/*  General-purpose I/O stuff:  */
-#define	SH4_PCTRA		0xff80002c
-#define	SH4_PDTRA		0xff800030
-#define	SH4_PCTRB		0xff800040
-#define	SH4_PDTRB		0xff800044
-#define	SH4_GPIOIC		0xff800048
+// Clock/occilation related
+#define SH4_CPG_FRQCR		0xffc00000	/* 16-bit */
+#define SH4_CPG_STBCR 		0xffc00004	/* 8-bit */
+#define SH4_CPG_WTCNT 		0xffc00008	/* 8/16-bit */
+#define SH4_CPG_WTCSR 		0xffc0000c	/* 8/16-bit */
+#define SH4_CPG_STBCR2 		0xffc00010	/* 8-bit */
 
-#ifdef UNSTABLE_DEVEL
-#define SH4_DEGUG
-/*  #define debug fatal  */
-#endif
 
 struct sh4_data {
 	/*  Store Queues:  */
@@ -114,18 +110,29 @@ struct sh4_data {
 	/*  Bus State Controller:  */
 	uint32_t	bsc_bcr1;
 	uint16_t	bsc_bcr2;
+	uint16_t	bsc_bcr3;	/*  SH7751R  */
 	uint32_t	bsc_wcr1;
 	uint32_t	bsc_wcr2;
+	uint32_t	bsc_wcr3;
 	uint32_t	bsc_mcr;
+	uint16_t	bsc_pcr;
 	uint16_t	bsc_rtcsr;
 	uint16_t	bsc_rtcor;
 	uint16_t	bsc_rfcr;
 
+	/*  CPG:  */
+	uint16_t	cpg_frqcr;
+	uint8_t		cpg_stbcr;
+	uint16_t	cpg_wtcnt;
+	uint16_t	cpg_wtcsr;
+	uint8_t		cpg_stbcr2;
+
 	/*  GPIO:  */
 	uint32_t	pctra;		/*  Port Control Register A  */
 	uint32_t	pdtra;		/*  Port Data Register A  */
 	uint32_t	pctrb;		/*  Port Control Register B  */
 	uint32_t	pdtrb;		/*  Port Data Register B  */
+	uint16_t	bsc_gpioic;
 
 	/*  PCIC (PCI controller):  */
 	struct pci_data	*pci_data;
@@ -155,8 +162,9 @@ struct sh4_data {
 	double		timer_hz[N_SH4_TIMERS];
 
 	/*  RTC:  */
-	uint32_t	rtc_reg[14];	/*  Excluding rcr1 and 2  */
+	uint32_t	rtc_reg[14];	/*  Excluding rcr1 and rcr2  */
 	uint8_t		rtc_rcr1;
+	uint8_t		rtc_rcr2;
 };
 
 
@@ -174,7 +182,7 @@ struct sh4_data {
  */
 static void sh4_timer_tick(struct timer *t, void *extra)
 {
-	struct sh4_data *d = extra;
+	struct sh4_data *d = (struct sh4_data *) extra;
 	int i;
 
 	/*  Fake RAM refresh:  */
@@ -197,7 +205,7 @@ static void sh4_timer_tick(struct timer *t, void *extra)
 			continue;
 
 		/*  Update the current count:  */
-		d->tcnt[i] -= d->timer_hz[i] / SH4_PSEUDO_TIMER_HZ;
+		d->tcnt[i] -= (uint32_t) (d->timer_hz[i] / SH4_PSEUDO_TIMER_HZ);
 
 		/*  Has the timer underflowed?  */
 		if ((int32_t)d->tcnt[i] < 0 && old >= 0) {
@@ -224,12 +232,12 @@ static void sh4_timer_tick(struct timer *t, void *extra)
 
 static void sh4_pcic_interrupt_assert(struct interrupt *interrupt)
 {
-	struct sh4_data *d = interrupt->extra;
+	struct sh4_data *d = (struct sh4_data *) interrupt->extra;
 	INTERRUPT_ASSERT(d->cpu_pcic_interrupt[interrupt->line]);
 }
 static void sh4_pcic_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct sh4_data *d = interrupt->extra;
+	struct sh4_data *d = (struct sh4_data *) interrupt->extra;
 	INTERRUPT_DEASSERT(d->cpu_pcic_interrupt[interrupt->line]);
 }
 
@@ -260,7 +268,7 @@ static void scif_reassert_interrupts(struct sh4_data *d)
 
 DEVICE_TICK(sh4)
 {
-	struct sh4_data *d = extra;
+	struct sh4_data *d = (struct sh4_data *) extra;
 	unsigned int i;
 
 	/*
@@ -362,7 +370,7 @@ void sh4_dmac_transfer(struct cpu *cpu, struct sh4_data *d, int channel)
 	src_delta *= transmit_size;
 	dst_delta *= transmit_size;
 
-#ifdef SH4_DEGUG
+#ifdef SH4_DEBUG
 	fatal("|SH4 DMA transfer, channel %i\n", channel);
 	fatal("|Source addr:      0x%08x (delta %i)\n", (int) sar, src_delta);
 	fatal("|Destination addr: 0x%08x (delta %i)\n", (int) dar, dst_delta);
@@ -378,6 +386,10 @@ void sh4_dmac_transfer(struct cpu *cpu, struct sh4_data *d, int channel)
 		 *  External Address Space => external device
 		 */
 
+		// Avoid compiler warnings about unused sar and dar.
+		(void)sar;
+		(void)dar;
+
 		/*  Note: No transfer is done here! It is up to the
 		    external device to do the transfer itself!  */
 		break;
@@ -774,7 +786,7 @@ DEVICE_ACCESS(sh4_utlb_da1)
 
 DEVICE_ACCESS(sh4_pcic)
 {
-	struct sh4_data *d = extra;
+	struct sh4_data *d = (struct sh4_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -828,7 +840,7 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != 0xac000000) {
 			fatal("sh4_pcic: SH4_PCICONF5 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
@@ -837,7 +849,7 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != 0x8c000000) {
 			fatal("sh4_pcic: SH4_PCICONF6 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
@@ -846,7 +858,7 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != ((64 - 1) << 20)) {
 			fatal("sh4_pcic: SH4_PCILSR0 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
@@ -855,7 +867,7 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != 0xac000000) {
 			fatal("sh4_pcic: SH4_PCILAR0 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
@@ -864,7 +876,7 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != ((64 - 1) << 20)) {
 			fatal("sh4_pcic: SH4_PCILSR1 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
@@ -873,15 +885,15 @@ DEVICE_ACCESS(sh4_pcic)
 		/*  Hardcoded to what OpenBSD/landisk uses:  */
 		if (writeflag == MEM_WRITE && idata != 0xac000000) {
 			fatal("sh4_pcic: SH4_PCILAR1 unknown value"
-			    " 0x%"PRIx32"\n", (uint32_t) idata);
+			    " 0x%" PRIx32"\n", (uint32_t) idata);
 			exit(1);
 		}
 		break;
 
 	case SH4_PCIMBR:
 		if (writeflag == MEM_WRITE && idata != SH4_PCIC_MEM) {
-			fatal("sh4_pcic: PCIMBR set to 0x%"PRIx32", not"
-			    " 0x%"PRIx32"? TODO\n", (uint32_t) idata,
+			fatal("sh4_pcic: PCIMBR set to 0x%" PRIx32", not"
+			    " 0x%" PRIx32"? TODO\n", (uint32_t) idata,
 			    (uint32_t) SH4_PCIC_MEM);
 			exit(1);
 		}
@@ -889,8 +901,8 @@ DEVICE_ACCESS(sh4_pcic)
 
 	case SH4_PCIIOBR:
 		if (writeflag == MEM_WRITE && idata != SH4_PCIC_IO) {
-			fatal("sh4_pcic: PCIIOBR set to 0x%"PRIx32", not"
-			    " 0x%"PRIx32"? TODO\n", (uint32_t) idata,
+			fatal("sh4_pcic: PCIIOBR set to 0x%" PRIx32", not"
+			    " 0x%" PRIx32"? TODO\n", (uint32_t) idata,
 			    (uint32_t) SH4_PCIC_IO);
 			exit(1);
 		}
@@ -932,8 +944,8 @@ DEVICE_ACCESS(sh4_pcic)
 
 DEVICE_ACCESS(sh4_sq)
 {
-	struct sh4_data *d = extra;
-	int i;
+	struct sh4_data *d = (struct sh4_data *) extra;
+	size_t i;
 
 	if (writeflag == MEM_WRITE) {
 		for (i=0; i<len; i++)
@@ -949,7 +961,7 @@ DEVICE_ACCESS(sh4_sq)
 
 DEVICE_ACCESS(sh4)
 {
-	struct sh4_data *d = extra;
+	struct sh4_data *d = (struct sh4_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int timer_nr = 0, dma_channel = 0;
 
@@ -966,7 +978,7 @@ DEVICE_ACCESS(sh4)
 			d->sdmr3 = v;
 		else
 			d->sdmr2 = v;
-		debug("[ sh4: sdmr%i set to 0x%04"PRIx16" ]\n",
+		debug("[ sh4: sdmr%i set to 0x%04" PRIx16" ]\n",
 		    relative_addr & 0x00040000? 3 : 2, v);
 		return 1;
 	}
@@ -1107,7 +1119,18 @@ DEVICE_ACCESS(sh4)
 	/********************************/
 	/*  UBC: User Break Controller  */
 
-	case 0xff200008:    /*  SH4_BBRA  */
+	case 0xff000014:    /*  SH4_UBC_BASRA  */
+	case 0xff000018:    /*  SH4_UBC_BASRB  */
+
+	case 0xff200000:    /*  SH4_UBC_BARA  */
+	case 0xff200004:    /*  SH4_UBC_BAMRA  */
+	case 0xff200008:    /*  SH4_UBC_BBRA  */
+	case 0xff20000c:    /*  SH4_UBC_BARB  */
+	case 0xff200010:    /*  SH4_UBC_BAMRB  */
+	case 0xff200014:    /*  SH4_UBC_BBRB  */
+	case 0xff200018:    /*  SH4_UBC_BDRB  */
+	case 0xff20001c:    /*  SH4_UBC_BDMRB  */
+	case 0xff200020:    /*  SH4_UBC_BRCR  */
 		/*  TODO  */
 		break;
 
@@ -1150,8 +1173,10 @@ DEVICE_ACCESS(sh4)
 
 	case SH4_TCOR2:
 		timer_nr ++;
+		// fall through
 	case SH4_TCOR1:
 		timer_nr ++;
+		// fall through
 	case SH4_TCOR0:
 		/*  Timer Constant Register  */
 		if (writeflag == MEM_READ)
@@ -1162,8 +1187,10 @@ DEVICE_ACCESS(sh4)
 
 	case SH4_TCNT2:
 		timer_nr ++;
+		// fall through
 	case SH4_TCNT1:
 		timer_nr ++;
+		// fall through
 	case SH4_TCNT0:
 		/*  Timer Counter Register  */
 		if (writeflag == MEM_READ)
@@ -1174,8 +1201,10 @@ DEVICE_ACCESS(sh4)
 
 	case SH4_TCR2:
 		timer_nr ++;
+		// fall through
 	case SH4_TCR1:
 		timer_nr ++;
+		// fall through
 	case SH4_TCR0:
 		/*  Timer Control Register  */
 		if (writeflag == MEM_READ) {
@@ -1208,7 +1237,7 @@ DEVICE_ACCESS(sh4)
 			if (idata & (TCR_ICPF | TCR_ICPE1 | TCR_ICPE0 |
 			    TCR_CKEG1 | TCR_CKEG0 | TCR_TPSC2)) {
 				fatal("Unimplemented SH4 timer control"
-				    " bits: 0x%08"PRIx32". Aborting.\n",
+				    " bits: 0x%08" PRIx32". Aborting.\n",
 				    (int) idata);
 				exit(1);
 			}
@@ -1231,13 +1260,20 @@ DEVICE_ACCESS(sh4)
 	/*  8 channels on SH7760                         */
 
 	case SH4_SAR7:	dma_channel ++;
+			// fall through
 	case SH4_SAR6:	dma_channel ++;
+			// fall through
 	case SH4_SAR5:	dma_channel ++;
+			// fall through
 	case SH4_SAR4:	dma_channel ++;
+			// fall through
 	case SH4_SAR3:	dma_channel ++;
+			// fall through
 	case SH4_SAR2:	dma_channel ++;
+			// fall through
 	case SH4_SAR1:	dma_channel ++;
-	case SH4_SAR0:	dma_channel ++;
+			// fall through
+	case SH4_SAR0:
 		if (writeflag == MEM_READ)
 			odata = cpu->cd.sh.dmac_sar[dma_channel];
 		else
@@ -1245,13 +1281,20 @@ DEVICE_ACCESS(sh4)
 		break;
 
 	case SH4_DAR7:	dma_channel ++;
+			// fall through
 	case SH4_DAR6:	dma_channel ++;
+			// fall through
 	case SH4_DAR5:	dma_channel ++;
+			// fall through
 	case SH4_DAR4:	dma_channel ++;
+			// fall through
 	case SH4_DAR3:	dma_channel ++;
+			// fall through
 	case SH4_DAR2:	dma_channel ++;
+			// fall through
 	case SH4_DAR1:	dma_channel ++;
-	case SH4_DAR0:	dma_channel ++;
+			// fall through
+	case SH4_DAR0:
 		if (writeflag == MEM_READ)
 			odata = cpu->cd.sh.dmac_dar[dma_channel];
 		else
@@ -1259,13 +1302,20 @@ DEVICE_ACCESS(sh4)
 		break;
 
 	case SH4_DMATCR7: dma_channel ++;
+			// fall through
 	case SH4_DMATCR6: dma_channel ++;
+			// fall through
 	case SH4_DMATCR5: dma_channel ++;
+			// fall through
 	case SH4_DMATCR4: dma_channel ++;
+			// fall through
 	case SH4_DMATCR3: dma_channel ++;
+			// fall through
 	case SH4_DMATCR2: dma_channel ++;
+			// fall through
 	case SH4_DMATCR1: dma_channel ++;
-	case SH4_DMATCR0: dma_channel ++;
+			// fall through
+	case SH4_DMATCR0:
 		if (writeflag == MEM_READ)
 			odata = cpu->cd.sh.dmac_tcr[dma_channel] & 0x00ffffff;
 		else {
@@ -1281,13 +1331,20 @@ DEVICE_ACCESS(sh4)
 		break;
 
 	case SH4_CHCR7:	dma_channel ++;
+			// fall through
 	case SH4_CHCR6:	dma_channel ++;
+			// fall through
 	case SH4_CHCR5:	dma_channel ++;
+			// fall through
 	case SH4_CHCR4:	dma_channel ++;
+			// fall through
 	case SH4_CHCR3:	dma_channel ++;
+			// fall through
 	case SH4_CHCR2:	dma_channel ++;
+			// fall through
 	case SH4_CHCR1:	dma_channel ++;
-	case SH4_CHCR0:	dma_channel ++;
+			// fall through
+	case SH4_CHCR0:
 		if (writeflag == MEM_READ) {
 			odata = cpu->cd.sh.dmac_chcr[dma_channel];
 		} else {
@@ -1302,6 +1359,15 @@ DEVICE_ACCESS(sh4)
 		}
 		break;
 
+	case SH4_DMAOR:
+		if (writeflag == MEM_READ) {
+			odata = cpu->cd.sh.dmaor;
+		} else {
+			// Only some bits are writable:
+			idata &= (DMAOR_DDT | DMAOR_PR1 | DMAOR_PR0 | DMAOR_DME);
+			cpu->cd.sh.dmaor = idata;
+		}
+		break;
 
 	/*************************************************/
 	/*  BSC: Bus State Controller                    */
@@ -1327,6 +1393,17 @@ DEVICE_ACCESS(sh4)
 			odata = d->bsc_bcr2;
 		break;
 
+	case SH4_BCR3:
+		if (len != sizeof(uint16_t)) {
+			fatal("Non-16-bit SH4_BCR3 access?\n");
+			exit(1);
+		}
+		if (writeflag == MEM_WRITE)
+			d->bsc_bcr3 = idata;
+		else
+			odata = d->bsc_bcr3;
+		break;
+
 	case SH4_WCR1:
 		if (writeflag == MEM_WRITE)
 			d->bsc_wcr1 = idata & 0x77777777;
@@ -1341,6 +1418,13 @@ DEVICE_ACCESS(sh4)
 			odata = d->bsc_wcr2;
 		break;
 
+	case SH4_WCR3:
+		if (writeflag == MEM_WRITE)
+			d->bsc_wcr3 = idata & 0x77777777;
+		else
+			odata = d->bsc_wcr3;
+		break;
+
 	case SH4_MCR:
 		if (writeflag == MEM_WRITE)
 			d->bsc_mcr = idata & 0xf8bbffff;
@@ -1348,6 +1432,13 @@ DEVICE_ACCESS(sh4)
 			odata = d->bsc_mcr;
 		break;
 
+	case SH4_PCR:
+		if (writeflag == MEM_WRITE)
+			d->bsc_pcr = idata;
+		else
+			odata = d->bsc_pcr;
+		break;
+
 	case SH4_RTCSR:
 		/*
 		 *  Refresh Time Control/Status Register. Called RTCSR in
@@ -1385,20 +1476,35 @@ DEVICE_ACCESS(sh4)
 	/*  GPIO:  General-purpose I/O controller  */
 
 	case SH4_PCTRA:
-		if (writeflag == MEM_WRITE)
+		if (writeflag == MEM_WRITE) {
 			d->pctra = idata;
-		else
+			
+			// Hack: Makes the Dreamcast BIOS pass "cable select"
+			// detection, it seems, without hanging in an endless
+			// loop.
+			d->pdtra |= 0x03;
+		} else {
 			odata = d->pctra;
+		}
 		break;
 
 	case SH4_PDTRA:
 		if (writeflag == MEM_WRITE) {
-			debug("[ sh4: pdtra: write: TODO ]\n");
+			// debug("[ sh4: pdtra: write 0x%08x (while pctra = 0x%08x) ]\n", (int)idata, (int)d->pctra);
 			d->pdtra = idata;
+
+			// Hack: Makes the Dreamcast BIOS pass "cable select"
+			// detection, it seems, without hanging in an endless
+			// loop.
+			if ((idata & 1) == 0 || (idata & 2) == 0)
+				d->pdtra &= ~3;
 		} else {
-			debug("[ sh4: pdtra: read: TODO ]\n");
+			// debug("[ sh4: pdtra: read ]\n");
 			odata = d->pdtra;
-odata = random();
+
+			// bits 8..9 on Dreamcast mean:
+			//  00 = VGA, 10 = RGB, 11 = composite.
+			odata |= (0 << 8);
 		}
 		break;
 
@@ -1419,6 +1525,13 @@ odata = random();
 		}
 		break;
 
+	case SH4_GPIOIC:
+		if (writeflag == MEM_WRITE)
+			d->bsc_gpioic = idata;
+		else
+			odata = d->bsc_gpioic;
+		break;
+
 
 	/****************************/
 	/*  SCI:  Serial Interface  */
@@ -1629,15 +1742,65 @@ odata = random();
 		}
 		break;
 
+	case SH4_SCIF_BASE + SCIF_FDR:
+		/*  Nr of bytes in the TX and RX fifos, respectively:  */
+		{
+			int chars_avail = console_charavail(d->scif_console_handle);
+			if (chars_avail > 255) {
+				fatal("[ SH4: Too many chars avail; dropping some ]\n");
+				chars_avail = 255;
+			}
+			
+			odata = chars_avail | (d->scif_tx_fifo_cursize << 8);
+		}
+		break;
+
+	case SH4_SCIF_BASE + SCIF_SPTR:
+		/*  TODO: Implement all bits.  */
+		odata = 0;
+		break;
+
 	case SH4_SCIF_BASE + SCIF_LSR:
 		/*  TODO: Implement all bits.  */
 		odata = 0;
 		break;
 
-	case SH4_SCIF_BASE + SCIF_FDR:
-		/*  Nr of bytes in the TX and RX fifos, respectively:  */
-		odata = (console_charavail(d->scif_console_handle)? 1 : 0)
-		    + (d->scif_tx_fifo_cursize << 8);
+
+	/*************************************************/
+
+	case SH4_CPG_FRQCR:	// 0xffc00000	16-bit
+		if (writeflag == MEM_WRITE)
+			d->cpg_frqcr = idata;
+		else
+			odata = d->cpg_frqcr;
+		break;
+
+	case SH4_CPG_STBCR: 	// 0xffc00004	8-bit
+		if (writeflag == MEM_WRITE)
+			d->cpg_stbcr = idata;
+		else
+			odata = d->cpg_stbcr;
+		break;
+
+	case SH4_CPG_WTCNT:	// 0xffc00008	8/16-bit
+		if (writeflag == MEM_WRITE)
+			d->cpg_wtcnt = idata;
+		else
+			odata = d->cpg_wtcnt;
+		break;
+
+	case SH4_CPG_WTCSR:	// 0xffc0000c	8/16-bit
+		if (writeflag == MEM_WRITE)
+			d->cpg_wtcsr = idata;
+		else
+			odata = d->cpg_wtcsr;
+		break;
+
+	case SH4_CPG_STBCR2:	// 0xffc00010	8-bit
+		if (writeflag == MEM_WRITE)
+			d->cpg_stbcr2 = idata;
+		else
+			odata = d->cpg_stbcr2;
 		break;
 
 
@@ -1676,6 +1839,18 @@ odata = random();
 		}
 		break;
 
+	case SH4_RCR2:
+		if (writeflag == MEM_READ)
+			odata = d->rtc_rcr2;
+		else {
+			d->rtc_rcr2 = idata;
+			// bit 1 (i.e. idata == 0x02) means reset.
+			if (idata != 0x02) {
+				debug("[ SH4: TODO: RTC RCR2 value 0x%02x ignored. ]\n", (int)idata);
+			}
+		}
+		break;
+
 
 	/*************************************************/
 
@@ -1686,8 +1861,9 @@ odata = random();
 			fatal("[ sh4: write to addr 0x%x: 0x%x ]\n",
 			    (int)relative_addr, (int)idata);
 		}
-#ifdef SH4_DEGUG
-		/*  exit(1);  */
+
+#ifdef SH4_DEBUG
+		exit(1);
 #endif
 	}
 
@@ -1705,7 +1881,7 @@ DEVINIT(sh4)
 	struct machine *machine = devinit->machine;
 	struct sh4_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct sh4_data)));
+	CHECK_ALLOCATION(d = (struct sh4_data *) malloc(sizeof(struct sh4_data)));
 	memset(d, 0, sizeof(struct sh4_data));
 
 
@@ -1717,7 +1893,7 @@ DEVINIT(sh4)
 	    SH4_REG_BASE, 0x01000000, dev_sh4_access, d, DM_DEFAULT, NULL);
 
 	/*  On-chip RAM/cache:  */
-	dev_ram_init(machine, 0x1e000000, 0x8000, DEV_RAM_RAM, 0x0);
+	dev_ram_init(machine, 0x1e000000, 0x8000, DEV_RAM_RAM, 0x0, NULL);
 
 	/*  0xe0000000: Store queues:  */
 	memory_device_register(machine->memory, "sh4_sq",
@@ -1747,13 +1923,13 @@ DEVINIT(sh4)
 	 *  0xf4000000	SH4_CCDA	D-Cache address array
 	 *  0xf5000000	SH4_CCDD	D-Cache data array
 	 *
-	 *  TODO: Implement more correct cache behaviour?
+	 *  TODO: Implement more correct cache behavior?
 	 */
 
-	dev_ram_init(machine, SH4_CCIA, SH4_ICACHE_SIZE * 2, DEV_RAM_RAM, 0x0);
-	dev_ram_init(machine, SH4_CCID, SH4_ICACHE_SIZE,     DEV_RAM_RAM, 0x0);
-	dev_ram_init(machine, SH4_CCDA, SH4_DCACHE_SIZE * 2, DEV_RAM_RAM, 0x0);
-	dev_ram_init(machine, SH4_CCDD, SH4_DCACHE_SIZE,     DEV_RAM_RAM, 0x0);
+	dev_ram_init(machine, SH4_CCIA, SH4_ICACHE_SIZE * 2, DEV_RAM_RAM, 0x0, NULL);
+	dev_ram_init(machine, SH4_CCID, SH4_ICACHE_SIZE,     DEV_RAM_RAM, 0x0, NULL);
+	dev_ram_init(machine, SH4_CCDA, SH4_DCACHE_SIZE * 2, DEV_RAM_RAM, 0x0, NULL);
+	dev_ram_init(machine, SH4_CCDD, SH4_DCACHE_SIZE,     DEV_RAM_RAM, 0x0, NULL);
 
 	/*  0xf2000000	SH4_ITLB_AA  */
 	memory_device_register(machine->memory, "sh4_itlb_aa", SH4_ITLB_AA,
@@ -1787,16 +1963,16 @@ DEVINIT(sh4)
 
 	/*  Register 16 PCIC interrupts:  */
 	for (i=0; i<N_PCIC_IRQS; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		snprintf(n, sizeof(n), "%s.pcic.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = sh4_pcic_interrupt_assert;
-		template.interrupt_deassert = sh4_pcic_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = sh4_pcic_interrupt_assert;
+		templ.interrupt_deassert = sh4_pcic_interrupt_deassert;
+		interrupt_handler_register(&templ);
 
 		snprintf(tmp, sizeof(tmp), "%s.irq[0x%x]",
 		    devinit->interrupt_path, SH4_INTEVT_IRQ0 + 0x20 * i);
@@ -1855,6 +2031,11 @@ DEVINIT(sh4)
 	}
 
 
+	// The RTC RCR2 register is "basically" initialized to 0x09, with
+	// some bit undefined. :-) According to the manual.
+	d->rtc_rcr2 = 0x09;
+
+
 	/*
 	 *  Bus State Controller initial values, according to the
 	 *  SH7760 manual:
@@ -1863,7 +2044,7 @@ DEVINIT(sh4)
 	d->bsc_bcr2 = 0x3ffc;
 	d->bsc_wcr1 = 0x77777777;
 	d->bsc_wcr2 = 0xfffeefff;
-
+	d->bsc_wcr3 = 0x77777777;
 
 	return 1;
 }
