@@ -42,21 +42,21 @@
 #include "misc.h"
 #include "net.h"
 
-#include "dp83932reg.h"
+#include "thirdparty/dp83932reg.h"
 
 
 #define	DEV_SN_LENGTH		0x1000
 
 struct sn_data {
+	struct nic_data	nic;
 	struct interrupt irq;
-	unsigned char	macaddr[6];
 	uint32_t	reg[SONIC_NREGS];
 };
 
 
 DEVICE_ACCESS(sn)
 {
-	struct sn_data *d = extra;
+	struct sn_data *d = (struct sn_data *) extra;
 	uint64_t idata = 0, odata = 0;
 	int regnr;
 
@@ -99,23 +99,24 @@ DEVINIT(sn)
 	size_t nlen = 55;
 	struct sn_data *d;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct sn_data)));
+	CHECK_ALLOCATION(d = (struct sn_data *) malloc(sizeof(struct sn_data)));
 	memset(d, 0, sizeof(struct sn_data));
 
 	INTERRUPT_CONNECT(devinit->interrupt_path, d->irq);
 
-	net_generate_unique_mac(devinit->machine, d->macaddr);
+	net_generate_unique_mac(devinit->machine, d->nic.mac_address);
 
-	CHECK_ALLOCATION(name2 = malloc(nlen));
+	CHECK_ALLOCATION(name2 = (char *) malloc(nlen));
 	snprintf(name2, nlen, "%s [%02x:%02x:%02x:%02x:%02x:%02x]",
-	    devinit->name, d->macaddr[0], d->macaddr[1], d->macaddr[2],
-	    d->macaddr[3], d->macaddr[4], d->macaddr[5]);
+	    devinit->name, d->nic.mac_address[0], d->nic.mac_address[1],
+	    d->nic.mac_address[2], d->nic.mac_address[3],
+	    d->nic.mac_address[4], d->nic.mac_address[5]);
 
 	memory_device_register(devinit->machine->memory, name2,
 	    devinit->addr, DEV_SN_LENGTH,
 	    dev_sn_access, (void *)d, DM_DEFAULT, NULL);
 
-	net_add_nic(devinit->machine->emul->net, d, d->macaddr);
+	net_add_nic(devinit->machine->emul->net, &d->nic);
 
 	return 1;
 }
