@@ -103,7 +103,7 @@ void v3_isa_interrupt_common(struct v3_data *d, int old_isa_assert)
  */
 void v3_isa_interrupt_assert(struct interrupt *interrupt)
 {
-	struct v3_data *d = interrupt->extra;
+	struct v3_data *d = (struct v3_data *) interrupt->extra;
 	int old_isa_assert, line = interrupt->line;
 	int mask = 1 << (line & 7);
 
@@ -126,7 +126,7 @@ void v3_isa_interrupt_assert(struct interrupt *interrupt)
  */
 void v3_isa_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct v3_data *d = interrupt->extra;
+	struct v3_data *d = (struct v3_data *) interrupt->extra;
 	int line = interrupt->line, mask = 1 << (line & 7);
 	int old_irr1 = d->pic1->irr, old_isa_assert;
 
@@ -152,7 +152,7 @@ DEVICE_ACCESS(v3_pci)
 {
 	uint64_t idata = 0, odata = 0;
 	int bus, dev, func, reg;
-	struct v3_data *d = extra;
+	struct v3_data *d = (struct v3_data *) extra;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len|MEM_PCI_LITTLE_ENDIAN);
@@ -186,7 +186,7 @@ DEVICE_ACCESS(v3_pci)
 
 DEVICE_ACCESS(v3)
 {
-	struct v3_data *d = extra;
+	struct v3_data *d = (struct v3_data *) extra;
 	uint64_t idata = 0, odata = 0;
 
 	if (writeflag == MEM_WRITE)
@@ -250,7 +250,7 @@ DEVINIT(v3)
 	char pci_irq_base[200];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct v3_data)));
+	CHECK_ALLOCATION(d = (struct v3_data *) malloc(sizeof(struct v3_data)));
 	memset(d, 0, sizeof(struct v3_data));
 
 	switch (devinit->machine->machine_type) {
@@ -276,30 +276,30 @@ DEVINIT(v3)
 	 *  dummy is used by re-asserts.
 	 */
 	for (i=0; i<17; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		char n[300];
 		snprintf(n, sizeof(n), "%s.v3.isa.%i",
 		    devinit->interrupt_path, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = n;
-		template.extra = d;
-		template.interrupt_assert = v3_isa_interrupt_assert;
-		template.interrupt_deassert = v3_isa_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = n;
+		templ.extra = d;
+		templ.interrupt_assert = v3_isa_interrupt_assert;
+		templ.interrupt_deassert = v3_isa_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*  Register two 8259 PICs:  */
 	snprintf(tmpstr, sizeof(tmpstr), "8259 irq=%s.v3.isa.16 addr=0x%llx",
 	    devinit->interrupt_path, (long long)(isa_port_base + 0x20));
-	d->pic1 = devinit->machine->isa_pic_data.pic1 =
+	d->pic1 = devinit->machine->isa_pic_data.pic1 = (struct pic8259_data *)
 	    device_add(devinit->machine, tmpstr);
 	d->ptr_to_pending_timer_interrupts =
 	    devinit->machine->isa_pic_data.pending_timer_interrupts;
 
 	snprintf(tmpstr, sizeof(tmpstr), "8259 irq=%s.v3.isa.2 addr=0x%llx",
 	    devinit->interrupt_path, (long long)(isa_port_base + 0xa0));
-	d->pic2 = devinit->machine->isa_pic_data.pic2 =
+	d->pic2 = devinit->machine->isa_pic_data.pic2 = (struct pic8259_data *)
 	    device_add(devinit->machine, tmpstr);
 
 	snprintf(isa_irq_base, sizeof(isa_irq_base), "%s.v3",
