@@ -46,10 +46,10 @@
 #include "misc.h"
 #include "timer.h"
 
-#include "bcureg.h"
-#include "vripreg.h"
-#include "vrkiureg.h"
-#include "vr_rtcreg.h"
+#include "thirdparty/bcureg.h"
+#include "thirdparty/vripreg.h"
+#include "thirdparty/vrkiureg.h"
+#include "thirdparty/vr_rtcreg.h"
 
 
 /*  #define debug fatal  */
@@ -94,7 +94,7 @@ struct vr41xx_data {
  */ 
 void vr41xx_vrip_interrupt_assert(struct interrupt *interrupt)
 {
-	struct vr41xx_data *d = interrupt->extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) interrupt->extra;
 	int line = interrupt->line;
 	if (line < 16)
 		d->sysint1 |= (1 << line);
@@ -105,7 +105,7 @@ void vr41xx_vrip_interrupt_assert(struct interrupt *interrupt)
 }
 void vr41xx_vrip_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct vr41xx_data *d = interrupt->extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) interrupt->extra;
 	int line = interrupt->line;
 	if (line < 16)
 		d->sysint1 &= ~(1 << line);
@@ -122,7 +122,7 @@ void vr41xx_vrip_interrupt_deassert(struct interrupt *interrupt)
  */ 
 void vr41xx_giu_interrupt_assert(struct interrupt *interrupt)
 {
-	struct vr41xx_data *d = interrupt->extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) interrupt->extra;
 	int line = interrupt->line;
 	d->giuint |= (1 << line);
 	if (d->giuint & d->giumask)
@@ -130,7 +130,7 @@ void vr41xx_giu_interrupt_assert(struct interrupt *interrupt)
 }
 void vr41xx_giu_interrupt_deassert(struct interrupt *interrupt)
 {
-	struct vr41xx_data *d = interrupt->extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) interrupt->extra;
 	int line = interrupt->line;
 	d->giuint &= ~(1 << line);
 	if (!(d->giuint & d->giumask))
@@ -392,14 +392,14 @@ static void vr41xx_keytick(struct cpu *cpu, struct vr41xx_data *d)
  */
 static void timer_tick(struct timer *timer, void *extra)
 {
-	struct vr41xx_data *d = extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) extra;
 	d->pending_timer_interrupts ++;
 }
 
 
 DEVICE_TICK(vr41xx)
 {
-	struct vr41xx_data *d = extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) extra;
 
 	if (d->pending_timer_interrupts > 0)
 		INTERRUPT_ASSERT(d->timer_irq);
@@ -466,7 +466,7 @@ static uint64_t vr41xx_kiu(struct cpu *cpu, int ofs, uint64_t idata,
 	default:
 		if (writeflag == MEM_WRITE)
 			debug("[ vr41xx KIU: unimplemented write to offset "
-			    "0x%x, data=0x%016"PRIx64" ]\n", ofs,
+			    "0x%x, data=0x%016" PRIx64" ]\n", ofs,
 			    (uint64_t) idata);
 		else
 			debug("[ vr41xx KIU: unimplemented read from offset "
@@ -479,15 +479,14 @@ static uint64_t vr41xx_kiu(struct cpu *cpu, int ofs, uint64_t idata,
 
 DEVICE_ACCESS(vr41xx)
 {
-	struct vr41xx_data *d = extra;
+	struct vr41xx_data *d = (struct vr41xx_data *) extra;
 	uint64_t idata = 0, odata = 0;
-	int regnr;
 	int revision = 0;
 
 	if (writeflag == MEM_WRITE)
 		idata = memory_readmax64(cpu, data, len);
 
-	regnr = relative_addr / sizeof(uint64_t);
+	// int regnr = relative_addr / sizeof(uint64_t);
 
 	/*  KIU ("Keyboard Interface Unit") is handled separately.  */
 	if (relative_addr >= d->kiu_offset &&
@@ -590,6 +589,7 @@ DEVICE_ACCESS(vr41xx)
 			switch (relative_addr) {
 			case 0xc0:
 				odata = (tv.tv_sec & 1) << 15;
+				odata += (uint64_t)tv.tv_usec * 32768 / 1000000;
 				break;
 			case 0xc2:
 				odata = (tv.tv_sec >> 1) & 0xffff;
@@ -637,11 +637,11 @@ DEVICE_ACCESS(vr41xx)
 	default:
 		if (writeflag == MEM_WRITE)
 			debug("[ vr41xx: unimplemented write to address "
-			    "0x%"PRIx64", data=0x%016"PRIx64" ]\n",
+			    "0x%" PRIx64", data=0x%016" PRIx64" ]\n",
 			    (uint64_t) relative_addr, (uint64_t) idata);
 		else
 			debug("[ vr41xx: unimplemented read from address "
-			    "0x%"PRIx64" ]\n", (uint64_t) relative_addr);
+			    "0x%" PRIx64" ]\n", (uint64_t) relative_addr);
 	}
 
 ret:
@@ -677,7 +677,7 @@ struct vr41xx_data *dev_vr41xx_init(struct machine *machine,
 	char tmps[300];
 	int i;
 
-	CHECK_ALLOCATION(d = malloc(sizeof(struct vr41xx_data)));
+	CHECK_ALLOCATION(d = (struct vr41xx_data *) malloc(sizeof(struct vr41xx_data)));
 	memset(d, 0, sizeof(struct vr41xx_data));
 
 	/*  Connect to MIPS irq 2:  */
@@ -689,32 +689,32 @@ struct vr41xx_data *dev_vr41xx_init(struct machine *machine,
 	 *  Register VRIP interrupt lines 0..25:
 	 */
 	for (i=0; i<=25; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		snprintf(tmps, sizeof(tmps), "%s.cpu[%i].vrip.%i",
 		    machine->path, machine->bootstrap_cpu, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = tmps;
-		template.extra = d;
-		template.interrupt_assert = vr41xx_vrip_interrupt_assert;
-		template.interrupt_deassert = vr41xx_vrip_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = tmps;
+		templ.extra = d;
+		templ.interrupt_assert = vr41xx_vrip_interrupt_assert;
+		templ.interrupt_deassert = vr41xx_vrip_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	/*
 	 *  Register GIU interrupt lines 0..31:
 	 */
 	for (i=0; i<32; i++) {
-		struct interrupt template;
+		struct interrupt templ;
 		snprintf(tmps, sizeof(tmps), "%s.cpu[%i].vrip.%i.giu.%i",
 		    machine->path, machine->bootstrap_cpu, VRIP_INTR_GIU, i);
-		memset(&template, 0, sizeof(template));
-		template.line = i;
-		template.name = tmps;
-		template.extra = d;
-		template.interrupt_assert = vr41xx_giu_interrupt_assert;
-		template.interrupt_deassert = vr41xx_giu_interrupt_deassert;
-		interrupt_handler_register(&template);
+		memset(&templ, 0, sizeof(templ));
+		templ.line = i;
+		templ.name = tmps;
+		templ.extra = d;
+		templ.interrupt_assert = vr41xx_giu_interrupt_assert;
+		templ.interrupt_deassert = vr41xx_giu_interrupt_deassert;
+		interrupt_handler_register(&templ);
 	}
 
 	d->cpumodel = cpumodel;
@@ -747,7 +747,7 @@ struct vr41xx_data *dev_vr41xx_init(struct machine *machine,
 	case 4181:
 		baseaddr = 0xa000000;
 		dev_ram_init(machine, 0xb000000, 0x1000000, DEV_RAM_MIRROR,
-		    0xa000000);
+		    0xa000000, NULL);
 		break;
 	case 4122:
 	case 4131:
@@ -775,7 +775,7 @@ struct vr41xx_data *dev_vr41xx_init(struct machine *machine,
 	 */
 	if (cpumodel == 4131) {
 		snprintf(tmps, sizeof(tmps), "ns16550 irq=%s.cpu[%i].vrip.%i "
-		    "addr=0x%"PRIx64" name2=siu", machine->path,
+		    "addr=0x%" PRIx64" name2=siu", machine->path,
 		    machine->bootstrap_cpu, VRIP_INTR_SIU,
 		    (uint64_t) (baseaddr+0x800));
 		device_add(machine, tmps);
@@ -799,7 +799,7 @@ struct vr41xx_data *dev_vr41xx_init(struct machine *machine,
 	/*  Some machines (?) use ISA space at 0x15000000 instead of
 	    0x14000000, eg IBM WorkPad Z50.  */
 	dev_ram_init(machine, 0x15000000, 0x1000000, DEV_RAM_MIRROR,
-	    0x14000000);
+	    0x14000000, NULL);
 
 	return d;
 }
