@@ -1,5 +1,29 @@
 /*
- *  $Id: fb_include.c,v 1.3 2006/10/17 07:55:53 debug Exp $
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
+ *
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright  
+ *     notice, this list of conditions and the following disclaimer in the 
+ *     documentation and/or other materials provided with the distribution.
+ *  3. The name of the author may not be used to endorse or promote products
+ *     derived from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+ *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE   
+ *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ *  SUCH DAMAGE.
+ *   
  *
  *  Included from dev_fb.c.
  *
@@ -65,7 +89,6 @@
 void REDRAW(struct vfb_data *d, int addr, int len)
 {
 	int x, y, pixel, npixels;
-	long color_r, color_g, color_b;
 	long color;
 
 #ifndef FB_SCALEDOWN
@@ -75,25 +98,21 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 	y = pixel / d->xsize;
 	x = pixel % d->xsize;
 
-	/*  How many framebuffer pixels?  */
+	/*  How many framebuffer pixels? (len doesn't include the last pixel)  */
 	npixels = len * 8 / d->bit_depth;
-	if (npixels == 0)
-		npixels = 1;
+	npixels += d->bit_depth >= 8 ? 1 : (8 / d->bit_depth);
 
 	if (d->bit_depth < 8) {
 		for (pixel=0; pixel<npixels; pixel++) {
-			int fb_addr, c, r, g, b;
-			color_r = color_g = color_b = 0;
-
-			fb_addr = (y * d->xsize + x) * d->bit_depth;
-			/*  fb_addr is now which _bit_ in
-			    the framebuffer  */
+			int c, r, g, b;
+			size_t fb_addr = (y * d->xsize + x) * d->bit_depth;
+			/*  fb_addr is now which _bit_ in the framebuffer  */
 
 			c = d->framebuffer[fb_addr >> 3];
 			fb_addr &= 7;
 
 			/*  HPC is reverse:  */
-			if (d->vfb_type == VFB_HPC)
+			if (d->vfb_type == VFB_HPC || d->vfb_type == VFB_REVERSEBITS)
 				fb_addr = 8 - d->bit_depth - fb_addr;
 
 			c = (c >> fb_addr) & ((1<<d->bit_depth) - 1);
@@ -108,11 +127,10 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 		}
 	} else if (d->bit_depth == 8) {
 		for (pixel=0; pixel<npixels; pixel++) {
-			int fb_addr, c, r, g, b;
-			color_r = color_g = color_b = 0;
-
-			fb_addr = y * d->xsize + x;
+			int c, r, g, b;
+			size_t fb_addr = y * d->xsize + x;
 			/*  fb_addr is now which byte in framebuffer  */
+
 			c = d->framebuffer[fb_addr];
 			r = d->rgb_palette[c*3 + 0];
 			g = d->rgb_palette[c*3 + 1];
@@ -123,10 +141,8 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 		}
 	} else {	/*  d->bit_depth > 8  */
 		for (pixel=0; pixel<npixels; pixel++) {
-			int fb_addr, r, g, b;
-			color_r = color_g = color_b = 0;
-
-			fb_addr = (y * d->xsize + x) * d->bit_depth;
+			int r, g, b;
+			size_t fb_addr = (y * d->xsize + x) * d->bit_depth;
 			/*  fb_addr is now which byte in framebuffer  */
 
 			/*  > 8 bits color.  */
@@ -208,7 +224,7 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 	if (d->bit_depth < 8) {
 		for (pixel=0; pixel<npixels; pixel++) {
 			int subx, suby, r, g, b;
-			color_r = color_g = color_b = 0;
+			int64_t color_r = 0, color_g = 0, color_b = 0;
 			for (suby=0; suby<scaledown; suby++)
 			    for (subx=0; subx<scaledown; subx++) {
 				int fb_x, fb_y, fb_addr, c;
@@ -224,7 +240,7 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 				fb_addr &= 7;
 
 				/*  HPC is reverse:  */
-				if (d->vfb_type == VFB_HPC)
+				if (d->vfb_type == VFB_HPC || d->vfb_type == VFB_REVERSEBITS)
 					fb_addr = 8 - d->bit_depth - fb_addr;
 
 				c = (c >> fb_addr) & ((1<<d->bit_depth) - 1);
@@ -248,7 +264,7 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 	} else if (d->bit_depth == 8) {
 		for (pixel=0; pixel<npixels; pixel++) {
 			int subx, suby, r, g, b;
-			color_r = color_g = color_b = 0;
+			int64_t color_r = 0, color_g = 0, color_b = 0;
 			for (suby=0; suby<scaledown; suby++)
 			    for (subx=0; subx<scaledown; subx++) {
 				int fb_x, fb_y, fb_addr, c;
@@ -276,7 +292,7 @@ void REDRAW(struct vfb_data *d, int addr, int len)
 		/*  Generic > 8 bit bit-depth:  */
 		for (pixel=0; pixel<npixels; pixel++) {
 			int subx, suby, r, g, b;
-			color_r = color_g = color_b = 0;
+			int64_t color_r = 0, color_g = 0, color_b = 0;
 			for (suby=0; suby<scaledown; suby++)
 			    for (subx=0; subx<scaledown; subx++) {
 				int fb_x, fb_y, fb_addr;
