@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -37,8 +37,11 @@
 #include "machine.h"
 #include "misc.h"
 
-#include "dc7085.h"	/*  for port names  */
-#include "lk201.h"
+#include "thirdparty/dc7085.h"	/*  for port names  */
+#include "thirdparty/lk201.h"
+
+
+// #define debug fatal
 
 
 /*
@@ -148,26 +151,17 @@ void lk201_convert_ascii_to_keybcode(struct lk201_data *d, unsigned char ch)
 
 /*
  *  lk201_send_mouse_update_sequence():
- *
- *  mouse_x, _y, _buttons contains the coordinates on the host's display, the
- *  "goal" of where we want to move.
- *
- *  d->mouse_x, _y, _buttons contain the last values transmitted to the
- *  emulated machine.
  */
-static int lk201_send_mouse_update_sequence(struct lk201_data *d, int mouse_x,
-	int mouse_y, int mouse_buttons, int mouse_fb_nr)
+static int lk201_send_mouse_update_sequence(struct lk201_data *d, int xdelta,
+	int ydelta, int mouse_buttons, int mouse_fb_nr)
 {
-	int xsign, xdelta, ysign, ydelta, m;
-
-	xdelta = mouse_x - d->mouse_x;
-	ydelta = mouse_y - d->mouse_y;
+	int xsign, ysign, m;
 
 	/*  If no change, then don't send any update!  */
 	if (xdelta == 0 && ydelta == 0 && d->mouse_buttons == mouse_buttons)
 		return 0;
 
-	m = 20;
+	m = 50;
 
 	if (xdelta > m)
 		xdelta = m;
@@ -178,8 +172,6 @@ static int lk201_send_mouse_update_sequence(struct lk201_data *d, int mouse_x,
 	if (ydelta < -m)
 		ydelta = -m;
 
-	d->mouse_x += xdelta;
-	d->mouse_y += ydelta;
 	d->mouse_buttons = mouse_buttons;
 
 	/*
@@ -226,59 +218,55 @@ static int lk201_send_mouse_update_sequence(struct lk201_data *d, int mouse_x,
  *  lk201_tick():
  *
  *  This function should be called "every now and then".
- *  If a key is available from the keyboard, add it to the rx queue.
+ *  If keys are available from the keyboard, add them to the rx queue.
  *  If other bits are set, an interrupt might need to be caused.
  */
 void lk201_tick(struct machine *machine, struct lk201_data *d)
 {
-	int mouse_x, mouse_y, mouse_buttons, mouse_fb_nr;
+	/*
+	 *  Different machines seem to use different ports for their
+	 *  serial console:
+	 *
+	 *  DEC MIPSMATE 5100 uses the keyboard port.
+	 *  DECstation 3100 (PMAX) and 5000/2000 (3MAX) use the printer port.
+	 *  Others seem to use the comm port.
+	 */
+	int port = DCCOMM_PORT;
 
-	if (console_charavail(d->console_handle)) {
+	if (machine->machine_type == MACHINE_PMAX) {
+		switch (machine->machine_subtype) {
+		case MACHINE_DEC_MIPSMATE_5100:
+			port = DCKBD_PORT;
+			break;
+		case MACHINE_DEC_PMAX_3100:
+		case MACHINE_DEC_3MAX_5000:
+			port = DCPRINTER_PORT;
+			break;
+		default:
+			port = DCCOMM_PORT;
+		}
+	}
+
+	if (d->use_fb)
+		port = DCKBD_PORT;
+
+	while (d->space_available_in_queue(d->add_data, port) &&
+	    console_charavail(d->console_handle)) {
 		unsigned char ch = console_readchar(d->console_handle);
 		if (d->use_fb)
 			lk201_convert_ascii_to_keybcode(d, ch);
-		else {
-			/*
-			 *  This is ugly, but necessary because different
-			 *  machines seem to use different ports for their
-			 *  serial console:
-			 *
-			 *  DEC MIPSMATE 5100 uses the keyboard port.
-			 *  DECstation 3100 (PMAX) and 5000/2000 (3MAX) use
-			 *  the printer port.
-			 *  Others seem to use the comm port.
-			 */
-			if (machine->machine_type == MACHINE_PMAX) {
-				switch (machine->machine_subtype) {
-				case MACHINE_DEC_MIPSMATE_5100:
-					d->add_to_rx_queue(d->add_data,
-					    ch, DCKBD_PORT);
-					break;
-				case MACHINE_DEC_PMAX_3100:
-				case MACHINE_DEC_3MAX_5000:
-					d->add_to_rx_queue(d->add_data,
-					    ch, DCPRINTER_PORT);
-					break;
-				default:
-					d->add_to_rx_queue(d->add_data,
-					    ch, DCCOMM_PORT);
-				}
-			} else {
-				d->add_to_rx_queue(d->add_data,
-				    ch, DCCOMM_PORT);
-			}
-		}
+		else
+			d->add_to_rx_queue(d->add_data, ch, port);
 	}
 
 	/*  Don't do mouse updates if we're running in serial console mode:  */
 	if (!d->use_fb)
 		return;
 
-	console_getmouse(&mouse_x, &mouse_y, &mouse_buttons,
-	    &mouse_fb_nr);
+	int dx, dy, mouse_buttons, mouse_fb_nr;
+	console_getmouse(&dx, &dy, &mouse_buttons, &mouse_fb_nr);
 
-	lk201_send_mouse_update_sequence(d, mouse_x, mouse_y,
-	    mouse_buttons, mouse_fb_nr);
+	lk201_send_mouse_update_sequence(d, dx, dy, mouse_buttons, mouse_fb_nr);
 }
 
 
@@ -290,6 +278,7 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
 			/*  Simply print the character to stdout:  */
 			console_putchar(d->console_handle, idata);
 		} else {
+			debug("[ lk201: writing data to KBD: 0x%02x ]\n", idata);
 			switch (idata) {
 			case LK_LED_DISABLE:	/*  0x11  */
 				break;
@@ -297,6 +286,12 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
 				break;
 			case LK_BELL_ENABLE:	/*  0x23  */
 				break;
+			case 0x41:
+				d->add_to_rx_queue(d->add_data, 0xa, DCKBD_PORT);
+				d->add_to_rx_queue(d->add_data, 0xb, DCKBD_PORT);
+				d->add_to_rx_queue(d->add_data, 0xc, DCKBD_PORT);
+				d->add_to_rx_queue(d->add_data, 0xd, DCKBD_PORT);
+				break;
 			case LED_1:
 			case LED_2:
 				break;
@@ -318,12 +313,11 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
 				 *	3 = LK443
 				 *	4 = LK421
 				 *
-				 *  TODO: What about the second byte?
+				 *  TODO: What about the second byte? 0x22 or
+				 *  0x00?
 				 */
-				d->add_to_rx_queue(d->add_data,
-				    0x01, DCKBD_PORT);
-				d->add_to_rx_queue(d->add_data,
-				    0x22, DCKBD_PORT);
+				d->add_to_rx_queue(d->add_data, 0x01, DCKBD_PORT);
+				d->add_to_rx_queue(d->add_data, 0x00, DCKBD_PORT);
 				break;
 			case LK_DEFAULTS:	/*  0xd3  */
 				/*  TODO?  */
@@ -341,16 +335,18 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
 				    0x00, DCKBD_PORT);
 				break;
 			default:
-				debug("[ lk201: keyboard control: 0x%x ]\n",
+				debug("[ lk201: unimplemented keyboard control: 0x%x ]\n",
 				    idata);
 			}
 		}
 		break;
 	case DCMOUSE_PORT:		/*  port 1  */
-		debug("[ lk201: writing data to MOUSE: 0x%x", idata);
-		if (idata == MOUSE_INCREMENTAL) {
+		debug("[ lk201: writing data to MOUSE: 0x%x ", idata);
+		switch (idata) {
+		case MOUSE_INCREMENTAL:
 			d->mouse_mode = MOUSE_INCREMENTAL;
-		} else if (idata == MOUSE_SELF_TEST) {
+			break;
+		case MOUSE_SELF_TEST:
 			/*
 			 *  Mouse self-test:
 			 *
@@ -359,14 +355,16 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
 			 *        0x2, according to NetBSD/pmax. But the
 			 *        other bits and bytes?
 			 */
-			debug(" (mouse self-test request)");
+			debug("(mouse self-test request)");
 			d->add_to_rx_queue(d->add_data,
 			    0xa0 | d->mouse_revision, DCMOUSE_PORT);
 			d->add_to_rx_queue(d->add_data, 0x02, DCMOUSE_PORT);
 			d->add_to_rx_queue(d->add_data, 0x00, DCMOUSE_PORT);
 			d->add_to_rx_queue(d->add_data, 0x00, DCMOUSE_PORT);
-		} else
-			debug(" UNKNOWN byte; TODO");
+			break;
+		default:
+			debug("UNKNOWN byte; TODO");
+		}
 		debug(" ]\n");
 		break;
 	case DCCOMM_PORT:		/*  port 2  */
@@ -383,11 +381,13 @@ void lk201_tx_data(struct lk201_data *d, int port, int idata)
  *  Initialize lk201 keyboard/mouse settings.
  */
 void lk201_init(struct lk201_data *d, int use_fb,
+	int (*space_available_in_queue)(void *,int),
 	void (*add_to_rx_queue)(void *,int,int),
 	int console_handle, void *add_data)
 {
 	memset(d, 0, sizeof(struct lk201_data));
 
+	d->space_available_in_queue = space_available_in_queue;
 	d->add_to_rx_queue = add_to_rx_queue;
 	d->add_data = add_data;
 
@@ -395,5 +395,17 @@ void lk201_init(struct lk201_data *d, int use_fb,
 	d->mouse_mode = 0;
 	d->mouse_revision = 0;	/*  0..15  */
 	d->console_handle = console_handle;
+
+	/*
+	 *  Power up self-test result, as per
+	 *  https://www.netbsd.org/docs/Hardware/Machines/DEC/lk201.html#power_up
+	 */
+	if (d->use_fb) {
+		d->add_to_rx_queue(d->add_data, 0x01, DCKBD_PORT);
+		d->add_to_rx_queue(d->add_data, 0x00, DCKBD_PORT);
+		d->add_to_rx_queue(d->add_data, 0x00, DCKBD_PORT);
+		d->add_to_rx_queue(d->add_data, 0x00, DCKBD_PORT);
+	}
 }
 
+
