@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2006-2008  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2006-2009  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -25,8 +25,6 @@
  *  SUCH DAMAGE.
  *
  *
- *  $Id: bootblock_apple.c,v 1.1.2.1 2008/01/18 19:12:31 debug Exp $
- *
  *  Apple bootblock handling.
  *
  *  TODO: This is just a quick hack skeleton. Doesn't really work yet.
@@ -57,7 +55,6 @@ int apple_load_bootblock(struct machine *m, struct cpu *cpu,
 {
 	unsigned char buf[0x8000];
 	int res, partnr, n_partitions = 0, n_hfs_partitions = 0;
-	uint64_t hfs_start, hfs_length;
 
 	res = diskimage_access(m, disk_id, disk_type, 0, 0x0, buf, sizeof(buf));
 	if (!res) {
@@ -83,8 +80,9 @@ int apple_load_bootblock(struct machine *m, struct cpu *cpu,
 
 		if (strcmp((char *)buf + ofs + 0x30, "Apple_HFS") == 0) {
 			n_hfs_partitions ++;
-			hfs_start = 512 * start;
-			hfs_length = 512 * length;
+			// uint64_t hfs_start = 512 * start;
+			// uint64_t hfs_length = 512 * length;
+			// TODO.
 		}
 
 		/*  Any more partitions?  */
