@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -27,9 +27,9 @@
  *
  *  ISO9660 CD-ROM "bootblock" handling.
  *
- *  There is really no bootblock; instead, the file which is to be booted
- *  is extracted into a temporary file, and started as if it was given
- *  as a normal file argument on the command line.
+ *  Despite the name, there is actually no bootblock; instead, the file which
+ *  is to be booted is extracted into a temporary file, and started as if it
+ *  was given as a normal file argument on the command line.
  *
  *  TODO: This is really ugly. It's a quick hack. All the magic constants
  *        need to be replaced with real code!
@@ -50,6 +50,7 @@
 #include "machine.h"
 #include "misc.h"
 
+
 /*  #define ISO_DEBUG  */
 
 
@@ -99,12 +100,13 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 	int disk_id, int disk_type, int iso_type, unsigned char *buf,
 	int *n_loadp, char ***load_namesp)
 {
-	int filenr, dirlen, res = 0, res2, iadd = DEBUG_INDENTATION, found_dir;
-	uint64_t dirofs, fileofs, filelen;
+	int filenr, dirlen, res = 0, res2, found_dir;
+	uint64_t dirofs, fileofs;
+	ssize_t filelen;
 	unsigned char *dirbuf = NULL, *dp, *match_entry = NULL, *filebuf = NULL;
 	char *p, *filename_orig, *filename, *tmpfname = NULL;
 	char **new_array;
-	char *tmpdir = getenv("TMPDIR");
+	const char *tmpdir = getenv("TMPDIR");
 	int tmpfile_handle;
 
 	if (tmpdir == NULL)
@@ -114,11 +116,49 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 	filename_orig = filename;
 
 	debug("ISO9660 boot:\n");
-	debug_indentation(iadd);
+	debug_indentation(1);
 
 	debug_print_volume_id_and_filename(iso_type, buf, filename);
 
 
+	/*
+	 *  Hack: If there is something like this at offset 0x9000:
+	 *
+	 *  "MKI Wed Nov 29 17:56:55 2000\n"
+	 *  "mkisofs 1.13 -r -l -C 0,11702 /home/LC2000-CDR"
+	 *
+	 *  and the disk offset is zero, then it can PROBABLY be assumed that
+	 *  offset 11702 is better. (This is counted in 2048 byte sectors.)
+	 */
+	{
+		unsigned char tmpbuf[0x800];
+		uint64_t base_offset = diskimage_get_baseoffset(m, disk_id, disk_type);
+		res2 = diskimage_access(m, disk_id, disk_type, 0, base_offset + 0x9000, tmpbuf, 0x800);
+		if (!res2) {
+			fatal("Couldn't read MKI part after the ISO header of the disk image. Aborting.\n");
+			goto ret;
+		}
+
+		int64_t suggestedBaseOffset = 0;
+		tmpbuf[sizeof(tmpbuf)-1] = '\0';
+		char* found = strstr((char*)tmpbuf, "MKI ");
+		if (found != NULL) {
+			found = strstr(found, " -C 0,");
+			if (found != NULL) {
+				suggestedBaseOffset = strtoll(found + 6, NULL, 0);
+			}
+		}
+
+		int64_t currentBaseOffset = diskimage_get_baseoffset(m, disk_id, disk_type);
+		suggestedBaseOffset *= 2048;
+		if (currentBaseOffset == 0 && suggestedBaseOffset != currentBaseOffset) 
+		{
+			debug("NOTE: automagically adjusting to use base offset %lli\n", (long long)suggestedBaseOffset);
+			diskimage_set_baseoffset(m, disk_id, disk_type, suggestedBaseOffset);
+		}
+	}
+	
+
 	/*
 	 *  Traverse the directory structure to find the kernel.
 	 */
@@ -134,7 +174,7 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 	debug("root = %i bytes at 0x%llx\n", dirlen, (long long)dirofs);
 #endif
 
-	CHECK_ALLOCATION(dirbuf = malloc(dirlen));
+	CHECK_ALLOCATION(dirbuf = (unsigned char *) malloc(dirlen));
 	res2 = diskimage_access(m, disk_id, disk_type, 0, dirofs, dirbuf,
 	    dirlen);
 	if (!res2) {
@@ -191,7 +231,7 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 
 		dp += nlen;
 
-		/*  16-bit aligned lenght:  */
+		/*  16-bit aligned length:  */
 		if (nlen & 1)
 			dp ++;
 
@@ -205,7 +245,7 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 	if (p != NULL) {
 		char *blah = filename_orig;
 
-		fatal("could not find '%s' in /", filename);
+		fatal("could not find '%s' (1) in /", filename);
 
 		/*  Print the first part of the filename:  */
 		while (blah != filename)
@@ -219,7 +259,7 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 
 	/*  Free the old dirbuf, and allocate a new one:  */
 	free(dirbuf);
-	CHECK_ALLOCATION(dirbuf = malloc(512));
+	CHECK_ALLOCATION(dirbuf = (unsigned char *) malloc(512));
 
 	for (;;) {
 		size_t len, i;
@@ -230,6 +270,7 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 			/*  debug("realign dirofs = 0x%llx\n", dirofs);  */
 		}
 
+		// debug("dirofs = %lli\n", (long long)dirofs);
 		res2 = diskimage_access(m, disk_id, disk_type, 0, dirofs,
 		    dirbuf, 256);
 		if (!res2) {
@@ -239,8 +280,10 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 
 		dp = dirbuf;
 		len = dp[0];
-		if (len < 2)
+		if (len < 2) {
+			// debug("dir too short\n");
 			break;
+		}
 
 		/*
 		 *  TODO: Actually parse the directory entry!
@@ -248,15 +291,18 @@ int iso_load_bootblock(struct machine *m, struct cpu *cpu,
 		 *  Haha, this must be rewritten.
 		 */
 
-#if 0
+#ifdef ISO_DEBUG
 /*  hahahaha  */
 printf("filename = '%s'\n", filename);
 {
-int j;
+size_t j;
 for (j=32; j<len; j++)
 	printf("%c", dp[j] >= ' ' && dp[j] < 128? dp[j] : '.');
 printf("\n");
 }
+
+// Hahaha. Yes, it's horrible. (Updated 2011-06-09.)
+// Yes it is. (2018-06-18.) :-(
 #endif
 
 		for (i=32; i<len; i++) {
@@ -272,7 +318,7 @@ printf("\n");
 						    "now... (BUG)\n");
 						exit(1);
 					}
-					CHECK_ALLOCATION(match_entry =
+					CHECK_ALLOCATION(match_entry = (unsigned char *)
 					    malloc(512));
 					memcpy(match_entry, dp, 512);
 					break;
@@ -288,7 +334,7 @@ printf("\n");
 	if (match_entry == NULL) {
 		char *blah = filename_orig;
 
-		fatal("could not find '%s' in /", filename);
+		fatal("could not find '%s' (2) in /", filename);
 
 		/*  Print the first part of the filename:  */
 		while (blah != filename)
@@ -307,9 +353,9 @@ printf("\n");
 	/*  debug("filelen=%llx fileofs=%llx\n", (long long)filelen,
 	    (long long)fileofs);  */
 
-	CHECK_ALLOCATION(filebuf = malloc(filelen));
+	CHECK_ALLOCATION(filebuf = (unsigned char *) malloc(filelen));
 
-	CHECK_ALLOCATION(tmpfname = malloc(300));
+	CHECK_ALLOCATION(tmpfname = (char *) malloc(300));
 	snprintf(tmpfname, 300, "%s/gxemul.XXXXXXXXXXXX", tmpdir);
 
 	res2 = diskimage_access(m, disk_id, disk_type, 0, fileofs, filebuf,
@@ -322,13 +368,13 @@ printf("\n");
 	tmpfile_handle = mkstemp(tmpfname);
 	if (tmpfile_handle < 0) {
 		fatal("could not create %s\n", tmpfname);
-		exit(1);
+		goto ret;
 	}
 
 	if (write(tmpfile_handle, filebuf, filelen) != filelen) {
 		fatal("could not write to %s\n", tmpfname);
 		perror("write");
-		exit(1);
+		goto ret;
 	}
 
 	close(tmpfile_handle);
@@ -337,14 +383,14 @@ printf("\n");
 
 	/*  Add the temporary filename to the load_namesp array:  */
 	(*n_loadp)++;
-	CHECK_ALLOCATION(new_array = malloc(sizeof(char *) * (*n_loadp)));
+	CHECK_ALLOCATION(new_array = (char **) malloc(sizeof(char *) * (*n_loadp)));
 	memcpy(new_array, *load_namesp, sizeof(char *) * (*n_loadp));
 	*load_namesp = new_array;
 
 	/*  This adds a Backspace char in front of the filename; this
 	    is a special hack which causes the file to be removed once
 	    it has been loaded.  */
-	CHECK_ALLOCATION(tmpfname = realloc(tmpfname, strlen(tmpfname) + 2));
+	CHECK_ALLOCATION(tmpfname = (char *) realloc(tmpfname, strlen(tmpfname) + 2));
 	memmove(tmpfname + 1, tmpfname, strlen(tmpfname) + 1);
 	tmpfname[0] = 8;
 
@@ -367,7 +413,7 @@ ret:
 
 	free(filename_orig);
 
-	debug_indentation(-iadd);
+	debug_indentation(-1);
 	return res;
 }
 
