@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -29,8 +29,11 @@
  *
  *  TODO:  diskimage_remove()? This would be useful for floppies in PC-style
  *	   machines, where disks may need to be swapped during boot etc.
+ *
+ *  TODO:  expose do_fsync as a command line option?
  */
 
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,11 +47,11 @@
 #include "misc.h"
 
 
+int do_fsync = 0;
+
 /*  #define debug fatal  */
 
-extern int single_step;
-
-static char *diskimage_types[] = DISKIMAGE_TYPES;
+static const char *diskimage_types[] = DISKIMAGE_TYPES;
 
 
 /**************************************************************************/
@@ -122,37 +125,62 @@ int diskimage_exist(struct machine *machine, int id, int type)
  *  Opens an overlay data file and its corresponding bitmap file, and adds
  *  the overlay to a disk image.
  */
-void diskimage_add_overlay(struct diskimage *d, char *overlay_basename)
+bool diskimage_add_overlay(struct diskimage *d, char *overlay_basename, bool remove_after_open)
 {
 	struct diskimage_overlay overlay;
 	size_t bitmap_name_len = strlen(overlay_basename) + 20;
 	char *bitmap_name;
 
-	CHECK_ALLOCATION(bitmap_name = malloc(bitmap_name_len));
+	CHECK_ALLOCATION(bitmap_name = (char *) malloc(bitmap_name_len));
 	snprintf(bitmap_name, bitmap_name_len, "%s.map", overlay_basename);
 
 	CHECK_ALLOCATION(overlay.overlay_basename = strdup(overlay_basename));
 	overlay.f_data = fopen(overlay_basename, d->writable? "r+" : "r");
 	if (overlay.f_data == NULL) {
 		perror(overlay_basename);
-		exit(1);
+
+		if (remove_after_open) {
+			unlink(overlay_basename);
+			unlink(bitmap_name);
+		}
+
+		free(bitmap_name);
+		return false;
 	}
 
 	overlay.f_bitmap = fopen(bitmap_name, d->writable? "r+" : "r");
 	if (overlay.f_bitmap == NULL) {
 		perror(bitmap_name);
 		fprintf(stderr, "Please create the map file first.\n");
-		exit(1);
+		fclose(overlay.f_data);
+
+		if (remove_after_open) {
+			unlink(overlay_basename);
+			unlink(bitmap_name);
+		}
+
+		free(bitmap_name);
+		return false;
 	}
 
 	d->nr_of_overlays ++;
 
-	CHECK_ALLOCATION(d->overlays = realloc(d->overlays,
+	CHECK_ALLOCATION(d->overlays = (struct diskimage_overlay *) realloc(d->overlays,
 	    sizeof(struct diskimage_overlay) * d->nr_of_overlays));
 
 	d->overlays[d->nr_of_overlays - 1] = overlay;
 
+	if (remove_after_open) {
+		if (unlink(overlay_basename) != 0)
+			perror(overlay_basename);
+
+		if (unlink(bitmap_name) != 0)
+			perror(bitmap_name);
+	}
+
 	free(bitmap_name);
+
+	return true;
 }
 
 
@@ -162,18 +190,15 @@ void diskimage_add_overlay(struct diskimage *d, char *overlay_basename)
  *  Recalculate a disk's size by stat()-ing it.
  *  d is assumed to be non-NULL.
  */
-void diskimage_recalc_size(struct diskimage *d)
+bool diskimage_recalc_size(struct diskimage *d)
 {
 	struct stat st;
 	int res;
-	off_t size = 0;
+	int64_t size = 0;
 
 	res = stat(d->fname, &st);
-	if (res) {
-		fprintf(stderr, "[ diskimage_recalc_size(): could not stat "
-		    "'%s' ]\n", d->fname);
-		return;
-	}
+	if (res)
+		return false;
 
 	size = st.st_size;
 
@@ -187,10 +212,49 @@ void diskimage_recalc_size(struct diskimage *d)
 		size = 762048000;
 
 	d->total_size = size;
-	d->ncyls = d->total_size / 1048576;
 
-	/*  TODO: There is a mismatch between d->ncyls and d->cylinders,
-	    SCSI-based stuff usually doesn't care.  TODO: Fix this.  */
+	if ((d->total_size == 720*1024 || d->total_size == 1474560
+	    || d->total_size == 2949120 || d->total_size == 1228800)
+	    && d->type == DISKIMAGE_UNKNOWN)
+		d->type = DISKIMAGE_FLOPPY;
+
+	switch (d->type) {
+	case DISKIMAGE_FLOPPY:
+		if (d->total_size < 737280) {
+			debugmsg(SUBSYS_DISK, "", VERBOSITY_WARNING,
+			    "TODO: small (non-80-cylinder) floppies?");
+		}
+
+		if (!d->chs_override) {
+			d->cylinders = 80;
+			d->heads = 2;
+			d->sectors_per_track = d->nr_of_logical_blocks * d->logical_block_size
+				/ (d->cylinders * d->heads * 512);
+		}
+		break;
+
+	default:/*  Non-floppies:  */
+		if (!d->chs_override) {
+			d->heads = 16;
+			d->sectors_per_track = 63;
+
+			int64_t bytespercyl = d->heads * d->sectors_per_track * 512;
+			d->cylinders = size / bytespercyl;
+			if (d->cylinders * bytespercyl < size)
+				d->cylinders ++;
+		}
+	}
+
+	// printf("c=%i h=%i s=%i\n", d->cylinders, d->heads, d->sectors_per_track);
+
+	// Update size to full cylinders.
+	size = d->heads * d->sectors_per_track * d->cylinders * 512;
+
+	d->nr_of_logical_blocks = size / d->logical_block_size;
+	if (size & (d->logical_block_size - 1))
+		d->nr_of_logical_blocks ++;
+
+	return true;
 }
 
 
@@ -206,7 +270,7 @@ int64_t diskimage_getsize(struct machine *machine, int id, int type)
 
 	while (d != NULL) {
 		if (d->type == type && d->id == id)
-			return d->total_size;
+			return d->nr_of_logical_blocks * d->logical_block_size;
 		d = d->next;
 	}
 	return -1;
@@ -226,12 +290,38 @@ int64_t diskimage_get_baseoffset(struct machine *machine, int id, int type)
 	while (d != NULL) {
 		if (d->type == type && d->id == id)
 			return d->override_base_offset;
+
 		d = d->next;
 	}
 	return -1;
 }
 
 
+/*
+ *  diskimage_set_baseoffset():
+ *
+ *  Sets the base offset for a disk image. Useful e.g. when booting directly
+ *  from NetBSD/dreamcast or Linux/dreamcast ISO images.
+ */
+void diskimage_set_baseoffset(struct machine *machine, int id, int type, int64_t offset)
+{
+	struct diskimage *d = machine->first_diskimage;
+
+	while (d != NULL) {
+		if (d->type == type && d->id == id) {
+			d->override_base_offset = offset;
+			return;
+		}
+		
+		d = d->next;
+	}
+
+	fatal("diskimage_set_baseoffset(): disk id %i (type %i) not found?\n",
+	    id, diskimage_types[type]);
+	exit(1);
+}
+
+
 /*
  *  diskimage_getchs():
  *
@@ -347,6 +437,9 @@ static void overlay_set_block_in_use(struct diskimage *d,
 		fprintf(stderr, "Could not write to bitmap file. Aborting.\n");
 		exit(1);
 	}
+	
+	if (do_fsync)
+		fsync(fileno(d->overlays[overlay_nr].f_bitmap));
 }
 
 
@@ -395,7 +488,12 @@ static size_t fwrite_helper(off_t offset, unsigned char *buf,
 			return 0;
 		}
 
-		return fwrite(buf, 1, len, d->f);
+		size_t written = fwrite(buf, 1, len, d->f);
+
+		if (do_fsync)
+			fsync(fileno(d->f));
+
+		return written;
 	}
 
 	if ((len & (OVERLAY_BLOCK_SIZE-1)) != 0) {
@@ -419,15 +517,24 @@ static size_t fwrite_helper(off_t offset, unsigned char *buf,
 		int res = my_fseek(d->overlays[overlay_nr].f_data,
 		    curofs, SEEK_SET);
 		if (res != 0) {
-			fatal("[ diskimage__internal_access(): fseek()"
-			    " failed on disk id %i \n", d->id);
+			fatal("[ diskimage: fwrite_helper(): fseek()"
+			    " failed on disk id %i ]\n", d->id);
 			return 0;
 		}
 
 		lenwritten = fwrite(buf, 1, OVERLAY_BLOCK_SIZE,
 		    d->overlays[overlay_nr].f_data);
+		if (lenwritten != OVERLAY_BLOCK_SIZE) {
+			fatal("[ diskimage: fwrite_helper(): fwrite to"
+			    " overlay failed on disk id %i ]\n", d->id);
+			exit(1);
+		}
+
 		buf += OVERLAY_BLOCK_SIZE;
 
+		if (do_fsync)
+			fsync(fileno(d->overlays[overlay_nr].f_data));
+
 		/*  Mark this block in the last overlay as in use:  */
 		overlay_set_block_in_use(d, overlay_nr, curofs);
 	}
@@ -548,23 +655,26 @@ int diskimage__internal_access(struct diskimage *d, int writeflag,
 		else
 			lendone = fread_helper(offset, buf, len, d);
 
-		if (lendone < (ssize_t)len)
+		if (lendone >= 0 && lendone < (ssize_t)len)
 			memset(buf + lendone, 0, len - lendone);
 	}
 
-	/*  Incomplete data transfer? Then return failure:  */
-	if (lendone != (ssize_t)len) {
-#ifdef UNSTABLE_DEVEL
-		fatal
-#else
-		debug
-#endif
-		    ("[ diskimage__internal_access(): disk_id %i, offset %lli"
+	/*
+	 *  Incomplete data transfer?
+	 *
+	 *  If we could not read anything at all, then return failure.
+	 *  If we could read a partial block, then return success (with
+	 *  the resulting buffer zero-filled at the end).
+	 */
+#if 0
+// TODO: check against full cylinder-size instead
+	if (lendone <= 0) {
+		fatal("[ diskimage__internal_access(): disk_id %i, offset %lli"
 		    ", transfer not completed. len=%i, len_done=%i ]\n",
 		    d->id, (long long)offset, (int)len, (int)lendone);
 		return 0;
 	}
-
+#endif
 	return 1;
 }
 
@@ -606,6 +716,19 @@ int diskimage_access(struct machine *machine, int id, int type, int writeflag,
 }
 
 
+int get_default_disk_type_for_machine(struct machine *machine)
+{
+	if (machine->machine_type == MACHINE_PMAX ||
+	    machine->machine_type == MACHINE_ARC ||
+	    machine->machine_type == MACHINE_SGI ||
+	    machine->machine_type == MACHINE_LUNA88K ||
+	    machine->machine_type == MACHINE_MVME88K)
+		return DISKIMAGE_SCSI;
+
+	return DISKIMAGE_IDE;
+}
+
+
 /*
  *  diskimage_add():
  *
@@ -628,21 +751,25 @@ int diskimage_access(struct machine *machine, int id, int type, int writeflag,
  *	0-7	force a specific SCSI ID number
  *
  *  machine is assumed to be non-NULL.
- *  Returns an integer >= 0 identifying the disk image.
+ *
+ *  Returns an integer >= 0 identifying the disk image. On error, a value less
+ *  than zero is returned.
  */
 int diskimage_add(struct machine *machine, char *fname)
 {
 	struct diskimage *d, *d2;
 	int id = 0, override_heads=0, override_spt=0;
-	int64_t bytespercyl, override_base_offset=0;
+	int64_t override_base_offset=0;
 	char *cp;
 	int prefix_b=0, prefix_c=0, prefix_d=0, prefix_f=0, prefix_g=0;
 	int prefix_i=0, prefix_r=0, prefix_s=0, prefix_t=0, prefix_id=-1;
 	int prefix_o=0, prefix_V=0;
+	bool prefix_R = false;
 
 	if (fname == NULL) {
-		fprintf(stderr, "diskimage_add(): NULL ptr\n");
-		return 0;
+		debugmsg(SUBSYS_DISK, "diskimage_add()", VERBOSITY_ERROR,
+		    "NULL ptr");
+		return -1;
 	}
 
 	/*  Get prefix from fname:  */
@@ -691,7 +818,7 @@ int diskimage_add(struct machine *machine, char *fname)
 					fatal("Bad geometry: heads=%i "
 					    "spt=%i\n", override_heads,
 					    override_spt);
-					exit(1);
+					return -1;
 				}
 				break;
 			case 'i':
@@ -706,11 +833,14 @@ int diskimage_add(struct machine *machine, char *fname)
 				if (*fname == ':' || *fname == ';')
 					fname ++;
 				if (override_base_offset < 0) {
-					fatal("Bad base offset: %"PRIi64
+					fatal("Bad base offset: %" PRIi64
 					    "\n", override_base_offset);
-					exit(1);
+					return -1;
 				}
 				break;
+			case 'R':
+				prefix_R = true;
+				break;
 			case 'r':
 				prefix_r = 1;
 				break;
@@ -728,24 +858,15 @@ int diskimage_add(struct machine *machine, char *fname)
 			default:
 				fprintf(stderr, "diskimage_add(): invalid "
 				    "prefix char '%c'\n", c);
-				exit(1);
+				return -1;
 			}
 		}
 	}
 
 	/*  Allocate a new diskimage struct:  */
-	CHECK_ALLOCATION(d = malloc(sizeof(struct diskimage)));
+	CHECK_ALLOCATION(d = (struct diskimage *) malloc(sizeof(struct diskimage)));
 	memset(d, 0, sizeof(struct diskimage));
 
-	/*  Default to IDE disks...  */
-	d->type = DISKIMAGE_IDE;
-
-	/*  ... but some machines use SCSI by default:  */
-	if (machine->machine_type == MACHINE_PMAX ||
-	    machine->machine_type == MACHINE_ARC ||
-	    machine->machine_type == MACHINE_MVME88K)
-		d->type = DISKIMAGE_SCSI;
-
 	if (prefix_i + prefix_f + prefix_s > 1) {
 		fprintf(stderr, "Invalid disk image prefix(es). You can"
 		    "only use one of i, f, and s\nfor each disk image.\n");
@@ -759,34 +880,50 @@ int diskimage_add(struct machine *machine, char *fname)
 	if (prefix_s)
 		d->type = DISKIMAGE_SCSI;
 
+	if (prefix_V && prefix_R) {
+		debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+		    "'V' and 'R' prefixes can not be combined!");
+		return -1;
+	}
+
 	/*  Special case: Add an overlay for an already added disk image:  */
 	if (prefix_V) {
 		struct diskimage *dx = machine->first_diskimage;
 
 		if (prefix_id < 0) {
-			fprintf(stderr, "The 'V' disk image prefix requires"
-			    " a disk ID to also be supplied.\n");
-			exit(1);
+			debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+			    "The 'V' disk image prefix requires"
+			    " a disk ID to also be supplied.");
+			return -1;
 		}
 
+		if (d->type == DISKIMAGE_UNKNOWN)
+			d->type = get_default_disk_type_for_machine(machine);
+
 		while (dx != NULL) {
 			if (d->type == dx->type && prefix_id == dx->id)
 				break;
 			dx = dx->next;
 		}
 
-		if (dx == NULL) {
-			fprintf(stderr, "Bad ID supplied for overlay?\n");
-			exit(1);
-		}
-
-		diskimage_add_overlay(dx, fname);
-
 		/*  Free the preliminary d struct:  */
 		free(d);
 
-		/*  Don't add any disk image. This is an overlay!  */
-		return -1;
+		if (dx == NULL) {
+			fprintf(stderr, "Bad ID supplied for overlay?\n");
+			return -1;
+		}
+
+		if (!diskimage_add_overlay(dx, fname, false)) {
+			debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+			    "could not add overlay name '%s'",
+			    fname);
+			return -1;
+		}
+
+		/*  Return the ID of the disk image, even though no disk
+		    image was added.  */
+		return dx->id;
 	}
 
 	/*  Add the new disk image in the disk image chain:  */
@@ -845,64 +982,76 @@ int diskimage_add(struct machine *machine, char *fname)
 		}
 	}
 
-	diskimage_recalc_size(d);
-
-	if ((d->total_size == 720*1024 || d->total_size == 1474560
-	    || d->total_size == 2949120 || d->total_size == 1228800)
-	    && !prefix_i && !prefix_s)
-		d->type = DISKIMAGE_FLOPPY;
-
-	switch (d->type) {
-	case DISKIMAGE_FLOPPY:
-		if (d->total_size < 737280) {
-			fatal("\nTODO: small (non-80-cylinder) floppies?\n\n");
-			exit(1);
-		}
-		d->cylinders = 80;
-		d->heads = 2;
-		d->sectors_per_track = d->total_size / (d->cylinders *
-		    d->heads * 512);
-		break;
-	default:/*  Non-floppies:  */
-		d->heads = 16;
-		d->sectors_per_track = 63;
-		if (prefix_g) {
-			d->chs_override = 1;
-			d->heads = override_heads;
-			d->sectors_per_track = override_spt;
-		}
-		bytespercyl = d->heads * d->sectors_per_track * 512;
-		d->cylinders = d->total_size / bytespercyl;
-		if (d->cylinders * bytespercyl < d->total_size)
-			d->cylinders ++;
+	if (prefix_g) {
+		d->chs_override = 1;
+		d->heads = override_heads;
+		d->sectors_per_track = override_spt;
 	}
 
+	if (!diskimage_recalc_size(d))
+		debugmsg(SUBSYS_DISK, "", VERBOSITY_WARNING,
+		    "diskimage_recalc_size() failed!");
+
+	if (d->type == DISKIMAGE_UNKNOWN)
+		d->type = get_default_disk_type_for_machine(machine);
+
 	d->rpms = 3600;
 
 	if (prefix_b)
 		d->is_boot_device = 1;
 
+	bool readable = access(fname, R_OK) == 0? 1 : 0;
 	d->writable = access(fname, W_OK) == 0? 1 : 0;
 
-	if (d->is_a_cdrom || prefix_r)
-		d->writable = 0;
+	if (!readable) {
+		debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+		    "could not access '%s': %s", fname, strerror(errno));
+		return -1;
+	}
 
-	d->f = fopen(fname, d->writable? "r+" : "r");
+	if (d->is_a_cdrom || prefix_r) {
+		d->writable = 0;
+	} else if (!d->writable) {
+		if (prefix_R) {
+			d->writable = 1;
+		} else {
+			debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+			    "'%s' is read-only in the host file system.\n"
+			    "Use the r: prefix to indicate that the file should be treated\n"
+			    "as read-only, the R: prefix to use a temporary overlay (allowing\n"
+			    "guest operating systems to write to that temporary overlay), or\n"
+			    "make the file writable.",
+			    d->fname);
+			return -1;
+		}
+	}
+
+	d->f = fopen(fname, d->writable && !prefix_R ? "r+" : "r");
 	if (d->f == NULL) {
-		char *errmsg = malloc(200 + strlen(fname));
-		snprintf(errmsg, 200+strlen(fname),
-		    "could not fopen %s for reading%s", fname,
-		    d->writable? " and writing" : "");
-		perror(errmsg);
-		exit(1);
+		debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+		    "could not open '%s' for reading%s: %s",
+		    fname,
+		    d->writable? " and writing" : "",
+		    strerror(errno));
+		return -1;
 	}
 
 	/*  Calculate which ID to use:  */
 	if (prefix_id == -1) {
-		int free = 0, collision = 1;
+		int start = 0;
+		int delta = 1;
+		
+		if (machine->machine_type == MACHINE_LUNA88K) {
+			start = 6;
+			delta = -1;
+		}
+		
+		int free = start;
+		
+		bool collision = true;
 
 		while (collision) {
-			collision = 0;
+			collision = false;
 			d2 = machine->first_diskimage;
 			while (d2 != NULL) {
 				/*  (don't compare against ourselves :)  */
@@ -911,7 +1060,7 @@ int diskimage_add(struct machine *machine, char *fname)
 					continue;
 				}
 				if (d2->id == free && d2->type == d->type) {
-					collision = 1;
+					collision = true;
 					break;
 				}
 				d2 = d2->next;
@@ -919,7 +1068,12 @@ int diskimage_add(struct machine *machine, char *fname)
 			if (!collision)
 				id = free;
 			else
-				free ++;
+				free += delta;
+		}
+		
+		if (id < 0) {
+			fprintf(stderr, "Too many IDs used?\n");
+			exit(1);
 		}
 	} else {
 		id = prefix_id;
@@ -941,6 +1095,39 @@ int diskimage_add(struct machine *machine, char *fname)
 
 	d->id = id;
 
+	if (prefix_R) {
+		char tmpname[1000];
+		const char* tmpdir = getenv("TMPDIR");
+		if (tmpdir == NULL)
+			tmpdir = "/tmp";
+		
+		int old_umask = umask(077);
+
+		// TODO: Not secure. Both the .img and the .img.map file
+		// need to be created by the process itself to avoid a race.
+		// NOTE/TODO: Ideally fopen mode "x" (from C11) for exclusive open.
+		snprintf(tmpname, sizeof(tmpname), "%s/gxemul.%i.temp-diskimage-overlay.%i.img.map",
+		    tmpdir, d->id, getpid());
+
+		FILE* f_img_map = fopen(tmpname, "w");
+		fclose(f_img_map);
+
+		snprintf(tmpname, sizeof(tmpname), "%s/gxemul.%i.temp-diskimage-overlay.%i.img",
+		    tmpdir, d->id, getpid());
+
+		FILE* f_img = fopen(tmpname, "w");
+		fclose(f_img);
+
+		umask(old_umask);
+
+		if (!diskimage_add_overlay(d, tmpname, true)) {
+			debugmsg(SUBSYS_DISK, "", VERBOSITY_ERROR,
+			    "could not add overlay name '%s'",
+			    fname);
+			return -1;
+		}
+	}
+
 	return id;
 }
 
@@ -1054,12 +1241,12 @@ int diskimage_is_a_tape(struct machine *machine, int id, int type)
  */
 void diskimage_dump_info(struct machine *machine)
 {
-	int i, iadd = DEBUG_INDENTATION;
 	struct diskimage *d = machine->first_diskimage;
 
 	while (d != NULL) {
-		debug("diskimage: %s\n", d->fname);
-		debug_indentation(iadd);
+		debugmsg(SUBSYS_MACHINE, "diskimage", VERBOSITY_INFO, "%s", d->fname);
+
+		debug_indentation(1);
 
 		switch (d->type) {
 		case DISKIMAGE_SCSI:
@@ -1080,28 +1267,31 @@ void diskimage_dump_info(struct machine *machine)
 		debug(" id %i, ", d->id);
 		debug("%s, ", d->writable? "read/write" : "read-only");
 
+		int64_t s = d->nr_of_logical_blocks * d->logical_block_size;
+
 		if (d->type == DISKIMAGE_FLOPPY)
-			debug("%lli KB", (long long) (d->total_size / 1024));
+			debug("%lli KB", (long long) (s / 1024));
 		else
-			debug("%lli MB", (long long) (d->total_size / 1048576));
+			debug("%lli MB", (long long) (s / 1048576));
 
 		if (d->type == DISKIMAGE_FLOPPY || d->chs_override)
 			debug(" (CHS=%i,%i,%i)", d->cylinders, d->heads,
 			    d->sectors_per_track);
 		else
-			debug(" (%lli sectors)", (long long)
-			   (d->total_size / 512));
+			debug(" (%lli %i-byte blocks)", (long long)d->nr_of_logical_blocks, d->logical_block_size);
 
 		if (d->is_boot_device)
 			debug(" (BOOT)");
 		debug("\n");
 
-		for (i=0; i<d->nr_of_overlays; i++) {
+		debug_indentation(1);
+
+		for (int i=0; i<d->nr_of_overlays; i++) {
 			debug("overlay %i: %s\n",
 			    i, d->overlays[i].overlay_basename);
 		}
 
-		debug_indentation(-iadd);
+		debug_indentation(-2);
 
 		d = d->next;
 	}
