@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -30,6 +30,9 @@
  *  TODO:  There are LOTS of ugly magic values in this module. These should
  *         be replaced by proper defines.
  *
+ *  TODO:  scsi_transfer_free could save away previous allocations, and
+ *         scsi_transfer_allocbuf could reuse them (for given sizes).
+ *
  *  TODO:  There's probably a bug in the tape support:
  *         Let's say there are 10240 bytes left in a file, and 10240
  *         bytes are read. Then feof() is not true yet (?), so the next
@@ -49,7 +52,7 @@
 #include "misc.h"
 
 
-static char *diskimage_types[] = DISKIMAGE_TYPES;
+static const char *diskimage_types[] = DISKIMAGE_TYPES;
 static struct scsi_transfer *first_free_scsi_transfer_alloc = NULL;
 
 
@@ -69,7 +72,7 @@ struct scsi_transfer *scsi_transfer_alloc(void)
 		p = first_free_scsi_transfer_alloc;
 		first_free_scsi_transfer_alloc = p->next_free;
 	} else {
-		p = malloc(sizeof(struct scsi_transfer));
+		p = (struct scsi_transfer *) malloc(sizeof(struct scsi_transfer));
 		if (p == NULL) {
 			fprintf(stderr, "scsi_transfer_alloc(): out "
 			    "of memory\n");
@@ -129,15 +132,17 @@ void scsi_transfer_allocbuf(size_t *lenp, unsigned char **pp, size_t want_len,
 	unsigned char *p = (*pp);
 
 	if (p != NULL) {
-		printf("WARNING! scsi_transfer_allocbuf(): old pointer "
-		    "was not NULL, freeing it now\n");
+		debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_WARNING,
+		    "WARNING! scsi_transfer_allocbuf(): old pointer "
+		    "was not NULL, freeing it now");
 		free(p);
 	}
 
 	(*lenp) = want_len;
-	if ((p = malloc(want_len)) == NULL) {
-		fprintf(stderr, "scsi_transfer_allocbuf(): out of "
-		    "memory trying to allocate %li bytes\n", (long)want_len);
+	if ((p = (unsigned char *) malloc(want_len)) == NULL) {
+		debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_ERROR,
+		    "scsi_transfer_allocbuf(): out of "
+		    "memory trying to allocate %li bytes", (long)want_len);
 		exit(1);
 	}
 
@@ -185,8 +190,10 @@ static void diskimage__switch_tape(struct diskimage *d)
 
 	d->f = fopen(tmpfname, d->writable? "r+" : "r");
 	if (d->f == NULL) {
-		fprintf(stderr, "[ diskimage__switch_tape(): could not "
-		    "(re)open '%s' ]\n", tmpfname);
+		debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_ERROR,
+		    "diskimage__switch_tape(): could not "
+		    "(re)open '%s'", tmpfname);
+
 		/*  TODO: return error  */
 	}
 	d->tape_offset = 0;
@@ -223,7 +230,8 @@ int diskimage_scsicommand(struct cpu *cpu, int id, int type,
 	struct diskimage *d;
 
 	if (machine == NULL) {
-		fatal("[ diskimage_scsicommand(): machine == NULL ]\n");
+		debugmsg_cpu(cpu, SUBSYS_DISK, "scsi",
+		    VERBOSITY_ERROR, "machine == NULL");
 		return 0;
 	}
 
@@ -233,24 +241,33 @@ int diskimage_scsicommand(struct cpu *cpu, int id, int type,
 			break;
 		d = d->next;
 	}
-	if (d == NULL) {
-		fprintf(stderr, "[ diskimage_scsicommand(): %s "
-		    " id %i not connected? ]\n", diskimage_types[type], id);
-	}
 
 	if (xferp->cmd == NULL) {
-		fatal("[ diskimage_scsicommand(): cmd == NULL ]\n");
+		debugmsg_cpu(cpu, SUBSYS_DISK, "scsi",
+		    VERBOSITY_ERROR, "cmd == NULL");
 		return 0;
 	}
 
 	if (xferp->cmd_len < 1) {
-		fatal("[ diskimage_scsicommand(): cmd_len == %i ]\n",
+		debugmsg_cpu(cpu, SUBSYS_DISK, "scsi",
+		    VERBOSITY_ERROR, "cmd_len = %i",
 		    xferp->cmd_len);
 		return 0;
 	}
 
-	debug("[ diskimage_scsicommand(id=%i) cmd=0x%02x: ",
-	    id, xferp->cmd[0]);
+	if (d == NULL) {
+		debugmsg_cpu(cpu, SUBSYS_DISK, "scsi",
+		    VERBOSITY_INFO, "%s id %i not connected!",
+		    diskimage_types[type], id);
+
+		// TODO: How to return suitable error?
+		diskimage__return_default_status_and_message(xferp);
+		xferp->status[0] = 0x02;	// ?
+		return 0;
+	}
+
+	// TODO: debugmsg:ify the rest of the debug messages...
+	debug("[ diskimage_scsicommand(id=%i) cmd=0x%02x: ", id, xferp->cmd[0]);
 
 #if 0
 	fatal("[ diskimage_scsicommand(id=%i) cmd=0x%02x len=%i:",
@@ -259,23 +276,7 @@ int diskimage_scsicommand(struct cpu *cpu, int id, int type,
 		fatal(" %02x", xferp->cmd[i]);
 	fatal("\n");
 if (xferp->cmd_len > 7 && xferp->cmd[5] == 0x11)
-	single_step = ENTER_SINGLE_STEPPING;
-#endif
-
-#if 0
-{
-	static FILE *f = NULL;
-	if (f == NULL)
-		f = fopen("scsi_log.txt", "w"); 
-	if (f != NULL) {
-		int i;
-		fprintf(f, "id=%i cmd =", id);
-		for (i=0; i<xferp->cmd_len; i++)
-			fprintf(f, " %02x", xferp->cmd[i]);
-		fprintf(f, "\n");
-		fflush(f);
-	}
-}
+	single_step = true;
 #endif
 
 	switch (xferp->cmd[0]) {
@@ -287,8 +288,9 @@ if (xferp->cmd_len > 7 && xferp->cmd[5] == 0x11)
 
 		/*  TODO: bits 765 of buf[1] contains the LUN  */
 		if (xferp->cmd[1] != 0x00)
-			fatal("WARNING: TEST_UNIT_READY with cmd[1]=0x%02x"
-			    " not yet implemented\n", (int)xferp->cmd[1]);
+			debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_WARNING,
+			    "WARNING: TEST_UNIT_READY with cmd[1]=0x%02x"
+			    " not yet implemented", (int)xferp->cmd[1]);
 
 		diskimage__return_default_status_and_message(xferp);
 		break;
@@ -298,16 +300,16 @@ if (xferp->cmd_len > 7 && xferp->cmd[5] == 0x11)
 		if (xferp->cmd_len != 6)
 			debug(" (weird len=%i)", xferp->cmd_len);
 		if (xferp->cmd[1] != 0x00) {
-			debug("WARNING: INQUIRY with cmd[1]=0x%02x not yet "
-			    "implemented\n", (int)xferp->cmd[1]);
-
-			break;
+			debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_WARNING,
+			    "WARNING: INQUIRY with cmd[1]=0x%02x not yet "
+			    "implemented", (int)xferp->cmd[1]);
 		}
 
 		/*  Return values:  */
 		retlen = xferp->cmd[4];
 		if (retlen < 36) {
-			fatal("WARNING: SCSI inquiry len=%i, <36!\n", retlen);
+			debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_WARNING,
+			    "WARNING: SCSI inquiry len=%i, <36!", retlen);
 			retlen = 36;
 		}
 
@@ -330,13 +332,14 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		memcpy(xferp->data_in+8,  "GXemul  ", 8);
 		if (diskimage_getname(cpu->machine, id,
 		    type, namebuf, sizeof(namebuf))) {
-			size_t i;
-			for (i=0; i<sizeof(namebuf); i++)
-				if (namebuf[i] == 0) {
-					for (; i<sizeof(namebuf); i++)
-						namebuf[i] = ' ';
+			for (size_t j=0; j<sizeof(namebuf); j++) {
+				if (namebuf[j] == 0) {
+					for (; j<sizeof(namebuf); j++)
+						namebuf[j] = ' ';
 					break;
 				}
+			}
+			
 			memcpy(xferp->data_in+16, namebuf, 16);
 		} else
 			memcpy(xferp->data_in+16, "DISK            ", 16);
@@ -424,9 +427,7 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 
 		diskimage_recalc_size(d);
 
-		size = d->total_size / d->logical_block_size;
-		if (d->total_size & (d->logical_block_size-1))
-			size ++;
+		size = d->nr_of_logical_blocks - 1;
 
 		xferp->data_in[0] = (size >> 24) & 255;
 		xferp->data_in[1] = (size >> 16) & 255;
@@ -526,11 +527,10 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		case 4:		/*  rigid disk geometry page  */
 			xferp->data_in[q + 0] = pagecode;
 			xferp->data_in[q + 1] = 22;
-			xferp->data_in[q + 2] = (d->ncyls >> 16) & 255;
-			xferp->data_in[q + 3] = (d->ncyls >> 8) & 255;
-			xferp->data_in[q + 4] = d->ncyls & 255;
+			xferp->data_in[q + 2] = (d->cylinders >> 16) & 255;
+			xferp->data_in[q + 3] = (d->cylinders >> 8) & 255;
+			xferp->data_in[q + 4] = d->cylinders & 255;
 			xferp->data_in[q + 5] = d->heads;
-
 			xferp->data_in[q + 20] = (d->rpms >> 8) & 255;
 			xferp->data_in[q + 21] = d->rpms & 255;
 			break;
@@ -550,15 +550,16 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 			    & 255;
 			xferp->data_in[q + 7] = d->logical_block_size & 255;
 
-			xferp->data_in[q + 8] = (d->ncyls >> 8) & 255;
-			xferp->data_in[q + 9] = d->ncyls & 255;
+			xferp->data_in[q + 8] = (d->cylinders >> 8) & 255;
+			xferp->data_in[q + 9] = d->cylinders & 255;
 
 			xferp->data_in[q + 28] = (d->rpms >> 8) & 255;
 			xferp->data_in[q + 29] = d->rpms & 255;
 			break;
 		default:
-			fatal("[ MODE_SENSE for page %i is not yet "
-			    "implemented! ]\n", pagecode);
+			debugmsg(SUBSYS_DISK, "scsi", VERBOSITY_WARNING,
+			    "MODE_SENSE for page %i is not yet "
+			    "implemented!", pagecode);
 		}
 
 		break;
@@ -667,9 +668,9 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 			xferp->status[0] = 0x02;	/*  CHECK CONDITION  */
 
 			d->filemark = 1;
-		} else
-			diskimage__internal_access(d, 0, ofs,
-			    xferp->data_in, size);
+		} else {
+			/* int result = */  diskimage__internal_access(d, 0, ofs, xferp->data_in, size);
+		}
 
 		if (d->is_a_tape && d->f != NULL)
 			d->tape_offset = ftello(d->f);
@@ -726,19 +727,13 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 			return 2;
 		}
 
-		debug(", data_out != NULL, OK :-)");
-
 		debug("WRITE ofs=%i size=%i offset=%i\n", (int)ofs,
 		    (int)size, (int)xferp->data_out_offset);
 
-		diskimage__internal_access(d, 1, ofs,
-		    xferp->data_out, size);
+		/*  int result = */  diskimage__internal_access(d, 1, ofs, xferp->data_out, size);
 
 		/*  TODO: how about return code?  */
 
-		/*  Is this really necessary?  */
-		/*  fsync(fileno(d->f));  */
-
 		diskimage__return_default_status_and_message(xferp);
 		break;
 
@@ -749,6 +744,9 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 			debug(" (weird len=%i)", xferp->cmd_len);
 
 		/*  TODO: actualy care about cmd[]  */
+		/*  TODO: Move to diskimage.cc and make sure that both
+			d->f is fsynced AND any overlays and overlay bitmap
+			files are fsynced too!  */
 		fsync(fileno(d->f));
 
 		diskimage__return_default_status_and_message(xferp);
@@ -940,9 +938,7 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 
 		diskimage_recalc_size(d);
 
-		size = d->total_size / d->logical_block_size;
-		if (d->total_size & (d->logical_block_size-1))
-			size ++;
+		size = d->nr_of_logical_blocks;
 
 		xferp->data_in[0] = (size >> 24) & 255;
 		xferp->data_in[1] = (size >> 16) & 255;
@@ -986,7 +982,7 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		break;
 
 	case SCSICDROM_READ_DISCINFO:
-		/*  (Patch from Hvard Eidnes.)  */
+		/*  (Patch from Haavard Eidnes.)  */
 		debug("CDROM_READ_DISCINFO, cmd[1]=0x%02x", xferp->cmd[1]);
 		retlen = 34;
 		scsi_transfer_allocbuf(&xferp->data_in_len,
@@ -1005,18 +1001,16 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		xferp->data_in[9] = 1;   /*  num sessions, msb  */
 		xferp->data_in[10] = 0;  /*  first_track_last_session_msb  */
 		xferp->data_in[11] = 0;  /*  last_track_last_session_msb  */
-		{
-			int i;
-			/*  Lead-in data, for completed cd-rom:  */
-			for (i=16; i<=23; i++)
-				xferp->data_in[i] = 0xff;
-		}
+
+		/*  Lead-in data, for completed cd-rom:  */
+		for (size_t j=16; j<=23; j++)
+			xferp->data_in[j] = 0xff;
 
 		diskimage__return_default_status_and_message(xferp);
 		break;
 
 	case SCSICDROM_READ_TRACKINFO:
-		/*  (Patch from Hvard Eidnes.)  */
+		/*  (Patch from Haavard Eidnes.)  */
 		debug("CDROM_READ_TRACKINFO");
 		retlen = 36;
 		scsi_transfer_allocbuf(&xferp->data_in_len,
@@ -1040,15 +1034,12 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		xferp->data_in[6] = 0x81;     /*  trk info: RT + trk mode  */
 		xferp->data_in[7] = 0x2;      /*  last rec=valid, next w=not
 						  valid */
-		{
-			/*
-			 *  track start, next writable, free blcks,
-			 *  blocking factor
-			 */
-			int i;
-			for(i=8; i<=23; i++)
-				xferp->data_in[i] = 0;
-		}
+		/*
+		 *  track start, next writable, free blcks,
+		 *  blocking factor
+		 */
+		for(size_t j=8; j<=23; j++)
+			xferp->data_in[j] = 0;
 
 		/*  Track size:  */
 		xferp->data_in[24] = (size >> 24) & 0xff;
@@ -1057,11 +1048,8 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		xferp->data_in[27] = size & 0xff;
 
 		/*  Last recorded address, only for dvd; zero out the rest:  */
-		{
-			int i;
-			for (i=28; i<=35; i++)
-				xferp->data_in[i] = 0;
-		}
+		for (size_t k=28; k<=35; k++)
+			xferp->data_in[k] = 0;
 
 		diskimage__return_default_status_and_message(xferp);
 		break;
@@ -1100,13 +1088,13 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 			debug("[ setting logical_block_size to %i ]\n",
 			    d->logical_block_size);
 		} else {
-			int i;
+			int j;
 			fatal("[ unknown MODE_SELECT: cmd =");
-			for (i=0; i<(ssize_t)xferp->cmd_len; i++)
-				fatal(" %02x", xferp->cmd[i]);
+			for (j=0; j<(ssize_t)xferp->cmd_len; j++)
+				fatal(" %02x", xferp->cmd[j]);
 			fatal(", data_out =");
-			for (i=0; i<(ssize_t)xferp->data_out_len; i++)
-				fatal(" %02x", xferp->data_out[i]);
+			for (j=0; j<(ssize_t)xferp->data_out_len; j++)
+				fatal(" %02x", xferp->data_out[j]);
 			fatal(" ]");
 		}
 
@@ -1152,13 +1140,13 @@ xferp->data_in[4] = 0x2c - 4;	/*  Additional length  */
 		break;
 
 	default:
-		fatal("[ UNIMPLEMENTED SCSI command 0x%02x, disk id=%i ]\n",
+		debug("[ UNIMPLEMENTED SCSI command 0x%02x, disk id=%i",
 		    xferp->cmd[0], id);
-		exit(1);
+		// exit(1);
 	}
+
 	debug(" ]\n");
 
 	return 1;
 }
 
-
