@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -26,18 +26,23 @@
  *
  *
  *  COMMENT: a.out executable file support
+ *
+ *  Also: b.out was an a.out-like format used by i960 systems.
  */
 
 /*  Note: Included from file.c.  */
 
 
-#include "exec_aout.h"
+#include "thirdparty/exec_aout.h"
+#include "thirdparty/exec_bout.h"
 
 
-#define	AOUT_FLAG_DECOSF1		1
-#define	AOUT_FLAG_FROM_BEGINNING	2
-#define	AOUT_FLAG_VADDR_ZERO_HACK	4
-#define	AOUT_FLAG_NO_SIZES		8
+#define	AOUT_FLAG_DECOSF1				1
+#define	AOUT_FLAG_FROM_BEGINNING			2
+#define	AOUT_FLAG_VADDR_ZERO_HACK			4
+#define	AOUT_FLAG_NO_SIZES				8
+#define	AOUT_FLAG_DATA_AT_END_MAY_BE_OMITTED		16
+#define	AOUT_FLAG_I960_BOUT				32
 
 struct aout_symbol {
 	uint32_t	strindex;
@@ -60,6 +65,7 @@ static void file_load_aout(struct machine *m, struct memory *mem,
 	uint64_t *entrypointp, int arch, int *byte_orderp)
 {
 	struct exec aout_header;
+	struct bout_exec bout_header;
 	FILE *f;
 	int len;
 	int encoding = ELFDATA2LSB;
@@ -78,7 +84,24 @@ static void file_load_aout(struct machine *m, struct memory *mem,
 		exit(1);
 	}
 
-	if (flags & AOUT_FLAG_DECOSF1) {
+	if (flags & AOUT_FLAG_I960_BOUT) {
+		len = fread(&bout_header, 1, sizeof(bout_header), f);
+		if (len != sizeof(bout_header)) {
+			fprintf(stderr, "%s: not a complete b.out image\n",
+			    filename);
+			exit(1);
+		}
+
+		unencode(entry, &bout_header.a_entry, uint32_t);
+		debug("b.out, entry point 0x%08lx\n", (long)entry);
+		vaddr = entry;
+
+		unencode(textsize, &bout_header.a_text, uint32_t);
+		unencode(datasize, &bout_header.a_data, uint32_t);
+		debug("text + data = %i + %i bytes\n", textsize, datasize);
+
+		unencode(symbsize, &bout_header.a_syms, uint32_t);
+	} else if (flags & AOUT_FLAG_DECOSF1) {
 		if (fread(&buf, 1, 32, f) != 32) {
 			perror(filename);
 			exit(1);
@@ -151,7 +174,7 @@ static void file_load_aout(struct machine *m, struct memory *mem,
 			m->cpus[0]->memory_rw(m->cpus[0], mem, vaddr + len2,
 			    &buf[len2], len-len2, MEM_WRITE, NO_EXCEPTIONS);
 		} else {
-			if (flags & AOUT_FLAG_DECOSF1)
+			if (flags & AOUT_FLAG_DATA_AT_END_MAY_BE_OMITTED)
 				break;
 			else {
 				fprintf(stderr, "could not read from %s,"
@@ -174,7 +197,7 @@ static void file_load_aout(struct machine *m, struct memory *mem,
 		off_t oldpos;
 
 		debug("symbols: %i bytes @ 0x%x\n", symbsize, (int)ftello(f));
-		CHECK_ALLOCATION(syms = malloc(symbsize));
+		CHECK_ALLOCATION(syms = (unsigned char *) malloc(symbsize));
 		len = fread(syms, 1, symbsize, f);
 		if (len != symbsize) {
 			fprintf(stderr, "error reading symbols from %s\n",
@@ -187,14 +210,14 @@ static void file_load_aout(struct machine *m, struct memory *mem,
 		strings_len = ftello(f) - oldpos;
 		fseek(f, oldpos, SEEK_SET);
 		debug("strings: %i bytes @ 0x%x\n", strings_len,(int)ftello(f));
-		CHECK_ALLOCATION(string_symbols = malloc(strings_len));
+		CHECK_ALLOCATION(string_symbols = (char *) malloc(strings_len));
 		if (fread(string_symbols, 1, strings_len, f) != strings_len) {
 			fprintf(stderr, "Could not read symbols from %s?\n", filename);
 			perror("fread");
 			exit(1);
 		}
 
-		aout_symbol_ptr = (struct aout_symbol *) syms;
+		aout_symbol_ptr = (struct aout_symbol *) (void*) syms;
 		n_symbols = symbsize / sizeof(struct aout_symbol);
 		i = 0;
 		while (i < n_symbols) {
