@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -31,11 +31,11 @@
 /*  Note: Included from file.c.  */
 
 
-#include "exec_elf.h"
+#include "thirdparty/exec_elf.h"
 
 /*  ELF machine types as strings: (same as exec_elf.h)  */
 #define N_ELF_MACHINE_TYPES     89
-static char *elf_machine_type[N_ELF_MACHINE_TYPES] = {
+static const char *elf_machine_type[N_ELF_MACHINE_TYPES] = {
         "NONE", "M32", "SPARC", "386",				/*  0..3  */
         "68K", "88K", "486", "860",				/*  4..7  */  
         "MIPS", "S370", "MIPS_RS3_LE", "RS6000",		/*  8..11  */
@@ -93,9 +93,11 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 	Elf64_Sym sym64;
 	int ofs;
 	int chunk_len = 1024, align_len;
-	char *symbol_strings = NULL; size_t symbol_length = 0;
-	char *s;
-	Elf32_Sym *symbols_sym32 = NULL;  int n_symbols = 0;
+	char *symbol_strings = NULL;
+	size_t symbol_length = 0;
+	const char *s;
+	int n_symbols = 0;
+	Elf32_Sym *symbols_sym32 = NULL;
 	Elf64_Sym *symbols_sym64 = NULL;
 
 	f = fopen(filename, "r");
@@ -202,6 +204,12 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 
 	ok = 0;
 	switch (arch) {
+	case ARCH_M88K:
+		switch (emachine) {
+		case EM_88K:
+			ok = 1;
+		}
+		break;
 	case ARCH_ALPHA:
 		switch (emachine) {
 		case EM_ALPHA:
@@ -232,14 +240,14 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		case EM_PARISC:
 			ok = 1;
 		}
-		break;
+		break;  */
 	case ARCH_I960:
 		switch (emachine) {
 		case EM_960:
 			ok = 1;
 		}
 		break;
-	case ARCH_IA64:
+	/*  case ARCH_IA64:
 		switch (emachine) {
 		case EM_IA_64:
 			ok = 1;
@@ -250,13 +258,13 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		case EM_68K:
 			ok = 1;
 		}
-		break;  */
+		break;
 	case ARCH_M32R:
 		switch (emachine) {
 		case EM_M32R:
 			ok = 1;
 		}
-		break;
+		break;*/
 	case ARCH_MIPS:
 		switch (emachine) {
 		case EM_MIPS:
@@ -271,20 +279,26 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 			ok = 1;
 		}
 		break;
+	case ARCH_RISCV:
+		switch (emachine) {
+		case 243:
+			ok = 1;
+		}
+		break;
 	case ARCH_SH:
 		switch (emachine) {
 		case EM_SH:
 			ok = 1;
 		}
 		break;
-	case ARCH_SPARC:
+	/*case ARCH_SPARC:
 		switch (emachine) {
 		case EM_SPARC:
 		case EM_SPARCV9:
 			ok = 1;
 		}
 		break;
-	/*  case ARCH_X86:
+	case ARCH_X86:
 		switch (emachine) {
 		case EM_386:
 		case EM_486:
@@ -298,7 +312,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		}
 		break;  */
 	default:
-		fatal("file.c: INTERNAL ERROR: Unimplemented arch!\n");
+		fatal("file.c: INTERNAL ERROR: Unimplemented arch %i!\n", arch);
 	}
 	if (!ok) {
 		fprintf(stderr, "%s: this is a ", filename);
@@ -318,9 +332,9 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 	    encoding == ELFDATA2LSB? "LSB (LE)" : "MSB (BE)", s);
 
 	if (elf64)
-		debug("%016"PRIx64"\n", (uint64_t) eentry);
+		debug("%016" PRIx64"\n", (uint64_t) eentry);
 	else
-		debug("%08"PRIx32"\n", (uint32_t) eentry);
+		debug("%08" PRIx32"\n", (uint32_t) eentry);
 
 	/*
 	 *  SH64: 32-bit instruction encoding?
@@ -392,34 +406,34 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 			if (p_type == PT_LOAD)
 				debug("load");
 			else
-				debug("0x%08"PRIx32, (uint32_t) p_type);
+				debug("0x%08" PRIx32, (uint32_t) p_type);
 
-			debug(") @ 0x%"PRIx64", vaddr 0x", (uint64_t) p_offset);
+			debug(") @ 0x%" PRIx64", vaddr 0x", (uint64_t) p_offset);
 
 			if (elf64)
-				debug("%016"PRIx64, (uint64_t) p_vaddr);
+				debug("%016" PRIx64, (uint64_t) p_vaddr);
 			else
-				debug("%08"PRIx32, (uint32_t) p_vaddr);
+				debug("%08" PRIx32, (uint32_t) p_vaddr);
 
-			debug(" len=0x%"PRIx64"\n", (uint64_t) p_memsz);
+			debug(" len=0x%" PRIx64"\n", (uint64_t) p_memsz);
 
 			if (p_vaddr != p_paddr) {
 				if (elf64)
-					debug("NOTE: vaddr (0x%"PRIx64") and "
-					    "paddr (0x%"PRIx64") differ; using "
+					debug("NOTE: vaddr (0x%" PRIx64") and "
+					    "paddr (0x%" PRIx64") differ; using "
 					    "vaddr\n", (uint64_t) p_vaddr,
 					    (uint64_t) p_paddr);
 				else
-					debug("NOTE: vaddr (0x%08"PRIx32") and "
-					    "paddr (0x%08"PRIx32") differ; usin"
+					debug("NOTE: vaddr (0x%08" PRIx32") and "
+					    "paddr (0x%08" PRIx32") differ; usin"
 					    "g vaddr\n", (uint32_t) p_vaddr,
 					    (uint32_t)p_paddr);
 			}
 
 			if (p_memsz < p_filesz) {
 				fprintf(stderr, "%s: memsz < filesz. TODO: how"
-				    " to handle this? memsz=%016"PRIx64
-				    " filesz=%016"PRIx64"\n", filename,
+				    " to handle this? memsz=%016" PRIx64
+				    " filesz=%016" PRIx64"\n", filename,
 				    (uint64_t) p_memsz, (uint64_t) p_filesz);
 				exit(1);
 			}
@@ -435,9 +449,8 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 			ofs = 0;  len = chunk_len = align_len;
 			while (ofs < (int64_t)p_filesz && len==chunk_len) {
 				unsigned char *ch;
-				int i = 0;
 
-				CHECK_ALLOCATION(ch = malloc(chunk_len));
+				CHECK_ALLOCATION(ch = (unsigned char *) malloc(chunk_len));
 
 				/*  Switch to larger size, if possible:  */
 				if (align_len < 0x10000 &&
@@ -445,28 +458,29 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 					align_len = 0x10000;
 					len = chunk_len = align_len;
 					free(ch);
-					CHECK_ALLOCATION(ch=malloc(chunk_len));
+					CHECK_ALLOCATION(ch=(unsigned char *)malloc(chunk_len));
 				} else if (align_len < 0x1000 &&
 				    ((p_vaddr + ofs) & 0xfff)==0) {
 					align_len = 0x1000;
 					len = chunk_len = align_len;
 					free(ch);
-					CHECK_ALLOCATION(ch=malloc(chunk_len));
+					CHECK_ALLOCATION(ch=(unsigned char *)malloc(chunk_len));
 				}
 
 				len = fread(&ch[0], 1, chunk_len, f);
 				if (ofs + len > (int64_t)p_filesz)
 					len = p_filesz - ofs;
 
-				while (i < len) {
+				int j = 0;
+				while (j < len) {
 					size_t len_to_copy;
-					len_to_copy = (i + align_len) <= len?
-					    align_len : len - i;
+					len_to_copy = (j + align_len) <= len?
+					    align_len : len - j;
 					m->cpus[0]->memory_rw(m->cpus[0], mem,
-					    p_vaddr + ofs, &ch[i], len_to_copy,
+					    p_vaddr + ofs, &ch[j], len_to_copy,
 					    MEM_WRITE, NO_EXCEPTIONS);
 					ofs += align_len;
-					i += align_len;
+					j += align_len;
 				}
 
 				free(ch);
@@ -485,7 +499,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		off_t sh_offset;
 		int n_entries;	/*  for reading the symbol / string tables  */
 
-		/*  debug("section header %i at %016"PRIx64"\n", i,
+		/*  debug("section header %i at %016" PRIx64"\n", i,
 		    (uint64_t) eshoff+i*eshentsize);  */
 
 		fseek(f, eshoff + i * eshentsize, SEEK_SET);
@@ -537,7 +551,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 			sh_entsize = sizeof(Elf32_Sym);
 
 		if (sh_type == SHT_SYMTAB) {
-			size_t len;
+			size_t len2;
 			n_entries = sh_size / sh_entsize;
 
 			fseek(f, sh_offset, SEEK_SET);
@@ -546,29 +560,29 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 				if (symbols_sym64 != NULL)
 					free(symbols_sym64);
 
-				CHECK_ALLOCATION(symbols_sym64 =
+				CHECK_ALLOCATION(symbols_sym64 = (Elf64_Sym *)
 				    malloc(sh_size));
 
-				len = fread(symbols_sym64, 1, sh_entsize *
+				len2 = fread(symbols_sym64, 1, sh_entsize *
 				    n_entries, f);
 			} else {
 				if (symbols_sym32 != NULL)
 					free(symbols_sym32);
 
-				CHECK_ALLOCATION(symbols_sym32 =
+				CHECK_ALLOCATION(symbols_sym32 = (Elf32_Sym *)
 				    malloc(sh_size));
 
-				len = fread(symbols_sym32, 1,
+				len2 = fread(symbols_sym32, 1,
 				    sh_entsize * n_entries, f);
 			}
 
-			if (len != sh_size) {
+			if (len2 != sh_size) {
 				fprintf(stderr, "could not read symbols from "
 				    "%s\n", filename);
 				exit(1);
 			}
 
-			debug("%i symbol entries at 0x%"PRIx64"\n",
+			debug("%i symbol entries at 0x%" PRIx64"\n",
 			    (int) n_entries, (uint64_t) sh_offset);
 
 			n_symbols = n_entries;
@@ -583,22 +597,20 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		 */
 
 		if (sh_type == SHT_STRTAB && sh_size > symbol_length) {
-			size_t len;
-
 			if (symbol_strings != NULL)
 				free(symbol_strings);
 
-			CHECK_ALLOCATION(symbol_strings = malloc(sh_size + 1));
+			CHECK_ALLOCATION(symbol_strings = (char *) malloc(sh_size + 1));
 
 			fseek(f, sh_offset, SEEK_SET);
-			len = fread(symbol_strings, 1, sh_size, f);
-			if (len != sh_size) {
+			size_t len2 = fread(symbol_strings, 1, sh_size, f);
+			if (len2 != sh_size) {
 				fprintf(stderr, "could not read symbols from "
 				    "%s\n", filename);
 				exit(1);
 			}
 
-			debug("%i bytes of symbol strings at 0x%"PRIx64"\n",
+			debug("%i bytes of symbol strings at 0x%" PRIx64"\n",
 			    (int) sh_size, (uint64_t) sh_offset);
 
 			symbol_strings[sh_size] = '\0';
@@ -628,7 +640,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 				unencode(size,    &sym32.st_size, Elf32_Word);
 			}
 
-			/*  debug("symbol info=0x%02x addr=0x%016"PRIx64
+			/*  debug("symbol info=0x%02x addr=0x%016" PRIx64
 			    " (%i) '%s'\n", st_info, (uint64_t) addr,
 			    st_name, symbol_strings + st_name);  */
 
@@ -637,7 +649,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 
 			if (addr != 0) /* && ((st_info >> 4) & 0xf)
 			    >= STB_GLOBAL) */ {
-				/*  debug("symbol info=0x%02x addr=0x%016"PRIx64
+				/*  debug("symbol info=0x%02x addr=0x%016" PRIx64
 				    " '%s'\n", st_info, (uint64_t) addr,
 				    symbol_strings + st_name);  */
 				add_symbol_name(&m->symbol_context,
@@ -648,9 +660,9 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 			if (strcmp(symbol_strings + st_name, "_gp") == 0) {
 				debug("found _gp address: 0x");
 				if (elf64)
-					debug("%016"PRIx64"\n", (uint64_t)addr);
+					debug("%016" PRIx64"\n", (uint64_t)addr);
 				else
-					debug("%08"PRIx32"\n", (uint32_t)addr);
+					debug("%08" PRIx32"\n", (uint32_t)addr);
 				*gpp = addr;
 			}
 		}
@@ -703,7 +715,7 @@ static void file_load_elf(struct machine *m, struct memory *mem,
 		    ((uint64_t)b[5] << 16) + ((uint64_t)b[6] << 8) +
 		    (uint64_t)b[7];
 
-		debug("entrypoint 0x%016"PRIx64", toc_base 0x%016"PRIx64"\n",
+		debug("entrypoint 0x%016" PRIx64", toc_base 0x%016" PRIx64"\n",
 		    (uint64_t) *entrypointp, (uint64_t) toc_base);
 		if (tocp != NULL)
 			*tocp = toc_base;
