@@ -193,6 +193,7 @@ void dev_decxmi_init(struct memory *mem, uint64_t baseaddr);
 #define	VFB_DEC_VFB02		3
 #define	VFB_DEC_MAXINE		4
 #define	VFB_PLAYSTATION2	5
+#define	VFB_REVERSEBITS		6
 /*  Extra flags:  */
 #define	VFB_REVERSE_START	0x10000
 struct vfb_data {
@@ -368,7 +369,7 @@ void dev_px_init(struct machine *machine, struct memory *mem,
 int dev_ram_access(struct cpu *cpu, struct memory *mem, uint64_t relative_addr,
 	unsigned char *data, size_t len, int writeflag, void *);
 void dev_ram_init(struct machine *machine, uint64_t baseaddr, uint64_t length,
-	int mode, uint64_t otheraddr, const char* name = NULL);
+	int mode, uint64_t otheraddr, const char* name);
 
 /*  dev_scc.c:  */
 #define	DEV_SCC_LENGTH			0x1000
@@ -406,6 +407,7 @@ void dev_sgi_re_init(struct machine *machine, struct memory *mem, uint64_t basea
 int dev_sgi_de_access(struct cpu *cpu, struct memory *mem,
 	uint64_t relative_addr, unsigned char *data, size_t len,
 	int writeflag, void *);
+struct sgi_re_data;
 void dev_sgi_de_init(struct memory *mem, uint64_t baseaddr, struct sgi_re_data *);
 // SGI O2 Memory Transfer Engine:
 #define	DEV_SGI_MTE_LENGTH		0x1000
@@ -520,7 +522,7 @@ struct lk201_data {
                         
         int                     mouse_mode;
         int                     mouse_revision;         /*  0..15  */  
-        int                     mouse_x, mouse_y, mouse_buttons;
+        int                     mouse_buttons;
 };
 void lk201_tick(struct machine *, struct lk201_data *); 
 void lk201_tx_data(struct lk201_data *, int port, int idata);
