@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -28,6 +28,7 @@
  *  Machine registry.
  */
 
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -44,9 +45,11 @@
 #include "misc.h"
 #include "settings.h"
 #include "symbol.h"
-#include "useremul.h"
 
 
+extern bool debugmsg_executing_noninteractively;
+extern bool single_step;
+
 /*  This is initialized by machine_init():  */
 struct machine_entry *first_machine_entry = NULL;
 
@@ -60,7 +63,7 @@ struct machine *machine_new(char *name, struct emul *emul, int id)
 {
 	struct machine *m;
 
-	CHECK_ALLOCATION(m = malloc(sizeof(struct machine)));
+	CHECK_ALLOCATION(m = (struct machine *) malloc(sizeof(struct machine)));
 	memset(m, 0, sizeof(struct machine));
 
 	/*  Pointer back to the emul object that this machine belongs to:  */
@@ -70,7 +73,7 @@ struct machine *machine_new(char *name, struct emul *emul, int id)
 		CHECK_ALLOCATION(m->name = strdup(name));
 
 	/*  Full path, e.g. "machine[0]":  */
-	CHECK_ALLOCATION(m->path = malloc(20));
+	CHECK_ALLOCATION(m->path = (char *) malloc(20));
 	snprintf(m->path, 20, "machine[%i]", id);
 
 	/*  Sane default values:  */
@@ -82,7 +85,7 @@ struct machine *machine_new(char *name, struct emul *emul, int id)
 	m->prom_emulation = 1;
 	m->allow_instruction_combinations = 1;
 	m->byte_order_override = NO_BYTE_ORDER_OVERRIDE;
-	m->boot_kernel_filename = "";
+	m->boot_kernel_filename = strdup("");
 	m->boot_string_argument = NULL;
 	m->x11_md.scaledown = 1;
 	m->x11_md.scaleup = 1;
@@ -130,7 +133,7 @@ void machine_destroy(struct machine *machine)
 		cpu_destroy(machine->cpus[i]);
 
 	if (machine->name != NULL)
-		free(machine->name);
+	 	free(machine->name);
 
 	if (machine->path != NULL)
 		free(machine->path);
@@ -153,7 +156,7 @@ void machine_destroy(struct machine *machine)
  *  Also, any errors/warnings are printed using fatal()/debug().
  */
 int machine_name_to_type(char *stype, char *ssubtype,
-	int *type, int *subtype, int *arch)
+	int *type, int *subtype)
 {
 	struct machine_entry *me;
 	int i, j, k, nmatches = 0;
@@ -168,7 +171,6 @@ int machine_name_to_type(char *stype, char *ssubtype,
 			if (strcasecmp(me->aliases[i], stype) == 0) {
 				/*  Found a type:  */
 				*type = me->machine_type;
-				*arch = me->arch;
 
 				if (me->n_subtypes == 0)
 					return 1;
@@ -210,9 +212,7 @@ int machine_name_to_type(char *stype, char *ssubtype,
 				if (strcasecmp(ssubtype, me->subtype[j]->
 				    aliases[k]) == 0) {
 					*type = me->machine_type;
-					*arch = me->arch;
-					*subtype = me->subtype[j]->
-					    machine_subtype;
+					*subtype = me->subtype[j]->machine_subtype;
 					nmatches ++;
 				}
 
@@ -252,9 +252,9 @@ void machine_add_breakpoint_string(struct machine *machine, char *str)
 {
 	int n = machine->breakpoints.n + 1;
 
-	CHECK_ALLOCATION(machine->breakpoints.string =
+	CHECK_ALLOCATION(machine->breakpoints.string = (char **)
 	    realloc(machine->breakpoints.string, n * sizeof(char *)));
-	CHECK_ALLOCATION(machine->breakpoints.addr =
+	CHECK_ALLOCATION(machine->breakpoints.addr = (uint64_t *)
 	    realloc(machine->breakpoints.addr, n * sizeof(uint64_t)));
 	CHECK_ALLOCATION(machine->breakpoints.string[machine->breakpoints.n] =
 	    strdup(optarg));
@@ -281,13 +281,13 @@ void machine_add_tickfunction(struct machine *machine, void (*func)
 {
 	int n = machine->tick_functions.n_entries;
 
-	CHECK_ALLOCATION(machine->tick_functions.ticks_till_next = realloc(
+	CHECK_ALLOCATION(machine->tick_functions.ticks_till_next = (int *) realloc(
 	    machine->tick_functions.ticks_till_next, (n+1) * sizeof(int)));
-	CHECK_ALLOCATION(machine->tick_functions.ticks_reset_value = realloc(
+	CHECK_ALLOCATION(machine->tick_functions.ticks_reset_value = (int *) realloc(
 	    machine->tick_functions.ticks_reset_value, (n+1) * sizeof(int)));
-	CHECK_ALLOCATION(machine->tick_functions.f = realloc(
+	CHECK_ALLOCATION(machine->tick_functions.f = (void (**)(struct cpu*,void*)) realloc(
 	    machine->tick_functions.f, (n+1) * sizeof(void *)));
-	CHECK_ALLOCATION(machine->tick_functions.extra = realloc(
+	CHECK_ALLOCATION(machine->tick_functions.extra = (void **) realloc(
 	    machine->tick_functions.extra, (n+1) * sizeof(void *)));
 
 	/*
@@ -323,7 +323,7 @@ void machine_statistics_init(struct machine *machine, char *fname)
 {
 	int n_fields = 0;
 	char *pcolon = fname;
-	char *mode = "a";	/*  Append by default  */
+	const char *mode = "a";	/*  Append by default  */
 
 	machine->allow_instruction_combinations = 0;
 
@@ -333,7 +333,7 @@ void machine_statistics_init(struct machine *machine, char *fname)
 	}
 
 	machine->statistics.enabled = 1;
-	CHECK_ALLOCATION(machine->statistics.fields = malloc(1));
+	CHECK_ALLOCATION(machine->statistics.fields = (char *) malloc(1));
 	machine->statistics.fields[0] = '\0';
 
 	while (*pcolon && *pcolon != ':')
@@ -353,7 +353,7 @@ void machine_statistics_init(struct machine *machine, char *fname)
 		case 'v':
 		case 'i':
 		case 'p':
-			CHECK_ALLOCATION(machine->statistics.fields = realloc(
+			CHECK_ALLOCATION(machine->statistics.fields = (char *) realloc(
 			    machine->statistics.fields, strlen(
 			    machine->statistics.fields) + 2));
 			machine->statistics.fields[n_fields ++] = *fname;
@@ -397,25 +397,27 @@ void machine_dumpinfo(struct machine *m)
 		debug("  (nr of NICs: %i)", m->nr_of_nics);
 	debug("\n");
 
-	debug("memory: %i MB", m->physical_ram_in_mb);
+	char meminfo[2000];
+	snprintf(meminfo, sizeof(meminfo), "%i MB", m->physical_ram_in_mb);
+
 	if (m->memory_offset_in_mb != 0)
-		debug(" (offset by %i MB)", m->memory_offset_in_mb);
+		snprintf(meminfo + strlen(meminfo), sizeof(meminfo) - strlen(meminfo),
+		    " (offset by %i MB)", m->memory_offset_in_mb);
 	if (m->random_mem_contents)
-		debug(", randomized contents");
-	debug("\n");
+		snprintf(meminfo + strlen(meminfo), sizeof(meminfo) - strlen(meminfo),
+		    ", randomized content");
+
+	debugmsg(SUBSYS_MACHINE, "memory", VERBOSITY_INFO, "%s", meminfo);
 
 	if (!m->prom_emulation)
 		debug("PROM emulation disabled\n");
 
 	for (i=0; i<m->ncpus; i++)
-		cpu_dumpinfo(m, m->cpus[i]);
+		cpu_dumpinfo(m, m->cpus[i], true);
 
 	if (m->ncpus > 1)
 		debug("Bootstrap cpu is nr %i\n", m->bootstrap_cpu);
 
-	if (m->slow_serial_interrupts_hack_for_linux)
-		debug("Using slow_serial_interrupts_hack_for_linux\n");
-
 	if (m->x11_md.in_use) {
 		debug("Using X11");
 		if (m->x11_md.scaledown > 1)
@@ -442,48 +444,36 @@ void machine_dumpinfo(struct machine *m)
 /*
  *  machine_setup():
  *
- *  This (rather large) function initializes memory, registers, and/or devices
+ *  This function initializes memory, registers, and/or devices
  *  required by specific machine emulations.
  */
 void machine_setup(struct machine *machine)
 {
-	struct memory *mem;
 	struct machine_entry *me;
-
-	/*  Abreviation:  :-)  */
 	struct cpu *cpu = machine->cpus[machine->bootstrap_cpu];
 
-	machine->bootdev_id = diskimage_bootdev(machine,
-	    &machine->bootdev_type);
-
-	mem = cpu->mem;
+	machine->bootdev_id = diskimage_bootdev(machine, &machine->bootdev_type);
 	machine->machine_name = NULL;
 
 	/*  TODO: Move this somewhere else?  */
 	if (machine->boot_string_argument == NULL) {
 		switch (machine->machine_type) {
 		case MACHINE_ARC:
-			machine->boot_string_argument = "-aN";
+			machine->boot_string_argument = strdup("-aN");
 			break;
 		case MACHINE_CATS:
-			machine->boot_string_argument = "-A";
+			machine->boot_string_argument = strdup("-A");
 			break;
 		case MACHINE_PMAX:
-			machine->boot_string_argument = "-a";
+			machine->boot_string_argument = strdup("-a");
 			break;
 		default:
 			/*  Important, because boot_string_argument should
 			    not be set to NULL:  */
-			machine->boot_string_argument = "";
+			machine->boot_string_argument = strdup("");
 		}
 	}
 
-
-	/*
-	 *  If the machine has a setup function in src/machines/machine_*.c
-	 *  then use that one, otherwise use the old hardcoded stuff here:
-	 */
-
 	me = first_machine_entry;
 	while (me != NULL) {
 		if (machine->machine_type == me->machine_type &&
@@ -491,6 +481,7 @@ void machine_setup(struct machine *machine)
 			me->setup(machine, cpu);
 			break;
 		}
+
 		me = me->next;
 	}
 
@@ -499,20 +490,65 @@ void machine_setup(struct machine *machine)
 		exit(1);
 	}
 
+	switch (cpu->cpu_family->arch) {
+	case ARCH_MIPS:
+		// Register devices at the "typical" kseg1 base:
+		// 0xffffffff80000000 (kseg0) could work too but it is more
+		// likely that devices are accessed using their uncached address.
+		// TODO: How about "high" mapping for 64-bit systems?
+		machine_add_devices_as_symbols(machine, 0xffffffffa0000000);
+		break;
+	}
+	
 	if (machine->machine_name != NULL)
-		debug("machine: %s", machine->machine_name);
+		debugmsg(SUBSYS_MACHINE, "model", VERBOSITY_INFO, "%s", machine->machine_name);
 
 	if (machine->emulated_hz > 0)
-		debug(" (%.2f MHz)", (float)machine->emulated_hz / 1000000);
-	debug("\n");
+		debugmsg(SUBSYS_MACHINE, "clock speed", VERBOSITY_INFO,
+		    "%.2f MHz", (float)machine->emulated_hz / 1000000);
 
-	if (machine->bootstr != NULL) {
-		debug("bootstring%s: %s", (machine->bootarg!=NULL &&
-		    strlen(machine->bootarg) >= 1)? "(+bootarg)" : "",
-		    machine->bootstr);
-		if (machine->bootarg != NULL && strlen(machine->bootarg) >= 1)
-			debug(" %s", machine->bootarg);
-		debug("\n");
+	if (machine->bootstr != NULL && machine->bootstr[0]) {
+		bool has_bootarg = machine->bootarg != NULL && machine->bootarg[0];
+		debugmsg(SUBSYS_MACHINE,
+		    has_bootarg ? "bootstring & bootarg" : "bootstring",
+		    VERBOSITY_INFO,
+		    "%s %s",
+		    machine->bootstr,
+		    has_bootarg ? machine->bootarg : "");
+	}
+}
+
+
+/*
+ *  machine_add_devices_as_symbols():
+ *
+ *  Registers device addresses as symbols, to make the debugging experience
+ *  somewhat nicer. offset should be e.g. 0xffffffff80000000 and
+ *  0xffffffffa0000000 on MIPS.
+ */
+void machine_add_devices_as_symbols(struct machine *machine, uint64_t offset)
+{
+	struct memory* mem = machine->cpus[machine->bootstrap_cpu]->mem;
+	int i;
+
+	for (i = 0; i < mem->n_mmapped_devices; ++i) {
+		char name[1000];
+
+		// Hm. Perhaps adding all "ram" devices looks too cluttered.
+		// if (strcmp(mem->devices[i].name, "ram") == 0)
+		//	continue;
+
+		snprintf(name, sizeof(name),
+		    "(%s@0x%llx)", mem->devices[i].name,
+		    (long long)mem->devices[i].baseaddr);
+
+		add_symbol_name(
+		    &machine->symbol_context,
+		    mem->devices[i].baseaddr + offset,
+		    mem->devices[i].length,
+		    name,
+		    0,
+		    0);
 	}
 }
 
@@ -610,50 +646,55 @@ void machine_default_cputype(struct machine *m)
  *  around N_SAFE_DYNTRANS_LIMIT instructions will be run by the dyntrans
  *  system.)
  *
- *  Return value is 1 if any CPU in this machine is still running,
- *  or 0 if all CPUs are stopped.
+ *  Return value is true if any CPU in this machine is still running,
+ *  false if all CPUs are stopped.
  */
-int machine_run(struct machine *machine)
+bool machine_run(struct machine *machine)
 {
 	struct cpu **cpus = machine->cpus;
-	int ncpus = machine->ncpus, cpu0instrs = 0, i, te;
+	int ncpus = machine->ncpus;
+	bool any_running = false;
 
-	for (i=0; i<ncpus; i++) {
+	for (int i=0; i<ncpus; i++) {
 		if (cpus[i]->running) {
-			int instrs_run = cpus[i]->run_instr(cpus[i]);
-			if (i == 0)
-				cpu0instrs += instrs_run;
+			any_running = true;
+			cpus[i]->run_instr(cpus[i]);
 		}
 	}
 
+	if (!any_running)
+		return false;
+
 	/*
 	 *  Hardware 'ticks':  (clocks, interrupt sources...)
 	 *
-	 *  Here, cpu0instrs is the number of instructions executed on cpu0.
+	 *  This assumes that the CPUs ran approximately N_SAFE_DYNTRANS_LIMIT
+	 *  instructions (or "ticks") in the core dyntrans loop.
 	 *
 	 *  TODO: This should be redesigned into some "mainbus" stuff instead!
 	 */
 
-	for (te=0; te<machine->tick_functions.n_entries; te++) {
-		machine->tick_functions.ticks_till_next[te] -= cpu0instrs;
+	for (int te=0; te<machine->tick_functions.n_entries; te++) {
+		machine->tick_functions.ticks_till_next[te] -=
+		    single_step ? 1 : N_SAFE_DYNTRANS_LIMIT;
+
 		if (machine->tick_functions.ticks_till_next[te] <= 0) {
 			while (machine->tick_functions.ticks_till_next[te]<=0) {
 				machine->tick_functions.ticks_till_next[te] +=
-				    machine->tick_functions.
-				    ticks_reset_value[te];
-			}
+				    machine->tick_functions.ticks_reset_value[te];
 
-			machine->tick_functions.f[te](cpus[0],
-			    machine->tick_functions.extra[te]);
+				machine->tick_functions.f[te](cpus[0],
+			    	    machine->tick_functions.extra[te]);
+			}
 		}
 	}
 
 	/*  Is any CPU still alive?  */
-	for (i=0; i<ncpus; i++)
+	for (int i=0; i<ncpus; i++)
 		if (cpus[i]->running)
-			return 1;
+			return true;
 
-	return 0;
+	return false;
 }
 
 
@@ -668,16 +709,14 @@ int machine_run(struct machine *machine)
  *  passed as arguments to this function, such as alias names and machine
  *  subtypes.
  */
-struct machine_entry *machine_entry_new(const char *name, int arch,
-	int oldstyle_type)
+struct machine_entry *machine_entry_new(const char *name, int oldstyle_type)
 {
 	struct machine_entry *me;
 
-	CHECK_ALLOCATION(me = malloc(sizeof(struct machine_entry)));
+	CHECK_ALLOCATION(me = (struct machine_entry *) malloc(sizeof(struct machine_entry)));
 	memset(me, 0, sizeof(struct machine_entry));
 
 	me->name = name;
-	me->arch = arch;
 	me->machine_type = oldstyle_type;
 	me->n_aliases = 0;
 	me->aliases = NULL;
@@ -697,10 +736,10 @@ void machine_entry_add_alias(struct machine_entry *me, const char *name)
 {
 	me->n_aliases ++;
 
-	CHECK_ALLOCATION(me->aliases = realloc(me->aliases,
+	CHECK_ALLOCATION(me->aliases = (char **) realloc(me->aliases,
 	    sizeof(char *) * me->n_aliases));
 
-	me->aliases[me->n_aliases - 1] = (char *) name;
+	me->aliases[me->n_aliases - 1] = strdup(name);
 }
 
 
@@ -720,11 +759,13 @@ void machine_entry_add_subtype(struct machine_entry *me, const char *name,
 	struct machine_entry_subtype *mes;
 
 	/*  Allocate a new subtype struct:  */
-	CHECK_ALLOCATION(mes = malloc(sizeof(struct machine_entry_subtype)));
+	CHECK_ALLOCATION(mes = (struct machine_entry_subtype *)
+	    malloc(sizeof(struct machine_entry_subtype)));
 
 	/*  Add the subtype to the machine entry:  */
 	me->n_subtypes ++;
-	CHECK_ALLOCATION(me->subtype = realloc(me->subtype, sizeof(struct
+	CHECK_ALLOCATION(me->subtype = (struct
+	    machine_entry_subtype **) realloc(me->subtype, sizeof(struct
 	    machine_entry_subtype *) * me->n_subtypes));
 	me->subtype[me->n_subtypes - 1] = mes;
 
@@ -745,7 +786,7 @@ void machine_entry_add_subtype(struct machine_entry *me, const char *name,
 			break;
 
 		mes->n_aliases ++;
-		CHECK_ALLOCATION(mes->aliases =
+		CHECK_ALLOCATION(mes->aliases = (char **)
 		    realloc(mes->aliases, sizeof(char *) * mes->n_aliases));
 
 		mes->aliases[mes->n_aliases - 1] = s;
@@ -760,15 +801,10 @@ void machine_entry_add_subtype(struct machine_entry *me, const char *name,
  *
  *  Inserts a new machine_entry into the machine entries list.
  */
-void machine_entry_register(struct machine_entry *me, int arch)
+void machine_entry_register(struct machine_entry *me)
 {
 	struct machine_entry *prev, *next;
 
-	/*  Only insert it if the architecture is implemented in this
-	    emulator configuration:  */
-	if (cpu_family_ptr_by_number(arch) == NULL)
-		return;
-
 	prev = NULL;
 	next = first_machine_entry;
 
@@ -799,7 +835,7 @@ void machine_entry_register(struct machine_entry *me, int arch)
 void machine_list_available_types_and_cpus(void)
 {
 	struct machine_entry *me;
-	int iadd = DEBUG_INDENTATION * 2;
+	int iadd = 2;
 
 	debug("Available CPU types:\n\n");
 
@@ -819,15 +855,14 @@ void machine_list_available_types_and_cpus(void)
 		fatal("No machines defined!\n");
 
 	while (me != NULL) {
-		int i, j, iadd = DEBUG_INDENTATION;
+		int i, j, iadd2 = 1;
 
-		debug("%s [%s] (", me->name,
-		    cpu_family_ptr_by_number(me->arch)->name);
+		debug("%s (", me->name);
 		for (i=0; i<me->n_aliases; i++)
 			debug("%s\"%s\"", i? ", " : "", me->aliases[i]);
 		debug(")\n");
 
-		debug_indentation(iadd);
+		debug_indentation(iadd2);
 		for (i=0; i<me->n_subtypes; i++) {
 			struct machine_entry_subtype *mes;
 			mes = me->subtype[i];
@@ -838,7 +873,7 @@ void machine_list_available_types_and_cpus(void)
 				    mes->aliases[j]);
 			debug(")\n");
 		}
-		debug_indentation(-iadd);
+		debug_indentation(-iadd2);
 
 		me = me->next;
 	}
@@ -850,9 +885,6 @@ void machine_list_available_types_and_cpus(void)
 	    "or subtype, not the real name.\n");
 
 	debug("\n");
-
-	useremul_list_emuls();
-	debug("Userland emulation is not really working yet.\n");
 }
 
 
