@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -44,17 +44,16 @@
 
 MACHINE_SETUP(algor)
 {
-	struct pci_data *pci_bus;
 	char tmpstr[300];
 
 	machine->emulated_hz = 166560000;
 
 	switch (machine->machine_subtype) {
 	case MACHINE_ALGOR_P4032:
-		machine->machine_name = "Algor P4032";
+		machine->machine_name = strdup("Algor P4032");
 		break;
 	case MACHINE_ALGOR_P5064:
-		machine->machine_name = "Algor P5064";
+		machine->machine_name = strdup("Algor P5064");
 		break;
 	default:fatal("Unimplemented Algor machine.\n");
 		exit(1);
@@ -69,7 +68,7 @@ MACHINE_SETUP(algor)
 	 *  2 = ISA
 	 */
 
-	pci_bus = device_add(machine, "v3");
+	/*  struct pci_data *pci_bus = (struct pci_data *) */  device_add(machine, "v3");
 
 	device_add(machine, "algor addr=0x1ff00000");
 
@@ -125,7 +124,7 @@ MACHINE_DEFAULT_CPU(algor)
 
 MACHINE_REGISTER(algor)
 {
-	MR_DEFAULT(algor, "Algor evaluation board", ARCH_MIPS, MACHINE_ALGOR);
+	MR_DEFAULT(algor, "Algor evaluation board", MACHINE_ALGOR);
 
 	machine_entry_add_alias(me, "algor");
 
