@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -41,7 +41,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "cyclone_boot.h"
+#include "thirdparty/cyclone_boot.h"
 
 
 MACHINE_SETUP(cats)
@@ -51,7 +51,7 @@ MACHINE_SETUP(cats)
 	char bs[300], tmpstr[400];
 	int boot_id = machine->bootdev_id >= 0? machine->bootdev_id : 0;
 
-	machine->machine_name = "CATS evaluation board";
+	machine->machine_name = strdup("CATS evaluation board");
 
 	if (machine->emulated_hz == 0)
 		machine->emulated_hz = 50000000;
@@ -62,22 +62,22 @@ MACHINE_SETUP(cats)
 
 	snprintf(tmpstr, sizeof(tmpstr), "footbridge irq=%s.cpu[%i].irq"
 	    " addr=0x42000000", machine->path, machine->bootstrap_cpu);
-	pci_bus = device_add(machine, tmpstr);
+	pci_bus = (struct pci_data *) device_add(machine, tmpstr);
 
 	/*  DC21285_ROM_BASE (256 KB at 0x41000000)  */
-	dev_ram_init(machine, 0x41000000, 256 * 1024, DEV_RAM_RAM, 0);
+	dev_ram_init(machine, 0x41000000, 256 * 1024, DEV_RAM_RAM, 0, NULL);
 
 	/*  NetBSD, OpenBSD, and Linux (?) clean their caches here:  */
-	dev_ram_init(machine, 0x50000000, 0x10000, DEV_RAM_RAM, 0);
+	dev_ram_init(machine, 0x50000000, 0x10000, DEV_RAM_RAM, 0, NULL);
 
 	/*  Interrupt ack space?  */
-	dev_ram_init(machine, 0x80000000, 0x1000, DEV_RAM_RAM, 0);
+	dev_ram_init(machine, 0x80000000, 0x1000, DEV_RAM_RAM, 0, NULL);
 
 	/*  Linux uses 0xc0000000 as phys.:  */
-	dev_ram_init(machine, 0xc0000000, 0x20000000, DEV_RAM_MIRROR, 0x0);
+	dev_ram_init(machine, 0xc0000000, 0x20000000, DEV_RAM_MIRROR, 0x0, NULL);
 
 	/*  OpenBSD reboot needs 0xf??????? to be mapped to phys.:  */
-	dev_ram_init(machine, 0xf0000000, 0x1000000, DEV_RAM_MIRROR, 0x0);
+	dev_ram_init(machine, 0xf0000000, 0x1000000, DEV_RAM_MIRROR, 0x0, NULL);
 
 	bus_pci_add(machine, pci_bus, machine->memory, 0xc0, 8, 0, "s3_virge");
 
@@ -146,8 +146,7 @@ MACHINE_DEFAULT_RAM(cats)
 
 MACHINE_REGISTER(cats)
 {
-	MR_DEFAULT(cats, "CATS evaluation board", ARCH_ARM,
-	    MACHINE_CATS);
+	MR_DEFAULT(cats, "CATS evaluation board", MACHINE_CATS);
 
 	machine_entry_add_alias(me, "cats");
 
