@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -39,8 +39,8 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "hpc_bootinfo.h"
-#include "vripreg.h"
+#include "thirdparty/hpc_bootinfo.h"
+#include "thirdparty/vripreg.h"
 
 
 MACHINE_SETUP(hpcarm)
@@ -66,7 +66,7 @@ MACHINE_SETUP(hpcarm)
 
 	case MACHINE_HPCARM_IPAQ:
 		/*  SA-1110 206MHz  */
-		machine->machine_name = "Compaq iPAQ H3600";
+		machine->machine_name = strdup("Compaq iPAQ H3600");
 		hpc_fb_addr = 0x48200000;	/*  TODO  */
 		hpc_fb_xsize = 240;
 		hpc_fb_ysize = 320;
@@ -87,9 +87,9 @@ MACHINE_SETUP(hpcarm)
 	case MACHINE_HPCARM_JORNADA720:
 	case MACHINE_HPCARM_JORNADA728:
 		/*  SA-1110 206MHz  */
-		machine->machine_name = (machine->machine_subtype ==
-		    MACHINE_HPCARM_JORNADA720) ?
-		    "Jornada 720" : "Jornada 728";
+		machine->machine_name =
+		    ((machine->machine_subtype == MACHINE_HPCARM_JORNADA720) ?
+		    "Jornada 720" : "Jornada 728");
 		hpc_fb_addr = 0x48200000;
 		hpc_fb_xsize = 640;
 		hpc_fb_ysize = 240;
@@ -121,10 +121,10 @@ MACHINE_SETUP(hpcarm)
 
 	/*  Physical RAM at 0xc0000000:  */
 	dev_ram_init(machine, 0xc0000000, 0x20000000,
-	    DEV_RAM_MIRROR, 0x0);
+	    DEV_RAM_MIRROR, 0x0, NULL);
 
 	/*  Cache flush region:  */
-	dev_ram_init(machine, 0xe0000000, 0x10000, DEV_RAM_RAM, 0x0);
+	dev_ram_init(machine, 0xe0000000, 0x10000, DEV_RAM_RAM, 0x0, NULL);
 
 	if (hpc_fb_addr != 0) {
 		dev_fb_init(machine, machine->memory, hpc_fb_addr, VFB_HPC,
@@ -214,7 +214,7 @@ MACHINE_DEFAULT_RAM(hpcarm)
 
 MACHINE_REGISTER(hpcarm)
 {
-	MR_DEFAULT(hpcarm, "Handhelp ARM (HPCarm)", ARCH_ARM, MACHINE_HPCARM);
+	MR_DEFAULT(hpcarm, "Handhelp ARM (HPCarm)", MACHINE_HPCARM);
 
 	machine_entry_add_alias(me, "hpcarm");
 
