@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -52,13 +52,13 @@ MACHINE_SETUP(iq80321)
 	 *  for more details about the memory map.
 	 */
 
-	machine->machine_name = "Intel IQ80321";
+	machine->machine_name = strdup("Intel IQ80321");
 
 	cpu->cd.arm.coproc[6] = arm_coproc_i80321_6;
 
 	snprintf(tmpstr, sizeof(tmpstr), "i80321 irq=%s.cpu[%i].irq "
 	    "addr=0xffffe000", machine->path, machine->bootstrap_cpu);
-	pci = device_add(machine, tmpstr);
+	pci = (struct pci_data *) device_add(machine, tmpstr);
 
 	snprintf(tmpstr, sizeof(tmpstr), "ns16550 irq=%s.cpu[%i].irq."
 	    "i80321.%i addr=0xfe800000 name2='serial console'",
@@ -66,12 +66,12 @@ MACHINE_SETUP(iq80321)
 	device_add(machine, tmpstr);
 
 	/*  0xa0000000 = physical ram, 0xc0000000 = uncached  */
-	dev_ram_init(machine, 0xa0000000, 0x20000000, DEV_RAM_MIRROR, 0x0);
-	dev_ram_init(machine, 0xc0000000, 0x20000000, DEV_RAM_MIRROR, 0x0);
+	dev_ram_init(machine, 0xa0000000, 0x20000000, DEV_RAM_MIRROR, 0x0, NULL);
+	dev_ram_init(machine, 0xc0000000, 0x20000000, DEV_RAM_MIRROR, 0x0, NULL);
 
 	/*  0xe0000000 and 0xff000000 = cache flush regions  */
-	dev_ram_init(machine, 0xe0000000, 0x100000, DEV_RAM_RAM, 0x0);
-	dev_ram_init(machine, 0xff000000, 0x100000, DEV_RAM_RAM, 0x0);
+	dev_ram_init(machine, 0xe0000000, 0x100000, DEV_RAM_RAM, 0x0, NULL);
+	dev_ram_init(machine, 0xff000000, 0x100000, DEV_RAM_RAM, 0x0, NULL);
 
 	device_add(machine, "iq80321_7seg addr=0xfe840000");
 
@@ -102,7 +102,7 @@ MACHINE_DEFAULT_CPU(iq80321)
 
 MACHINE_REGISTER(iq80321)
 {
-	MR_DEFAULT(iq80321, "Intel IQ80321", ARCH_ARM, MACHINE_IQ80321);
+	MR_DEFAULT(iq80321, "Intel IQ80321", MACHINE_IQ80321);
 
 	machine_entry_add_alias(me, "iq80321");
 }
