@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -49,27 +49,27 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "netbsd_iyonix_bootconfig.h"
-#include "i80321reg.h"
+#include "thirdparty/netbsd_iyonix_bootconfig.h"
+#include "thirdparty/i80321reg.h"
 
 
 MACHINE_SETUP(iyonix)
 {
 	char tmpstr[1000];
 	uint32_t bootblock_addr;
-	struct bootconfig bootconfig;
+	struct bootconfig bootconfigData;
 
-	machine->machine_name = "Iyonix";
+	machine->machine_name = strdup("Iyonix");
 
 	cpu->cd.arm.coproc[6] = arm_coproc_i80321_6;
 
 	/*  TODO: Is it flash, or just RAM?  */
 	dev_ram_init(machine, 0xa0000000,
-	    machine->physical_ram_in_mb * 1048576, DEV_RAM_MIRROR, 0x0);
+	    machine->physical_ram_in_mb * 1048576, DEV_RAM_MIRROR, 0x0, NULL);
 
 	/*  Uncached RAM?  */
 	dev_ram_init(machine, 0xc0000000,
-	    machine->physical_ram_in_mb * 1048576, DEV_RAM_MIRROR, 0x0);
+	    machine->physical_ram_in_mb * 1048576, DEV_RAM_MIRROR, 0x0, NULL);
 
 	snprintf(tmpstr, sizeof(tmpstr), "i80321 irq=%s.cpu[%i].irq "
 	    "addr=0x%08x", machine->path, machine->bootstrap_cpu,
@@ -100,16 +100,16 @@ MACHINE_SETUP(iyonix)
 
 	bootblock_addr = machine->physical_ram_in_mb * 1048576 - 65536 - 16384;
 	cpu->cd.arm.r[0] = bootblock_addr;
-	memset(&bootconfig, 0, sizeof(bootconfig));
+	memset(&bootconfigData, 0, sizeof(bootconfigData));
 
 	store_32bit_word_in_host(cpu, (unsigned char *)
-	    &bootconfig.magic, BOOTCONFIG_MAGIC);
+	    &bootconfigData.magic, BOOTCONFIG_MAGIC);
 	store_32bit_word_in_host(cpu, (unsigned char *)
-	    &bootconfig.version, BOOTCONFIG_VERSION);
+	    &bootconfigData.version, BOOTCONFIG_VERSION);
 
 	/*  TODO: More fields.  */
 
-	store_buf(cpu, bootblock_addr, (char *)&bootconfig, sizeof(bootconfig));
+	store_buf(cpu, bootblock_addr, (char *)&bootconfigData, sizeof(bootconfigData));
 }
 
 
@@ -127,7 +127,7 @@ MACHINE_DEFAULT_RAM(iyonix)
 
 MACHINE_REGISTER(iyonix)
 {
-	MR_DEFAULT(iyonix, "Iyonix", ARCH_ARM, MACHINE_IYONIX);
+	MR_DEFAULT(iyonix, "Iyonix", MACHINE_IYONIX);
 
 	machine_entry_add_alias(me, "iyonix");
 
