@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2006-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2006-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -55,8 +55,8 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "sh4_exception.h"
-#include "sh4_scireg.h"
+#include "thirdparty/sh4_exception.h"
+#include "thirdparty/sh4_scireg.h"
 
 
 /*  This is not really implemented yet: (experimental)  */
@@ -71,7 +71,7 @@ MACHINE_SETUP(landisk)
 #endif
 	char tmpstr[300];
 
-	machine->machine_name = "Landisk USL-5P";
+	machine->machine_name = strdup("Landisk USL-5P");
 
 	/*  266.67 MHz SH4 CPU clock:  */
 	if (machine->emulated_hz == 0)
@@ -81,7 +81,7 @@ MACHINE_SETUP(landisk)
 	machine->cpus[machine->bootstrap_cpu]->cd.sh.pclock = 33333333;
 
 	/*  Note: 64 MB RAM at 0x0c000000, not at 0x00000000.  */
-	dev_ram_init(machine, 0x0c000000, 64 * 1048576, DEV_RAM_RAM, 0x0);
+	dev_ram_init(machine, 0x0c000000, 64 * 1048576, DEV_RAM_RAM, 0x0, NULL);
 
 	/*  wdc0 at obio0 port 0x14000000-0x1400000f irq 10  */
 	snprintf(tmpstr, sizeof(tmpstr), "wdc irq=%s.cpu[%i].irq[0x%x]"
@@ -90,7 +90,7 @@ MACHINE_SETUP(landisk)
 	device_add(machine, tmpstr);
 
 	/*  rsclock0 at shb0: RS5C313 real time clock  */
-	snprintf(tmpstr, sizeof(tmpstr), "rs5c313 addr=0x%"PRIx64,
+	snprintf(tmpstr, sizeof(tmpstr), "rs5c313 addr=0x%" PRIx64,
 	    (uint64_t) SCI_DEVICE_BASE);
 	device_add(machine, tmpstr);
 
@@ -136,7 +136,7 @@ MACHINE_DEFAULT_RAM(landisk)
 
 MACHINE_REGISTER(landisk)
 {
-	MR_DEFAULT(landisk, "Landisk", ARCH_SH, MACHINE_LANDISK);
+	MR_DEFAULT(landisk, "Landisk", MACHINE_LANDISK);
 	me->set_default_ram = machine_default_ram_landisk;
 
 	machine_entry_add_alias(me, "landisk");
