@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -59,7 +59,7 @@ MACHINE_SETUP(macppc)
 	uint64_t b, a;
 	int i;
 
-	machine->machine_name = "Macintosh (PPC)";
+	machine->machine_name = strdup("Macintosh (PPC)");
 	if (machine->emulated_hz == 0)
 		machine->emulated_hz = 40000000;
 
@@ -141,10 +141,8 @@ MACHINE_SETUP(macppc)
 	 *  r5 = OpenFirmware entry point.  NOTE: See
 	 *  cpu_ppc.c for the rest of this semi-ugly hack.
 	 */
-	dev_ram_init(machine, cpu->cd.ppc.of_emul_addr,
-	    0x1000, DEV_RAM_RAM, 0x0);
-	store_32bit_word(cpu, cpu->cd.ppc.of_emul_addr,
-	    0x44ee0002);
+	dev_ram_init(machine, cpu->cd.ppc.of_emul_addr, 0x1000, DEV_RAM_RAM, 0x0, NULL);
+	store_32bit_word(cpu, cpu->cd.ppc.of_emul_addr, 0x44ee0002);
 	cpu->cd.ppc.gpr[5] = cpu->cd.ppc.of_emul_addr;
 
 #if 0
@@ -185,7 +183,7 @@ MACHINE_DEFAULT_RAM(macppc)
 
 MACHINE_REGISTER(macppc)
 {
-	MR_DEFAULT(macppc, "Macintosh", ARCH_PPC, MACHINE_MACPPC);
+	MR_DEFAULT(macppc, "Macintosh", MACHINE_MACPPC);
 
 	machine_entry_add_alias(me, "macppc");
 
