@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -46,7 +46,7 @@ MACHINE_SETUP(netwinder)
 	char tmpstr[300];
 	struct pci_data *pci_bus;
 
-	machine->machine_name = "NetWinder";
+	machine->machine_name = strdup("NetWinder");
 
 	if (machine->physical_ram_in_mb > 256)
 		fprintf(stderr, "WARNING! Real NetWinders cannot"
@@ -57,7 +57,7 @@ MACHINE_SETUP(netwinder)
 
 	snprintf(tmpstr, sizeof(tmpstr), "footbridge irq=%s.cpu[%i].irq"
 	    " addr=0x42000000", machine->path, machine->bootstrap_cpu);
-	pci_bus = device_add(machine, tmpstr);
+	pci_bus = (struct pci_data *) device_add(machine, tmpstr);
 
 	if (machine->x11_md.in_use) {
 		bus_pci_add(machine, pci_bus, machine->memory,
@@ -85,7 +85,7 @@ MACHINE_DEFAULT_RAM(netwinder)
 
 MACHINE_REGISTER(netwinder)
 {
-	MR_DEFAULT(netwinder, "NetWinder", ARCH_ARM, MACHINE_NETWINDER);
+	MR_DEFAULT(netwinder, "NetWinder", MACHINE_NETWINDER);
 
 	machine_entry_add_alias(me, "netwinder");
 
