@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -60,7 +60,7 @@ MACHINE_SETUP(playstation2)
 	time_t timet;
 	struct tm *tm_ptr;
 
-	machine->machine_name = "Playstation 2";
+	machine->machine_name = strdup("Playstation 2");
 	cpu->byte_order = EMUL_LITTLE_ENDIAN;
 
 	if (machine->physical_ram_in_mb != 32)
@@ -89,7 +89,7 @@ MACHINE_SETUP(playstation2)
 	device_add(machine, "ps2_ether addr=0x14001000");
 
 	/*  TODO: how much?  */
-	dev_ram_init(machine, 0x1c000000, 4 * 1048576, DEV_RAM_RAM, 0);
+	dev_ram_init(machine, 0x1c000000, 4 * 1048576, DEV_RAM_RAM, 0, NULL);
 
 	/*  OHCI at SBUS irq 1:  */
 	snprintf(tmpstr, sizeof(tmpstr), "ohci addr=0x1f801600 irq="
@@ -111,7 +111,7 @@ MACHINE_SETUP(playstation2)
 
 
 	tmplen = 1000;
-	CHECK_ALLOCATION(tmp = malloc(tmplen));
+	CHECK_ALLOCATION(tmp = (char *) malloc(tmplen));
 
 	add_symbol_name(&machine->symbol_context,
 	    PLAYSTATION2_SIFBIOS, 0x10000, "[SIFBIOS entry]", 0, 0);
@@ -175,7 +175,7 @@ MACHINE_DEFAULT_RAM(playstation2)
 
 MACHINE_REGISTER(playstation2)
 {
-	MR_DEFAULT(playstation2, "Playstation 2", ARCH_MIPS, MACHINE_PS2);
+	MR_DEFAULT(playstation2, "Playstation 2", MACHINE_PS2);
 
 	machine_entry_add_alias(me, "playstation2");
 	machine_entry_add_alias(me, "ps2");
