@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -39,26 +39,29 @@
 #include "machine.h"
 #include "machine_pmax.h"
 #include "memory.h"
+#include "mips_cpu_types.h"
 #include "misc.h"
 
-#include "dec_prom.h"
-#include "dec_bootinfo.h"
-#include "dec_5100.h"
-#include "dec_kn01.h"
-#include "dec_kn02.h"  
-#include "dec_kn03.h"
-#include "dec_kmin.h"
-#include "dec_maxine.h"
+#include "thirdparty/dec_prom.h"
+#include "thirdparty/dec_bootinfo.h"
+#include "thirdparty/dec_5100.h"
+#include "thirdparty/dec_kn01.h"
+#include "thirdparty/dec_kn02.h"  
+#include "thirdparty/dec_kn03.h"
+#include "thirdparty/dec_kmin.h"
+#include "thirdparty/dec_maxine.h"
 
 #define	BOOTARG_BUFLEN		2000
 
 
 MACHINE_SETUP(pmax)
 {
-	char *framebuffer_console_name, *serial_console_name, *init_bootpath;
+	const char *framebuffer_console_name, *serial_console_name;
+	char bootpath[200];
+	char *init_bootpath;
 	int color_fb_flag, i;
 	int boot_scsi_boardnumber = 3, boot_net_boardnumber = 3;
-	char *turbochannel_default_gfx_card = "PMAG-BA";
+	const char *turbochannel_default_gfx_card = "PMAG-BA";
 		/*  PMAG-AA, -BA, -CA/DA/EA/FA, -JA, -RO, PMAGB-BA  */
 	struct xx {
 		struct btinfo_magic a;
@@ -86,7 +89,7 @@ MACHINE_SETUP(pmax)
 	case MACHINE_DEC_PMAX_3100:		/*  type  1, KN01  */
 		/*  Supposed to have 12MHz or 16.67MHz R2000 CPU, R2010 FPC,
 		    R2020 Memory coprocessor  */
-		machine->machine_name = "DEC PMAX 3100 (KN01)";
+		machine->machine_name = strdup("DEC PMAX 3100 (KN01)");
 
 		/*  12 MHz for 2100, 16.67 MHz for 3100  */
 		if (machine->emulated_hz == 0)
@@ -151,7 +154,7 @@ MACHINE_SETUP(pmax)
 	case MACHINE_DEC_3MAX_5000:		/*  type  2, KN02  */
 		/*  Supposed to have 25MHz R3000 CPU, R3010 FPC,  */
 		/*  and a R3220 Memory coprocessor  */
-		machine->machine_name = "DECstation 5000/200 (3MAX, KN02)";
+		machine->machine_name = strdup("DECstation 5000/200 (3MAX, KN02)");
 
 		if (machine->emulated_hz == 0)
 			machine->emulated_hz = 25000000;
@@ -249,18 +252,18 @@ MACHINE_SETUP(pmax)
 		break;
 
 	case MACHINE_DEC_3MIN_5000:		/*  type 3, KN02BA  */
-		machine->machine_name = "DECstation 5000/112 or 145 (3MIN,"
-		    " KN02BA)";
+		machine->machine_name = strdup("DECstation 5000/112, 125, or 145 (3MIN,"
+		    " KN02BA)");
 		if (machine->emulated_hz == 0)
 			machine->emulated_hz = 33000000;
 		if (machine->physical_ram_in_mb > 128)
 			fprintf(stderr, "WARNING! Real 3MIN machines cannot "
 			    "have more than 128MB RAM. Continuing anyway.\n");
 
-		/*  KMIN interrupts:  */
-fatal("TODO: Legacy rewrite\n");
-abort();
-//		machine->md_interrupt = kmin_interrupt;
+		/*  KN02BA "3min" TurboChannel interrupt controller:  */
+		snprintf(tmpstr, sizeof(tmpstr), "kn02ba irq=%s.cpu[%i].%i",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC3);
+		device_add(machine, tmpstr);
 
 		/*
 		 *  tc0 at mainbus0: 12.5 MHz clock  (0x10000000,slotsize=64MB)
@@ -279,27 +282,32 @@ abort();
 		 *	SCSI ID 7			(0x1c300000) slot 12
 		 *  dma for asc0			(0x1c380000) slot 14
 		 */
-fatal("TODO: dec_ioasic legacy rewrite\n");
-abort();
-//		machine->md_int.dec_ioasic_data = dev_dec_ioasic_init(cpu,
-//		    mem, 0x1c000000, 0);
-fatal("TODO: kmin dev_le_init.\n");
-abort();
+
+////fatal("TODO: kmin dev_le_init.\n");
+////abort();
 //		dev_le_init(machine, mem, 0x1c0c0000, 0, 0,
 //		    KMIN_INTR_LANCE + 8, 4 * 65536);
-		dev_scc_init(machine, mem, 0x1c100000, KMIN_INTR_SCC_0 + 8,
+
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i.kn02ba.0x%x",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC3, KMIN_INTR_SCC_0);
+		dev_scc_init(machine, mem, 0x1c100000, tmpstr,
 		    machine->x11_md.in_use, 0, 1);
-		dev_scc_init(machine, mem, 0x1c180000, KMIN_INTR_SCC_1 + 8,
+
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i.kn02ba.0x%x",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC3, KMIN_INTR_SCC_1);
+		dev_scc_init(machine, mem, 0x1c180000, tmpstr,
 		    machine->x11_md.in_use, 1, 1);
-fatal("TODO: mc146818 irq\n");
-abort();
-//		dev_mc146818_init(machine, mem, 0x1c200000, 
-//KMIN_INTR_CLOCK + 8,
-//		    MC146818_DEC, 1);
-fatal("TODO: kmin asc init\n");
-abort();
-//		dev_asc_init(machine, mem, 0x1c300000, KMIN_INTR_SCSI +8,
-//		    NULL, DEV_ASC_DEC, NULL, NULL);
+
+
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i.kn02ba.0x%x",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC3, KMIN_INTR_CLOCK);
+		dev_mc146818_init(machine, mem,
+		    KMIN_SYS_CLOCK, tmpstr, MC146818_DEC, 1);
+		    
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i.kn02ba.0x%x",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC3, KMIN_INTR_SCSI);
+		dev_asc_init(machine, mem, 0x1c300000, tmpstr,
+		    NULL, DEV_ASC_DEC, NULL, NULL);
 
 		/*
 		 *  TURBOchannel slots 0, 1, and 2 are free for
@@ -308,21 +316,31 @@ abort();
 		 *
 		 *  TODO: irqs 
 		 */
-fatal("TODO: turbochannel init rewrite!\n");
-abort();
-#if 0
-		dev_turbochannel_init(machine, mem, 0, 0x10000000, 0x103fffff,
+
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC0);
+		dev_turbochannel_init(machine, mem, 0,
+		    KMIN_PHYS_TC_0_START, KMIN_PHYS_TC_0_END,
 		    machine->n_gfx_cards >= 1?
-			turbochannel_default_gfx_card : "", KMIN_INT_TC0);
+			turbochannel_default_gfx_card : "",
+		    tmpstr);
 
-		dev_turbochannel_init(machine, mem, 1, 0x14000000, 0x143fffff,
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC1);
+		dev_turbochannel_init(machine, mem, 0,
+		    KMIN_PHYS_TC_1_START, KMIN_PHYS_TC_1_END,
 		    machine->n_gfx_cards >= 2?
-		    turbochannel_default_gfx_card : "", KMIN_INT_TC1);
+			turbochannel_default_gfx_card : "",
+		    tmpstr);
 
-		dev_turbochannel_init(machine, mem, 2, 0x18000000, 0x183fffff,
+		snprintf(tmpstr, sizeof(tmpstr), "%s.cpu[%i].%i",
+		    machine->path, machine->bootstrap_cpu, KMIN_INT_TC2);
+		dev_turbochannel_init(machine, mem, 0,
+		    KMIN_PHYS_TC_2_START, KMIN_PHYS_TC_2_END,
 		    machine->n_gfx_cards >= 3?
-			turbochannel_default_gfx_card : "", KMIN_INT_TC2);
-#endif
+			turbochannel_default_gfx_card : "",
+		    tmpstr);
+
 		/*  (kmin shared irq numbers (IP) are offset by +8 in the
 		    emulator)  */
 		/*  kmin_csr = dev_kmin_init(cpu, mem, KMIN_REG_INTR);  */
@@ -332,7 +350,7 @@ abort();
 		break;
 
 	case MACHINE_DEC_3MAXPLUS_5000:	/*  type 4, KN03  */
-		machine->machine_name = "DECsystem 5900 or 5000 (3MAX+) (KN03)";
+		machine->machine_name = strdup("DECsystem 5900 or 5000 (3MAX+) (KN03)");
 
 		/*  5000/240 (KN03-GA, R3000): 40 MHz  */
 		/*  5000/260 (KN05-NB, R4000): 60 MHz  */
@@ -431,7 +449,7 @@ abort();
 		break;
 
 	case MACHINE_DEC_5800:		/*  type 5, KN5800  */
-		machine->machine_name = "DECsystem 5800";
+		machine->machine_name = strdup("DECsystem 5800");
 
 		/*  TODO: this is incorrect, banks multiply by 8 etc  */
 		if (machine->physical_ram_in_mb < 48)
@@ -468,7 +486,7 @@ abort();
 		break;
 
 	case MACHINE_DEC_5400:		/*  type 6, KN210  */
-		machine->machine_name = "DECsystem 5400 (KN210)";
+		machine->machine_name = strdup("DECsystem 5400 (KN210)");
 		/*
 		 *  Misc. info from the KN210 manual:
 		 *
@@ -502,8 +520,8 @@ abort();
 		break;
 
 	case MACHINE_DEC_MAXINE_5000:	/*  type 7, KN02CA  */
-		machine->machine_name = "Personal DECstation 5000/xxx "
-		    "(MAXINE) (KN02CA)";
+		machine->machine_name = strdup("Personal DECstation 5000/xxx "
+		    "(MAXINE) (KN02CA)");
 		if (machine->emulated_hz == 0)
 			machine->emulated_hz = 33000000;
 
@@ -578,8 +596,8 @@ fatal("TODO: xine dev_le_init rewrite\n");
 abort();
 //		dev_le_init(machine, mem, 0x1c0c0000, 0, 0,
 //		    XINE_INTR_LANCE +8, 4*65536);
-		dev_scc_init(machine, mem, 0x1c100000,
-		    XINE_INTR_SCC_0 +8, machine->x11_md.in_use, 0, 1);
+//		dev_scc_init(machine, mem, 0x1c100000,
+//		    XINE_INTR_SCC_0 +8, machine->x11_md.in_use, 0, 1);
 fatal("TODO: mc146818 irq\n");
 abort();
 //		dev_mc146818_init(machine, mem, 0x1c200000,
@@ -594,7 +612,7 @@ abort();
 		break;
 
 	case MACHINE_DEC_5500:	/*  type 11, KN220  */
-		machine->machine_name = "DECsystem 5500 (KN220)";
+		machine->machine_name = strdup("DECsystem 5500 (KN220)");
 
 		/*
 		 *  According to NetBSD's pmax ports page:
@@ -642,7 +660,7 @@ abort();
 		break;
 
 	case MACHINE_DEC_MIPSMATE_5100:	/*  type 12  */
-		machine->machine_name = "DEC MIPSMATE 5100 (KN230)";
+		machine->machine_name = strdup("DEC MIPSMATE 5100 (KN230)");
 
 		if (machine->emulated_hz == 0)
 			machine->emulated_hz = 20000000;
@@ -659,7 +677,7 @@ abort();
 
 		/*  KN230 mainbus / interrupt controller:  */
 		snprintf(tmpstr, sizeof(tmpstr),
-		    "kn230 addr=0x%"PRIx64, (uint64_t) KN230_SYS_ICSR);
+		    "kn230 addr=0x%" PRIx64, (uint64_t) KN230_SYS_ICSR);
 		device_add(machine, tmpstr);
 
 		/*
@@ -710,7 +728,14 @@ abort();
 	 *  it was 0x1e......, so we need this hack:
 	 */
 	dev_ram_init(machine, 0xa0000000, 0x20000000,
-	    DEV_RAM_MIRROR | DEV_RAM_MIGHT_POINT_TO_DEVICES, 0x0);
+	    DEV_RAM_MIRROR | DEV_RAM_MIGHT_POINT_TO_DEVICES, 0x0, NULL);
+
+
+	if (machine->machine_type == MACHINE_PMAX &&
+	    cpu->cd.mips.cpu_type.mmu_model == MMU3K)
+		add_symbol_name(&machine->symbol_context,
+		    0x9fff0000, 0x10000, "r2k3k_cache", 0, 0);
+
 
 	if (!machine->prom_emulation)
 		return;
@@ -778,17 +803,14 @@ abort();
 	 *  slot number of the boot device.
 	 *
 	 *  'rz' for disks, 'tz' for tapes.
-	 *
-	 *  TODO:  Make this nicer.
 	 */
 	{
-		char bootpath[200];
 #if 0
 		if (machine->machine_subtype == MACHINE_DEC_PMAX_3100)
 			strlcpy(bootpath, "rz(0,0,0)", sizeof(bootpath));
 		else
 #endif
-			strlcpy(bootpath, "5/rz1/", sizeof(bootpath));
+			strlcpy(bootpath, "X/rzY/", sizeof(bootpath));
 
 		if (machine->bootdev_id < 0 || machine->force_netboot) {
 			/*  tftp boot:  */
@@ -806,7 +828,7 @@ abort();
 		init_bootpath = bootpath;
 	}
 
-	CHECK_ALLOCATION(machine->bootarg = malloc(BOOTARG_BUFLEN));
+	CHECK_ALLOCATION(machine->bootarg = (char *) malloc(BOOTARG_BUFLEN));
 	strlcpy(machine->bootarg, init_bootpath, BOOTARG_BUFLEN);
 	if (strlcat(machine->bootarg, machine->boot_kernel_filename,
 	    BOOTARG_BUFLEN) > BOOTARG_BUFLEN) {
@@ -814,7 +836,7 @@ abort();
 		exit(1);
 	}
 
-	machine->bootstr = "boot";
+	machine->bootstr = strdup("boot");
 
 	store_string(cpu, DEC_PROM_INITIAL_ARGV+0x10, machine->bootstr);
 	store_string(cpu, DEC_PROM_INITIAL_ARGV+0x70, machine->bootarg);
@@ -851,21 +873,20 @@ abort();
 
 	store_buf(cpu, BOOTINFO_ADDR, (char *)&xx, sizeof(xx));
 
-	CHECK_ALLOCATION(machine->md.pmax =
+	CHECK_ALLOCATION(machine->md.pmax = (struct machine_pmax *)
 	    malloc(sizeof(struct machine_pmax)));
 	memset(machine->md.pmax, 0, sizeof(struct machine_pmax));
 
 	/*  The system's memmap:  */
-	CHECK_ALLOCATION(machine->md.pmax->memmap =
+	CHECK_ALLOCATION(machine->md.pmax->memmap = (struct dec_memmap *)
 	    malloc(sizeof(struct dec_memmap)));
 	store_32bit_word_in_host(cpu,
 	    (unsigned char *)&machine->md.pmax->memmap->pagesize, 4096);
-	{
-		unsigned int i;
-		for (i=0; i<sizeof(machine->md.pmax->memmap->bitmap); i++)
-			machine->md.pmax->memmap->bitmap[i] = ((int)i * 4096*8 <
-			    1048576*machine->physical_ram_in_mb)? 0xff : 0x00;
-	}
+
+	for (unsigned int j=0; j<sizeof(machine->md.pmax->memmap->bitmap); j++)
+		machine->md.pmax->memmap->bitmap[j] = (j * 4096*8 <
+		    1048576*machine->physical_ram_in_mb)? 0xff : 0x00;
+
 	store_buf(cpu, DEC_MEMMAP_ADDR,
 	    (char *)machine->md.pmax->memmap, sizeof(struct dec_memmap));
 
@@ -885,7 +906,7 @@ abort();
 	 */
 	{
 		char tmps[300];
-		snprintf(tmps, sizeof(tmps), "cca=%"PRIx32,
+		snprintf(tmps, sizeof(tmps), "cca=%" PRIx32,
 		    (uint32_t) (DEC_DECCCA_BASEADDR + 0xa0000000ULL));
 		add_environment_string(cpu, tmps, &addr);
 	}
@@ -898,14 +919,14 @@ abort();
 		tmps[sizeof(tmps)-1] = '\0';
 		add_environment_string(cpu, tmps, &addr);
 
-		snprintf(tmps, sizeof(tmps), "bitmap=0x%"PRIx32, (uint32_t)
+		snprintf(tmps, sizeof(tmps), "bitmap=0x%" PRIx32, (uint32_t)
 		    ( (DEC_MEMMAP_ADDR + sizeof(uint32_t) /* skip the
 			page size and point to the memmap */
 		    ) & 0xffffffffULL) );
 		tmps[sizeof(tmps)-1] = '\0';
 		add_environment_string(cpu, tmps, &addr);
 
-		snprintf(tmps, sizeof(tmps), "bitmaplen=0x%"PRIx32, (uint32_t)
+		snprintf(tmps, sizeof(tmps), "bitmaplen=0x%" PRIx32, (uint32_t)
 		    ( machine->physical_ram_in_mb * 1048576 / 4096 / 8) );
 		tmps[sizeof(tmps)-1] = '\0';
 		add_environment_string(cpu, tmps, &addr);
@@ -959,7 +980,7 @@ MACHINE_DEFAULT_RAM(pmax)
 
 MACHINE_REGISTER(pmax)
 {
-	MR_DEFAULT(pmax, "DECstation/DECsystem", ARCH_MIPS, MACHINE_PMAX);
+	MR_DEFAULT(pmax, "DECstation/DECsystem", MACHINE_PMAX);
 
 	machine_entry_add_alias(me, "decstation");
 	machine_entry_add_alias(me, "decsystem");
