@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -50,7 +50,7 @@ MACHINE_SETUP(pmppc)
 	/*
 	 *  NetBSD/pmppc (http://www.netbsd.org/ports/pmppc/)
 	 */
-	machine->machine_name = "Artesyn's PM/PPC board";
+	machine->machine_name = strdup("Artesyn's PM/PPC board");
 
 	/*  Bogus default speed = 33 MHz  */
 	if (machine->emulated_hz == 0)
@@ -60,7 +60,7 @@ MACHINE_SETUP(pmppc)
 	device_add(machine, "pmppc");
 
 	/*  PCI and Interrupt controller:  */
-	pci_data = device_add(machine, "cpc700");
+	pci_data = (struct pci_data *) device_add(machine, "cpc700");
 
 	/*  RTC at "ext int 5" = "int 25" in IBM jargon, int
 	    31-25 = 6 for the rest of us.  */
@@ -81,7 +81,7 @@ MACHINE_DEFAULT_CPU(pmppc)
 
 MACHINE_REGISTER(pmppc)
 {
-	MR_DEFAULT(pmppc, "Artesyn's PM/PPC board", ARCH_PPC, MACHINE_PMPPC);
+	MR_DEFAULT(pmppc, "Artesyn's PM/PPC board", MACHINE_PMPPC);
 
 	machine_entry_add_alias(me, "pmppc");
 }
