@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -48,13 +48,13 @@ MACHINE_SETUP(prep)
 	char tmpstr[300];
 
 	struct pci_data *pci_data;
-	char *model_name = "";
+	const char *model_name = "";
 
 	switch (machine->machine_subtype) {
 
 	case MACHINE_PREP_IBM6050:
 		machine->machine_name =
-		    "PowerPC Reference Platform, IBM 6050/6070";
+		    strdup("PowerPC Reference Platform, IBM 6050/6070");
 		model_name = "IBM PPS Model 6050/6070 (E)";
 
 		if (machine->emulated_hz == 0)
@@ -66,7 +66,7 @@ MACHINE_SETUP(prep)
 
 		snprintf(tmpstr, sizeof(tmpstr), "eagle irq=%s.cpu[%i]",
 		    machine->path, machine->bootstrap_cpu);
-		pci_data = device_add(machine, tmpstr);
+		pci_data = (struct pci_data *) device_add(machine, tmpstr);
 
 		bus_pci_add(machine, pci_data, machine->memory,
 		    0, 13, 0, "dec21143");
@@ -78,7 +78,7 @@ MACHINE_SETUP(prep)
 		break;
 
 	case MACHINE_PREP_MVME2400:
-		machine->machine_name = "PowerPC Reference Platform, MVME2400";
+		machine->machine_name = strdup("PowerPC Reference Platform, MVME2400");
 
 		/*  TODO: _EXACT_ model name for mvme2400?  */
 		model_name = "MOT MVME2400";
@@ -89,7 +89,7 @@ MACHINE_SETUP(prep)
 
 		snprintf(tmpstr, sizeof(tmpstr), "eagle irq=%s.cpu[%i]",
 		    machine->path, machine->bootstrap_cpu);
-		pci_data = device_add(machine, tmpstr);
+		pci_data = (struct pci_data *) device_add(machine, tmpstr);
 
 		break;
 
@@ -176,7 +176,7 @@ MACHINE_DEFAULT_RAM(prep)
 
 MACHINE_REGISTER(prep)
 {
-	MR_DEFAULT(prep, "PowerPC Reference Platform", ARCH_PPC, MACHINE_PREP);
+	MR_DEFAULT(prep, "PowerPC Reference Platform", MACHINE_PREP);
 
 	machine_entry_add_alias(me, "prep");
 	me->set_default_ram = machine_default_ram_prep;
