@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2005-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2005-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -51,11 +51,12 @@
 
 #include "cpu.h"
 #include "device.h"
+#include "devices.h"
 #include "machine.h"
 #include "memory.h"
 #include "misc.h"
 
-#include "sh4_exception.h"
+#include "thirdparty/sh4_exception.h"
 
 #include "testmachine/dev_cons.h"
 #include "testmachine/dev_disk.h"
@@ -74,7 +75,7 @@
  */
 static void default_test(struct machine *machine, struct cpu *cpu)
 {
-	char tmpstr[1000];
+	char tmpstr[2000];
 	char base_irq[1000];
 	char end_of_base_irq[50];
 
@@ -87,7 +88,7 @@ static void default_test(struct machine *machine, struct cpu *cpu)
 	 *  ARM and SH are dealt with here.
 	 */
 
-	switch (machine->arch) {
+	switch (cpu->cpu_family->arch) {
 
 	case ARCH_ARM:
 		snprintf(end_of_base_irq, sizeof(end_of_base_irq), ".irq");
@@ -105,95 +106,60 @@ static void default_test(struct machine *machine, struct cpu *cpu)
 	snprintf(base_irq, sizeof(base_irq), "%s.cpu[%i]%s",
 	    machine->path, machine->bootstrap_cpu, end_of_base_irq);
 
-	snprintf(tmpstr, sizeof(tmpstr), "irqc addr=0x%"PRIx64" irq=%s",
+	snprintf(tmpstr, sizeof(tmpstr), "irqc addr=0x%" PRIx64" irq=%s",
 	    (uint64_t) DEV_IRQC_ADDRESS, base_irq);
 	device_add(machine, tmpstr);
 
 
 	/*  Now, add the other devices:  */
 
-	snprintf(tmpstr, sizeof(tmpstr), "cons addr=0x%"PRIx64
+	snprintf(tmpstr, sizeof(tmpstr), "cons addr=0x%" PRIx64
 	    " irq=%s.irqc.2 in_use=%i",
-	    (uint64_t) DEV_CONS_ADDRESS, base_irq, machine->arch != ARCH_SH);
+	    (uint64_t) DEV_CONS_ADDRESS, base_irq, cpu->cpu_family->arch != ARCH_SH);
 	machine->main_console_handle = (size_t)device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "mp addr=0x%"PRIx64" irq=%s%sirqc.6",
+	snprintf(tmpstr, sizeof(tmpstr), "mp addr=0x%" PRIx64" irq=%s%sirqc.6",
 	    (uint64_t) DEV_MP_ADDRESS,
 	    end_of_base_irq[0]? end_of_base_irq + 1 : "",
 	    end_of_base_irq[0]? "." : "");
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "fbctrl addr=0x%"PRIx64,
+	snprintf(tmpstr, sizeof(tmpstr), "fbctrl addr=0x%" PRIx64,
 	    (uint64_t) DEV_FBCTRL_ADDRESS);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "disk addr=0x%"PRIx64,
+	snprintf(tmpstr, sizeof(tmpstr), "disk addr=0x%" PRIx64,
 	    (uint64_t) DEV_DISK_ADDRESS);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "ether addr=0x%"PRIx64" irq=%s.irqc.3",
+	snprintf(tmpstr, sizeof(tmpstr), "ether addr=0x%" PRIx64" irq=%s.irqc.3",
 	    (uint64_t) DEV_ETHER_ADDRESS, base_irq);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "rtc addr=0x%"PRIx64" irq=%s.irqc.4",
+	snprintf(tmpstr, sizeof(tmpstr), "rtc addr=0x%" PRIx64" irq=%s.irqc.4",
 	    (uint64_t) DEV_RTC_ADDRESS, base_irq);
 	device_add(machine, tmpstr);
 }
 
 
-MACHINE_SETUP(barealpha)
-{
-	machine->machine_name = "Generic \"bare\" Alpha machine";
-}
-
-
-MACHINE_SETUP(testalpha)
-{
-	machine->machine_name = "Alpha test machine";
-
-	default_test(machine, cpu);
-}
-
-
-MACHINE_DEFAULT_CPU(barealpha)
-{
-	machine->cpu_name = strdup("21264");
-}
-
-
-MACHINE_DEFAULT_CPU(testalpha)
-{
-	machine->cpu_name = strdup("21264");
-}
-
-
-MACHINE_REGISTER(barealpha)
-{
-	MR_DEFAULT(barealpha, "Generic \"bare\" Alpha machine",
-	    ARCH_ALPHA, MACHINE_BAREALPHA);
-
-	machine_entry_add_alias(me, "barealpha");
-}
-
-
-MACHINE_REGISTER(testalpha)
-{
-	MR_DEFAULT(testalpha, "Test-machine for Alpha",
-	    ARCH_ALPHA, MACHINE_TESTALPHA);
-
-	machine_entry_add_alias(me, "testalpha");
-}
-
-
 MACHINE_SETUP(barearm)
 {
-	machine->machine_name = "Generic \"bare\" ARM machine";
+	machine->machine_name = strdup("Generic \"bare\" ARM machine");
+
+#if 1
+	// An experiment with running a particular firmware image on a device;
+	// move some other place when/if it works?
+	cpu->byte_order = EMUL_BIG_ENDIAN;
+
+	dev_ram_init(machine, 0xa0000000, 128 * 1048576, DEV_RAM_MIRROR
+		| DEV_RAM_MIGHT_POINT_TO_DEVICES, 0x00000000, "ram_mirror");
+#endif
 }
 
 
 MACHINE_SETUP(testarm)
 {
-	machine->machine_name = "ARM test machine";
+	machine->machine_name = strdup("ARM test machine");
 
 	default_test(machine, cpu);
 
@@ -224,8 +190,7 @@ MACHINE_DEFAULT_CPU(testarm)
 
 MACHINE_REGISTER(barearm)
 {
-	MR_DEFAULT(barearm, "Generic \"bare\" ARM machine",
-	    ARCH_ARM, MACHINE_BAREARM);
+	MR_DEFAULT(barearm, "Generic \"bare\" ARM machine", MACHINE_BAREARM);
 
 	machine_entry_add_alias(me, "barearm");
 }
@@ -233,66 +198,42 @@ MACHINE_REGISTER(barearm)
 
 MACHINE_REGISTER(testarm)
 {
-	MR_DEFAULT(testarm, "Test-machine for ARM", ARCH_ARM, MACHINE_TESTARM);
+	MR_DEFAULT(testarm, "Test-machine for ARM", MACHINE_TESTARM);
 
 	machine_entry_add_alias(me, "testarm");
 }
 
 
 
-MACHINE_SETUP(barem32r)
+MACHINE_SETUP(barei960)
 {
-	machine->machine_name = "Generic \"bare\" M32R machine";
+	machine->machine_name = strdup("Generic \"bare\" i960 machine");
 }
 
 
-MACHINE_SETUP(testm32r)
+MACHINE_DEFAULT_CPU(barei960)
 {
-	machine->machine_name = "M32R test machine";
-
-	default_test(machine, cpu);
+	machine->cpu_name = strdup("i960CA");
 }
 
 
-MACHINE_DEFAULT_CPU(barem32r)
+MACHINE_REGISTER(barei960)
 {
-	machine->cpu_name = strdup("M32R");
-}
+	MR_DEFAULT(barei960, "Generic \"bare\" i960 machine", MACHINE_BAREI960);
 
-
-MACHINE_DEFAULT_CPU(testm32r)
-{
-	machine->cpu_name = strdup("M32R");
-}
-
-
-MACHINE_REGISTER(barem32r)
-{
-	MR_DEFAULT(barem32r, "Generic \"bare\" M32R machine",
-	    ARCH_M32R, MACHINE_BAREM32R);
-
-	machine_entry_add_alias(me, "barem32r");
-}
-
-
-MACHINE_REGISTER(testm32r)
-{
-	MR_DEFAULT(testm32r, "Test-machine for M32R",
-	    ARCH_M32R, MACHINE_TESTM32R);
-
-	machine_entry_add_alias(me, "testm32r");
+	machine_entry_add_alias(me, "barei960");
 }
 
 
 MACHINE_SETUP(barem88k)
 {
-	machine->machine_name = "Generic \"bare\" M88K machine";
+	machine->machine_name = strdup("Generic \"bare\" M88K machine");
 }
 
 
 MACHINE_SETUP(testm88k)
 {
-	machine->machine_name = "M88K test machine";
+	machine->machine_name = strdup("M88K test machine");
 
 	default_test(machine, cpu);
 }
@@ -313,7 +254,7 @@ MACHINE_DEFAULT_CPU(testm88k)
 MACHINE_REGISTER(barem88k)
 {
 	MR_DEFAULT(barem88k, "Generic \"bare\" M88K machine",
-	    ARCH_M88K, MACHINE_BAREM88K);
+	    MACHINE_BAREM88K);
 
 	machine_entry_add_alias(me, "barem88k");
 }
@@ -322,7 +263,7 @@ MACHINE_REGISTER(barem88k)
 MACHINE_REGISTER(testm88k)
 {
 	MR_DEFAULT(testm88k, "Test-machine for M88K",
-	    ARCH_M88K, MACHINE_TESTM88K);
+	    MACHINE_TESTM88K);
 
 	machine_entry_add_alias(me, "testm88k");
 }
@@ -330,7 +271,7 @@ MACHINE_REGISTER(testm88k)
 
 MACHINE_SETUP(baremips)
 {
-	machine->machine_name = "Generic \"bare\" MIPS machine";
+	machine->machine_name = strdup("Generic \"bare\" MIPS machine");
 	cpu->byte_order = EMUL_BIG_ENDIAN;
 }
 
@@ -354,32 +295,32 @@ MACHINE_SETUP(testmips)
 
 	char tmpstr[300];
 
-	machine->machine_name = "MIPS test machine";
+	machine->machine_name = strdup("MIPS test machine");
 	cpu->byte_order = EMUL_BIG_ENDIAN;
 
-	snprintf(tmpstr, sizeof(tmpstr), "cons addr=0x%"PRIx64" irq=%s."
+	snprintf(tmpstr, sizeof(tmpstr), "cons addr=0x%" PRIx64" irq=%s."
 	    "cpu[%i].2", (uint64_t) DEV_CONS_ADDRESS, machine->path,
 	    machine->bootstrap_cpu);
 	machine->main_console_handle = (size_t)device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "mp addr=0x%"PRIx64" irq=6",
+	snprintf(tmpstr, sizeof(tmpstr), "mp addr=0x%" PRIx64" irq=6",
 	    (uint64_t) DEV_MP_ADDRESS);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "fbctrl addr=0x%"PRIx64,
+	snprintf(tmpstr, sizeof(tmpstr), "fbctrl addr=0x%" PRIx64,
 	    (uint64_t) DEV_FBCTRL_ADDRESS);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "disk addr=0x%"PRIx64,
+	snprintf(tmpstr, sizeof(tmpstr), "disk addr=0x%" PRIx64,
 	    (uint64_t) DEV_DISK_ADDRESS);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "ether addr=0x%"PRIx64" irq=%s."
+	snprintf(tmpstr, sizeof(tmpstr), "ether addr=0x%" PRIx64" irq=%s."
 	    "cpu[%i].3", (uint64_t) DEV_ETHER_ADDRESS, machine->path,
 	    machine->bootstrap_cpu);
 	device_add(machine, tmpstr);
 
-	snprintf(tmpstr, sizeof(tmpstr), "rtc addr=0x%"PRIx64" irq=%s."
+	snprintf(tmpstr, sizeof(tmpstr), "rtc addr=0x%" PRIx64" irq=%s."
 	    "cpu[%i].4", (uint64_t) DEV_RTC_ADDRESS, machine->path,
 	    machine->bootstrap_cpu);
 	device_add(machine, tmpstr);
@@ -401,7 +342,7 @@ MACHINE_DEFAULT_CPU(testmips)
 MACHINE_REGISTER(baremips)
 {
 	MR_DEFAULT(baremips, "Generic \"bare\" MIPS machine",
-	    ARCH_MIPS, MACHINE_BAREMIPS);
+	    MACHINE_BAREMIPS);
 
 	machine_entry_add_alias(me, "baremips");
 }
@@ -410,7 +351,7 @@ MACHINE_REGISTER(baremips)
 MACHINE_REGISTER(testmips)
 {
 	MR_DEFAULT(testmips, "Test-machine for MIPS",
-	    ARCH_MIPS, MACHINE_TESTMIPS);
+	    MACHINE_TESTMIPS);
 
 	machine_entry_add_alias(me, "testmips");
 }
@@ -418,13 +359,13 @@ MACHINE_REGISTER(testmips)
 
 MACHINE_SETUP(bareppc)
 {
-	machine->machine_name = "Generic \"bare\" PPC machine";
+	machine->machine_name = strdup("Generic \"bare\" PPC machine");
 }
 
 
 MACHINE_SETUP(testppc)
 {
-	machine->machine_name = "PPC test machine";
+	machine->machine_name = strdup("PPC test machine");
 
 	default_test(machine, cpu);
 }
@@ -445,7 +386,7 @@ MACHINE_DEFAULT_CPU(testppc)
 MACHINE_REGISTER(bareppc)
 {
 	MR_DEFAULT(bareppc, "Generic \"bare\" PPC machine",
-	    ARCH_PPC, MACHINE_BAREPPC);
+	    MACHINE_BAREPPC);
 
 	machine_entry_add_alias(me, "bareppc");
 }
@@ -453,21 +394,63 @@ MACHINE_REGISTER(bareppc)
 
 MACHINE_REGISTER(testppc)
 {
-	MR_DEFAULT(testppc, "Test-machine for PPC", ARCH_PPC, MACHINE_TESTPPC);
+	MR_DEFAULT(testppc, "Test-machine for PPC", MACHINE_TESTPPC);
 
 	machine_entry_add_alias(me, "testppc");
 }
 
 
+MACHINE_SETUP(bareriscv)
+{
+	machine->machine_name = strdup("Generic \"bare\" RISC-V machine");
+}
+
+
+MACHINE_SETUP(testriscv)
+{
+	machine->machine_name = strdup("RISC-V test machine");
+
+	default_test(machine, cpu);
+}
+
+
+MACHINE_DEFAULT_CPU(bareriscv)
+{
+	machine->cpu_name = strdup("RV64GC");
+}
+
+
+MACHINE_DEFAULT_CPU(testriscv)
+{
+	machine->cpu_name = strdup("RV64GC");
+}
+
+
+MACHINE_REGISTER(bareriscv)
+{
+	MR_DEFAULT(bareriscv, "Generic \"bare\" RISC-V machine", MACHINE_BARERISCV);
+
+	machine_entry_add_alias(me, "bareriscv");
+}
+
+
+MACHINE_REGISTER(testriscv)
+{
+	MR_DEFAULT(testriscv, "Test-machine for RISC-V", MACHINE_TESTRISCV);
+
+	machine_entry_add_alias(me, "testriscv");
+}
+
+
 MACHINE_SETUP(baresh)
 {
-	machine->machine_name = "Generic \"bare\" SH machine";
+	machine->machine_name = strdup("Generic \"bare\" SH machine");
 }
 
 
 MACHINE_SETUP(testsh)
 {
-	machine->machine_name = "SH test machine";
+	machine->machine_name = strdup("SH test machine");
 
 	default_test(machine, cpu);
 }
@@ -488,7 +471,7 @@ MACHINE_DEFAULT_CPU(testsh)
 MACHINE_REGISTER(baresh)
 {
 	MR_DEFAULT(baresh, "Generic \"bare\" SH machine",
-	    ARCH_SH, MACHINE_BARESH);
+	    MACHINE_BARESH);
 
 	machine_entry_add_alias(me, "baresh");
 }
@@ -496,53 +479,10 @@ MACHINE_REGISTER(baresh)
 
 MACHINE_REGISTER(testsh)
 {
-	MR_DEFAULT(testsh, "Test-machine for SH", ARCH_SH, MACHINE_TESTSH);
+	MR_DEFAULT(testsh, "Test-machine for SH", MACHINE_TESTSH);
 
 	machine_entry_add_alias(me, "testsh");
 }
 
 
-MACHINE_SETUP(baresparc)
-{
-	machine->machine_name = "Generic \"bare\" SPARC machine";
-}
-
-
-MACHINE_SETUP(testsparc)
-{
-	machine->machine_name = "SPARC test machine";
-
-	default_test(machine, cpu);
-}
-
-
-MACHINE_DEFAULT_CPU(baresparc)
-{
-	machine->cpu_name = strdup("UltraSPARC");
-}
-
-
-MACHINE_DEFAULT_CPU(testsparc)
-{
-	machine->cpu_name = strdup("UltraSPARC");
-}
-
-
-MACHINE_REGISTER(baresparc)
-{
-	MR_DEFAULT(baresparc, "Generic \"bare\" SPARC machine",
-	    ARCH_SPARC, MACHINE_BARESPARC);
-
-	machine_entry_add_alias(me, "baresparc");
-}
-
-
-MACHINE_REGISTER(testsparc)
-{
-	MR_DEFAULT(testsparc, "Test-machine for SPARC",
-	    ARCH_SPARC, MACHINE_TESTSPARC);
-
-	machine_entry_add_alias(me, "testsparc");
-}
-
 
