@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -30,10 +30,10 @@
  *  (Read the README file in this directory for more details.)
  *
  *
- *  NOTE: The 'extra' argument used in many functions in this file is a pointer
- *  to something unique for each NIC (i.e. the NIC itself :-), so that if
- *  multiple NICs are emulated concurrently, they will not get packets that
- *  are meant for some other controller.
+ *  NOTE: The 'nic' argument used in many functions in this file is a pointer
+ *  to the nic_data for each NIC, so that if multiple NICs are emulated
+ *  concurrently, they will not get packets that are meant for some other
+ *  controller.
  */
 
 #include <stdio.h>
@@ -54,15 +54,12 @@
 #include "net.h"
 
 
-/*  #define debug fatal  */
-
-
 /*
  *  net_allocate_ethernet_packet_link():
  *
  *  This routine allocates an ethernet_packet_link struct, and adds it at
  *  the end of the packet chain.  A data buffer is allocated, and the data,
- *  extra, and len fields of the link are set.
+ *  nic, and len fields of the link are set.
  *
  *  Note: The data buffer is not zeroed.
  *
@@ -70,15 +67,16 @@
  *  failure.
  */
 struct ethernet_packet_link *net_allocate_ethernet_packet_link(
-	struct net *net, void *extra, size_t len)
+	struct net *net, struct nic_data *nic, size_t len)
 {
 	struct ethernet_packet_link *lp;
 
-	CHECK_ALLOCATION(lp = malloc(sizeof(struct ethernet_packet_link)));
+	CHECK_ALLOCATION(lp = (struct ethernet_packet_link *)
+	    malloc(sizeof(struct ethernet_packet_link)));
 
 	lp->len = len;
-	lp->extra = extra;
-	CHECK_ALLOCATION(lp->data = malloc(len));
+	lp->nic = nic;
+	CHECK_ALLOCATION(lp->data = (unsigned char *) malloc(len));
 
 	lp->next = NULL;
 
@@ -115,7 +113,7 @@ struct ethernet_packet_link *net_allocate_ethernet_packet_link(
  *  An ARP request with the same from and to IP addresses should be ignored.
  *  (This would be a host testing to see if there is an IP collision.)
  */
-static void net_arp(struct net *net, void *extra,
+static void net_arp(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len, int reverse)
 {
 	int q;
@@ -160,7 +158,7 @@ static void net_arp(struct net *net, void *extra,
 				break;
 
 			lp = net_allocate_ethernet_packet_link(
-			    net, extra, 60 + 14);
+			    net, nic, 60 + 14);
 
 			/*  Copy the old packet first:  */
 			memset(lp->data, 0, 60 + 14);
@@ -185,7 +183,7 @@ static void net_arp(struct net *net, void *extra,
 			break;
 		case 3:		/*  Reverse Request  */
 			lp = net_allocate_ethernet_packet_link(
-			    net, extra, 60 + 14);
+			    net, nic, 60 + 14);
 
 			/*  Copy the old packet first:  */
 			memset(lp->data, 0, 60 + 14);
@@ -226,14 +224,16 @@ static void net_arp(struct net *net, void *extra,
 		case 2:		/*  Reply  */
 		case 4:		/*  Reverse Reply  */
 		default:
-			fatal("[ net: ARP: UNIMPLEMENTED request type "
-			    "0x%04x ]\n", r);
+			debugmsg(SUBSYS_NET, "ARP", VERBOSITY_WARNING,
+	    		    "UNIMPLEMENTED request type 0x%04x", r);
 		}
 	} else {
-		fatal("[ net: ARP: UNIMPLEMENTED arp packet type: ");
-		for (i=0; i<len; i++)
-			fatal("%02x", packet[i]);
-		fatal(" ]\n");
+		if (len >= 6)
+			debugmsg(SUBSYS_NET, "ARP", VERBOSITY_WARNING,
+	   		    "UNIMPLEMENTED arp packet type: "
+	   		    "0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x ...",
+	   		    packet[0], packet[1], packet[2], packet[3],
+	   		    packet[4], packet[5]);
 	}
 }
 
@@ -241,7 +241,7 @@ static void net_arp(struct net *net, void *extra,
 /*
  *  net_ethernet_rx_avail():
  *
- *  Return 1 if there is a packet available for this 'extra' pointer, otherwise
+ *  Return 1 if there is a packet available for this nic, otherwise
  *  return 0.
  *
  *  Appart from actually checking for incoming packets from the outside world,
@@ -249,11 +249,20 @@ static void net_arp(struct net *net, void *extra,
  *  a return value telling us whether there is a packet or not, we don't
  *  actually get the packet.
  */
-int net_ethernet_rx_avail(struct net *net, void *extra)
+int net_ethernet_rx_avail(struct net *net, struct nic_data *nic)
 {
 	if (net == NULL)
 		return 0;
 
+	/*
+	 *  If we're using a tap device, check in with that and
+	 *  that's it.
+	 */
+	if (net->tapdev) {
+		net_tap_rx_avail(net);
+		return net_ethernet_rx(net, nic, NULL, NULL);
+	}
+
 	/*
 	 *  If the network is distributed across multiple emulator processes,
 	 *  then receive incoming packets from those processes.
@@ -271,17 +280,18 @@ int net_ethernet_rx_avail(struct net *net, void *extra)
 			if (res != -1) {
 				nreceived ++;
 
-				/*  fatal("[ incoming DISTRIBUTED packet, %i "
+				debugmsg(SUBSYS_NET, "RX avail", VERBOSITY_DEBUG,
+    				    "incoming DISTRIBUTED packet, %i "
 				    "bytes from %s:%d\n", res,
 				    inet_ntoa(si.sin_addr),
-				    ntohs(si.sin_port));  */
+				    ntohs(si.sin_port));
 
 				/*  Add the packet to all "our" NICs on this
 				    network:  */
 				for (i=0; i<net->n_nics; i++) {
 					struct ethernet_packet_link *lp;
 					lp = net_allocate_ethernet_packet_link(
-					    net, net->nic_extra[i], res);
+					    net, net->nic_data[i], res);
 					memcpy(lp->data, buf, res);
 				}
 			}
@@ -289,10 +299,10 @@ int net_ethernet_rx_avail(struct net *net, void *extra)
 	}
 
 	/*  IP protocol specific:  */
-	net_udp_rx_avail(net, extra);
-	net_tcp_rx_avail(net, extra);
+	net_udp_rx_avail(net, nic);
+	net_tcp_rx_avail(net, nic);
 
-	return net_ethernet_rx(net, extra, NULL, NULL);
+	return net_ethernet_rx(net, nic, NULL, NULL);
 }
 
 
@@ -308,11 +318,11 @@ int net_ethernet_rx_avail(struct net *net, void *extra)
  *  available, 0 is returned.
  *
  *  If packetp is NULL, then the search is aborted as soon as a packet with
- *  the correct 'extra' field is found, and a 1 is returned, but as packetp
+ *  the correct 'nic' field is found, and a 1 is returned, but as packetp
  *  is NULL we can't return the actual packet. (This is the internal form
  *  if net_ethernet_rx_avail().)
  */
-int net_ethernet_rx(struct net *net, void *extra,
+int net_ethernet_rx(struct net *net, struct nic_data *nic,
 	unsigned char **packetp, int *lenp)
 {
 	struct ethernet_packet_link *lp, *prev;
@@ -320,12 +330,12 @@ int net_ethernet_rx(struct net *net, void *extra,
 	if (net == NULL)
 		return 0;
 
-	/*  Find the first packet which has the right 'extra' field.  */
+	/*  Find the first packet which has the right 'nic' field.  */
 
 	lp = net->first_ethernet_packet;
 	prev = NULL;
 	while (lp != NULL) {
-		if (lp->extra == extra) {
+		if (lp->nic == nic) {
 			/*  We found a packet for this controller!  */
 			if (packetp == NULL || lenp == NULL)
 				return 1;
@@ -367,7 +377,7 @@ int net_ethernet_rx(struct net *net, void *extra,
  *  If the packet can be handled here, it will not necessarily be transmitted
  *  to the outside world.
  */
-void net_ethernet_tx(struct net *net, void *extra,
+void net_ethernet_tx(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
 	int i, eth_type, for_the_gateway;
@@ -375,25 +385,34 @@ void net_ethernet_tx(struct net *net, void *extra,
 	if (net == NULL)
 		return;
 
-	for_the_gateway = !memcmp(packet, net->gateway_ethernet_addr, 6);
-
 	/*  Drop too small packets:  */
 	if (len < 20) {
-		fatal("[ net_ethernet_tx: Warning: dropping tiny packet "
-		    "(%i bytes) ]\n", len);
+		debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+    		    "Warning: dropping tiny packet (%i bytes) ]\n", len);
 		return;
 	}
 
+	/*
+	 * If we're using a tap device, we send the packet that way
+	 * and that's it.
+	 */
+	if (net->tapdev) {
+		net_tap_tx(net, nic, packet, len);
+		return;
+	}
+
+	for_the_gateway = !memcmp(packet, net->gateway_ethernet_addr, 6);
+
 	/*
 	 *  Copy this packet to all other NICs on this network (except if
 	 *  it is aimed specifically at the gateway's ethernet address):
 	 */
-	if (!for_the_gateway && extra != NULL && net->n_nics > 0) {
+	if (!for_the_gateway && nic != NULL && net->n_nics > 0) {
 		for (i=0; i<net->n_nics; i++)
-			if (extra != net->nic_extra[i]) {
+			if (nic != net->nic_data[i]) {
 				struct ethernet_packet_link *lp;
 				lp = net_allocate_ethernet_packet_link(net,
-				    net->nic_extra[i], len);
+				    net->nic_data[i], len);
 
 				/*  Copy the entire packet:  */
 				memcpy(lp->data, packet, len);
@@ -420,24 +439,17 @@ void net_ethernet_tx(struct net *net, void *extra,
 	 *  (DHCP packets are let through, though.)
 	 */
 
-	if (!for_the_gateway && packet[0] != 0xff && packet[0] != 0x00)
+	if (!for_the_gateway && packet[0] != 0xff &&
+	    !(packet[0] == 0x00 && packet[5] == 0x00))
 		return;
 
-#if 0
-	fatal("[ net: ethernet: ");
-	for (i=0; i<6; i++)	fatal("%02x", packet[i]); fatal(" ");
-	for (i=6; i<12; i++)	fatal("%02x", packet[i]); fatal(" ");
-	for (i=12; i<14; i++)	fatal("%02x", packet[i]); fatal(" ");
-	for (i=14; i<len; i++)	fatal("%02x", packet[i]); fatal(" ]\n");
-#endif
-
 	eth_type = (packet[12] << 8) + packet[13];
 
 	/*  IP:  */
 	if (eth_type == ETHERTYPE_IP) {
 		/*  Routed via the gateway?  */
 		if (for_the_gateway) {
-			net_ip(net, extra, packet, len);
+			net_ip(net, nic, packet, len);
 			return;
 		}
 
@@ -445,16 +457,20 @@ void net_ethernet_tx(struct net *net, void *extra,
 		if (packet[0] == 0xff && packet[1] == 0xff &&
 		    packet[2] == 0xff && packet[3] == 0xff &&
 		    packet[4] == 0xff && packet[5] == 0xff) {
-			net_ip_broadcast(net, extra, packet, len);
+			net_ip_broadcast(net, nic, packet, len);
 			return;
 		}
 
 		if (net->n_nics < 2) {
-			fatal("[ net_ethernet_tx: IP packet not for gateway, "
-			    "and not broadcast: ");
-			for (i=0; i<14; i++)
-				fatal("%02x", packet[i]);
-			fatal(" ]\n");
+			debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+	    		    "IP packet not for gateway, "
+			    "and not broadcast: %02x %02x %02x %02x "
+			    "%02x %02x %02x %02x %02x "
+			    "%02x %02x %02x %02x %02x ",
+			    packet[0], packet[1], packet[2], packet[3],
+			    packet[4], packet[5], packet[6], packet[7],
+			    packet[8], packet[9],
+			    packet[10], packet[11], packet[12], packet[13]);
 		}
 		return;
 	}
@@ -462,34 +478,36 @@ void net_ethernet_tx(struct net *net, void *extra,
 	/*  ARP:  */
 	if (eth_type == ETHERTYPE_ARP) {
 		if (len != 42 && len != 60)
-			fatal("[ net_ethernet_tx: WARNING! unusual "
-			    "ARP len (%i) ]\n", len);
-		net_arp(net, extra, packet + 14, len - 14, 0);
+			debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+	    		    "unusual ARP len (%i)", len);
+		net_arp(net, nic, packet + 14, len - 14, 0);
 		return;
 	}
 
 	/*  RARP:  */
 	if (eth_type == ETHERTYPE_REVARP) {
-		net_arp(net, extra, packet + 14, len - 14, 1);
+		net_arp(net, nic, packet + 14, len - 14, 1);
 		return;
 	}
 
 	/*  Sprite:  */
 	if (eth_type == ETHERTYPE_SPRITE) {
 		/*  TODO.  */
-		fatal("[ net: TX: UNIMPLEMENTED Sprite packet ]\n");
+		debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+		    "UNIMPLEMENTED Sprite packet");
 		return;
 	}
 
 	/*  IPv6:  */
 	if (eth_type == ETHERTYPE_IPV6) {
 		/*  TODO.  */
-		fatal("[ net_ethernet_tx: IPv6 is not implemented yet! ]\n");
+		debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+		    "IPv6 is not yet implemented!");
 		return;
 	}
 
-	fatal("[ net_ethernet_tx: ethernet packet type 0x%04x not yet "
-	    "implemented ]\n", eth_type);
+	debugmsg(SUBSYS_NET, "TX", VERBOSITY_WARNING,
+	    "ethernet packet type 0x%04x not yet implemented", eth_type);
 }
 
 
@@ -566,25 +584,6 @@ static void parse_resolvconf(struct net *net)
 			net->nameserver_known = 1;
 			break;
 		}
-
-	for (i=0; i<len; i++)
-		if (strncmp(buf+i, "domain", 6) == 0) {
-			/*  "domain" (1 or more whitespace) domain_name  */
-			i += 6;
-			while (i<len && (buf[i]==' ' || buf[i]=='\t'))
-				i++;
-			if (i >= len)
-				break;
-
-			start = i;
-			while (i<len && buf[i]!='\n' && buf[i]!='\r')
-				i++;
-			if (i < len)
-				buf[i] = '\0';
-			/*  fatal("DOMAIN='%s'\n", buf + start);  */
-			CHECK_ALLOCATION(net->domain_name = strdup(buf+start));
-			break;
-		}
 }
 
 
@@ -594,21 +593,28 @@ static void parse_resolvconf(struct net *net)
  *  Add a NIC to a network. (All NICs on a network will see each other's
  *  packets.)
  */
-void net_add_nic(struct net *net, void *extra, unsigned char *macaddr)
+void net_add_nic(struct net *net, struct nic_data *nic)
 {
 	if (net == NULL)
 		return;
 
-	if (extra == NULL) {
-		fprintf(stderr, "net_add_nic(): extra = NULL\n");
+	if (nic == NULL) {
+		fprintf(stderr, "net_add_nic(): nic = NULL\n");
 		exit(1);
 	}
 
-	net->n_nics ++;
-	CHECK_ALLOCATION(net->nic_extra = realloc(net->nic_extra, sizeof(void *)
-	    * net->n_nics));
+	/*
+	 *  Set up some of the basics for this NIC.  We assume the
+	 *  device has set up all of the other fields.
+	 */
+	nic->net = net;
+	nic->promiscuous_mode = 0;
 
-	net->nic_extra[net->n_nics - 1] = extra;
+	net->n_nics++;
+	CHECK_ALLOCATION(net->nic_data = (struct nic_data **)
+	    realloc(net->nic_data, sizeof(struct nic_data *) * net->n_nics));
+
+	net->nic_data[net->n_nics - 1] = nic;
 }
 
 
@@ -621,7 +627,7 @@ void net_add_nic(struct net *net, void *extra, unsigned char *macaddr)
  */
 static void net_gateway_init(struct net *net)
 {
-	unsigned char *p = (void *) &net->netmask_ipv4;
+	unsigned char *p = (unsigned char *) &net->netmask_ipv4;
 	uint32_t x;
 	int xl;
 
@@ -653,37 +659,44 @@ static void net_gateway_init(struct net *net)
  */
 void net_dumpinfo(struct net *net)
 {
-	int iadd = DEBUG_INDENTATION;
+	int iadd = 1;
 	struct remote_net *rnp;
 
-	debug("net:\n");
+	debugmsg(SUBSYS_NET, "", VERBOSITY_INFO, "");
+
+	debug_indentation(iadd);
+
+	if (net->tapdev) {
+		debugmsg(SUBSYS_NET, "tap", VERBOSITY_INFO,
+		    "using device %s", net->tapdev);
+		debug_indentation(-iadd);
+		return;
+	}
+
+	debugmsg(SUBSYS_NET, "simulated network", VERBOSITY_INFO, "");
 
 	debug_indentation(iadd);
 
-	debug("simulated network: ");
 	net_debugaddr(&net->netmask_ipv4, NET_ADDR_IPV4);
 	debug("/%i", net->netmask_ipv4_len);
 
 	debug(" (max outgoing: TCP=%i, UDP=%i)\n",
 	    MAX_TCP_CONNECTIONS, MAX_UDP_CONNECTIONS);
 
-	debug("simulated gateway+nameserver: ");
+	debug("gateway+nameserver: ");
 	net_debugaddr(&net->gateway_ipv4_addr, NET_ADDR_IPV4);
 	debug(" (");
 	net_debugaddr(&net->gateway_ethernet_addr, NET_ADDR_ETHERNET);
 	debug(")\n");
 
 	if (!net->nameserver_known) {
-		debug("(could not determine nameserver)\n");
+		debug("(could not determine real nameserver from /etc/resolv.conf)\n");
 	} else {
-		debug("simulated nameserver uses real nameserver ");
+		debug("nameserver uses real nameserver ");
 		net_debugaddr(&net->nameserver_ipv4, NET_ADDR_IPV4);
 		debug("\n");
 	}
 
-	if (net->domain_name != NULL && net->domain_name[0])
-		debug("domain: %s\n", net->domain_name);
-
 	rnp = net->remote_nets;
 	if (net->local_port != 0)
 		debug("distributed network: local port = %i\n",
@@ -698,6 +711,7 @@ void net_dumpinfo(struct net *net)
 	debug_indentation(-iadd);
 
 	debug_indentation(-iadd);
+	debug_indentation(-iadd);
 }
 
 
@@ -714,9 +728,10 @@ void net_dumpinfo(struct net *net)
  *  Network settings are registered if settings_prefix is non-NULL.
  *  (The one calling net_init() is also responsible for calling net_deinit().)
  *
- *  On failure, exit() is called.
+ *  On failure, NULL is returned.
  */
 struct net *net_init(struct emul *emul, int init_flags,
+	const char *tapdev,
 	const char *ipv4addr, int netipv4len,
 	char **remote, int n_remote, int local_port,
 	const char *settings_prefix)
@@ -724,7 +739,7 @@ struct net *net_init(struct emul *emul, int init_flags,
 	struct net *net;
 	int res;
 
-	CHECK_ALLOCATION(net = malloc(sizeof(struct net)));
+	CHECK_ALLOCATION(net = (struct net *) malloc(sizeof(struct net)));
 	memset(net, 0, sizeof(struct net));
 
 	/*  Set the back pointer:  */
@@ -733,6 +748,22 @@ struct net *net_init(struct emul *emul, int init_flags,
 	/*  Sane defaults:  */
 	net->timestamp = 0;
 	net->first_ethernet_packet = net->last_ethernet_packet = NULL;
+	net->tapdev = NULL;
+	net->tap_fd = -1;
+
+	/*
+	 *  If we're using a tap device, attempt to initialize it and
+	 *  none of the other stuff.
+	 */
+	if (tapdev) {
+		if (!net_tap_init(net, tapdev)) {
+			free(net);
+			return NULL;
+		}
+
+		net_dumpinfo(net);
+		return net;
+	}
 
 #ifdef HAVE_INET_PTON
 	res = inet_pton(AF_INET, ipv4addr, &net->netmask_ipv4);
@@ -742,18 +773,20 @@ struct net *net_init(struct emul *emul, int init_flags,
 	if (res < 1) {
 		fprintf(stderr, "net_init(): could not parse IPv4 address"
 		    " '%s'\n", ipv4addr);
-		exit(1);
+		free(net);
+		return NULL;
 	}
 
 	if (netipv4len < 1 || netipv4len > 30) {
 		fprintf(stderr, "net_init(): extremely weird ipv4 "
 		    "network length (%i)\n", netipv4len);
-		exit(1);
+		free(net);
+		return NULL;
 	}
+
 	net->netmask_ipv4_len = netipv4len;
 
 	net->nameserver_known = 0;
-	net->domain_name = "";
 	parse_resolvconf(net);
 
 	/*  Distributed network? Then add remote hosts:  */
@@ -764,7 +797,8 @@ struct net *net_init(struct emul *emul, int init_flags,
 		net->local_port_socket = socket(AF_INET, SOCK_DGRAM, 0);
 		if (net->local_port_socket < 0) {
 			perror("socket");
-			exit(1);
+			free(net);
+			return NULL;
 		}
 
 		memset((char *)&si_self, 0, sizeof(si_self));
@@ -774,7 +808,8 @@ struct net *net_init(struct emul *emul, int init_flags,
 		if (bind(net->local_port_socket, (struct sockaddr *)&si_self,
 		    sizeof(si_self)) < 0) {
 			perror("bind");
-			exit(1);
+			free(net);
+			return NULL;
 		}
 
 		/*  Set the socket to non-blocking:  */
@@ -787,7 +822,7 @@ struct net *net_init(struct emul *emul, int init_flags,
 			struct hostent *hp;
 
 			/*  debug("adding '%s'\n", remote[n_remote]);  */
-			CHECK_ALLOCATION(rnp =
+			CHECK_ALLOCATION(rnp = (struct remote_net *)
 			    malloc(sizeof(struct remote_net)));
 			memset(rnp, 0, sizeof(struct remote_net));
 
@@ -802,8 +837,10 @@ struct net *net_init(struct emul *emul, int init_flags,
 			if (hp == NULL) {
 				fprintf(stderr, "could not resolve '%s'\n",
 				    rnp->name);
-				exit(1);
+				free(net);
+				return NULL;
 			}
+
 			memcpy(&rnp->ipv4_addr, hp->h_addr, hp->h_length);
 			free(rnp->name);
 
@@ -812,8 +849,10 @@ struct net *net_init(struct emul *emul, int init_flags,
 			if (strchr(rnp->name, ':') == NULL) {
 				fprintf(stderr, "Remote network '%s' is not "
 				    "'host:portnr'?\n", rnp->name);
-				exit(1);
+				free(net);
+				return NULL;
 			}
+
 			rnp->portnr = atoi(strchr(rnp->name, ':') + 1);
 		}
 	}
