@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -28,6 +28,7 @@
  *  Internet Protocol related networking stuff.
  */
 
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -43,6 +44,8 @@
 
 
 /*  #define debug fatal  */
+static int net_ip_debug = 0;	// replace by debugmsg(....)
+
 
 
 /*
@@ -155,7 +158,7 @@ void net_ip_tcp_checksum(unsigned char *tcp_header, int chksumoffset,
  *	1c1d1e1f202122232425262728292a2b
  *	2c2d2e2f3031323334353637
  */
-static void net_ip_icmp(struct net *net, void *extra,
+static void net_ip_icmp(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
 	int type;
@@ -165,8 +168,9 @@ static void net_ip_icmp(struct net *net, void *extra,
 
 	switch (type) {
 	case 8:	/*  ECHO request  */
-		debug("[ ICMP echo ]\n");
-		lp = net_allocate_ethernet_packet_link(net, extra, len);
+		debugmsg(SUBSYS_NET, "ICMP", VERBOSITY_DEBUG, "ECHO request");
+
+		lp = net_allocate_ethernet_packet_link(net, nic, len);
 
 		/*  Copy the old packet first:  */
 		memcpy(lp->data + 12, packet + 12, len - 12);
@@ -193,7 +197,7 @@ static void net_ip_icmp(struct net *net, void *extra,
 
 		break;
 	default:
-		fatal("[ net: ICMP type %i not yet implemented ]\n", type);
+		debugmsg(SUBSYS_NET, "ICMP", VERBOSITY_WARNING, "type %i not yet implemented", type);
 	}
 }
 
@@ -225,7 +229,7 @@ static void tcp_closeconnection(struct net *net, int con_id)
  *  This creates an ethernet packet for the guest OS with an ACK to the
  *  initial SYN packet.
  */
-void net_ip_tcp_connectionreply(struct net *net, void *extra,
+void net_ip_tcp_connectionreply(struct net *net, struct nic_data *nic,
 	int con_id, int connecting, unsigned char *data, int datalen, int rst)
 {
 	struct ethernet_packet_link *lp;
@@ -238,7 +242,7 @@ void net_ip_tcp_connectionreply(struct net *net, void *extra,
 	net->tcp_connections[con_id].tcp_id ++;
 	tcp_length = 20 + option_len + datalen;
 	ip_len = 20 + tcp_length;
-	lp = net_allocate_ethernet_packet_link(net, extra, 14 + ip_len);
+	lp = net_allocate_ethernet_packet_link(net, nic, 14 + ip_len);
 
 	/*  Ethernet header:  */
 	memcpy(lp->data + 0, net->tcp_connections[con_id].ethernet_address, 6);
@@ -339,15 +343,12 @@ void net_ip_tcp_connectionreply(struct net *net, void *extra,
 	net_ip_tcp_checksum(lp->data + 34, 16, tcp_length,
 	    lp->data + 26, lp->data + 30, 0);
 
-#if 0
-	{
-		int i;
+	if (net_ip_debug) {
 		fatal("[ net_ip_tcp_connectionreply(%i): ", connecting);
-		for (i=0; i<ip_len+14; i++)
+		for (int i=0; i<ip_len+14; i++)
 			fatal("%02x", lp->data[i]);
 		fatal(" ]\n");
 	}
-#endif
 
 	if (connecting)
 		net->tcp_connections[con_id].outside_seqnr ++;
@@ -376,7 +377,7 @@ void net_ip_tcp_connectionreply(struct net *net, void *extra,
  *	http://www.networksorcery.com/enp/protocol/tcp.htm
  *	http://www.tcpipguide.com/free/t_TCPIPTransmissionControlProtocolTCP.htm
  */
-static void net_ip_tcp(struct net *net, void *extra,
+static void net_ip_tcp(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
 	int con_id, free_con_id, i, res;
@@ -388,12 +389,12 @@ static void net_ip_tcp(struct net *net, void *extra,
 	struct timeval tv;
 	int send_ofs;
 
-#if 0
-	fatal("[ net: TCP: ");
-	for (i=0; i<26; i++)
-		fatal("%02x", packet[i]);
-	fatal(" ");
-#endif
+	if (net_ip_debug) {
+		fatal("[ net: TCP: ");
+		for (i=0; i<26; i++)
+			fatal("%02x", packet[i]);
+		fatal(" ");
+	}
 
 	srcport = (packet[34] << 8) + packet[35];
 	dstport = (packet[36] << 8) + packet[37];
@@ -403,12 +404,11 @@ static void net_ip_tcp(struct net *net, void *extra,
 	acknr   = (packet[42] << 24) + (packet[43] << 16)
 		+ (packet[44] << 8) + packet[45];
 
-#if 0
-	fatal("%i.%i.%i.%i:%i -> %i.%i.%i.%i:%i, seqnr=%lli acknr=%lli ",
-	    packet[26], packet[27], packet[28], packet[29], srcport,
-	    packet[30], packet[31], packet[32], packet[33], dstport,
-	    (long long)seqnr, (long long)acknr);
-#endif
+	if (net_ip_debug)
+		fatal("%i.%i.%i.%i:%i -> %i.%i.%i.%i:%i, seqnr=%lli acknr=%lli ",
+		    packet[26], packet[27], packet[28], packet[29], srcport,
+		    packet[30], packet[31], packet[32], packet[33], dstport,
+		    (long long)seqnr, (long long)acknr);
 
 	data_offset = (packet[46] >> 4) * 4 + 34;
 	/*  data_offset is now data offset within packet :-)  */
@@ -423,27 +423,26 @@ static void net_ip_tcp(struct net *net, void *extra,
 	checksum = (packet[50] << 8) + packet[51];
 	urgptr   = (packet[52] << 8) + packet[53];
 
-#if 0
-	fatal(urg? "URG " : "");
-	fatal(ack? "ACK " : "");
-	fatal(psh? "PSH " : "");
-	fatal(rst? "RST " : "");
-	fatal(syn? "SYN " : "");
-	fatal(fin? "FIN " : "");
+	if (net_ip_debug) {
+		fatal(urg? "URG " : "");
+		fatal(ack? "ACK " : "");
+		fatal(psh? "PSH " : "");
+		fatal(rst? "RST " : "");
+		fatal(syn? "SYN " : "");
+		fatal(fin? "FIN " : "");
 
-	fatal("window=0x%04x checksum=0x%04x urgptr=0x%04x ",
-	    window, checksum, urgptr);
+		fatal("window=0x%04x checksum=0x%04x urgptr=0x%04x ", window, checksum, urgptr);
 
-	fatal("options=");
-	for (i=34+20; i<data_offset; i++)
-		fatal("%02x", packet[i]);
+		fatal("options=");
+		for (i=34+20; i<data_offset; i++)
+			fatal("%02x", packet[i]);
 
-	fatal(" data=");
-	for (i=data_offset; i<len; i++)
-		fatal("%02x", packet[i]);
+		fatal(" data=");
+		for (i=data_offset; i<len; i++)
+			fatal("%02x", packet[i]);
 
-	fatal(" ]\n");
-#endif
+		fatal(" ]\n");
+	}
 
 	net_ip_tcp_checksum(packet + 34, 16, len - 34,
 		packet + 26, packet + 30, 0);
@@ -585,7 +584,7 @@ static void net_ip_tcp(struct net *net, void *extra,
 
 	if (rst) {
 		debug("[ 'rst': disconnecting TCP connection %i ]\n", con_id);
-		net_ip_tcp_connectionreply(net, extra, con_id, 0, NULL, 0, 1);
+		net_ip_tcp_connectionreply(net, nic, con_id, 0, NULL, 0, 1);
 		tcp_closeconnection(net, con_id);
 		return;
 	}
@@ -596,7 +595,7 @@ static void net_ip_tcp(struct net *net, void *extra,
 		    "connection %i ]\n", con_id);
 
 		/*  Send an RST?  (TODO, this is wrong...)  */
-		net_ip_tcp_connectionreply(net, extra, con_id, 0, NULL, 0, 1);
+		net_ip_tcp_connectionreply(net, nic, con_id, 0, NULL, 0, 1);
 
 		/*  ... and forget about this connection:  */
 		tcp_closeconnection(net, con_id);
@@ -610,7 +609,7 @@ static void net_ip_tcp(struct net *net, void *extra,
 
 		/*  Send an ACK:  */
 		net->tcp_connections[con_id].state = TCP_OUTSIDE_CONNECTED;
-		net_ip_tcp_connectionreply(net, extra, con_id, 0, NULL, 0, 0);
+		net_ip_tcp_connectionreply(net, nic, con_id, 0, NULL, 0, 0);
 		net->tcp_connections[con_id].state = TCP_OUTSIDE_DISCONNECTED2;
 		return;
 	}
@@ -620,7 +619,7 @@ static void net_ip_tcp(struct net *net, void *extra,
 		    con_id);
 
 		/*  Send ACK:  */
-		net_ip_tcp_connectionreply(net, extra, con_id, 0, NULL, 0, 0);
+		net_ip_tcp_connectionreply(net, nic, con_id, 0, NULL, 0, 0);
 		net->tcp_connections[con_id].state = TCP_OUTSIDE_DISCONNECTED2;
 
 		/*  Return and send FIN:  */
@@ -725,7 +724,7 @@ debug("  all acked\n");
 
 ret:
 	/*  Send an ACK (or FIN) to the guest OS:  */
-	net_ip_tcp_connectionreply(net, extra, con_id, 0, NULL, 0, 0);
+	net_ip_tcp_connectionreply(net, nic, con_id, 0, NULL, 0, 0);
 }
 
 
@@ -746,7 +745,7 @@ ret:
  *	srcport=fffc dstport=0035 length=0028 chksum=76b6
  *	43e20100000100000000000003667470066e6574627364036f726700001c0001
  */
-static void net_ip_udp(struct net *net, void *extra,
+static void net_ip_udp(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
 	int con_id, free_con_id, i, srcport, dstport, udp_len;
@@ -797,19 +796,17 @@ static void net_ip_udp(struct net *net, void *extra,
 	else {
 		debug("NEW");
 		if (free_con_id < 0) {
-			int i;
 			int64_t oldest = net->
 			    udp_connections[0].last_used_timestamp;
 			free_con_id = 0;
 
 			debug(", NO FREE SLOTS, REUSING OLDEST ONE");
-			for (i=0; i<MAX_UDP_CONNECTIONS; i++)
-				if (net->udp_connections[i].
-				    last_used_timestamp < oldest) {
-					oldest = net->udp_connections[i].
-					    last_used_timestamp;
-					free_con_id = i;
+			for (int j=0; j<MAX_UDP_CONNECTIONS; j++)
+				if (net->udp_connections[j].last_used_timestamp < oldest) {
+					oldest = net->udp_connections[j].last_used_timestamp;
+					free_con_id = j;
 				}
+
 			close(net->udp_connections[free_con_id].socket);
 		}
 		con_id = free_con_id;
@@ -884,28 +881,34 @@ static void net_ip_udp(struct net *net, void *extra,
  *
  *  Handle an IP packet, coming from the emulated NIC.
  */
-void net_ip(struct net *net, void *extra, unsigned char *packet, int len)
+void net_ip(struct net *net, struct nic_data *nic, unsigned char *packet,
+	int len)
 {
-#if 1
 	int i;
 
-	debug("[ net: IP: ");
-	debug("ver=%02x ", packet[14]);
-	debug("tos=%02x ", packet[15]);
-	debug("len=%02x%02x ", packet[16], packet[17]);
-	debug("id=%02x%02x ",  packet[18], packet[19]);
-	debug("ofs=%02x%02x ", packet[20], packet[21]);
-	debug("ttl=%02x ", packet[22]);
-	debug("p=%02x ", packet[23]);
-	debug("sum=%02x%02x ", packet[24], packet[25]);
-	debug("src=%02x%02x%02x%02x ",
-	    packet[26], packet[27], packet[28], packet[29]);
-	debug("dst=%02x%02x%02x%02x ",
-	    packet[30], packet[31], packet[32], packet[33]);
-	for (i=34; i<len; i++)
-		debug("%02x", packet[i]);
-	debug(" ]\n");
-#endif
+	if (ENOUGH_VERBOSITY(SUBSYS_NET, VERBOSITY_DEBUG)) {
+		char s[2000];
+		
+		snprintf(s, sizeof(s), "ver=%02x ", packet[14]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "tos=%02x ", packet[15]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "len=%02x%02x ", packet[16], packet[17]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "id=%02x%02x ",  packet[18], packet[19]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ofs=%02x%02x ", packet[20], packet[21]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ttl=%02x ", packet[22]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "p=%02x ", packet[23]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "sum=%02x%02x ", packet[24], packet[25]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "src=%02x%02x%02x%02x ",
+		    packet[26], packet[27], packet[28], packet[29]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "dst=%02x%02x%02x%02x ",
+		    packet[30], packet[31], packet[32], packet[33]);
+		for (i=34; i<len; i++) {
+			if (strlen(s) > sizeof(s) - 5)
+				break;
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+
+		debugmsg(SUBSYS_NET, "IP", VERBOSITY_DEBUG, "%s", s);
+	}
 
 	/*  Cut off overflowing tail data:  */
 	if (len > 14 + packet[16]*256 + packet[17])
@@ -915,21 +918,23 @@ void net_ip(struct net *net, void *extra, unsigned char *packet, int len)
 		/*  IPv4:  */
 		switch (packet[23]) {
 		case 1:	/*  ICMP  */
-			net_ip_icmp(net, extra, packet, len);
+			net_ip_icmp(net, nic, packet, len);
 			break;
 		case 6:	/*  TCP  */
-			net_ip_tcp(net, extra, packet, len);
+			net_ip_tcp(net, nic, packet, len);
 			break;
 		case 17:/*  UDP  */
-			net_ip_udp(net, extra, packet, len);
+			net_ip_udp(net, nic, packet, len);
 			break;
-		default:
-			fatal("[ net: IP: UNIMPLEMENTED protocol %i ]\n",
+		default:debugmsg(SUBSYS_NET, "IP", VERBOSITY_WARNING,
+			    "UNIMPLEMENTED protocol %i",
 			    packet[23]);
 		}
-	} else
-		fatal("[ net: IP: UNIMPLEMENTED ip, first byte = 0x%02x ]\n",
+	} else {
+		debugmsg(SUBSYS_NET, "IP", VERBOSITY_WARNING,
+		    "UNIMPLEMENTED ip version, first byte = 0x%02x",
 		    packet[14]);
+	}
 }
 
 
@@ -938,41 +943,41 @@ void net_ip(struct net *net, void *extra, unsigned char *packet, int len)
  *
  *  Handle an IPv4 DHCP broadcast packet, coming from the emulated NIC.
  *
- *  Read http://www.ietf.org/rfc/rfc2131.txt for details on DHCP.
+ *  Read http://tools.ietf.org/html/rfc2131 for details on DHCP.
  *  (And http://users.telenet.be/mydotcom/library/network/dhcp.htm.)
  */
-static void net_ip_broadcast_dhcp(struct net *net, void *extra,
+static void net_ip_broadcast_dhcp(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
-	/*
-	 *  TODO
-	 */
-#if 1
 	struct ethernet_packet_link *lp;
-	int i;
+        int i, reply_len;
 
-	fatal("[ net: IPv4 DHCP: ");
-#if 1
-	fatal("ver=%02x ", packet[14]);
-	fatal("tos=%02x ", packet[15]);
-	fatal("len=%02x%02x ", packet[16], packet[17]);
-	fatal("id=%02x%02x ",  packet[18], packet[19]);
-	fatal("ofs=%02x%02x ", packet[20], packet[21]);
-	fatal("ttl=%02x ", packet[22]);
-	fatal("p=%02x ", packet[23]);
-	fatal("sum=%02x%02x ", packet[24], packet[25]);
-#endif
-	fatal("src=%02x%02x%02x%02x ",
-	    packet[26], packet[27], packet[28], packet[29]);
-	fatal("dst=%02x%02x%02x%02x ",
-	    packet[30], packet[31], packet[32], packet[33]);
-#if 0
-	for (i=34; i<len; i++)
-		fatal("%02x", packet[i]);
-#endif
+	if (ENOUGH_VERBOSITY(SUBSYS_NET, VERBOSITY_DEBUG)) {
+		char s[4000];
+		
+		snprintf(s, sizeof(s), "ver=%02x ", packet[14]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "tos=%02x ", packet[15]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "len=%02x%02x ", packet[16], packet[17]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "id=%02x%02x ",  packet[18], packet[19]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ofs=%02x%02x ", packet[20], packet[21]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ttl=%02x ", packet[22]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "p=%02x ", packet[23]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "sum=%02x%02x ", packet[24], packet[25]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "src=%02x%02x%02x%02x ",
+		    packet[26], packet[27], packet[28], packet[29]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "dst=%02x%02x%02x%02x ",
+		    packet[30], packet[31], packet[32], packet[33]);
+		for (i=34; i<len; i++) {
+			if (strlen(s) > sizeof(s) - 5)
+				break;
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+
+		debugmsg(SUBSYS_NET, "IPv4 DHCP", VERBOSITY_DEBUG, "%s", s);
+	}
 
 	if (len < 34 + 8 + 236) {
-		fatal("[ DHCP packet too short? Len=%i ]\n", len);
+		debugmsg(SUBSYS_NET, "IPv4 DHCP", VERBOSITY_ERROR, "packet too short? len=%i", len);
 		return;
 	}
 
@@ -984,6 +989,7 @@ static void net_ip_broadcast_dhcp(struct net *net, void *extra,
 	 *         0000000000102030405060...0000...638253633501...000
 	 */
 
+#if 0
 	fatal("op=%02x ", packet[42]);
 	fatal("htype=%02x ", packet[43]);
 	fatal("hlen=%02x ", packet[44]);
@@ -1004,20 +1010,27 @@ static void net_ip_broadcast_dhcp(struct net *net, void *extra,
 	for (i=70; i<70+16; i++)
 		fatal("%02x", packet[i]);
 	/*
-   |                          sname   (64)                         |
-   |                          file    (128)                        |
+	   |                          sname   (64)                         |
+	   |                          file    (128)                        |
 	 */
 	fatal(" ]\n");
+#endif
 
-	lp = net_allocate_ethernet_packet_link(net, extra, len);
+        reply_len = 307;
+        lp = net_allocate_ethernet_packet_link(net, nic, reply_len);
 
-	/*  Copy the old packet first:  */
-	memcpy(lp->data, packet, len);
+        /*  From old packet, copy everything before options field:  */
+        memcpy(lp->data, packet, 278);
 
 	/*  We are sending to the client, from the gateway:  */
 	memcpy(lp->data + 0, packet + 6, 6);
 	memcpy(lp->data + 6, net->gateway_ethernet_addr, 6);
 
+	/*  Set IP length:  */
+	lp->data[16] = (reply_len - 14) >> 8;
+	lp->data[17] = (reply_len - 14) & 0xff;
+
+        /*  Set IP addresses:  */
 	memcpy(lp->data + 26, &net->gateway_ipv4_addr[0], 4);
 	lp->data[30] = 0xff;
 	lp->data[31] = 0xff;
@@ -1028,11 +1041,15 @@ static void net_ip_broadcast_dhcp(struct net *net, void *extra,
 	memcpy(lp->data + 34, packet + 36, 2);
 	memcpy(lp->data + 36, packet + 34, 2);
 
+	/*  Set UDP length:  */
+	lp->data[38] = (reply_len - 34) >> 8;
+	lp->data[39] = (reply_len - 34) & 0xff;
+
 	/*  Client's (yiaddr) IPv4 address:  */
 	lp->data[58] = 10;
 	lp->data[59] = 0;
 	lp->data[60] = 0;
-	lp->data[61] = 1;
+	lp->data[61] = 1;	/*  TODO  */
 
 	/*  Server's IPv4 address:  (giaddr)  */
 	memcpy(lp->data + 66, &net->gateway_ipv4_addr[0], 4);
@@ -1042,53 +1059,86 @@ static void net_ip_broadcast_dhcp(struct net *net, void *extra,
 
 	snprintf((char *)lp->data + 70+16+64, 8, "gxemul");
 
+	/*  Options field at offset 278:  */
+	lp->data[278] = 99;
+	lp->data[279] = 130;
+	lp->data[280] = 83;
+	lp->data[281] = 99;
+
+	/*  DHCP options, http://tools.ietf.org/html/rfc1533  */
+	lp->data[282] = 1; /* subnet mask */
+	lp->data[283] = 4;
+	lp->data[284] = 255;
+	lp->data[285] = 0;
+	lp->data[286] = 0;
+	lp->data[287] = 0;
+	lp->data[288] = 3; /* router */
+	lp->data[289] = 4;
+	memcpy(lp->data + 290, &net->gateway_ipv4_addr[0], 4);
+	lp->data[294] = 6; /* domain name server */
+	lp->data[295] = 4;
+	memcpy(lp->data + 296, &net->gateway_ipv4_addr[0], 4);
+	lp->data[300] = 54; /* server identifier */
+	lp->data[301] = 4;
+	memcpy(lp->data + 302, &net->gateway_ipv4_addr[0], 4);
+	lp->data[306] = 255; /* end */
+
 	/*  Recalculate IP header checksum:  */
 	net_ip_checksum(lp->data + 14, 10, 20);
 
 	/*  ... and the UDP checksum:  */
-	net_ip_tcp_checksum(lp->data + 34, 6, len - 34 - 8,
+        net_ip_tcp_checksum(lp->data + 34, 6, reply_len - 34,
 	    lp->data + 26, lp->data + 30, 1);
 
+	if (ENOUGH_VERBOSITY(SUBSYS_NET, VERBOSITY_DEBUG)) {
+		char s[4000];
+		s[0] = '\0';
 
-/*  Debug dump:  */
-packet = lp->data;
-	fatal("[ net: IPv4 DHCP REPLY: ");
-	for (i=0; i<14; i++)
-		fatal("%02x", packet[i]);
-	fatal("ver=%02x ", packet[14]);
-	fatal("tos=%02x ", packet[15]);
-	fatal("len=%02x%02x ", packet[16], packet[17]);
-	fatal("id=%02x%02x ",  packet[18], packet[19]);
-	fatal("ofs=%02x%02x ", packet[20], packet[21]);
-	fatal("ttl=%02x ", packet[22]);
-	fatal("p=%02x ", packet[23]);
-	fatal("sum=%02x%02x ", packet[24], packet[25]);
-	fatal("src=%02x%02x%02x%02x ",
-	    packet[26], packet[27], packet[28], packet[29]);
-	fatal("dst=%02x%02x%02x%02x ",
-	    packet[30], packet[31], packet[32], packet[33]);
-	fatal("op=%02x ", packet[42]);
-	fatal("htype=%02x ", packet[43]);
-	fatal("hlen=%02x ", packet[44]);
-	fatal("hops=%02x ", packet[45]);
-	fatal("xid=%02x%02x%02x%02x ", packet[46], packet[47],
-	    packet[48], packet[49]);
-	fatal("secs=%02x%02x ", packet[50], packet[51]);
-	fatal("flags=%02x%02x ", packet[52], packet[53]);
-	fatal("ciaddr=%02x%02x%02x%02x ", packet[54], packet[55],
-	    packet[56], packet[57]);
-	fatal("yiaddr=%02x%02x%02x%02x ", packet[58], packet[59],
-	    packet[60], packet[61]);
-	fatal("siaddr=%02x%02x%02x%02x ", packet[62], packet[63],
-	    packet[64], packet[65]);
-	fatal("giaddr=%02x%02x%02x%02x ", packet[66], packet[67],
-	    packet[68], packet[69]);
-	fatal("chaddr=");
-	for (i=70; i<70+16; i++)
-		fatal("%02x", packet[i]);
-	fatal(" ]\n");
+		packet = lp->data;
 
-#endif
+		for (i=0; i<14; i++) {
+			if (strlen(s) > sizeof(s) - 5)
+				break;
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ver=%02x ", packet[14]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "tos=%02x ", packet[15]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "len=%02x%02x ", packet[16], packet[17]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "id=%02x%02x ",  packet[18], packet[19]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ofs=%02x%02x ", packet[20], packet[21]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ttl=%02x ", packet[22]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "p=%02x ", packet[23]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "sum=%02x%02x ", packet[24], packet[25]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "src=%02x%02x%02x%02x ",
+		    packet[26], packet[27], packet[28], packet[29]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "dst=%02x%02x%02x%02x ",
+		    packet[30], packet[31], packet[32], packet[33]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "op=%02x ", packet[42]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "htype=%02x ", packet[43]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "hlen=%02x ", packet[44]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "hops=%02x ", packet[45]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "xid=%02x%02x%02x%02x ",
+			packet[46], packet[47], packet[48], packet[49]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "secs=%02x%02x ", packet[50], packet[51]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "flags=%02x%02x ", packet[52], packet[53]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ciaddr=%02x%02x%02x%02x ",
+			packet[54], packet[55], packet[56], packet[57]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "yiaddr=%02x%02x%02x%02x ",
+			packet[58], packet[59], packet[60], packet[61]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "siaddr=%02x%02x%02x%02x ",
+			packet[62], packet[63], packet[64], packet[65]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "giaddr=%02x%02x%02x%02x ",
+			packet[66], packet[67], packet[68], packet[69]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "chaddr=");
+		for (i=70; i<70+16; i++) {
+			if (strlen(s) > sizeof(s) - 5)
+				break;
+
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+
+		debugmsg(SUBSYS_NET, "IPv4 DHCP REPLY", VERBOSITY_DEBUG, "%s", s);
+	}
 }
 
 
@@ -1098,31 +1148,36 @@ packet = lp->data;
  *  Handle an IP broadcast packet, coming from the emulated NIC.
  *  (This is usually a DHCP request, or similar.)
  */
-void net_ip_broadcast(struct net *net, void *extra,
+void net_ip_broadcast(struct net *net, struct nic_data *nic,
 	unsigned char *packet, int len)
 {
-	unsigned char *p = (void *) &net->netmask_ipv4;
+	unsigned char *p = (unsigned char *) &net->netmask_ipv4;
 	uint32_t x, y;
 	int i, xl, warning = 0, match = 0;
 
-#if 0
-	fatal("[ net: IP BROADCAST: ");
-	fatal("ver=%02x ", packet[14]);
-	fatal("tos=%02x ", packet[15]);
-	fatal("len=%02x%02x ", packet[16], packet[17]);
-	fatal("id=%02x%02x ",  packet[18], packet[19]);
-	fatal("ofs=%02x%02x ", packet[20], packet[21]);
-	fatal("ttl=%02x ", packet[22]);
-	fatal("p=%02x ", packet[23]);
-	fatal("sum=%02x%02x ", packet[24], packet[25]);
-	fatal("src=%02x%02x%02x%02x ",
-	    packet[26], packet[27], packet[28], packet[29]);
-	fatal("dst=%02x%02x%02x%02x ",
-	    packet[30], packet[31], packet[32], packet[33]);
-	for (i=34; i<len; i++)
-		fatal("%02x", packet[i]);
-	fatal(" ]\n");
-#endif
+	if (ENOUGH_VERBOSITY(SUBSYS_NET, VERBOSITY_DEBUG)) {
+		char s[4000];
+		
+		snprintf(s, sizeof(s), "ver=%02x ", packet[14]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "tos=%02x ", packet[15]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "len=%02x%02x ", packet[16], packet[17]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "id=%02x%02x ",  packet[18], packet[19]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ofs=%02x%02x ", packet[20], packet[21]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ttl=%02x ", packet[22]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "p=%02x ", packet[23]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "sum=%02x%02x ", packet[24], packet[25]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "src=%02x%02x%02x%02x ",
+		    packet[26], packet[27], packet[28], packet[29]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "dst=%02x%02x%02x%02x ",
+		    packet[30], packet[31], packet[32], packet[33]);
+		for (i=34; i<len; i++) {
+			if (strlen(s) > sizeof(s) - 5)
+				break;
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+
+		debugmsg(SUBSYS_NET, "ip BROADCAST", VERBOSITY_DEBUG, "%s", s);
+	}
 
 	/*  Check for 10.0.0.255 first, maybe some guest OSes think that
 	    it's a /24 network, regardless of what it actually is.  */
@@ -1161,27 +1216,44 @@ void net_ip_broadcast(struct net *net, void *extra,
 	    packet[23] == 0x11 &&			/*  UDP  */
 	    packet[34] == 0 && packet[35] == 68 &&	/*  DHCP client  */
 	    packet[36] == 0 && packet[37] == 67) {	/*  DHCP server  */
-		net_ip_broadcast_dhcp(net, extra, packet, len);
+		net_ip_broadcast_dhcp(net, nic, packet, len);
+		return;
+	}
+
+	/*  NetBSD seems to send this during shutdown:  */
+	if (packet[14] == 0x45 &&			/*  IPv4  */
+	    packet[23] == 0x11 &&			/*  UDP  */
+	    packet[34] == 0xff && packet[35] == 0xff &&	/*  0xffff  */
+	    packet[36] == 0x00 && packet[37] == 0x6f) {	/*  ipx-in-ip  */
+		// Let's ignore for now.
 		return;
 	}
 
 	/*  Unknown packet:  */
-	fatal("[ net: UNIMPLEMENTED IP BROADCAST: ");
-	fatal("ver=%02x ", packet[14]);
-	fatal("tos=%02x ", packet[15]);
-	fatal("len=%02x%02x ", packet[16], packet[17]);
-	fatal("id=%02x%02x ",  packet[18], packet[19]);
-	fatal("ofs=%02x%02x ", packet[20], packet[21]);
-	fatal("ttl=%02x ", packet[22]);
-	fatal("p=%02x ", packet[23]);
-	fatal("sum=%02x%02x ", packet[24], packet[25]);
-	fatal("src=%02x%02x%02x%02x ",
-	    packet[26], packet[27], packet[28], packet[29]);
-	fatal("dst=%02x%02x%02x%02x ",
-	    packet[30], packet[31], packet[32], packet[33]);
-	for (i=34; i<len; i++)
-		fatal("%02x", packet[i]);
-	fatal(" ]\n");
+	if (ENOUGH_VERBOSITY(SUBSYS_NET, VERBOSITY_ERROR)) {
+		char s[4000];
+		
+		snprintf(s, sizeof(s), "ver=%02x ", packet[14]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "tos=%02x ", packet[15]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "len=%02x%02x ", packet[16], packet[17]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "id=%02x%02x ",  packet[18], packet[19]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ofs=%02x%02x ", packet[20], packet[21]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "ttl=%02x ", packet[22]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "p=%02x ", packet[23]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "sum=%02x%02x ", packet[24], packet[25]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "src=%02x%02x%02x%02x ",
+		    packet[26], packet[27], packet[28], packet[29]);
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), "dst=%02x%02x%02x%02x ",
+		    packet[30], packet[31], packet[32], packet[33]);
+		for (i=34; i<len; i++) {
+			if (strlen(s) > sizeof(s) - 50)
+				break;
+			snprintf(s+strlen(s), sizeof(s)-strlen(s), "%02x", packet[i]);
+		}
+		snprintf(s+strlen(s), sizeof(s)-strlen(s), " (match=%i)", match);
+
+		debugmsg(SUBSYS_NET, "ip UNIMPLEMENTED BROADCAST", VERBOSITY_ERROR, "%s", s);
+	}
 }
 
 
@@ -1190,7 +1262,7 @@ void net_ip_broadcast(struct net *net, void *extra,
  *
  *  Receive any available UDP packets (from the outside world).
  */
-void net_udp_rx_avail(struct net *net, void *extra)
+void net_udp_rx_avail(struct net *net, struct nic_data *nic)
 {
 	int received_packets_this_tick = 0;
 	int max_packets_this_tick = 200;
@@ -1294,7 +1366,7 @@ void net_udp_rx_avail(struct net *net, void *extra)
 
 			ip_len = 20 + this_packets_data_length;
 
-			lp = net_allocate_ethernet_packet_link(net, extra,
+			lp = net_allocate_ethernet_packet_link(net, nic,
 			    14 + 20 + this_packets_data_length);
 
 			/*  Ethernet header:  */
@@ -1349,7 +1421,7 @@ void net_udp_rx_avail(struct net *net, void *extra)
  *
  *  Receive any available TCP packets (from the outside world).
  */
-void net_tcp_rx_avail(struct net *net, void *extra)
+void net_tcp_rx_avail(struct net *net, struct nic_data *nic)
 {
 	int received_packets_this_tick = 0;
 	int max_packets_this_tick = 200;
@@ -1375,7 +1447,7 @@ void net_tcp_rx_avail(struct net *net, void *extra)
 
 		if (net->tcp_connections[con_id].incoming_buf == NULL)
 			CHECK_ALLOCATION(net->tcp_connections[con_id].
-			    incoming_buf = malloc(TCP_INCOMING_BUF_LEN));
+			    incoming_buf = (unsigned char *) malloc(TCP_INCOMING_BUF_LEN));
 
 		if (net->tcp_connections[con_id].state >=
 		    TCP_OUTSIDE_DISCONNECTED)
@@ -1413,7 +1485,7 @@ void net_tcp_rx_avail(struct net *net, void *extra)
 			net->tcp_connections[con_id].state =
 			    TCP_OUTSIDE_CONNECTED;
 			debug("CHANGING TO TCP_OUTSIDE_CONNECTED\n");
-			net_ip_tcp_connectionreply(net, extra, con_id, 1,
+			net_ip_tcp_connectionreply(net, nic, con_id, 1,
 			    NULL, 0, 0);
 		}
 
@@ -1445,7 +1517,7 @@ void net_tcp_rx_avail(struct net *net, void *extra)
 				    net->tcp_connections[con_id].
 				    incoming_buf_seqnr;
 
-				net_ip_tcp_connectionreply(net, extra, con_id,
+				net_ip_tcp_connectionreply(net, nic, con_id,
 				    0, net->tcp_connections[con_id].
 				    incoming_buf,
 				    net->tcp_connections[con_id].
@@ -1487,21 +1559,21 @@ void net_tcp_rx_avail(struct net *net, void *extra)
 			memcpy(net->tcp_connections[con_id].incoming_buf,
 			    buf, res);
 
-			net_ip_tcp_connectionreply(net, extra, con_id, 0,
+			net_ip_tcp_connectionreply(net, nic, con_id, 0,
 			    buf, res, 0);
 		} else if (res == 0) {
 			net->tcp_connections[con_id].state =
 			    TCP_OUTSIDE_DISCONNECTED;
 			debug("CHANGING TO TCP_OUTSIDE_DISCONNECTED, read"
 			    " res=0\n");
-			net_ip_tcp_connectionreply(net, extra, con_id, 0,
+			net_ip_tcp_connectionreply(net, nic, con_id, 0,
 			    NULL, 0, 0);
 		} else {
 			net->tcp_connections[con_id].state =
 			    TCP_OUTSIDE_DISCONNECTED;
 			fatal("CHANGING TO TCP_OUTSIDE_DISCONNECTED, "
 			    "read res<=0, errno = %i\n", errno);
-			net_ip_tcp_connectionreply(net, extra, con_id, 0,
+			net_ip_tcp_connectionreply(net, nic, con_id, 0,
 			    NULL, 0, 0);
 		}
 
