@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2004-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2004-2018  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -50,7 +50,7 @@
 void net_debugaddr(void *addr, int type)
 {
 	int i;
-	unsigned char *p = addr;
+	unsigned char *p = (unsigned char *) addr;
 
 	switch (type) {
 
@@ -87,17 +87,14 @@ void net_debugaddr(void *addr, int type)
  */
 void net_generate_unique_mac(struct machine *machine, unsigned char *macbuf)
 {
-	int x, y;
-
 	if (macbuf == NULL || machine == NULL) {
 		fatal("**\n**  net_generate_unique_mac(): NULL ptr\n**\n");
 		return;
 	}
 
-	x = machine->serial_nr;
-	y = machine->nr_of_nics;
-
-	macbuf[0] = 0x10;
+	// Special case: real SGI machines have 0x08 as their first
+	// byte, so let's mimic that.
+	macbuf[0] = machine->machine_type == MACHINE_SGI ? 0x08 : 0x10;
 	macbuf[1] = 0x20;
 	macbuf[2] = 0x30;
 	macbuf[3] = 0;
@@ -121,8 +118,7 @@ void net_generate_unique_mac(struct machine *machine, unsigned char *macbuf)
  *  Send a simple UDP packet to a real (physical) host. Used for distributed
  *  network simulations.
  */
-void send_udp(struct in_addr *addrp, int portnr, unsigned char *packet,
-	size_t len)
+void send_udp(struct in_addr *addrp, int portnr, unsigned char *packet, size_t len)
 {
 	int s;
 	struct sockaddr_in si;
