@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2003-2009  Anders Gavare.  All rights reserved.
+ *  Copyright (C) 2003-2021  Anders Gavare.  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
  *  modification, are permitted provided that the following conditions are met:
@@ -47,11 +47,11 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "dec_prom.h"
-#include "dec_5100.h"
-#include "dec_kn01.h"
-#include "dec_kn02.h"
-#include "dec_kn03.h"
+#include "thirdparty/dec_prom.h"
+#include "thirdparty/dec_5100.h"
+#include "thirdparty/dec_kn01.h"
+#include "thirdparty/dec_kn02.h"
+#include "thirdparty/dec_kn03.h"
 
 
 extern int quiet_mode;
@@ -102,17 +102,14 @@ int dec_jumptable_func(struct cpu *cpu, int vector)
 	switch (vector) {
 	case 0x0:	/*  reset()  */
 		/*  TODO  */
-		cpu->machine->exit_without_entering_debugger = 1;
 		cpu->running = 0;
 		break;
 	case 0x10:	/*  restart()  */
 		/*  TODO  */
-		cpu->machine->exit_without_entering_debugger = 1;
 		cpu->running = 0;
 		break;
 	case 0x18:	/*  reinit()  */
 		/*  TODO  */
-		cpu->machine->exit_without_entering_debugger = 1;
 		cpu->cd.mips.gpr[MIPS_GPR_V0] = 0;
 		break;
 	case 0x30:	/*  open()  */
@@ -130,13 +127,13 @@ int dec_jumptable_func(struct cpu *cpu, int vector)
 		cpu->cd.mips.gpr[MIPS_GPR_V0] = 1;
 		break;
 	case 0x38:	/*  read(handle, ptr, length)  */
-		cpu->cd.mips.gpr[MIPS_GPR_V0] = -1;
+		cpu->cd.mips.gpr[MIPS_GPR_V0] = (uint64_t) -1;
 		if ((int32_t)cpu->cd.mips.gpr[MIPS_GPR_A2] > 0) {
 			int disk_id = diskimage_bootdev(cpu->machine, NULL);
 			int res;
 			unsigned char *tmp_buf;
 
-			CHECK_ALLOCATION(tmp_buf =
+			CHECK_ALLOCATION(tmp_buf = (unsigned char *)
 			    malloc(cpu->cd.mips.gpr[MIPS_GPR_A2]));
 
 			res = diskimage_access(cpu->machine, disk_id,
@@ -232,8 +229,8 @@ int decstation_prom_emul(struct cpu *cpu)
 	int vector = cpu->pc & 0xfff;
 	int callback = (cpu->pc & 0xf000)? 1 : 0;
 	unsigned char buf[100];
-	unsigned char ch1, ch2, ch3;
-	uint64_t tmpaddr, slot_base = 0x10000000, slot_size = 0;
+	unsigned char ch1, ch2;
+	uint64_t slot_base = 0x10000000, slot_size = 0;
 
 	if (!callback) {
 		vector = dec_jumptable_func(cpu, vector);
@@ -258,7 +255,7 @@ int decstation_prom_emul(struct cpu *cpu)
 		if ((signed char)ch1 > (signed char)ch2)
 			cpu->cd.mips.gpr[MIPS_GPR_V0] = 1;
 		if ((signed char)ch1 < (signed char)ch2)
-			cpu->cd.mips.gpr[MIPS_GPR_V0] = -1;
+			cpu->cd.mips.gpr[MIPS_GPR_V0] = (uint64_t) -1;
 		break;
 	case 0x14:		/*  strlen():  */
 		i = 0;
@@ -274,37 +271,32 @@ int decstation_prom_emul(struct cpu *cpu)
 		break;
 	case 0x28:		/*  gets()  */
 		/*  debug("[ DEC PROM gets() ]\n");  */
-		tmpaddr = cpu->cd.mips.gpr[MIPS_GPR_A0];
+		// tmpaddr = cpu->cd.mips.gpr[MIPS_GPR_A0];
 		i = 0;
 
-		/*  TODO: Make this not hang (block) the entire emulator  */
-
 		do {
-			while ((ch = console_readchar(
-			    cpu->machine->main_console_handle)) < 1)
+			/*  TODO: Make this not hang (block) the entire emulator  */
+			while ((ch = console_readchar(cpu->machine->main_console_handle)) < 1)
 				;
+
 			if (ch == '\r')
 				ch = '\n';
+
 			ch2 = ch;
 
 			if (ch == '\b') {
 				if (i > 0) {
-					console_putchar(cpu->machine->
-					    main_console_handle, ch2);
-					console_putchar(cpu->machine->
-					    main_console_handle, ' ');
-					console_putchar(cpu->machine->
-					    main_console_handle, ch2);
+					console_putchar(cpu->machine->main_console_handle, ch2);
+					console_putchar(cpu->machine->main_console_handle, ' ');
+					console_putchar(cpu->machine->main_console_handle, ch2);
 				}
 			} else
-				console_putchar(cpu->machine->
-				    main_console_handle, ch2);
+				console_putchar(cpu->machine->main_console_handle, ch2);
 
 			fflush(stdout);
 
 			if (ch == '\n') {
-				/*  It seems that trailing newlines
-				    are not included in the buffer.  */
+				/*  It seems that trailing newlines are not included in the buffer.  */
 			} else if (ch != '\b') {
 				cpu->memory_rw(cpu, cpu->mem, (int32_t)
 				    cpu->cd.mips.gpr[MIPS_GPR_A0] + i,
@@ -381,8 +373,7 @@ int decstation_prom_emul(struct cpu *cpu)
 #endif
 					}
 
-					ch2 = argdata =
-					    cpu->cd.mips.gpr[argreg];
+					ch2 = argdata = cpu->cd.mips.gpr[argreg];
 
 					switch (ch) {
 					case 'c':
@@ -399,21 +390,15 @@ int decstation_prom_emul(struct cpu *cpu)
 						break;
 					case 's':
 						/*  Print a "%s" string.  */
-						j = 0; ch3 = '\n';
+						j = 0;
 						while (ch2) {
-							ch2 = mem_readchar(cpu,
-							    argreg, j++);
+							ch2 = mem_readchar(cpu, argreg, j++);
 							if (ch2) {
 								snprintf(
-								    printfbuf +
-								    strlen(
-								    printfbuf),
-								    sizeof(
-								    printfbuf)-
-								    1-strlen(
-								    printfbuf),
-								    "%c", ch2);
-								ch3 = ch2;
+								    printfbuf + strlen(printfbuf),
+								    sizeof(printfbuf) - 1 - strlen(printfbuf),
+								    "%c",
+								    ch2);
 							}
 						}
 						break;
@@ -421,8 +406,7 @@ int decstation_prom_emul(struct cpu *cpu)
 					argreg ++;
 					break;
 				default:
-					printf("[ unknown printf format char"
-					    " '%c' ]", ch);
+					printf("[ unknown printf format char '%c' ]", ch);
 				}
 				break;
 			case '\0':
@@ -470,7 +454,7 @@ int decstation_prom_emul(struct cpu *cpu)
 			int res;
 			unsigned char *tmp_buf;
 
-			CHECK_ALLOCATION(tmp_buf =
+			CHECK_ALLOCATION(tmp_buf = (unsigned char *)
 			    malloc(cpu->cd.mips.gpr[MIPS_GPR_A2]));
 
 			res = diskimage_access(cpu->machine, disk_id,
@@ -596,7 +580,6 @@ int decstation_prom_emul(struct cpu *cpu)
 		break;
 	case 0x9c:		/*  halt()  */
 		debug("[ DEC PROM halt() ]\n");
-		cpu->machine->exit_without_entering_debugger = 1;
 		cpu->running = 0;
 		break;
 	case 0xa4:		/*  gettcinfo()  */
@@ -635,17 +618,15 @@ int decstation_prom_emul(struct cpu *cpu)
 		switch ((int32_t)cpu->cd.mips.gpr[MIPS_GPR_A0]) {
 		case 'h':
 			debug("DEC PROM: rex('h') ==> halt\n");
-			cpu->machine->exit_without_entering_debugger = 1;
 			cpu->running = 0;
 			break;
 		case 'b':
 			debug("DEC PROM: rex('b') ==> reboot: TODO "
 			    "(halting CPU instead)\n");
-			cpu->machine->exit_without_entering_debugger = 1;
 			cpu->running = 0;
 			break;
 		default:
-			fatal("DEC prom emulation: unknown rex() a0=0x%"PRIx64
+			fatal("DEC prom emulation: unknown rex() a0=0x%" PRIx64
 			    " ('%c')\n",
 			    (int64_t) cpu->cd.mips.gpr[MIPS_GPR_A0],
 			    (char) cpu->cd.mips.gpr[MIPS_GPR_A0]);
@@ -656,14 +637,14 @@ int decstation_prom_emul(struct cpu *cpu)
 		cpu_register_dump(cpu->machine, cpu, 1, 0x1);
 		printf("a0 points to: ");
 		for (i=0; i<40; i++) {
-			unsigned char ch = '\0';
+			unsigned char chTmp = '\0';
 			cpu->memory_rw(cpu, cpu->mem, (int32_t)
-			    cpu->cd.mips.gpr[MIPS_GPR_A0] + i, &ch,
-			    sizeof(ch), MEM_READ, CACHE_DATA | NO_EXCEPTIONS);
-			if (ch >= ' ' && ch < 126)
-				printf("%c", ch);
+			    cpu->cd.mips.gpr[MIPS_GPR_A0] + i, &chTmp,
+			    sizeof(chTmp), MEM_READ, CACHE_DATA | NO_EXCEPTIONS);
+			if (chTmp >= ' ' && chTmp < 126)
+				printf("%c", chTmp);
 			else
-				printf("[%02x]", ch);
+				printf("[%02x]", chTmp);
 		}
 		printf("\n");
 		fatal("PROM emulation: unimplemented callback vector 0x%x\n",
