@@ -42,7 +42,7 @@
 #include "memory.h"
 #include "misc.h"
 
-#include "sh4_exception.h"
+#include "thirdparty/sh4_exception.h"
 
 
 /*
