@@ -60,7 +60,7 @@ int symbol_nsymbols(struct symbol_context *sc)
  *
  *  NOTE:  This is O(n).
  */
-int get_symbol_addr(struct symbol_context *sc, char *symbol, uint64_t *addr)
+int get_symbol_addr(struct symbol_context *sc, const char *symbol, uint64_t *addr)
 {
 	struct symbol *s;
 
@@ -133,7 +133,7 @@ char *get_symbol_name_and_n_args(struct symbol_context *sc, uint64_t addr,
 					    "%s", s->name);
 				else
 					snprintf(symbol_buf, SYMBOLBUF_MAX,
-					    "%s+0x%"PRIx64, s->name, (uint64_t)
+					    "%s+0x%" PRIx64, s->name, (uint64_t)
 					    (addr - s->addr));
 				if (offset != NULL)
 					*offset = addr - s->addr;
@@ -151,14 +151,20 @@ char *get_symbol_name_and_n_args(struct symbol_context *sc, uint64_t addr,
 			s = sc->first_symbol + ofs;
 
 			/*  Found a match?  */
-			if (addr >= s->addr && addr < s->addr + s->len) {
+			if (addr >= s->addr && addr <= s->addr + (s->len - 1)) {
 				if (addr == s->addr)
 					snprintf(symbol_buf, SYMBOLBUF_MAX,
-					    "%s", s->name);
+					    "%s%s%s",
+					    color_symbol_ptr(),
+					    s->name,
+					    color_normal_ptr());
 				else
 					snprintf(symbol_buf, SYMBOLBUF_MAX,
-					    "%s+0x%"PRIx64, s->name, (uint64_t)
-					    (addr - s->addr));
+					    "%s%s%s+0x%" PRIx64,
+					    color_symbol_ptr(),
+					    s->name,
+					    color_normal_ptr(),
+					    (uint64_t) (addr - s->addr));
 
 				if (offset != NULL)
 					*offset = addr - s->addr;
@@ -197,7 +203,7 @@ char *get_symbol_name(struct symbol_context *sc, uint64_t addr, uint64_t *offs)
  *  Add a symbol to the symbol list.
  */
 void add_symbol_name(struct symbol_context *sc,
-	uint64_t addr, uint64_t len, char *name, int type, int n_args)
+	uint64_t addr, uint64_t len, const char *name, int type, int n_args)
 {
 	struct symbol *s;
 
@@ -251,7 +257,7 @@ void add_symbol_name(struct symbol_context *sc,
 	if ((addr >> 32) == 0 && (addr & 0x80000000ULL))
 		addr |= 0xffffffff00000000ULL;
 
-	CHECK_ALLOCATION(s = malloc(sizeof(struct symbol)));
+	CHECK_ALLOCATION(s = (struct symbol *) malloc(sizeof(struct symbol)));
 	memset(s, 0, sizeof(struct symbol));
 
 	s->name = symbol_demangle_cplusplus(name);
@@ -318,7 +324,7 @@ void symbol_readfile(struct symbol_context *sc, char *fname)
 		addr = strtoull(b1, NULL, 16);
 		len  = strtoull(b2, NULL, 16);
 		type = b3[0];
-		/*  printf("addr=%016"PRIx64" len=%016"PRIx64" type=%i\n",
+		/*  printf("addr=%016" PRIx64" len=%016" PRIx64" type=%i\n",
 		    addr, len, type);  */
 
 		if (type == 't' || type == 'r' || type == 'g')
@@ -366,7 +372,7 @@ void symbol_recalc_sizes(struct symbol_context *sc)
 	struct symbol *tmp_ptr;
 	int i;
 
-	CHECK_ALLOCATION(tmp_array = malloc(sizeof (struct symbol) *
+	CHECK_ALLOCATION(tmp_array = (struct symbol *) malloc(sizeof (struct symbol) *
 	    sc->n_symbols));
 
 	/*  Copy first_symbol --> tmp_array, and remove the old
